// CPPDecl.h:  C++ Declaration representation classes

#ifndef INCL_CPPDECL_H
#define INCL_CPPDECL_H

#include <assert.h>
#include "CPPDeclIncomplete.h"
#include "CPPTypeIncomplete.h"
#include "CPPExprIncomplete.h"
#include "StrListIncomplete.h"
#include "CPPConst.h"
#include "JLStr.h"

// Note: All declarations are named, and they may name either a name for
// a type or a name for an object of some type.

///////////////////////////////////////////////////////////////////////
//  CPPDecl class
///////////////////////////////////////////////////////////////////////

class CPPDecl : public VRefCount
{
protected:
   // ctor -- Only created by derived classes
   CPPDecl(
      const JLStr& name_,
      CPPConst::AccessSpecifier accessSpecifier_
   );

public:
   // dtor
   virtual ~CPPDecl();

   // access methods
   CPPConst::AccessSpecifier GetAccessSpecifier() const
   { assertValid(); return accessSpecifier; }

   //
   // virtual Methods to be overridden by derived classes
   //

   // Argument name list.  Zero for non-function declarations
   virtual StrListRefI GetArgNameList() const { return StrListRefI(); }

   // Access to name
   const JLStr& GetName() const { return name; }

   // Are the two declarations identical?
   virtual bool SameDecl(const CPPDecl& rhs) const;

   // Does the declaration define a type or an object?
   virtual bool IsType() const = 0;
   // If the declaration defines a type, is it a typedef?
   virtual bool IsTypedef() const = 0;

   // Get the type of the name
   virtual CPPTypeRefI GetType() const = 0;

   // Get the storage class for the declaration
   virtual CPPConst::StorageClass GetStorageClass() const
   { assertValid(); return CPPConst::scNONE; }

   // Get the declaration modifiers for the declaration
   virtual CPPConst::DeclModifier GetDeclModifier() const
   { assertValid(); return CPPConst::dmNONE; }

   // Get the initializer for the declaration.  Returns
   // a zero-ref if there is no initializer.
   virtual CPPExprRefI GetInitializer() const
   { assertValid(); return CPPExprRefI(); }

protected:
   #ifndef NDEBUG
      enum { CPPDeclMagic = 0x398F };
      short magic;
   #endif
   void assertValid() const { assert(magic == CPPDeclMagic); }

private:
   // Do not allow copy/assignment
   // If you implement these, do not copy the refcount!
   CPPDecl(const CPPDecl&);
   operator=(const CPPDecl&);

   // Declaration name.  This is a copy of the scope key.
   JLStr name;

   // access (public/protected/private) in the containing scope.
   CPPConst::AccessSpecifier accessSpecifier;
};

///////////////////////////////////////////////////////////////////////
// CPPDeclTypedef declaration class.  This class holds declarations
// of typedefs, which name an arbitrary type.  This always refers to
// a type with no derivation, so for example the following declaration:
//      typedef T* foo;
// (assuming T is a type) must first create a derived type of (T*), 
// and then refer to it.
// 
// This class also holds implicit type declarations for class, struct,
// union, and enum declarations.
///////////////////////////////////////////////////////////////////////
class CPPDeclType : public CPPDecl
{
public:
   // ctor/dtor
   CPPDeclType(
      const JLStr& name_,
      CPPConst::AccessSpecifier accessSpecifier_,
      CPPTypeRefI declType_,
      bool isTypedef_
   ) :
      CPPDecl(name_, accessSpecifier_),
      declType(declType_),
      isTypedef(isTypedef_)
   {
   }
   virtual ~CPPDeclType() { assertValid(); }


   // access methods
   // Are the two declarations identical?
   virtual bool SameDecl(const CPPDecl& rhs) const;

   // Does the declaration define a type or an object?
   virtual bool IsType() const { assertValid(); return true; }
   virtual bool IsTypedef() const { assertValid(); return isTypedef; }

   // Get the type of the name
   virtual CPPTypeRefI GetType() const { assertValid(); return declType; }

private:
   // Do not allow copy/assignment
   CPPDeclType(const CPPDeclType&);
   operator=(const CPPDeclType&);

   CPPTypeRefI declType;
   bool isTypedef;
};


///////////////////////////////////////////////////////////////////////
// CPPDeclObject declaration class.  This class holds declarations
// of objects.  Since all types are stored elsewhere, this
// is pretty simple -- it just associates an object name with its type,
// storage class, and declaration modifiers;
///////////////////////////////////////////////////////////////////////
class CPPDeclObject : public CPPDecl
{
public:
   // ctor/dtor
   CPPDeclObject(
      const JLStr& name_,
      CPPConst::AccessSpecifier accessSpecifier_,
      CPPTypeRefI declType_,
      CPPConst::StorageClass storageClass_,
      CPPConst::DeclModifier declModifier_,
      CPPExprRefI initializer_ = CPPExprRefI(),
      StrListRefI argNameList_ = StrListRefI()
   ) :
      CPPDecl(name_, accessSpecifier_),
      declType(declType_),
      storageClass(storageClass_),
      declModifier(declModifier_),
      initializer(initializer_),
      argNameList(argNameList_)
   { }

   virtual ~CPPDeclObject()  { assertValid(); }

   // access methods

   // argument name list.  Zero for non-function declarations
   virtual StrListRefI GetArgNameList() const { return argNameList; }

   // Are the two declarations identical?
   virtual bool SameDecl(const CPPDecl& rhs) const;

   // Does the declaration define a type or an object?
   virtual bool IsType() const { return false; }
   virtual bool IsTypedef() const { return false; }

   // Get the type of the name
   virtual CPPTypeRefI GetType() const { assertValid(); return declType; }

   // Get the storage class for the declaration
   virtual CPPConst::StorageClass GetStorageClass() const
   { assertValid(); return storageClass; }

   // Get the declaration modifiers for the declaration
   virtual CPPConst::DeclModifier GetDeclModifier() const
   { assertValid(); return declModifier; }

   // Get the initializer for the declaration (only valid for object decl)
   virtual CPPExprRefI GetInitializer() const
   { assertValid(); return initializer; }

private:
   // Do not allow copy/assignment
   CPPDeclObject(const CPPDeclObject&);
   operator=(const CPPDeclObject&);

   CPPTypeRefI declType;
   CPPConst::StorageClass storageClass;
   CPPConst::DeclModifier declModifier;
   CPPExprRefI initializer;
   StrListRefI argNameList;
};



#endif
