// CPPExpr.cpp: Expression respresentation.

#include "CPPExpr.h"
#include "CPPDecl.h"
#include "CPPTypeIndirection.h"
#include "CPPScope.h"
#include "CPPFormatOperator.h"
#include <iostream>

////////////////////////////////////////////////////////////////////
// Binary expressions
////////////////////////////////////////////////////////////////////
bool 
CPPExprBinary::EvaluateConstant(
   double& result,
   bool& resultIsInteger
) const
{
   double lhsVal, rhsVal;
   bool rhsIsInteger, lhsIsInteger;
   if (
      lhs->EvaluateConstant(lhsVal, lhsIsInteger) && 
      rhs->EvaluateConstant(rhsVal, rhsIsInteger)
   )
   {
      resultIsInteger = (lhsIsInteger && rhsIsInteger);

      switch (opToken)
      {
      case TOK_PLUS:
         result = lhsVal + rhsVal;
         break;
      case TOK_MINUS:
         result = lhsVal - rhsVal;
         break;
      case TOK_AMPERSAND:
         result = (long)lhsVal & (long)rhsVal;
         resultIsInteger = true;
         break;
      case TOK_NOTEQUAL:
         result = (lhsVal != rhsVal);
         resultIsInteger = true;
         break;
      case TOK_EQUAL:
         result = (lhsVal == rhsVal);
         resultIsInteger = true;
         break;
      case TOK_BITWISEXOR:
         result = (long)lhsVal ^ (long)rhsVal;
         resultIsInteger = true;
         break;
      case TOK_BITWISEOR:
         result = (long)lhsVal | (long)rhsVal;
         resultIsInteger = true;
         break;
      case TOK_AND:
         result = lhsVal && rhsVal;
         resultIsInteger = true;
         break;
      case TOK_OR:
         result = lhsVal || rhsVal;
         resultIsInteger = true;
         break;
      case TOK_STAR:
         result = lhsVal * rhsVal;
         break;
      case TOK_DIVIDE:
         if (rhsVal == 0.0)
         {
            return false;
         } else {
            result = lhsVal / rhsVal;
         }
         break;
      case TOK_MOD:
         if ((long)rhsVal == 0)
         {
            return false;
         } else {
            result = (long)lhsVal % (long)rhsVal;
         }
         resultIsInteger = true;
         break;
      case TOK_LESSTHAN:
         result = lhsVal < rhsVal;
         resultIsInteger = true;
         break;
      case TOK_GREATERTHAN:
         result = lhsVal > rhsVal;
         resultIsInteger = true;
         break;
      case TOK_LESSTHANOREQUALTO:
         result = lhsVal <= rhsVal;
         resultIsInteger = true;
         break;
      case TOK_GREATERTHANOREQUALTO:
         result = lhsVal >= rhsVal;
         resultIsInteger = true;
         break;
      case TOK_SHIFTLEFT:
         if (rhsVal < 0) rhsVal = 0;
         if (rhsVal > 32) rhsVal = 32;
         result = (long)lhsVal << (int)rhsVal;
         resultIsInteger = true;
         break;
      case TOK_SHIFTRIGHT:
         if (rhsVal < 0) rhsVal = 0;
         if (rhsVal > 32) rhsVal = 32;
         result = (long)lhsVal >> (int)rhsVal;
         resultIsInteger = true;
         break;
      default:
         // Things like assignment expressions are not constant
         return false;
      }

      if (resultIsInteger)
      {
         result = (long)result;
      }
      return true;
   }
   return false;
}


JLStr
CPPExprBinary::Format() const
{
   // Always surround with () to make precendence simple
   return 
      JLStr("(") + lhs->Format() + " " + FormatOperator(opToken) + " " + 
      rhs->Format() + ")";
}


CPPTypeRefI 
CPPExprBinary::GetType()
{
   CPPTypeRefI lhsType = lhs->GetType();
   if (lhsType.get() == 0)
   {
      return CPPTypeRefI();
   }

   // TODO: deal with type promotion function overloading

   // This only handles a->b where a is a pointer type, or
   // a.b where a is a class or reference
   switch (opToken)
   {
      case TOK_POINTERTO:
      {
         // Special case for pointer types
         lhsType = lhsType->StripCVQualifiers();
         if (lhsType != lhsType->StripPointer())
         {
            // When lhs is a pointer, then the type is the type of the rhs,
            // which was presumably looked up in the lhs scope.
            return rhs->GetType();
         }
         break;
      }
      case TOK_DOT:
      {
         // Special case for class member access
         // TODO: this won't work for p.~T() when p is non-class
         if (lhsType->StripCVQualifiers()->StripReference()->GetFlavor() == CPPType::Class)
         {
            // When lhs is a class, then the type is the type of the rhs,
            // which was presumably looked up in the lhs scope.
            return rhs->GetType();
         }
         break;
      }
   }
   return CPPTypeRefI();
}

////////////////////////////////////////////////////////////////////
// Unary expressions
////////////////////////////////////////////////////////////////////
bool 
CPPExprUnary::EvaluateConstant(
   double& result,
   bool& resultIsInteger
) const
{
   double rhsVal;
   bool rhsIsInteger;
   if (rhs->EvaluateConstant(rhsVal, rhsIsInteger))
   {
      resultIsInteger = rhsIsInteger;

      switch (opToken) {
      case TOK_AMPERSAND:
      case TOK_STAR:
         return false;
      case TOK_PLUS:
         result = rhsVal;
         break;
      case TOK_MINUS:
         result = -rhsVal;
         break;
      case TOK_TILDE:
         result = ~(long)rhsVal;
         resultIsInteger = true;
         break;
      case TOK_NOT:
         result = (rhsVal == 0);
         resultIsInteger = true;
         break;
      default:
         assert(0);
         return false;
      }
      return true;
   }
   return false;
}


JLStr 
CPPExprUnary::Format() const
{
   return FormatOperator(opToken) + rhs->Format();
}

////////////////////////////////////////////////////////////////////
// Conditional expressions
////////////////////////////////////////////////////////////////////
bool 
CPPExprConditional::EvaluateConstant(
   double& result,
   bool& resultIsInteger
) const
{
   double condVal;
   bool condIsInteger;
   if (condition->EvaluateConstant(condVal, condIsInteger))
   {
      double lhsVal, rhsVal;
      bool rhsIsInteger, lhsIsInteger;
      if (
         lhs->EvaluateConstant(lhsVal, lhsIsInteger) && 
         rhs->EvaluateConstant(rhsVal, rhsIsInteger)
      )
      {
         resultIsInteger = (lhsIsInteger && rhsIsInteger);
         result = (condVal != 0) ? lhsVal : rhsVal;
         return true;
      }
   }
   return false;  
}


JLStr 
CPPExprConditional::Format() const
{
   // Always surround with () to make precendence simple
   return 
      JLStr("(") + condition->Format() + " ? " + 
      lhs->Format() + " : " + rhs->Format();
}


////////////////////////////////////////////////////////////////////
// literal expressions
////////////////////////////////////////////////////////////////////
CPPExprLiteral::CPPExprLiteral(
   ANTLRTokenType tokType_,
   const JLStr& tokText_
) :
   CPPExpr(Literal),
   tokType(tokType_),
   tokText(tokText_)
{
}

bool
CPPExprLiteral::EvaluateConstant(
   double& result,
   bool& isInteger
) const
{
   result = 0;
   const char* rawText = tokText.c_str();
   bool isConstant = true;

   switch (tokType)
   {
   case TOK_OCTALINT:
      {
         for (const char *p = rawText; *p != 0; p++)
         {
            if (*p >= '0' && *p <= '7')
            {
               result = result * 8 + *p - '0';
            }
         }
      }
      isInteger = true;
      break;
   case TOK_DECIMALINT:
      {
         for (const char *p = rawText; *p != 0; p++)
         {
            if (*p >= '0' && *p <= '9')
            {
               result = result * 10 + *p - '0';
            }
         }
      }
      isInteger = true;
      break;
   case TOK_HEXADECIMALINT:
      {
         for (const char *p = rawText; *p != 0; p++)
         {
            if (*p >= '0' && *p <= '9')
            {
               result = result * 16 + *p - '0';
            } else if (*p >= 'a' && *p <= 'f')
            {
               result = result * 16 + *p - 'a' + 10;
            } else if (*p >= 'A' && *p <= 'F')
            {
               result = result * 16 + *p - 'A' + 10;
            }
         }
      }
      isInteger = true;
      break;
   case TOK_CHARACTER:
      {
         // Skip over the ''
         for (const char *p = rawText+1; *p != 0 && *p != '\''; p++)
         {
            result = result * 256 + *p;
         }
      }
      isInteger = true;
      break;
   case TOK_FLOATONE:
   case TOK_FLOATTWO:
      result = atof(rawText);
      isInteger = false;
      break;
   case TOK_BOOLTRUE:
      isInteger = true;
      result = 1;
      break;
   case TOK_BOOLFALSE:
      isInteger = true;
      result = 0;
      break;
   case TOK_STRING:
      isConstant = false;
      break;
   default:
      assert(0);
      isConstant = false;
      break;
   }

   return isConstant;
}

////////////////////////////////////////////////////////////////////
// sizeof expressions
////////////////////////////////////////////////////////////////////
bool 
CPPExprSizeofExpr::EvaluateConstant(
   double& result,
   bool& resultIsInteger
) const
{
   // We don't have a real sizeof() yet...
   cout << "Warning: unimplemented sizeof() was evaluated" << endl;

   result = 4;
   resultIsInteger = true;
   return true;
}


JLStr 
CPPExprSizeofExpr::Format() const
{
   return "sizeof(NYI)";
}

bool 
CPPExprSizeofType::EvaluateConstant(
   double& result,
   bool& resultIsInteger
) const
{
   // We don't have a real sizeof() yet...
   cout << "Warning: unimplemented sizeof() was evaluated" << endl;

   result = 4;
   resultIsInteger = true;
   return true;
}


JLStr 
CPPExprSizeofType::Format() const
{
   return "sizeof(NYI)";
}


////////////////////////////////////////////////////////////////////
// id expressions
////////////////////////////////////////////////////////////////////
bool 
CPPExprId::EvaluateConstant(
   double& result, 
   bool& resultIsInteger
) const
{
   CPPExprRef declExpr = decl->GetInitializer();
   if (declExpr.get() != 0)
   {
      return declExpr->EvaluateConstant(result, resultIsInteger);
   } else {
      return false;
   }
}


JLStr 
CPPExprId::Format() const
{
   return decl->GetName();
}


CPPTypeRefI 
CPPExprId::GetType()
{
   return decl->GetType();
}

////////////////////////////////////////////////////////////////////
// cast expressions
////////////////////////////////////////////////////////////////////
bool 
CPPExprCast::EvaluateConstant(
   double& retval, 
   bool& isInteger
) const
{
   // TODO: Reject evaluations that get cast to non-scalars
   return rhs->EvaluateConstant(retval, isInteger);
}


JLStr 
CPPExprCast::Format() const
{
   bool traditional = (castFlavor == Traditional);
   bool functionLike = (castFlavor == FunctionLike);
   const char* closing = "";

   JLStr retval;
   switch (castFlavor)
   {
   case Const:
      retval = "const_cast<";
      closing = ">";
      break;
   case Static:
      retval = "static_cast<";
      closing = ">";
      break;
   case Dynamic:
      retval = "dynamic_cast<";
      closing = ">";
      break;
   case Reinterpret:
      retval = "reinterpret_cast<";
      closing = ">";
      break;
   case Traditional:
      retval = "(";
      closing = ")";
      break;
   case FunctionLike:
      break;
   }

   retval += type->FormatName("") + closing;

   if (!traditional)
   {
      retval += "(";
   }

   retval += rhs->Format();

   if (!traditional)
   {
      retval += ")";
   }

   return retval;
}

////////////////////////////////////////////////////////////////////
// template placeholder expression
////////////////////////////////////////////////////////////////////
JLStr 
CPPExprTemplatePlaceholder::Format() const
{
   char buf[10];
   sprintf(buf, "%d", parameterIdx);
   char buf2[10];
   if (templateLevel != 0)
   {
      sprintf(buf2, "_%d", templateLevel);
   } else {
      buf2[0] = 0;
   }
   return JLStr("#") + buf + buf2;
}


////////////////////////////////////////////////////////////////////
// new expressions, all forms
////////////////////////////////////////////////////////////////////
JLStr 
CPPExprNew::Format() const
{
   JLStr retval = "new ";

   if (placementExpr.get() != 0)
   {
      retval += "(" + placementExpr->Format() + ") ";
   }

   // TODO: if this is important, get the array size expression
   // embedded in the first dimension of the type format somehow.
   retval += "(" + arrayType->FormatName("") + ") ";

   if (initExpr.get() != 0)
   {
      retval += "(" + initExpr->Format() + ")";
   }

   return retval;
}


////////////////////////////////////////////////////////////////////
// delete expressions
////////////////////////////////////////////////////////////////////
JLStr 
CPPExprDelete::Format() const
{
   return (deleteArray ? "delete[] " : "delete ") + exprToDelete->Format();
}


////////////////////////////////////////////////////////////////////
// function call expressions
////////////////////////////////////////////////////////////////////
JLStr 
CPPExprFunctionCall::Format() const
{
   JLStr retval = exprFunction->Format() + "(";
   if (exprArgList.get() != 0)
   {
      retval += exprArgList->Format();
   }
   retval += ")";
   return retval;
}

CPPTypeRefI 
CPPExprFunctionCall::GetType()
{
   // TODO: perform overload resolution when exprFunction is ExprIdLazy
   CPPTypeRefI functionType = exprFunction->GetType();
   if (
      functionType.get() != 0 && 
      functionType->IsFunctionType()
   )
   {
      return functionType->GetFunctionReturnType();
   } else {
      return CPPTypeRefI();
   }
}


////////////////////////////////////////////////////////////////////
// post-op expressions
////////////////////////////////////////////////////////////////////
JLStr 
CPPExprPostop::Format() const
{
   return lhs->Format() + FormatOperator(opToken);
}

////////////////////////////////////////////////////////////////////
// array subscript expressions
////////////////////////////////////////////////////////////////////
JLStr 
CPPExprArraySubscript::Format() const
{
   return lhs->Format() + "[" + exprSubscript->Format() + "]";
}


CPPTypeRefI 
CPPExprArraySubscript::GetType()
{
   CPPTypeRefI lhsType = lhs->GetType();
   if (lhsType.get() != 0)
   {
      switch (lhsType->GetFlavor())
      {
         case CPPType::Array:
            return lhsType->GetBaseType();
            break;
         case CPPType::Indirection:
            if (
               STATIC_CAST(CPPTypeIndirection*, lhsType.get())->GetIndirectType() ==
               CPPTypeIndirection::IndirectPointer
            )
            {
               return lhsType->GetBaseType();
            } else {
               // TODO: apply operator []
               return CPPTypeRefI();
            }
            break;
         default:
            break;
      }
   }
   return CPPTypeRefI();
}

