// CPPParserMembers.h: Declaration of C++ parser class members.
// This is not in the grammar file, to avoid rebuilding the grammar.

#ifndef INCL_CPPPARSERMEMBERS_H
#define INCL_CPPPARSERMEMBERS_H

   class ScopePopper;
   friend class CPPParser::ScopePopper;
   class ScopeSaver;
   friend class CPPParser::ScopeSaver;
   class TokenCollector;
   friend class CPPParser::TokenCollector;
   class TemplateLevelIncrementer;
   friend class CPPParser::TemplateLevelIncrementer;

public:
   // init 
   virtual void init()
   {
      ANTLRParser::init();
   }

   // dtor
   virtual ~CPPParser() {}

   // Get/set parser options
   enum Option {
      OptPragmaMSPacking,    // process MS-style #pragma packing
      OptMSTypedefHack,      // allow "typedef unsigned int size_t"
                             // "typedef unsigned short wchar_t"
                             // and "typedef int bool"
      OptMax                 // max option value+1
   };
   virtual void SetOption(Option option, bool value) = 0;
   virtual bool GetOption(Option option) const = 0;

   // Debugging and error handling (cpp_debug.cpp);
   // This is public so that symbol table can forward error reporting
   // to the parser
   virtual void ReportError(const JLStr& errMsg, const ANTLRTokenPtr& token);

   // Access to save/restore input stream
   ANTLRTokenBuffer* GetTokenBuffer() const { return inputTokens; }
   void SetTokenBuffer(ANTLRTokenBuffer* newBuffer) { inputTokens = newBuffer; }

   // Access required for changing token stream
   void prime_lookahead() { ANTLRParser::prime_lookahead(); }

protected:
   // Predicates
   bool IsScopeSegment(int k=1, bool allowTokTemplate = false);
   bool NestedBracketsAreFollowedByScope(int k=1);
   bool SkipTemplateQualifiers(int& kInOut);
   bool SkipNestedParens(int& kInOut);
   enum { MaxTemplateTokenScan = 2000 };

   // Debugging and error handling (cpp_debug.cpp);
   virtual void panic(char *);
   virtual void ReportExceptionError(AParserEX& ex);
   virtual void trace_output(const char* msg);

   // Get the pointer to the "void" type.
   CPPTypeRefI LookupVoidType() {
      return LookupBuiltinType(CPPConst::tsVOID, LT(1));
   }

   // Object to save scope state automatically, and restore it if the
   // block is exited (including exceptions) without first calling Commit().
   // The scope stack is restored by popping excess scopes from the stack.
   class ScopePopper {
      public:
         ScopePopper(CPPParser* parser_) : parser(parser_), committed(false)
         {
            state = parser->SaveScope();
         }
         ~ScopePopper() { if (!committed) Restore(); }
         void Commit() { committed = true; }
         void Restore() { parser->RestoreScope(state); }
         CPPScopeState Get() const { return state; }
      private:
         CPPParser *parser;
         CPPScopeState state;
         bool committed;
   };

   // This object is like ScopeSaver but more.  When created it will
   // copy scopes from the scope stack, down to the level indicated 
   // by the state parameter.  
   // Upon destruction, an uncommitted ScopeSaver will restore the
   // scope stack by popping scopes down to the reference state, and then
   // pushing scopes on the stack to recreate the original state.
   class ScopeSaver {
      public:
         ScopeSaver(CPPParser* parser_, CPPScopeState state_);
         ~ScopeSaver() { if (!committed) Restore(); }
         void Commit() { committed = true; }
         void Restore();
      private:
         CPPParser *parser;
         CPPScopeState state;
         CPPScopeStackRefI scopeStack;
         bool committed;
   };

   // Object interface to parser token collection functions.
   // Creating this object on the stack will guarantee stop of
   // token collection if an exception is thrown.
   class TokenCollector {
      public:
         TokenCollector(CPPParser* parser_);
         ~TokenCollector();
         void Clear();
         TokenListRefI Get() const { return tokenList; }
      private:
         CPPParser *parser;
         TokenListRefI tokenList;
   };

   // Class to increase the template nesting level when created, and
   // descrease it when destructed.
   class TemplateLevelIncrementer {
      public:
         TemplateLevelIncrementer(CPPParser* parser_) : 
            parser(parser_) 
         {
            parser->IncreaseTemplateLevel(); 
         }
         ~TemplateLevelIncrementer()
         {
            parser->DecreaseTemplateLevel();
         }
      private:
         CPPParser *parser;
   };

#endif
