// CPPParserPredicates.cpp: predicate methods for CPPParser.
// Note: none of these methods should directly access the symbol table.

#include "CPPParser.h"

///////////////////////////////////////////////////////////////////////
// IsScopeSegment:  Match an identifier optionally followed by nested <> 
// all followed by "::"
//
// Inputs:
//   int     k                   The lookahead slot to start scanning
//   bool    allowTokTemplate    If true, then allow TOK_TEMPLATE to
//                               start the scope segment
// Outputs: none
// Return:
//    bool    true if the test was successful
///////////////////////////////////////////////////////////////////////
bool
CPPParser::IsScopeSegment(
   int k,
   bool allowTokTemplate
)
{
   if (allowTokTemplate && LT(k)->getType() == TOK_TEMPLATE)
   {
      k++;
   }
   return
      LT(k)->getType() == TOK_ID &&
      NestedBracketsAreFollowedByScope(k+1);
}

///////////////////////////////////////////////////////////////////////
// NestedBracketsAreFollowedByScope: Skip over nested <> and determine
// that the following token is TOK_SCOPE.
//
// Inputs:
//   int     k                   The lookahead slot to start scanning
// Outputs: none
// Return:
//    bool    true if the test was successful
///////////////////////////////////////////////////////////////////////
bool
CPPParser::NestedBracketsAreFollowedByScope(
   int k
)
{
   return 
      (LT(k)->getType() != TOK_LESSTHAN || SkipTemplateQualifiers(k)) && 
      LT(k)->getType() == TOK_SCOPE;
}

///////////////////////////////////////////////////////////////////////
// SkipTemplateQualifiers: Skip over <...>.  This correctly handles 
// nested <> and (), e.g:
//    <T>
//    < (i>3) >
//    < T2<...> >
// but not
//    < i>3 >
//
// Inputs:
//    int&    kInOut      The index of the "<"
// Outputs:
//    int&    kInOut      if the return is true, then 
//                           the index of the token after ">"
//                        else
//                           kInOut is unchanged
// Return:
//    bool    true if the template qualifiers were successfully
//            matched and skipped.
///////////////////////////////////////////////////////////////////////
bool
CPPParser::SkipTemplateQualifiers(
   int& kInOut
)
{
   // Start after "<"
   int k = kInOut + 1;

   while ( LT(k)->getType() != TOK_GREATERTHAN ) // scan to end of <...>
   {
      switch ( LT(k)->getType())
      {
         case TOK_Eof:
            return false;
         case TOK_LESSTHAN:
            if (!SkipTemplateQualifiers(k))
            {
               return false;
            }
            break;
         case TOK_LPAREN:
            if (!SkipNestedParens(k))
            {
               return false;
            }
            break;
         default:
            k++;     // skip everything else
            break;
      }
      if (k > MaxTemplateTokenScan )
      {
         return false;
      }
   }

   // Update output argument to point past ">"
   kInOut = k + 1;
   return true;
}



///////////////////////////////////////////////////////////////////////
// SkipNestedParens: Skip over (...).  This correctly handles 
// nested (), e.g:
//    (i>3, (i>5))
//
// Inputs:
//    int&    kInOut      The index of the "("
// Outputs:
//    int&    kInOut      if the return is true, then 
//                           the index of the token after ")"
//                        else
//                           kInOut is unchanged
// Return:
//    bool    true if the (...) were successfully matched and skipped.
///////////////////////////////////////////////////////////////////////
bool
CPPParser::SkipNestedParens(
   int& kInOut
)
{
   // Start after "("
   int k = kInOut + 1;

   while ( LT(k)->getType() != TOK_RPAREN )   // scan to end of (...)
   {
      switch ( LT(k)->getType())
      {
         case TOK_Eof:
            return false;
         case TOK_LPAREN:
            if (!SkipNestedParens(k))
            {
               return false;
            }
            break;
         default:
            k++;     // skip everything else
            break;
      }
      if (k > MaxTemplateTokenScan )
      {
         return false;
      }
   }

   // Update output argument to point past ")"
   kInOut = k + 1;
   return true;
}

