// CPPParserSym.cpp: Implementation of CPPParser with symbols.
// This file mostly forwards all symbol-table-access methods
// to the symbol-table object, except for error-reporting.

#include "CPPParserSym.h"
#include "CPPSymTable.h"
#include "CPPExpr.h"
#include "CPPType.h"
#include "TokenListList.h"

///////////////////////////////////////////////////////////////////////
// ctor
///////////////////////////////////////////////////////////////////////
CPPParserSym::CPPParserSym(
   ANTLRTokenBuffer *input
) :
   CPPParser(input)
{
   options = NewOptionSet();
   symTable = new CPPSymTable(this);
   tokenCollectionLists = new TokenListList;
   markerStack = NewMarkerStack();
}

// ctor used for sub-processing (e.g., template expansion)
CPPParserSym::CPPParserSym(
   ANTLRTokenBuffer *input,
   CPPSymTableRefI symTable_,
   OptionSetRefI options_
) :
   CPPParser(input)
{
   options = options_;
   symTable = symTable_;
   tokenCollectionLists = new TokenListList;
   markerStack = NewMarkerStack();
}

///////////////////////////////////////////////////////////////////////
// initialization
///////////////////////////////////////////////////////////////////////
// virtual
void
CPPParserSym::init()
{
   CPPParser::init();
}

///////////////////////////////////////////////////////////////////////
// dtor
///////////////////////////////////////////////////////////////////////
// 
CPPParserSym::~CPPParserSym()
{
}

///////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////
// Debugging output methods
///////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////
void 
CPPParserSym::DumpScopes() const
{
   symTable->DumpScopes();
}

///////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////
// Methods forwarded to symbol table on behalf of grammar base class
///////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////
bool 
CPPParserSym::IsTypeName(const JLStr &name) const
{
   return symTable->IsTypeName(name);
}

///////////////////////////////////////////////////////////////////////
bool 
CPPParserSym::IsScopeName(const JLStr &name) const
{
   return symTable->IsScopeName(name);
}


///////////////////////////////////////////////////////////////////////
bool 
CPPParserSym::IsTemplateName(const JLStr &name) const
{
   return symTable->IsTemplateName(name);
}


///////////////////////////////////////////////////////////////////////
bool 
CPPParserSym::IsNamespaceName(const JLStr &name) const
{
   return symTable->IsNamespaceName(name);
}


///////////////////////////////////////////////////////////////////////
bool 
CPPParserSym::IsEnclosingClassName(const JLStr &name) const
{
   return symTable->IsEnclosingClassName(name);
}


///////////////////////////////////////////////////////////////////////
void
CPPParserSym::DeclareClassType(
   const JLStr& name,
   CPPConst::TypeSpecifier ts,
   CPPScopeRefI targetScope,
   const ANTLRTokenPtr& errTok,
   CPPTypeRefI& typeReturn,
   CPPDeclRefI& declReturn,
   JLStr& classNameReturn
)
{
   symTable->DeclareClassType(
      name, ts, targetScope, errTok, typeReturn, declReturn, classNameReturn
   );
}


///////////////////////////////////////////////////////////////////////
void 
CPPParserSym::DefineClassType(
   const JLStr& name, 
   CPPConst::TypeSpecifier ts,
   CPPScopeRefI targetScope,
   const ANTLRTokenPtr& errTok,
   CPPTypeRefI& typeReturn,
   CPPDeclRefI& declReturn,
   JLStr& classNameReturn
)
{
   symTable->DefineClassType(
      name, ts, targetScope, errTok, typeReturn, declReturn, classNameReturn
   );
}


///////////////////////////////////////////////////////////////////////
bool
CPPParserSym::DeclareClassTemplateType(
   const JLStr& name,
   CPPConst::TypeSpecifier ts, 
   CPPScopeRefI targetScope,
   TemplateParameterListRefI templateParameters,
   TemplateArgumentListRefI specializationArgList,
   const ANTLRTokenPtr& errTok,
   CPPTypeRefI& classTypeReturn,
   CPPTypeRefI& templateTypeReturn,
   CPPDeclRefI& templateDeclReturn, 
   JLStr& templateNameReturn
)
{
   return symTable->DeclareClassTemplateType(
      name,
      ts, 
      targetScope,
      templateParameters,
      specializationArgList,
      errTok,
      classTypeReturn,
      templateTypeReturn,
      templateDeclReturn, 
      templateNameReturn
   );
}

///////////////////////////////////////////////////////////////////////
bool
CPPParserSym::DefineClassTemplateType(
   const JLStr& name,
   CPPConst::TypeSpecifier ts, 
   CPPScopeRefI targetScope,
   TemplateParameterListRefI templateParameters,
   TemplateArgumentListRefI specializationArgList,
   const ANTLRTokenPtr& errTok,
   CPPTypeRefI& classTypeReturn,
   CPPTypeRefI& templateTypeReturn,
   CPPDeclRefI& templateDeclReturn, 
   JLStr& templateNameReturn
)
{
   return symTable->DefineClassTemplateType(
      name,
      ts, 
      targetScope,
      templateParameters,
      specializationArgList,
      errTok,
      classTypeReturn,
      templateTypeReturn,
      templateDeclReturn, 
      templateNameReturn
   );
}

///////////////////////////////////////////////////////////////////////
void 
CPPParserSym::AddBaseToClass(
   CPPTypeRefI classType,
   CPPTypeRefI baseClassType,
   CPPConst::BaseSpecifier baseSpecifier
)
{
   symTable->AddBaseToClass(classType, baseClassType, baseSpecifier);
}


///////////////////////////////////////////////////////////////////////
void
CPPParserSym::DeclareEnumType(
   const JLStr& name, 
   CPPScopeRefI targetScope,
   const ANTLRTokenPtr& errTok,
   CPPTypeRefI& typeReturn,
   CPPDeclRefI& declReturn,
   JLStr& enumNameReturn
)
{
   symTable->DeclareEnumType(
      name, targetScope, errTok, typeReturn, declReturn, enumNameReturn
   );
}

///////////////////////////////////////////////////////////////////////
void 
CPPParserSym::DefineEnumType(
   const JLStr& name, 
   CPPScopeRefI targetScope,
   const ANTLRTokenPtr& errTok,
   CPPTypeRefI& typeReturn,
   CPPDeclRefI& declReturn,
   JLStr& enumNameReturn
)
{
   symTable->DefineEnumType(
      name, targetScope, errTok, typeReturn, declReturn, enumNameReturn
   );
}

///////////////////////////////////////////////////////////////////////
CPPTypeRefI 
CPPParserSym::LookupPointerToType(
   CPPTypeRefI inputType
)
{
   return symTable->LookupPointerToType(inputType);
}

///////////////////////////////////////////////////////////////////////
CPPTypeRefI 
CPPParserSym::LookupReferenceToType(CPPTypeRefI inputType)
{
   return symTable->LookupReferenceToType(inputType);
}

///////////////////////////////////////////////////////////////////////
CPPTypeRefI 
CPPParserSym::LookupPointerToMemberOfType(
   CPPTypeRefI inputType,
   CPPTypeRefI classType,
   const ANTLRTokenPtr&  //errTok
)
{
   // Construct a type augmented by pointer-to-member
   return symTable->LookupPointerToMemberOfType(inputType, classType);
}

///////////////////////////////////////////////////////////////////////
CPPTypeRefI 
CPPParserSym::LookupCVQualifierOfType(
   CPPTypeRefI inputType, 
   CPPConst::CVQualifier cv
)
{
   return symTable->LookupCVQualifierOfType(inputType, cv);
}

///////////////////////////////////////////////////////////////////////
CPPTypeRefI 
CPPParserSym::LookupArrayOfType(
   CPPTypeRefI inputType, 
   CPPExprRefI exprDimension,
   const ANTLRTokenPtr& errTok
)
{
   return symTable->LookupArrayOfType(inputType, exprDimension, errTok);
}

///////////////////////////////////////////////////////////////////////
CPPTypeRefI 
CPPParserSym::LookupFunctionReturningType(
   CPPTypeRefI inputType,
   CPPConst::CallingConvention cc
)
{
   return symTable->LookupFunctionReturningType(inputType, cc);
}

///////////////////////////////////////////////////////////////////////
CPPTypeRefI 
CPPParserSym::LookupFunctionArgType(
   CPPTypeRefI inputType,
   CPPTypeRefI argType,
   const ANTLRTokenPtr& errTok
)
{
   if (argType == LookupVoidType())
   {
      ReportError("Void parameter type", errTok);
   }
   return symTable->LookupFunctionArgType(inputType, argType);
}

///////////////////////////////////////////////////////////////////////
CPPTypeRefI 
CPPParserSym::LookupFunctionEllipsesType(
   CPPTypeRefI inputType
)
{
   return symTable->LookupFunctionEllipsesType(inputType);
}

///////////////////////////////////////////////////////////////////////
CPPTypeRefI 
CPPParserSym::LookupBitfieldOfType(
   CPPTypeRefI inputType,
   CPPExprRefI exprFieldlen,
   const ANTLRTokenPtr& errTok
)
{
   double value;
   bool isInteger;
   int fieldLen = 1;
   if (!exprFieldlen->EvaluateConstant(value, isInteger))
   {
      ReportError("Non-constant expression for field length", errTok);
   } else if (!isInteger) {
      ReportError("Non-integral expression for field length", errTok);
   } else if (value <= 0) {
      char buf[200];
      sprintf(buf, "Field length (%d) is not positive", (int)value);
      ReportError(buf, errTok);
   } else {
      fieldLen = (int)value;
   }

   return symTable->LookupBitfieldOfType(inputType, fieldLen);
}

///////////////////////////////////////////////////////////////////////
CPPTypeRefI 
CPPParserSym::CreateTempType()
{
   return symTable->CreateTempType();
}

///////////////////////////////////////////////////////////////////////
CPPTypeRefI 
CPPParserSym::MergeTypeModifiers(
   CPPTypeRefI dst, 
   CPPTypeRefI src
)
{
   return symTable->MergeTypeModifiers(dst, src);
}

///////////////////////////////////////////////////////////////////////
CPPScopeState 
CPPParserSym::SaveScope() const
{
   return symTable->SaveScope();
}

///////////////////////////////////////////////////////////////////////
void 
CPPParserSym::RestoreScope(CPPScopeState state)
{
   symTable->RestoreScope(state);
}

///////////////////////////////////////////////////////////////////////
CPPScopeLevel CPPParserSym::PopScope()
{
   return symTable->PopScope();
}

///////////////////////////////////////////////////////////////////////
void CPPParserSym::PushScope(CPPScopeLevel scopeLevel)
{
   symTable->PushScope(scopeLevel);
}

///////////////////////////////////////////////////////////////////////
bool 
CPPParserSym::EnterNamedScope(const JLStr& className, bool resolved)
{
   return symTable->EnterNamedScope(className, resolved);
}

///////////////////////////////////////////////////////////////////////
void 
CPPParserSym::EnterClassScope(
   CPPTypeRefI theClass,
   bool resolved
) 
{
   symTable->EnterClassScope(theClass, resolved);
}


///////////////////////////////////////////////////////////////////////
void 
CPPParserSym::EnterScopeContext(
   CPPScopeRefI scope,
   bool resolved
)
{
   symTable->EnterScopeContext(scope, resolved);
}

///////////////////////////////////////////////////////////////////////
void 
CPPParserSym::EnterResolvedGlobalScope()
{
   symTable->EnterResolvedGlobalScope();
}

///////////////////////////////////////////////////////////////////////
void 
CPPParserSym::EnterLocalScope()
{
   symTable->EnterLocalScope();
}

///////////////////////////////////////////////////////////////////////
void 
CPPParserSym::EnterWorkScope()
{
   symTable->EnterWorkScope();
}

///////////////////////////////////////////////////////////////////////
void 
CPPParserSym::EnterScope(
   const CPPScopeRefI& scope,
   bool resolved
)
{
   symTable->EnterScope(scope, resolved);
}

///////////////////////////////////////////////////////////////////////
void 
CPPParserSym::EnterExprScope(
   CPPExprRefI expr,
   bool resolved
)
{
   symTable->EnterExprScope(expr, resolved);
}

///////////////////////////////////////////////////////////////////////
const JLStr& 
CPPParserSym::GetEnclosingClassName()
{
   return symTable->GetEnclosingClassName();
}

///////////////////////////////////////////////////////////////////////
JLStr 
CPPParserSym::GetUniqueName() const
{
   return symTable->GetUniqueName();
}

///////////////////////////////////////////////////////////////////////
CPPScopeRefI 
CPPParserSym::GetCurrentScope() const
{
   return symTable->GetCurrentScope();
}

///////////////////////////////////////////////////////////////////////
JLStr
CPPParserSym::GetCurrentScopeName() const
{
   return symTable->GetCurrentScopeName();
}

///////////////////////////////////////////////////////////////////////
void 
CPPParserSym::SetCurrentScopeAccess(CPPConst::AccessSpecifier as)
{
   symTable->SetCurrentScopeAccess(as);
}

///////////////////////////////////////////////////////////////////////
CPPTypeRefI
CPPParserSym::LookupEnum(
   const JLStr& name,
   const ANTLRTokenPtr& errTok
)
{
   CPPDeclRefI tmpDecl;
   CPPScopeRefI tmpScope;
   CPPTypeRefI typeReturn;
   if (!symTable->LookupEnum(name, tmpDecl, typeReturn, tmpScope))
   {
      // enum not found
      ReportError(
         "'" + name + "' is not an enum",
         errTok
      );
      // use "int" for declarator type...
      typeReturn = LookupBuiltinType(CPPConst::tsINT, errTok);
   }
   return typeReturn;
}

///////////////////////////////////////////////////////////////////////
CPPTypeRefI
CPPParserSym::LookupType(
   const JLStr& name,
   const ANTLRTokenPtr& errTok
)
{
   CPPDeclRefI tmpDecl;
   CPPScopeRefI tmpScope;
   CPPTypeRefI typeReturn;

   if (!symTable->LookupType(name, tmpDecl, typeReturn, tmpScope))
   {
      ReportError(
         "'" + name + "' does not name a type", 
         errTok
      );
      // Ensure that type is usable
      typeReturn = LookupVoidType();
   }
   return typeReturn;
}

///////////////////////////////////////////////////////////////////////
CPPExprRefI 
CPPParserSym::LookupExprId(
   const JLStr& name,
   const ANTLRTokenPtr& errTok
)
{
   return symTable->LookupExprId(name, errTok);
}

///////////////////////////////////////////////////////////////////////
CPPScopeRefI 
CPPParserSym::LookupScope(
   const JLStr& name,
   const ANTLRTokenPtr& errTok
)
{
   CPPScopeRefI scopeReturn;
   if (!symTable->LookupScope(name, scopeReturn))
   {
      ReportError("'" + name + "' does not name a scope", errTok);
      return GetCurrentScope();
   } else {
      return scopeReturn;
   }
}

///////////////////////////////////////////////////////////////////////
CPPDeclRefI 
CPPParserSym::DeclareObjectOrTypedef(
   const JLStr& name,
   CPPTypeRefI type,
   CPPConst::StorageClass sc,
   CPPConst::DeclModifier dm,
   CPPExprRefI initExpr,
   CPPScopeRefI scope,
   StrListRefI argNameList
)
{
   return symTable->DeclareObjectOrTypedef(
      name, type, sc, dm, initExpr, scope, argNameList
   );
}

///////////////////////////////////////////////////////////////////////
bool 
CPPParserSym::DeclareFunctionTemplate(
   const JLStr& name,
   CPPTypeRefI type,
   CPPConst::StorageClass sc,
   CPPConst::DeclModifier dm,
   CPPExprRefI initExpr,
   CPPScopeRefI scope,
   StrListRefI argNameList,
   TemplateParameterListRefI templateParameters,
   const ANTLRTokenPtr& errTok,
   CPPTypeRefI& templateTypeReturn,
   CPPDeclRefI& templateDeclReturn,
   JLStr& templateNameReturn
)
{
   return symTable->DeclareFunctionTemplate(
      name,
      type,
      sc,
      dm,
      initExpr,
      scope,
      argNameList,
      templateParameters,
      errTok,
      templateTypeReturn,
      templateDeclReturn,
      templateNameReturn
   );
}

///////////////////////////////////////////////////////////////////////
bool 
CPPParserSym::TypeUsesInnerTemplateParms(
   CPPTypeRefI type
)
{
   return symTable->TypeUsesInnerTemplateParms(type);
}

///////////////////////////////////////////////////////////////////////
void CPPParserSym::IncreaseTemplateLevel()
{
   symTable->IncreaseTemplateLevel();
}

void CPPParserSym::DecreaseTemplateLevel()
{
   symTable->DecreaseTemplateLevel();
}

int CPPParserSym::GetTemplateLevel() const
{
   return symTable->GetTemplateLevel();
}


///////////////////////////////////////////////////////////////////////
JLStr CPPParserSym::FormatTypeName(
   CPPTypeRefI cppType
) const
{
   return cppType->FormatName("");
}

