// CPPParserSymCollectTokens.cpp: collection of tokens during template parsing.

#include "CPPParserSym.h"
#include "TokenList.h"
#include "TokenListList.h"

///////////////////////////////////////////////////////////////////////
// Utility stack-of-ints
///////////////////////////////////////////////////////////////////////
class CPPParserSym::MarkerStack : 
   public VRefCount, 
   public JL_DList<int, allocator<int> >
{
};

// Returns a stack with one element initialized to zero.
//static 
CPPParserSym::MarkerStackRefI 
CPPParserSym::NewMarkerStack()
{
   MarkerStackRefI markerStack = new MarkerStack;
   markerStack->push_back(0);
   return markerStack;
}


///////////////////////////////////////////////////////////////////////
// MarkInputTokens: Mark the input token buffer so that (1) tokens are
// not discarded as they are consumed (2) the next-token position can
// be restored later if needed.
//
// Inputs: none
// Outputs: none
// Return: 
//    int    a position marker that can be used with RewindInputTokens
//           or UnmarkInputTokens
///////////////////////////////////////////////////////////////////////
//virtual 
int 
CPPParserSym::MarkInputTokens()
{
   int pos = ANTLRParser::MarkInputTokens();
   markerStack->push_back(0);
   return pos;
}

///////////////////////////////////////////////////////////////////////
// RewindInputTokens: Using a mark from MarkInputTokens(), rewind the
// token buffer to the previous position.
//
// Inputs:
//    int   marker    The position to rewind to; from MarkInputTokens()
// Outputs: none
// Return: none
///////////////////////////////////////////////////////////////////////
//virtual 
void 
CPPParserSym::RewindInputTokens(
   int marker
)
{
   ANTLRParser::RewindInputTokens(marker);
   int nbrTokens = markerStack->back();

   // Remove the number of tokens from each token collector that were
   // accrued since the last mark was made.
   if (tokenCollectionLists->size() != 0)
   {
      for (
         TokenListList::iterator iter = tokenCollectionLists->begin();
         iter != tokenCollectionLists->end();
         iter++
      )
      {
         for (int i = 0; i < nbrTokens; i++)
         {
            (*iter)->pop_back();
         }
      }
   }

   markerStack->pop_back();
}

///////////////////////////////////////////////////////////////////////
// UnmarkInputTokens: Invalidate a mark from MarkInputTokens(), by
// causing input tokens to be processed normally, as if the mark 
// had not been made.
//
// Inputs:
//    int   marker    The marked position to ignore.
// Outputs: none
// Return: none
///////////////////////////////////////////////////////////////////////
//virtual 
void 
CPPParserSym::UnmarkInputTokens(
   int marker
)
{
   ANTLRParser::UnmarkInputTokens(marker);

   // Remove the mark...
   int nbrTokens = markerStack->back();
   markerStack->pop_back();

   // ... but add its token count to the previous marker to
   // deal with nested backtracking.
   assert(markerStack->size() != 0);
   markerStack->back() += nbrTokens;
}

///////////////////////////////////////////////////////////////////////
// PushTokenCollectionList: Add a token-list that will accumulate
// tokens as they are consumed.
// Inputs:
//    const TokenListRefI&   tokenList   The list of tokens to add.
// Outputs: none
// Return: none
///////////////////////////////////////////////////////////////////////
//virtual 
void 
CPPParserSym::PushTokenCollectionList(
   const TokenListRefI& tokenList
)
{
   tokenCollectionLists->push_back(tokenList);
}


///////////////////////////////////////////////////////////////////////
// GetTokenCollectionList: Get the list of tokens use to accumulate
// tokens as they are consumed.
// Inputs: none
// Outputs: none
// Return: none
//    TokenListRefI   The list of tokens being used, or a zero-ref
//                    if no tokens are being collected.
///////////////////////////////////////////////////////////////////////
//virtual 
void
CPPParserSym::PopTokenCollectionList() const 
{
   assert(tokenCollectionLists->size() > 0);
   tokenCollectionLists->pop_back();
}


///////////////////////////////////////////////////////////////////////
// consume:  Override of the base class parser method that is called
// when each token is used by the parser.  This will add each used
// token to a list while collection is on.
///////////////////////////////////////////////////////////////////////
void 
CPPParserSym::consume()
{
   // Add the token to the end of each active token collection list
   if (tokenCollectionLists->size() != 0)
   {
      for (
         TokenListList::iterator iter = tokenCollectionLists->begin();
         iter != tokenCollectionLists->end();
         iter++
      )
      {
         ANTLRTokenPtr tmpPtr = LT(1);
         assert(tmpPtr.operator->() != 0);
         (*iter)->push_back(tmpPtr);
      }
   }

   assert(markerStack->size() != 0);
   markerStack->back()++;

   // Call the base class
   CPPParser::consume();
}

///////////////////////////////////////////////////////////////////////
// CPPParser::TokenCollector.  Utility class used to manage the
// collection of tokens.  This class, when created on the stack,
// will create a (nested) token collection process, which will be
// automatically stopped when the class is destroyed, including
// when an exception is thrown.
///////////////////////////////////////////////////////////////////////
CPPParser::TokenCollector::TokenCollector(
   CPPParser* parser_
) :
   parser(parser_)
{
   // Get a new token collection list
   tokenList = new TokenList;
   // Push the token list on the stack
   parser->PushTokenCollectionList(tokenList);
}

CPPParser::TokenCollector::~TokenCollector()
{
   // Restore the previous token collecting process
   parser->PopTokenCollectionList();
}

void CPPParser::TokenCollector::Clear()
{
   tokenList->clear();
}
