// CPPParserSymType.cpp: Type manipulation for CPPParserSym

#include "CPPParserSym.h"
#include "CPPSymTable.h"
#include "CPPScope.h"
#include "CPPTypeBuiltin.h"
#include "CPPDecl.h"
#include "CPPExpr.h"
#include "JLCast.h"
#include "CPPTypeClassBaseClass.h"

/////////////////////////////////////////////////////////////////////
// DeclareTemplateTypeParameter: Declare a template "type" parameter,
// for example, "T" in:
//     template <class T> ...
// 
// Inputs:
//    const JLStr&         name      The name of the class (may be empty).
//    TokenListRefI        tokens    A list of tokens for the template
//                                   parameter default value.
//    const ANTLRTokenPtr& errTok    A token for error-handling.
//    short                parameterIdx
//                                   The index of the template
//                                   parameter in its list.
//    short                templateLevel
//                                   The template nesting level.
// Outputs:
//    TemplateParameterRefI& templateParmReturn
//                                   Points to the new template
//                                   type parameter.
// Return: none
/////////////////////////////////////////////////////////////////////
// virtual 
void
CPPParserSym::DeclareTemplateTypeParameter(
   const JLStr& name,
   TokenListRefI tokens,
   const ANTLRTokenPtr& errTok,
   short parameterIdx,
   short templateLevel,
   TemplateParameterRefI& templateParmReturn
)
{
   CPPDeclRefI tmpDecl;
   CPPTypeRefI tmpType;
   CPPScopeRefI tmpScope;
   int tmpCount;

   if (symTable->LookupAnyDeclaration(
         name, 
         tmpDecl, 
         tmpType,
         tmpScope,
         tmpCount,
         false    // do not search scope stack
      )
   )
   {
      // The name must not already be in use in this scope for anything.
      ReportError(
         JLStr("Template parameter '") + errTok->getText() + "' is already defined",
         errTok
      );
   }

   symTable->DeclareTemplateTypeParameter(
      name, tokens, parameterIdx, templateLevel, templateParmReturn
   );
}


/////////////////////////////////////////////////////////////////////
// DeclareTemplateValueParameter: Declare a template "value" parameter,
// for example, "I" in:
//     template <int I> ...
// 
// Inputs:
//    const JLStr&         name      The name of the class.
//    TokenListRefI        tokens    A list of tokens for the template
//                                   parameter default value.
//    const ANTLRTokenPtr& errTok    A token for error-handling.
//    short                parameterIdx
//                                   The index of the template
//                                   parameter in its list.
//    short                templateLevel
//                                   The template nesting level.
//    CPPTypeRefI          parameterType
//                                   The type of the value parameter
// Outputs:
//    TemplateParameterRefI& templateParmReturn
//                                   Points to the new template
//                                   value parameter.
// Return: none
/////////////////////////////////////////////////////////////////////
// virtual 
void
CPPParserSym::DeclareTemplateValueParameter(
   const JLStr& name,
   TokenListRefI tokens,
   const ANTLRTokenPtr& errTok,
   short parameterIdx,
   short templateLevel,
   CPPTypeRefI parameterType,
   TemplateParameterRefI& templateParmReturn
)
{
   CPPDeclRefI tmpDecl;
   CPPTypeRefI tmpType;
   CPPScopeRefI tmpScope;
   int tmpCount;

   if (symTable->LookupAnyDeclaration(
         name, 
         tmpDecl, 
         tmpType,
         tmpScope,
         tmpCount,
         false    // do not search scope stack
      )
   )
   {
      // The name must not already be in use in this scope for anything.
      ReportError(
         JLStr("Template parameter '") + errTok->getText() + "' is already defined",
         errTok
      );
   }

   symTable->DeclareTemplateValueParameter(
      name, tokens, parameterIdx, templateLevel, parameterType, templateParmReturn
   );
}


/////////////////////////////////////////////////////////////////////
// DeclareEnumerator: Enter a new enumeration value into the 
// current scope.
// 
// Inputs:
//    const JLStr&          name      The enumerator name.
//    CPPTypeRefI           enumType  The const enumerator type
//    CPPExprRefI           expr      The enumerator value expression
//    const ANTLRTokenPtr&  errTok    A token for error handling
// Outputs: none
// Return:
//    int   The calculated enumerator value
/////////////////////////////////////////////////////////////////////
void
CPPParserSym::DeclareEnumerator(
   const JLStr& name, 
   CPPTypeRefI enumType,
   CPPExprRefI expr,
   const ANTLRTokenPtr& errTok
)
{
   // Evaluate the expression
   double dblValue;
   bool isInteger;
   if (!expr->EvaluateConstant(dblValue, isInteger))
   {
      ReportError("Non-constant expression for enumerator", errTok);
   } else if (!isInteger) {
      ReportError("Non-integral expression for enumerator", errTok);
   }

   // Declare the enumerator as a static enumeration value
   DeclareObjectOrTypedef(
      name,
      enumType,
      CPPConst::scSTATIC,
      CPPConst::dmNONE,
      expr,
      GetCurrentScope(),
      StrListRefI()
   );
}

/////////////////////////////////////////////////////////////////////
// LookupBuiltinType: Lookup a builtin type.  Make a new entry
// for the type if necessary.
//
// Inputs:
//    CPPConst::TypeSpecifier  ts   The builtin type specifier.
//
// Outputs: none
// Return:
//    CPPTypeRefI    Pointer to the builtin type.  Valid until the
//                   parser is reset or destroyed.  Will return
//                   a type representing "void" and report an error
//                   if the type specifiers are not valid.
/////////////////////////////////////////////////////////////////////
CPPTypeRefI
CPPParserSym::LookupBuiltinType(
   CPPConst::TypeSpecifier ts,
   const ANTLRTokenPtr& errTok
)
{
   if (!symTable->ValidateBuiltinType(ts))
   {
      JLStr errMsg = "Invalid builtin type: ";
      errMsg += symTable->FormatBuiltinTypeName(ts);
      ReportError(errMsg, errTok);
   }
   return symTable->LookupBuiltinType(ts);
}

