// CPPParserVirtual.h: Header file to declare parser class
// virtual functions.  Define PARSER_PURE_VIRTUAL before including
// this file:
//    #define PARSER_PURE_VIRTUAL = 0     // declare pure-virtual methods
//    #define PARSER_PURE_VIRTUAL         // declare virtual methods

// Note: This header file should not be gated, as it may be included more
// than once in the same translation unit.

//
// Overrides of guess mode token buffer management
//
virtual int MarkInputTokens() PARSER_PURE_VIRTUAL;
virtual void RewindInputTokens(int marker) PARSER_PURE_VIRTUAL;
virtual void UnmarkInputTokens(int marker) PARSER_PURE_VIRTUAL;

//
// Predicates that use the symbol table
//

// "name" is a type -- typedef/class/struct/union
virtual bool IsTypeName(const JLStr &name) const PARSER_PURE_VIRTUAL;

// "name" is some scope name -- class/struct/union/namespace
virtual bool IsScopeName(const JLStr &name) const PARSER_PURE_VIRTUAL;

// "name" is a templated declaration -- class/function
virtual bool IsTemplateName(const JLStr &name) const PARSER_PURE_VIRTUAL;

// "name" is a namespace
virtual bool IsNamespaceName(const JLStr &name) const PARSER_PURE_VIRTUAL;

// "name" is equivalent to the enclosing class
virtual bool IsEnclosingClassName(const JLStr &name) const PARSER_PURE_VIRTUAL;

//
// Type name formatting
//
virtual JLStr FormatTypeName(
   CPPTypeRefI cppType
) const  PARSER_PURE_VIRTUAL;


// 
// Builtin types
//

// Get the pointer to a builtin type.  Enter the type in the
// table if it does not exist.
virtual CPPTypeRefI 
LookupBuiltinType(
   CPPConst::TypeSpecifier ts,
   const ANTLRTokenPtr& errTok
) PARSER_PURE_VIRTUAL;

//
// Classes
//

// Declare a class type.  OK if class exists.
virtual void DeclareClassType(
   const JLStr& name,
   CPPConst::TypeSpecifier ts,
   CPPScopeRefI targetScope,
   const ANTLRTokenPtr& errTok,
   CPPTypeRefI& typeReturn,
   CPPDeclRefI& declReturn,
   JLStr& classNameReturn
) PARSER_PURE_VIRTUAL;
// Define a class type.  Error if class exists.
virtual void DefineClassType(
   const JLStr& name, 
   CPPConst::TypeSpecifier ts,
   CPPScopeRefI targetScope,
   const ANTLRTokenPtr& errTok,
   CPPTypeRefI& typeReturn,
   CPPDeclRefI& declReturn,
   JLStr& classNameReturn
) PARSER_PURE_VIRTUAL;

// Declare a class template
virtual bool DeclareClassTemplateType(
   const JLStr& name,
   CPPConst::TypeSpecifier ts, 
   CPPScopeRefI targetScope,
   TemplateParameterListRefI templateParameters,
   TemplateArgumentListRefI specializationArgList,
   const ANTLRTokenPtr& errTok,
   CPPTypeRefI& classTypeReturn,
   CPPTypeRefI& templateTypeReturn,
   CPPDeclRefI& templateDeclReturn, 
   JLStr& templateNameReturn
) PARSER_PURE_VIRTUAL;

// Define a class template
virtual bool DefineClassTemplateType(
   const JLStr& name,
   CPPConst::TypeSpecifier ts, 
   CPPScopeRefI targetScope,
   TemplateParameterListRefI templateParameters,
   TemplateArgumentListRefI specializationArgList,
   const ANTLRTokenPtr& errTok,
   CPPTypeRefI& classTypeReturn,
   CPPTypeRefI& templateTypeReturn,
   CPPDeclRefI& templateDeclReturn, 
   JLStr& templateNameReturn
) PARSER_PURE_VIRTUAL;

// Add a base class to a class type
virtual void AddBaseToClass(
   CPPTypeRefI classType,
   CPPTypeRefI baseClassType,
   CPPConst::BaseSpecifier baseSpecifier
) PARSER_PURE_VIRTUAL;


//
// Enums
//

// Declare an enumerator type.  OK if enumerator exists.
virtual void DeclareEnumType(
   const JLStr& name, 
   CPPScopeRefI targetScope,
   const ANTLRTokenPtr& errTok,
   CPPTypeRefI& typeReturn,
   CPPDeclRefI& declReturn,
   JLStr& enumNameReturn
) PARSER_PURE_VIRTUAL;
// Define an enumerator type.  Error if enumerator exists.
virtual void DefineEnumType(
   const JLStr& name, 
   CPPScopeRefI targetScope,
   const ANTLRTokenPtr& errTok,
   CPPTypeRefI& typeReturn,
   CPPDeclRefI& declReturn,
   JLStr& enumNameReturn
) PARSER_PURE_VIRTUAL;
virtual void DeclareEnumerator(
   const JLStr& name, 
   CPPTypeRefI enumType,
   CPPExprRefI exprEnumValue,
   const ANTLRTokenPtr& errTok
) PARSER_PURE_VIRTUAL;

//
// Type modifier manipulation.  These always use pointers to the 
// various representative classes so that they can be incomplete.
//

// Get Modified Type given the base type
virtual CPPTypeRefI LookupPointerToType(
   CPPTypeRefI inputType
) PARSER_PURE_VIRTUAL;
virtual CPPTypeRefI LookupReferenceToType(
   CPPTypeRefI inputType
) PARSER_PURE_VIRTUAL;
virtual CPPTypeRefI LookupPointerToMemberOfType(
   CPPTypeRefI inputType,
   CPPTypeRefI classType,
   const ANTLRTokenPtr& errTok
) PARSER_PURE_VIRTUAL;
virtual CPPTypeRefI LookupCVQualifierOfType(
   CPPTypeRefI inputType, 
   CPPConst::CVQualifier cv
) PARSER_PURE_VIRTUAL;
virtual CPPTypeRefI LookupArrayOfType(
   CPPTypeRefI inputType, 
   CPPExprRefI exprDimension,
   const ANTLRTokenPtr& errTok
) PARSER_PURE_VIRTUAL;
virtual CPPTypeRefI LookupFunctionReturningType(
   CPPTypeRefI inputType,
   CPPConst::CallingConvention cc
) PARSER_PURE_VIRTUAL;
virtual CPPTypeRefI LookupFunctionArgType(
   CPPTypeRefI inputType,
   CPPTypeRefI argType,
   const ANTLRTokenPtr& errTok
) PARSER_PURE_VIRTUAL;
virtual CPPTypeRefI LookupFunctionEllipsesType(
   CPPTypeRefI inputType
) PARSER_PURE_VIRTUAL;
virtual CPPTypeRefI LookupBitfieldOfType(
   CPPTypeRefI inputType,
   CPPExprRefI exprFieldlen,
   const ANTLRTokenPtr& errTok
) PARSER_PURE_VIRTUAL;

//
// Merging of type modifiers
//
virtual CPPTypeRefI CreateTempType() PARSER_PURE_VIRTUAL;
virtual CPPTypeRefI MergeTypeModifiers(
   CPPTypeRefI dst, 
   CPPTypeRefI src
) PARSER_PURE_VIRTUAL;

//
// Template parameters
// 
virtual void DeclareTemplateTypeParameter(
   const JLStr& name,
   TokenListRefI tokens,
   const ANTLRTokenPtr& errTok,
   short parameterIdx,
   short templateLevel,
   TemplateParameterRefI& templateParmReturn
) PARSER_PURE_VIRTUAL;

virtual void DeclareTemplateValueParameter(
   const JLStr& name,
   TokenListRefI tokens,
   const ANTLRTokenPtr& errTok,
   short parameterIdx,
   short templateLevel,
   CPPTypeRefI parameterType,
   TemplateParameterRefI& templateParmReturn
) PARSER_PURE_VIRTUAL;

//
// Template miscellany
//
virtual bool TypeUsesInnerTemplateParms(
   CPPTypeRefI type
) PARSER_PURE_VIRTUAL;

//
// Scope-related stuff
// The parser maintains a scope stack, which is searched for symbols
// from top to bottom, where the bottom of the scope stack is
// always the global scope.
//

// Save the current scope state, which is really just the number of
// scopes on the stack.  This is typically called at the start of 
// a complex declaration
virtual CPPScopeState SaveScope() const PARSER_PURE_VIRTUAL;

// Restore the scope by popping the scope stack back to a 
// pre-existing size. This is typically used at the end
// of a complex declaration.
virtual void RestoreScope(CPPScopeState state) PARSER_PURE_VIRTUAL;

// Pop a single scope off the scope stack
virtual CPPScopeLevel PopScope() PARSER_PURE_VIRTUAL;

// Push a single scope onto the scope stack
virtual void PushScope(CPPScopeLevel scopeLevel) PARSER_PURE_VIRTUAL;

// Enter an existing class/namespace scope.
// Will pull in any base classes as well.
// Use SaveScope()/RestoreScope() to exit the scope(s)
virtual bool EnterNamedScope(const JLStr& className, bool resolved) PARSER_PURE_VIRTUAL;

// Enter a scope context, meaing that the scope, it's parent
// scopes, and any base class scopes will be pulled in.
virtual void EnterScopeContext(CPPScopeRefI scope, bool resolved) PARSER_PURE_VIRTUAL;

// Enter the scope context of a class
virtual void EnterClassScope(CPPTypeRefI theClass, bool resolved) PARSER_PURE_VIRTUAL;

// Enter the global scope in "resolved" mode, meaning
// that global scope will be the only one reachable on the 
// scope stack.  This is used for resolving names like:
//    int x;
//    class A { A(); int x; };
//    A::A() { x = ::x ; } 
// Use SaveScope()/RestoreScope() to exit the global scope
virtual void EnterResolvedGlobalScope() PARSER_PURE_VIRTUAL;

// Enter an unnamed local scope (code block or function body)
// Use SaveScope()/RestoreScope() to exit the local scope
virtual void EnterLocalScope() PARSER_PURE_VIRTUAL;

// Enter an unnamed work scope, which is used to hold 
// the declaration following template <...>
// Use SaveScope()/RestoreScope() to exit the local scope
virtual void EnterWorkScope() PARSER_PURE_VIRTUAL;

// Enter a specific scope that we have a handle to
virtual void EnterScope(
   const CPPScopeRefI& scope,
   bool resolved
) PARSER_PURE_VIRTUAL;

// Enter the scope of the type of object designated by an expression
virtual void EnterExprScope(
   CPPExprRefI expr,
   bool resolved
) PARSER_PURE_VIRTUAL;

// Get the enclosing class name, if there is one
// Returns an empty string if there is no enclosing class scope
virtual const JLStr& GetEnclosingClassName() PARSER_PURE_VIRTUAL;

// Get a unique name for this scope
virtual JLStr GetUniqueName() const PARSER_PURE_VIRTUAL;

// Get the current scope
virtual CPPScopeRefI GetCurrentScope() const PARSER_PURE_VIRTUAL;
virtual JLStr GetCurrentScopeName() const PARSER_PURE_VIRTUAL;

//
// Scope access
//
virtual void SetCurrentScopeAccess(
   CPPConst::AccessSpecifier as
) PARSER_PURE_VIRTUAL;

//
// Lookup of stuff within the scope stack
//

// Find an enum declaration
virtual CPPTypeRefI LookupEnum(
   const JLStr& name,
   const ANTLRTokenPtr& errTok
) PARSER_PURE_VIRTUAL;

// Find a type declaration
virtual CPPTypeRefI LookupType(
   const JLStr& name,
   const ANTLRTokenPtr& errTok
) PARSER_PURE_VIRTUAL;

// Get an expression describing an identifier
virtual CPPExprRefI LookupExprId(
   const JLStr& name,
   const ANTLRTokenPtr& errTok
) PARSER_PURE_VIRTUAL;

// Find a scope by name
virtual CPPScopeRefI LookupScope(
   const JLStr& name,
   const ANTLRTokenPtr& errTok
) PARSER_PURE_VIRTUAL;

//
// Scope declaration stuff
//

// Enter a type or object declaration in the current scope
virtual CPPDeclRefI DeclareObjectOrTypedef(
   const JLStr& name,
   CPPTypeRefI type,
   CPPConst::StorageClass sc,
   CPPConst::DeclModifier dm,
   CPPExprRefI initExpr,
   CPPScopeRefI scope,
   StrListRefI argNameList
) PARSER_PURE_VIRTUAL;

virtual bool DeclareFunctionTemplate(
   const JLStr& name,
   CPPTypeRefI type,
   CPPConst::StorageClass sc,
   CPPConst::DeclModifier dm,
   CPPExprRefI initExpr,
   CPPScopeRefI scope,
   StrListRefI argNameList,
   TemplateParameterListRefI templateParameters,
   const ANTLRTokenPtr& errTok,
   CPPTypeRefI& templateTypeReturn,
   CPPDeclRefI& templateDeclReturn,
   JLStr& templateNameReturn
) PARSER_PURE_VIRTUAL; 

//
// handle #pragmas -- must use LT(i).
//
virtual void HandlePragma() PARSER_PURE_VIRTUAL;

//
// Collection of tokens during template processing.
// This is done so that the tokens comprising a template
// can be collected into a list parallel to the syntax
// scan of the template declaration.  These methods maintain
// a stack of token lists, all of which actively collect
// tokens while on the stack.
//
virtual void PushTokenCollectionList(const TokenListRefI& tokenList) PARSER_PURE_VIRTUAL;
virtual void PopTokenCollectionList() const PARSER_PURE_VIRTUAL;

//
// After a template declaration has been parsed, this
// method is called to set the template tokens
//
virtual void SetTokensInTemplateDeclaration(
   CPPTypeRefI templateType,
   TemplateParameterListRefI templateParameters,
   TokenListRefI tokens
) PARSER_PURE_VIRTUAL;

// 
// Creation of template specializations
// 
virtual void LookupSpecialization(
   const JLStr& name, 
   TemplateArgumentListRefI argList, 
   const ANTLRTokenPtr& errTok,
   JLStr& specializationNameReturn,
   CPPTypeRefI &specializationTypeReturn
) PARSER_PURE_VIRTUAL;

virtual JLStr FormatClassSpecializationName(
   const JLStr& name,
   TemplateArgumentListRefI specializationArgList
) PARSER_PURE_VIRTUAL;

virtual JLStr FormatFunctionSpecializationName(
   const JLStr& name,
   TemplateParameterListRefI templateParameterList,
   TemplateArgumentListRefI specializationArgList
) PARSER_PURE_VIRTUAL;

//
// Methods to build expressions
//
virtual CPPExprRefI MakeExprBinary(
   CPPExprRefI lhs,
   CPPExprRefI rhs,
   ANTLRTokenType opToken
)  PARSER_PURE_VIRTUAL;

virtual CPPExprRefI MakeExprUnary(
   CPPExprRefI rhs,
   ANTLRTokenType opToken
) PARSER_PURE_VIRTUAL;

virtual CPPExprRefI MakeExprConditional(
   CPPExprRefI condition,
   CPPExprRefI lhs,
   CPPExprRefI rhs
) PARSER_PURE_VIRTUAL;

virtual CPPExprRefI MakeExprLiteral(
   ANTLRTokenType tokType,
   const JLStr& tokText
) PARSER_PURE_VIRTUAL;

virtual CPPExprRefI MakeExprSizeofExpr(
   CPPExprRefI expr
) PARSER_PURE_VIRTUAL;

virtual CPPExprRefI MakeExprSizeofType(
   CPPTypeRefI type
) PARSER_PURE_VIRTUAL;

virtual CPPExprRefI MakeExprId(
   CPPDeclRefI decl
) PARSER_PURE_VIRTUAL;

virtual CPPExprRefI MakeExprNewCast(
   ANTLRTokenType castToken,
   CPPTypeRefI castType,
   CPPExprRefI expr
) PARSER_PURE_VIRTUAL;

virtual CPPExprRefI MakeExprTraditionalCast(
   CPPTypeRefI castType,
   CPPExprRefI expr
) PARSER_PURE_VIRTUAL;

virtual CPPExprRefI MakeExprFunctionLikeCast(
   CPPTypeRefI castType,
   CPPExprRefI expr
) PARSER_PURE_VIRTUAL;

virtual CPPExprRefI MakeExprNew(
   CPPExprRefI placementExpr, 
   CPPTypeRefI cppType, 
   CPPExprRefI arraySizeExpr, 
   CPPExprRefI initExpr,
   const ANTLRTokenPtr& errTok
) PARSER_PURE_VIRTUAL;

virtual CPPExprRefI MakeExprDelete(
   CPPExprRefI exprToDelete, 
   bool deleteArray
) PARSER_PURE_VIRTUAL;

virtual CPPExprRefI MakeExprFunctionCall(
   CPPExprRefI exprFunction, 
   CPPExprRefI exprArgList
) PARSER_PURE_VIRTUAL;

virtual CPPExprRefI MakeExprPostop(
   CPPExprRefI exprLhs, 
   ANTLRTokenType tt
) PARSER_PURE_VIRTUAL;

virtual CPPExprRefI MakeExprThis() PARSER_PURE_VIRTUAL;

virtual CPPExprRefI MakeExprArraySubscript(
   CPPExprRefI lhs,
   CPPExprRefI exprSubscript
) PARSER_PURE_VIRTUAL;

//
// Methods to manipulate string lists
//
virtual StrListRefI MakeStrList() PARSER_PURE_VIRTUAL;
virtual void AddStrToList(
   StrListRefI strList,
   const JLStr& str
) PARSER_PURE_VIRTUAL;

//
// Methods to manipulate template parameter lists
//
virtual TemplateParameterListRefI MakeTemplateParameterList() PARSER_PURE_VIRTUAL;
virtual void AddTemplateParameterToList(
   TemplateParameterListRefI parmList,
   TemplateParameterRefI parm
) PARSER_PURE_VIRTUAL;

//
// Methods to manipulate template argument lists
//
virtual TemplateArgumentListRefI MakeTemplateArgumentList() PARSER_PURE_VIRTUAL;
virtual TemplateArgumentRefI MakeTemplateArgumentType(
   CPPTypeRefI type
) PARSER_PURE_VIRTUAL;
virtual TemplateArgumentRefI MakeTemplateArgumentValue(
   CPPExprRefI value
) PARSER_PURE_VIRTUAL;
virtual void AddTemplateArgumentToList(
   TemplateArgumentListRefI argList,
   TemplateArgumentRefI arg
) PARSER_PURE_VIRTUAL;


//
// Methods to manage template nesting level
//
virtual void IncreaseTemplateLevel() PARSER_PURE_VIRTUAL;
virtual void DecreaseTemplateLevel() PARSER_PURE_VIRTUAL;
virtual int GetTemplateLevel() const PARSER_PURE_VIRTUAL;
