// CPPScope.h:  Definition of C++ scope base class.

#ifndef INCL_CPPSCOPE_H
#define INCL_CPPSCOPE_H

#include <assert.h>
#include "CPPDeclIncomplete.h"
#include "CPPScopeIncomplete.h"
#include "CPPTypeIncomplete.h"
#include "JLStr.h"
#include "CPPConst.h"


///////////////////////////////////////////////////////////////////////////////
// CPPScope class.  This is a common base class for the scope classes.
///////////////////////////////////////////////////////////////////////////////
class CPPScope : public VRefCount
{
public:
   // Scope flavor.
   enum Flavor { 
      ScopeInvalid,
      ScopeGlobal,         // File scope
      ScopeLocal,          // Unnamed code block (includes function body)
      ScopeClass,          // Named Class/struct/union
      ScopeNamespace,      // Named namespace
      ScopeFunction,       // Scope with function argument declarations
                           // does not include function body
      ScopeTemplate,       // Scope assoc. with parameterized template types
      ScopeWork            // Block used for template processing.  Differs from
                           // ScopeLocal in that it is considered a permanent scope
                           // for declaration purposes.
   };

   // ctor/dtor
   CPPScope(
      Flavor flavor_, 
      const JLStr& name_
   );
   virtual ~CPPScope();

   // 
   // Misc access methods
   //
   Flavor GetFlavor() const { assertValid(); return flavor; }
   // returned name reference is valid for lifetime of scope object
   const JLStr& GetName() const { assertValid(); return name; }

   //
   // Current access management
   //
   virtual CPPConst::AccessSpecifier GetCurrentAccess() const = 0;
   virtual void SetCurrentAccess(CPPConst::AccessSpecifier as) = 0;

   //
   // Declaration management
   // 
   // Add a declaration to the scope (cannot be removed)
   virtual void AddDecl(CPPDeclRefI decl) = 0;
   // See if a specific declaration is already in the table
   virtual bool HasDecl(const CPPDeclRefI decl) const = 0;

   // Iterating over all declarations in the order that they were made
   //
   // This is an opaque "iterator" that holds the real iterator.
   // It only needs to be big enough to hold the real iterator.
   // There is an assertion that checks this internally.
   // The only reason for this is to avoid exposing the
   // implementation of the declaration list.
   class AllDeclIterator { char dummy[8]; };
   // Gets the first declaration, or a zero-ref if there are none, 
   // and sets "iter" to point to the declaration.  If a zero-ref 
   // is returned, then the "iter" is invalid.
   virtual CPPDeclRefI FindAllDecls(AllDeclIterator& iter) const = 0;
   // Get the next declaration in declaration order
   virtual CPPDeclRefI NextAllDecl(AllDeclIterator& iter) const = 0;
   // Get the total number of declarations
   virtual int CountAllDecls() const = 0;

   // Find out how many declarations are under a name
   virtual int CountDeclsByName(const JLStr& name) const = 0;

   // Iterating over all declarations of a given name.
   //
   // This is an opaque "iterator" that holds the real iterator.
   // It only needs to be big enough to a pair of the real iterator.
   // There is an assertion that checks this internally.
   // The only reason for this is to avoid exposing the
   // implementation of the declaration table.
   class DeclIterator { char dummy[16]; };
   // Gets the first "name" declaration, or a zero-ref if there are 
   // none, and sets "iter" to point to the declaration.  If a zero-ref
   // is returned, then the "iter" is invalid.
   virtual CPPDeclRefI FindDeclByName(const JLStr& name, DeclIterator& iter) const = 0;
   // Given a DeclIterator returned by FindDeclByName() or
   // FindAllDecls, gets the next declaration, and advances 
   // "iter".  Returns a zero-ref if there are no more declarations.
   virtual CPPDeclRefI NextDecl(DeclIterator& iter) const = 0;

   // Finds a named type declaration, of which there should only be one.
   // Returns a zero-ref if there is no such type
   virtual CPPDeclRefI FindTypeDeclByName(
      const JLStr& name
   ) const = 0;

   // Finds a named class declaration, of which there should only be one.
   // Returns a zero-ref if there is no such class
   virtual CPPDeclRefI FindClassDeclByName(
      const JLStr& name
   ) const = 0;

   // Finds a named templated declaration (class/struct/union/function)
   // There should be at most one per scope.
   // Returns a zero-ref if there is no such templated declaration
   virtual CPPDeclRefI FindTemplatedDeclByName(
      const JLStr& name
   ) const = 0;

   //
   // Predicate support
   //
   // "name" is a type -- typedef/class/struct/union
   virtual bool IsTypeName(const JLStr &name) = 0;
   // "name" is some scope name -- class/struct/union/namespace
   virtual bool IsScopeName(const JLStr &name) = 0;
   // "name" is a templated declaration -- class/function
   virtual bool IsTemplatedName(const JLStr &name) = 0;
   // "name" is a namespace name
   virtual bool IsNamespaceName(const JLStr &name) = 0;


   //
   // Sub-scope management
   //
   // Add a named sub-scope to the scope.  Will assert if the scope is
   // already a child of a different scope.  Will assert if there is 
   // already a scope of that name.
   virtual void AddScope(const CPPScopeRefI& scope) = 0;
   // Add a sub-scope under an assumed name
   virtual void AddScope(const JLStr& name, const CPPScopeRefI& scope) = 0;
   // Remove a scope from this scope
   virtual void RemoveScope(const CPPScopeRefI& scope) = 0;
   // Find a sub-scope by name.
   // Returns a zero-ref if no such named scope exists.
   virtual CPPScopeRefI FindScope(
      const JLStr& name
   ) = 0;
   // Format the scope prefix string that represents this scope
   virtual JLStr FormatPrefix(bool formatGlobal) const = 0;

   // This is an opaque "iterator" that holds the real iterator.
   // It only needs to be big enough to hold the real iterator.
   // There is an assertion that checks this internally.
   // The only reason for this is to avoid exposing the
   // implementation of the symbol table in this header.
   class ScopeIterator { char dummy[16]; };
   // Gets the first declaration, or a zero-ref if there are 
   // none, and sets "iter" to point to the scope.  If a zero-ref
   // is returned, then the "iter" is invalid.
   virtual CPPScopeRefI FindAllScopes(ScopeIterator& iter) = 0;
   // Finds the number of sub-scopes in this scope
   virtual int CountAllScopes() const = 0;
   // Given a ScopeIterator returned by FindAllScopes()
   // gets the next scope, and advances "iter".  
   // Returns a zero-ref if there are no more scopes.
   virtual CPPScopeRefI NextScope(ScopeIterator& iter) const = 0;

   // 
   // Parent-scope management
   //
   // Get the parent scope.  Returns zero ref if there is no parent,
   // which will be the case for the global scope, or for scopes
   // that have not been assigned to a parent.
   CPPScopeRefI GetParentScope() const { assertValid(); return parentScope; }
   // Set the parent scope. Will assert if there is already a parent scope
   // that is different from the one being set.
   void SetParentScope(CPPScope* parentScope_)
   {
      assertValid(); 
      assert(parentScope == 0 || parentScope == parentScope_);
      parentScope = parentScope_;
   }
   // Change the parent scope for a scope that already has a parent,
   // and possibly change its name.
   // This is done during template processing to rearrange the relationship
   // of the declared class template scope to the containing scopes.
   void ChangeParentScope(CPPScope* newParent, const JLStr& newName);
   void ChangeParentScope(CPPScope* newParent);

   //
   // Misc stuff
   //
   // Get a unique name (for sub-scopes and declarations)
   static JLStr UniqueName();
   // Test for name generated by UniqueName()
   static bool IsGeneratedName(const JLStr& name);

private:
   JLStr name;                  // scope name
   Flavor flavor;               // scope variation
   CPPScope* parentScope;       // use dumb pointer for back-pointer to parent
   static long uniqueIdx;

   // Do not allow copy/assignment.  
   // If you implement these, do not copy the refcount!
   CPPScope(const CPPScope& rhs);
   CPPScope& operator=(const CPPScope& rhs);

   #ifndef NDEBUG
      enum { CPPScopeMagic = 0xE450 };
      unsigned short magic;
   #endif

protected:
   void assertValid() const { assert(magic == CPPScopeMagic); }
};


#endif
