// CPPScope.cpp: Implementation for CPPScope class

#ifdef _MSC_VER
   // Disable some warnings that appear commonly in Rogue Wave STL
   #pragma warning (disable:4786)    // long debug names
   #pragma warning (disable:4146)    // unary minus operator applied to unsigned type
   #pragma warning (disable:4018)    // '>' : signed/unsigned mismatch
#endif

#include <assert.h>
#include <stdio.h>
#include "CPPScopeNormal.h"
#include "CPPScopeContainers.h"
#include "CPPType.h"
#include "CPPTypeContainers.h"
#include "CPPDecl.h"
#include "CPPDeclContainers.h"

// The "real" iterator type used with DeclIterator
typedef pair<CPPDeclTable::iterator, CPPDeclTable::iterator> RealDeclIterator;

// The "real" iterator type used with AllDeclIterator
typedef CPPDeclTable::iterator RealAllDeclIterator;

// The "real" iterator type used with ScopeIterator
typedef CPPScopeTable::iterator RealScopeIterator;

////////////////////////////////////////////////////////////////////
// Constructor
////////////////////////////////////////////////////////////////////
CPPScopeNormal::CPPScopeNormal(
   Flavor flavor_,
   const JLStr& name_,
   CPPConst::AccessSpecifier defaultAccess
) :
   CPPScope(flavor_, name_),
   currentAccess(defaultAccess)
   #ifndef NDEBUG
      , magicNormal(CPPScopeMagicNormal)
   #endif
{
   declTable = new CPPDeclTable;
}

////////////////////////////////////////////////////////////////////
// Destructor
////////////////////////////////////////////////////////////////////
CPPScopeNormal::~CPPScopeNormal()
{
   assertValid();
   #ifndef NDEBUG
      magicNormal = 0;
   #endif
}

////////////////////////////////////////////////////////////////////
//                     Misc methods                               //
////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////
// GetCurrentAccess: Return the current access of the scope.
//
// Inputs: none
// Return: 
//    CPPConst::AccessSpecifier     the current access
////////////////////////////////////////////////////////////////////
//virtual 
CPPConst::AccessSpecifier 
CPPScopeNormal::GetCurrentAccess() const
{
   return currentAccess;
}

////////////////////////////////////////////////////////////////////
// SetCurrentAccess: Change the current access of the scope
//
// Inputs: 
//    CPPConst::AccessSpecifier   as    The new access
// Return: none
////////////////////////////////////////////////////////////////////
//virtual 
void 
CPPScopeNormal::SetCurrentAccess(
   CPPConst::AccessSpecifier as
)
{
   currentAccess = as;
}

////////////////////////////////////////////////////////////////////
// CPPScopeNormal::FormatPrefix:  Return a JLStr representing a 
//    the prefix string for this scope
//
// Inputs:
//    bool    formatGlobal   true to preprend "::" to the scope
// Outputs: none
// Return:
//     JLStr     A unique name
////////////////////////////////////////////////////////////////////
JLStr 
CPPScopeNormal::FormatPrefix(
   bool formatGlobal
) const
{
   assertValid();
   JLStr prefixStr;

   if (GetParentScope().get() != 0)
   {
      prefixStr = GetParentScope()->FormatPrefix(formatGlobal);
   }

   // OK, so this isn't object-oriented...
   switch (GetFlavor())
   {
      case ScopeGlobal:
         if (formatGlobal)
         {
            prefixStr = "::";
         }
         break;
      case ScopeClass:
      case ScopeNamespace:
         prefixStr += GetName();
         prefixStr += "::";
         break;
      case ScopeInvalid:
      case ScopeLocal:
      case ScopeFunction:
      case ScopeTemplate:
      default:
         break;
   }

   return prefixStr;
}

////////////////////////////////////////////////////////////////////
//                     Declaration management                     //
////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////
// CPPScopeNormal::AddDecl: Add a declaration to the scope.
//
// Inputs:
//    CPPDeclRefI          decl      The declaration to enter in the scope.
//                                Will be entered by the declaration name,
//                                which may be complex for operator functions.
// Outputs: none
// Return: none
//
// Note: If a duplicate declaration is already in the table, as
// determined by CPPDecl::operator==, then the entry is not
// made.
////////////////////////////////////////////////////////////////////
void 
CPPScopeNormal::AddDecl(
   CPPDeclRefI decl
)
{
   assertValid();
   if (!HasDecl(decl))
   {
      declTable->insert(CPPDeclTable::value_type(decl->GetName(), decl));
   }
}

////////////////////////////////////////////////////////////////////
// CPPScopeNormal::HasDecl: See if a specific declaration is in the scope.
//
// Inputs:
//    const CPPDeclRefI    decl   The declaration to look for.
//
// Outputs: none
// Return:
//    bool     true if the declaration is in the table, false o/w
//
// Note: CPPDecl::SameDecl() is used to determine equality of
//    the declarations.
////////////////////////////////////////////////////////////////////
bool
CPPScopeNormal::HasDecl(
   const CPPDeclRefI decl
) const
{
   assertValid();
   DeclIterator iter;
   for (
      CPPDeclRefI tmpDecl = FindDeclByName(decl->GetName(), iter);
      tmpDecl.get() != 0;
      tmpDecl = NextDecl(iter)
   )
   {
      if ( tmpDecl->SameDecl(*decl) )
      {
         return true;
      }
   }
   return false;
}

////////////////////////////////////////////////////////////////////
// CPPScopeNormal::FindAllDecls: return the first declaration 
//    and initialize an "iterator" to point to that declaration.
//
// Inputs: none
// Outputs:
//    AllDeclIterator& iter     If the return value is non-zero, this
//                              is an opaque "iterator" to pass to 
//                              NextDecl(), 
//
// Return:
//    CPPDeclRefI   The first declaration, or zero none found.
//
// Example: Get all of the declarations
//    CPPScopeNormal::DeclIterator iter;
//    for (
//        CPPDeclRefI decl = FindAllDecls(iter);
//        decl != 0;
//        decl = NextDecl(iter);
//    ) {
//        // do something with decl
//    }
////////////////////////////////////////////////////////////////////
CPPDeclRefI
CPPScopeNormal::FindAllDecls(
   AllDeclIterator& iter
) const
{
   assertValid();
   assert(sizeof(DeclIterator) >= sizeof(RealAllDeclIterator));

   RealAllDeclIterator realIter = declTable->begin();

   // Copy the real iterator to the address of "iter"
   new (&iter) RealAllDeclIterator(realIter);

   if (realIter == declTable->end())
   {
      return CPPDeclRefI();
   } else {
      return (*realIter).second;
   }
}

////////////////////////////////////////////////////////////////////
// CPPScopeNormal::NextAllDecl:  Given an "iterator" from FindAllDecls,
// return the next declaration.
//
// Inputs:
//    AllDeclIterator& iter     An "iterator" returned from FindAllDecls
// Outputs:
//    AllDeclIterator& iter     The "iterator" advanced to the next entry
//
// Return:
//    CPPDeclRefI   The next declaration, or zero if none found.
////////////////////////////////////////////////////////////////////
CPPDeclRefI
CPPScopeNormal::NextAllDecl(
   AllDeclIterator& iter
) const
{
   assertValid();
   // Make a correctly-typed alias of the iterator for simplicity
   RealAllDeclIterator& realIter = *(RealAllDeclIterator*)&iter;

   // Make sure that the caller did not overrun the iterator
   assert(realIter != declTable->end());

   // Advance the iterator
   realIter++;

   if (realIter == declTable->end())
   {
      return CPPDeclRefI();
   } else {
      return (*realIter).second;
   }
}

////////////////////////////////////////////////////////////////////
// CPPScopeNormal::CountAllDecls:
//    return the number of declarations in the scope.
//
// Inputs: none
// Outputs: none
//
// Return:
//    int    The number of declarations 
////////////////////////////////////////////////////////////////////
int
CPPScopeNormal::CountAllDecls() const
{
   assertValid();
   return (int)declTable->size();
}
   
////////////////////////////////////////////////////////////////////
// CPPScopeNormal::CountDeclsByName: Given the name for a declaration,
//    return the number of declarations under that name.
//
// Inputs:
//    const JLStr&   name   Name of the declaration.  This
//                          may be something other than an
//                          identifier for operators and
//                          type-casting methods.
// Outputs: none
//
// Return:
//    int    The number of declarations under that name
////////////////////////////////////////////////////////////////////
int
CPPScopeNormal::CountDeclsByName(
   const JLStr& name
) const
{
   assertValid();
   RealDeclIterator range = declTable->equal_range(name);
   return distance(range.first, range.second);
}
   
////////////////////////////////////////////////////////////////////
// CPPScopeNormal::FindDeclByName: Given the name for a declaration,
//    return the first declaration entered under that name,
//    and initialize an "iterator" to point to that declaration.
//
// Inputs:
//    const JLStr&   name   Name of the declaration.  This
//                           may be something other than an
//                           identifier for operators and
//                           type-casting methods.
// Outputs:
//    DeclIterator& iter     If the return value is non-zero, this
//                           is an opaque "iterator" to pass to 
//                           NextDecl(), 
//
// Return:
//    CPPDeclRefI     The first declaration, or zero if none found.
//
// Example: Get all of the declarations under a given name
//    CPPScopeNormal::DeclIterator iter;
//    for (
//        CPPDeclRefI decl = FindDeclByName("foo", iter);;
//        decl != 0;
//        decl = NextDecl(iter);
//    ) {
//        // do something with decl
//    }
////////////////////////////////////////////////////////////////////
CPPDeclRefI
CPPScopeNormal::FindDeclByName(
   const JLStr& name,
   DeclIterator& iter
) const
{
   assertValid();
   assert(sizeof(DeclIterator) >= sizeof(RealDeclIterator));

   RealDeclIterator range = declTable->equal_range(name);

   // Copy the real iterator to the address of "iter"
   new (&iter) RealDeclIterator(range);

   if (range.first == range.second)
   {
      return CPPDeclRefI();
   } else {
      return (*range.first).second;
   }
}

////////////////////////////////////////////////////////////////////
// CPPScopeNormal::NextDecl:  Given an "iterator" from FindDeclByName,
// return the next declaration.
//
// Inputs:
//    DeclIterator& iter     An "iterator" returned from FindDeclByName
// Outputs:
//    DeclIterator& iter     The "iterator" advanced to the next entry
//
// Return:
//    CPPDeclRefI   The next declaration, or zero if none found.
////////////////////////////////////////////////////////////////////
CPPDeclRefI
CPPScopeNormal::NextDecl(
   DeclIterator& iter
) const
{
   assertValid();
   // Make a correctly-typed alias of the iterator for simplicity
   RealDeclIterator& realIter = *(RealDeclIterator*)&iter;

   // Make sure that the caller did not overrun the iterator
   assert(realIter.first != realIter.second);

   // Advance the iterator
   realIter.first++;

   if (realIter.first == realIter.second)
   {
      return CPPDeclRefI();
   } else {
      return (*realIter.first).second;
   }
}

////////////////////////////////////////////////////////////////////
// CPPScopeNormal::FindTypeDeclByName: Given the name for a declaration,
//    return the declaration for a type entered under that name,
//    if there is one.
//
// Inputs:
//    const JLStr&   name   Name of the type.
// Outputs: none
// Return:
//    CPPDeclRefI   The type declaration, or zero if none found.
////////////////////////////////////////////////////////////////////
CPPDeclRefI
CPPScopeNormal::FindTypeDeclByName(
   const JLStr& name
) const
{
   assertValid();
   RealDeclIterator range = declTable->equal_range(name);

   for (
      CPPDeclTable::const_iterator iter = range.first;
      iter != range.second;
      iter++
   )
   {
      if ( (*iter).second->IsType())
      {
         return (*iter).second;
      }
   }

   // no type declaration found under that name
   return CPPDeclRefI();
}


////////////////////////////////////////////////////////////////////
// CPPScopeNormal::FindClassDeclByName: Given the name for a class,
//    return the declaration under that name, if there is one.
//
// Inputs:
//    const JLStr&   name   Name of the class.
// Outputs: none
// Return:
//    CPPDeclRefI  The class declaration, or zero if none found.
////////////////////////////////////////////////////////////////////
CPPDeclRefI
CPPScopeNormal::FindClassDeclByName(
   const JLStr& name
) const
{
   assertValid();
   CPPDeclRefI typeDecl = FindTypeDeclByName(name);
   if (
      typeDecl.get() == 0 ||
      typeDecl->GetType()->GetFlavor() != CPPType::Class
   )
   {
      // No class by that name
      return CPPDeclRefI();
   } else {
      return typeDecl;
   }
}

////////////////////////////////////////////////////////////////////
// CPPScopeNormal::FindTemplatedDeclByName: Given a name, return a
//    templated declaration under that name, if there is one.
//    The templated declaration may be either a class/struct/union
//    or a function.
//
// Inputs:
//    const JLStr&   name   Name of the class.
// Outputs: none
// Return:
//    CPPDeclRefI   The declaration, or zero if none found.
////////////////////////////////////////////////////////////////////
CPPDeclRefI
CPPScopeNormal::FindTemplatedDeclByName(
   const JLStr& name
) const
{
   assertValid();
   RealDeclIterator range = declTable->equal_range(name);

   for (
      CPPDeclTable::const_iterator iter = range.first;
      iter != range.second;
      iter++
   )
   {
      if ((*iter).second->GetType()->GetFlavor() == CPPType::Templated)
      {
         return (*iter).second;
      }
   }

   // no templated declaration found under that name
   return CPPDeclRefI();
}

////////////////////////////////////////////////////////////////////
//                     Scope management                           //
////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////
// CPPScopeNormal::AddScope: Add a sub-scope to the scope symbol table.
//    The scope will be stored using it's name as a key.
//
// Inputs:
//    const CPPScopeRef&    scope   The scope to enter in the table.
//
// Outputs: none
// Return: none
//
// Note: If the scope name is empty, an assertion will trip.  Use
//    a name from UniqueScopeName() for unnamed scopes.
////////////////////////////////////////////////////////////////////
void 
CPPScopeNormal::AddScope(
   const CPPScopeRefI& scope
)
{
   AddScope(scope->GetName(), scope);
}

////////////////////////////////////////////////////////////////////
// CPPScopeNormal::AddScope: Add a sub-scope to the scope symbol table.
// The scope will be stored using the given name as a key.
//
// Inputs:
//    const JLStr&          name    The name of the key
//    const CPPScopeRef&    scope   The scope to enter in the table.
//
// Outputs: none
// Return: none
//
// Note: If the scope name is empty, an assertion will trip.  Use
//    a name from UniqueScopeName() for unnamed scopes.
//
// Note: If a scope of that name is already in the table, then an
//    assertion will trip.  The caller should check for
//    duplicate entries first.
////////////////////////////////////////////////////////////////////
void 
CPPScopeNormal::AddScope(
   const JLStr& name, 
   const CPPScopeRefI& scope
)
{
   assertValid();
   if (scopeTable.get() == 0)
   {
      scopeTable = new CPPScopeTable;
   }

   // Don't allow empty names
   assert(!name.empty());
   // make sure the name is not already entered
   assert(scopeTable->find(name) == scopeTable->end());

   // Add the scope to the table and list
   scopeTable->insert(CPPScopeTable::value_type(name, scope));
   // Set the parent scope to be this scope
   scope->SetParentScope(this);
}


////////////////////////////////////////////////////////////////////
// CPPScopeNormal::RemoveScope.  Remove a subscope from this scope.
// Will assert if the scope is not a child of this scope.
//
// Inputs:
//    const CPPScopeRef&    scope   The scope to remove.
//
// Outputs: none
// Return: none
////////////////////////////////////////////////////////////////////
//virtual 
void 
CPPScopeNormal::RemoveScope(
   const CPPScopeRefI& scope
)
{
   assertValid();
   assert(scopeTable.get() != 0);

   CPPScopeTable::iterator iter = scopeTable->find(scope->GetName());
   assert(iter != scopeTable->end());
   scopeTable->erase(iter);
}


////////////////////////////////////////////////////////////////////
// CPPScopeNormal::FindScope: Find a sub-scope in the scope by name.
//
// Inputs:
//    const JLStr&       name   Name of the scope.
// Outputs: none
// Return:
//    CPPScopeRefI        The scope if found.  If not found, 
//                        then this is reference to a NULL ptr.
////////////////////////////////////////////////////////////////////
CPPScopeRefI
CPPScopeNormal::FindScope(
   const JLStr& name
)
{
   assertValid();
   if (scopeTable.get() == 0)
   {
      return CPPScopeRefI();
   }

   CPPScopeTable::const_iterator iter = scopeTable->find(name);
   if (iter == scopeTable->end())
   {
      return CPPScopeRefI();
   } else {
      return (*iter).second;
   }
}

////////////////////////////////////////////////////////////////////
// CPPScopeNormal::FindAllScopes: return the first sub-scope
//    and initialize an "iterator" to point to that scope.
//
// Inputs: none
// Outputs:
//    ScopeIterator& iter    If the return value is non-zero, this
//                           is an opaque "iterator" to pass to 
//                           NextScope(), 
//
// Return:
//    CPPScopeRefI   The first Scope, or zero none found.
//
// Example: Get all of the Scopes
//    CPPScopeNormal::ScopeIterator iter;
//    for (
//        CPPScopeRefI scope = FindAllScopes(iter);
//        scope.second != 0;
//        scope = NextScope(iter);
//    ) {
//        // do something with scope
//    }
////////////////////////////////////////////////////////////////////
CPPScopeRefI
CPPScopeNormal::FindAllScopes(
   ScopeIterator& iter
)
{
   assertValid();
   assert(sizeof(ScopeIterator) >= sizeof(RealScopeIterator));

   if (scopeTable.get() == 0)
   {
      return CPPScopeRefI();
   }

   RealScopeIterator realIter;
   realIter = scopeTable->begin();

   // Copy the real iterator to the address of "iter"
   new (&iter) RealScopeIterator(realIter);

   if (realIter == scopeTable->end())
   {
      return CPPScopeRefI();
   } else {
      return (*realIter).second;
   }
}

////////////////////////////////////////////////////////////////////
// CPPScopeNormal::NextScope:  Given an "iterator" from FindAllScopes,
// return the next Scope.
//
// Inputs:
//    ScopeIterator& iter     An "iterator" returned from FindAllScopes
// Outputs:
//    ScopeIterator& iter     The "iterator" advanced to the next entry
//
// Return:
//    CPPScope*    The next Scope, or zero if none found.
////////////////////////////////////////////////////////////////////
CPPScopeRefI
CPPScopeNormal::NextScope(
   ScopeIterator& iter
) const
{
   assertValid();
   if (scopeTable.get() == 0)
   {
      return CPPScopeRefI();
   }

   // Make a correctly-typed alias of the iterator for simplicity
   RealScopeIterator& realIter = *(RealScopeIterator*)&iter;

   // Make sure that the caller did not overrun the iterator
   assert(realIter != scopeTable->end());

   // Advance the iterator
   realIter++;

   if (realIter == scopeTable->end())
   {
      return CPPScopeRefI();
   } else {
      return (*realIter).second;
   }
}


////////////////////////////////////////////////////////////////////
// CPPScopeNormal::CountAllScopes: return the number of sub-scopes 
// in this scope.
//
// Inputs: none
// Outputs: none
// Return:
//    int      The number of sub-scopes
////////////////////////////////////////////////////////////////////
int 
CPPScopeNormal::CountAllScopes() const
{
   return (scopeTable.get() == 0) ? 0 : scopeTable->size();
}


////////////////////////////////////////////////////////////////////
//                     Predicate Support                          //
////////////////////////////////////////////////////////////////////

// "name" is a type -- typedef/class/struct/union
bool 
CPPScopeNormal::IsTypeName(
   const JLStr &name
)
{
   assertValid();
   return FindTypeDeclByName(name).get() != 0;
}

// "name" is some scope name -- class/struct/union/namespace
bool 
CPPScopeNormal::IsScopeName(
   const JLStr &name
)
{
   assertValid();
   return FindScope(name).get() != 0;
}

// "name" is a templated declaration -- class/function
bool 
CPPScopeNormal::IsTemplatedName(
   const JLStr &name
)
{
   assertValid();
   return FindTemplatedDeclByName(name).get() != 0;
}

// "name" is a namespace
bool 
CPPScopeNormal::IsNamespaceName(
   const JLStr &name
)
{
   assertValid();
   CPPScopeRefI scope = FindScope(name);
   return (scope.get() != 0 && scope->GetFlavor() == CPPScopeNormal::ScopeNamespace);
}


