// CPPScopeNormal.h: Declaration "normal" scope class and methods

#ifndef INCL_CPPSCOPENORMAL_H
#define INCL_CPPSCOPENORMAL_H

#include "CPPScope.h"

///////////////////////////////////////////////////////////////////////////////
// Implementation note: The parentScope pointer cannot be a ref-counted
// "smart" pointer, because that would create circular references that are
// hard to destroy.  Instead, the parentScope pointers are all "dumb"
// pointers.  Care must be taken to ensure that no scope points to
// a parent that is destroyed before itself.
///////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////
// CPPScope class.  This implements all flavors of scope but 
// CPPScopeTemplate
///////////////////////////////////////////////////////////////////////////////
class CPPScopeNormal : public CPPScope
{
public:

   // ctor/dtor
   CPPScopeNormal(
      Flavor flavor_,
      const JLStr& name_,
      CPPConst::AccessSpecifier defaultAccess = CPPConst::asPUBLIC
   );
   virtual ~CPPScopeNormal();

   //
   // Current access management
   //
   virtual CPPConst::AccessSpecifier GetCurrentAccess() const;
   virtual void SetCurrentAccess(CPPConst::AccessSpecifier as);

   //
   // Declaration management
   // 
   // Add a declaration to the scope (cannot be removed)
   virtual void AddDecl(CPPDeclRefI decl);
   // See if a specific declaration is already in the table
   virtual bool HasDecl(const CPPDeclRefI decl) const;

   // Iterating over all declarations in the order that they were made
   virtual CPPDeclRefI FindAllDecls(AllDeclIterator& iter) const;
   virtual CPPDeclRefI NextAllDecl(AllDeclIterator& iter) const;
   virtual int CountAllDecls() const;

   // Find out how many declarations are under a name
   virtual int CountDeclsByName(const JLStr& name) const;

   // Iterating over all declarations of a given name.
   virtual CPPDeclRefI FindDeclByName(const JLStr& name, DeclIterator& iter) const;
   virtual CPPDeclRefI NextDecl(DeclIterator& iter) const;

   // Finds a named type declaration, of which there should only be one.
   // Returns a zero-ref if there is no such type
   virtual CPPDeclRefI FindTypeDeclByName(
      const JLStr& name
   ) const;

   // Finds a named class declaration, of which there should only be one.
   // Returns a zero-ref if there is no such class
   virtual CPPDeclRefI FindClassDeclByName(
      const JLStr& name
   ) const;

   // Finds a named templated declaration (class/struct/union/function)
   // There should be at most one per scope.
   // Returns a zero-ref if there is no such templated declaration
   virtual CPPDeclRefI FindTemplatedDeclByName(
      const JLStr& name
   ) const;

   //
   // Predicate support
   //
   // "name" is a type -- typedef/class/struct/union
   virtual bool IsTypeName(const JLStr &name);
   // "name" is some scope name -- class/struct/union/namespace
   virtual bool IsScopeName(const JLStr &name);
   // "name" is a templated declaration -- class/function
   virtual bool IsTemplatedName(const JLStr &name);
   // "name" is a namespace name
   virtual bool IsNamespaceName(const JLStr &name);


   //
   // Sub-scope management
   //
   // Add a named sub-scope to the scope.  Will assert if the scope is
   // already a child of a different scope.  Will assert if there is 
   // already a scope of that name.
   virtual void AddScope(const CPPScopeRefI& scope);
   // Add a sub-scope under an assumed name
   virtual void AddScope(const JLStr& name, const CPPScopeRefI& scope);
   // Remove a scope from this scope
   virtual void RemoveScope(const CPPScopeRefI& scope);
   // Find a sub-scope by name.
   // Returns a zero-ref if no such named scope exists.
   virtual CPPScopeRefI FindScope(
      const JLStr& name
   );
   // Format the scope prefix string that represents this scope
   virtual JLStr FormatPrefix(bool formatGlobal) const;

   // Gets the first declaration, or a zero-ref if there are 
   // none, and sets "iter" to point to the scope.  If a zero-ref
   // is returned, then the "iter" is invalid.
   virtual CPPScopeRefI FindAllScopes(ScopeIterator& iter);
   // Finds the number of sub-scopes in this scope
   virtual int CountAllScopes() const;
   // Given a ScopeIterator returned by FindAllScopes()
   // gets the next scope, and advances "iter".  
   // Returns a zero-ref if there are no more scopes.
   virtual CPPScopeRefI NextScope(ScopeIterator& iter) const;

private:
   CPPDeclTableRefI declTable;  // mapping from decl names to declarations
   CPPScopeTableRefI scopeTable;
   CPPConst::AccessSpecifier currentAccess;

   // Do not allow copy/assignment.  
   CPPScopeNormal(const CPPScopeNormal& rhs);
   CPPScopeNormal& operator=(const CPPScopeNormal& rhs);

   #ifndef NDEBUG
      enum { CPPScopeMagicNormal = 0x9A32 };
      unsigned short magicNormal;
   #endif

   void assertValid() const { assert(magicNormal == CPPScopeMagicNormal); }
};


#endif
