// CPPScopeTemplate.h: Definition of special template placeholder scope

#ifndef INCL_CPPSCOPETEMPLATE_H
#define INCL_CPPSCOPETEMPLATE_H

#include "CPPScope.h"

///////////////////////////////////////////////////////////////////////////////
// CPPScopeTemplate: 
// Special subclass of CPPScope that supports name lookup for
// templated declarations.  For example, in:
//    template <class T> class A { T::I i; };
// The symbol "I" must be found in the T's scope, even though the
// type "T" is parameterized and thus not defined.  Essentially,
// this scope always reports that the desired name is found.
///////////////////////////////////////////////////////////////////////////////
class CPPScopeTemplate : public CPPScope
{
public:
   CPPScopeTemplate();
   virtual ~CPPScopeTemplate();

   // Misc access methods
   virtual CPPConst::AccessSpecifier GetCurrentAccess() const;
   virtual void SetCurrentAccess(CPPConst::AccessSpecifier as);

   // Reject attempts to add decls
   virtual void AddDecl(CPPDeclRefI) { assert(0);}

   // This test should not be performed.
   virtual bool HasDecl(const CPPDeclRefI decl) const { assert(0); return false; }

   // Gets the first declaration.  This special scope always
   // returns one object declaration for any name.
   virtual CPPDeclRefI FindDeclByName(
      const JLStr& name, 
      DeclIterator& iter
   ) const;
   // Given a DeclIterator returned by FindDeclByName() or
   // FindAllDecls, gets the next declaration, but there
   // never is one for the special template scope.
   virtual CPPDeclRefI NextDecl(DeclIterator&) const { assertValid(); return CPPDeclRefI(); }

   // Find out how many declarations are under a name
   virtual int CountDeclsByName(const JLStr&) const { assertValid(); return 1; }

   // Find all declarations
   virtual CPPDeclRefI FindAllDecls(AllDeclIterator& iter) const { return CPPDeclRefI(); }
   // Get the next declaration in declaration order
   virtual CPPDeclRefI NextAllDecl(AllDeclIterator& iter) const  { assert(0); return CPPDeclRefI(); }
   // Get the total number of declarations
   virtual int CountAllDecls() const { return 1; }


   // Finds a named type declaration.
   // This function always returns a placeholder as a type declaration
   virtual CPPDeclRefI FindTypeDeclByName(const JLStr&) const;

   // Finds a named class declaration.
   // This function always returns a placeholder as a type declaration
   virtual CPPDeclRefI FindClassDeclByName(const JLStr&) const;

   // Finds a named templated declaration (class/struct/union/function).
   // This function always returns a placeholder as a templated declaration
   virtual CPPDeclRefI FindTemplatedDeclByName(const JLStr&) const;

   // Find a sub-scope by name.
   // This special template scope always reports exactly one scope
   // regardless of the name.
   virtual CPPScopeRefI FindScope(const JLStr& name);
   // Format the scope prefix string that represents this scope
   virtual JLStr FormatPrefix(bool formatGlobal) const;

   // Return this scope
   virtual CPPScopeRefI FindAllScopes(ScopeIterator&) { assertValid(); return this; }
   // Always returns one scope
   virtual int CountAllScopes() const { assertValid(); return 1; }
   virtual CPPScopeRefI NextScope(ScopeIterator&) const { assertValid(); return CPPScopeRefI(); }

   // reject attempts to add scopes
   virtual void AddScope(const JLStr&, const CPPScopeRefI&) { assert(0); }
   virtual void AddScope(const CPPScopeRefI&) { assert(0); }
   virtual void RemoveScope(const CPPScopeRefI& scope) { assert(0); }

   //
   // Predicate support
   //
   // "name" is a type -- always false for special template scope
   virtual bool IsTypeName(const JLStr &name) { assertValid(); return false; }
   // "name" is some scope name -- always false for special template scope
   virtual bool IsScopeName(const JLStr &name) { assertValid(); return false; }
   // "name" is a templated declaration --  always false for special template scope
   virtual bool IsTemplatedName(const JLStr &name) { assertValid(); return false; }
   // "name" is a namespace name
   virtual bool IsNamespaceName(const JLStr &name) { assertValid(); return false; }

private:
   void InitStatics();

   // Do not allow copy/assignment.  
   CPPScopeTemplate(const CPPScopeTemplate& rhs);
   CPPScopeTemplate& operator=(const CPPScopeTemplate& rhs);

   // Data used for search return values.
   static CPPDeclRefI theObjectDecl;
   static CPPDeclRefI theTypeDecl;
   #ifndef NDEBUG
      enum { CPPScopeMagicTemplate = 0x9A32 };
      unsigned short magicTemplate;
   #endif

   void assertValid() const { assert(magicTemplate == CPPScopeMagicTemplate); }
};


#endif
