// CPPSymTable.h:  Implementation of C++ parser symbol table

#ifndef INCL_CPPSYMTABLE_H
#define INCL_CPPSYMTABLE_H

#include "CPPSymTableIncomplete.h"
#include "CPPDeclIncomplete.h"
#include "CPPExprIncomplete.h"
#include "CPPScopeIncomplete.h"
#include "CPPTypeIncomplete.h"
#include "StrListIncomplete.h"
#include "TokenListIncomplete.h"
#include "TemplateParameterListIncomplete.h"
#include "TemplateArgumentListIncomplete.h"
#include "CPPConst.h"
#include "CPPScopeState.h"
#include "CPPScopeLevel.h"
#include "JLStr.h"
#include "AToken.h"
#include "ATokPtr.h"

// Suppress warnings in MSVC++
#ifdef _MSC_VER
   #pragma warning (disable:4786)    // long debug names
   #pragma warning (disable:4146)    // unary minus operator applied to unsigned type
   #pragma warning (disable:4018)    // '>' : signed/unsigned mismatch
#endif

class CPPParserSym;

class CPPSymTable : public VRefCount
{
public:
   CPPSymTable(CPPParserSym* parser_);
   ~CPPSymTable();

   // Access to internal symbols -- do not modify return values!
   CPPBuiltinTypeTableRefI 
   GetBuiltinTypes() const { return builtinTypes; }

   CPPScopeRefI 
   GetGlobalScope() const { return globalScope; }

   //
   // Predicates
   // 
   bool IsTypeName(const JLStr &name) const;
   bool IsScopeName(const JLStr &name) const;
   bool IsTemplateName(const JLStr &name) const;
   bool IsNamespaceName(const JLStr &name) const;
   bool IsEnclosingClassName(const JLStr &name) const;

   //
   // Derived type manipulation
   //

   CPPTypeRefI 
   LookupPointerToType(
      CPPTypeRefI inputType
   );

   CPPTypeRefI 
   LookupReferenceToType(
      CPPTypeRefI inputType
   );

   CPPTypeRefI 
   LookupPointerToMemberOfType(
      CPPTypeRefI inputType,
      CPPTypeRefI classType
   );

   CPPTypeRefI 
   LookupCVQualifierOfType(
      CPPTypeRefI inputType, 
      CPPConst::CVQualifier cv
   );

   CPPTypeRefI 
   LookupArrayOfType(
      CPPTypeRefI inputType, 
      CPPExprRefI exprDimension,
      const ANTLRTokenPtr& errTok
   );

   CPPTypeRefI 
   LookupFunctionReturningType(
      CPPTypeRefI inputType,
      CPPConst::CallingConvention cc
   );

   CPPTypeRefI 
   LookupFunctionArgType(
      CPPTypeRefI inputType,
      CPPTypeRefI argType
   );

   CPPTypeRefI 
   LookupFunctionEllipsesType(
      CPPTypeRefI inputType
   );

   CPPTypeRefI 
   LookupBitfieldOfType(
      CPPTypeRefI inputType,
      int fieldlen
   );


   //
   // Manipulation of temporary type trees
   // 
   CPPTypeRefI 
   CreateTempType();

   CPPTypeRefI
   MergeTypeModifiers(
      CPPTypeRefI dst, 
      CPPTypeRefI src
   );

   //
   // Builtin types
   //
   CPPTypeRefI
   LookupBuiltinType(
      CPPConst::TypeSpecifier ts
   );

   bool
   ValidateBuiltinType(
      CPPConst::TypeSpecifier ts
   );

   CPPTypeRefI
   MakeBuiltinType(
      CPPConst::TypeSpecifier ts
   );

   JLStr 
   FormatBuiltinTypeName(
      CPPConst::TypeSpecifier ts
   );


   //
   // Scope stack
   //
   CPPScopeState
   SaveScope() const;

   void 
   RestoreScope(
      CPPScopeState state
   );

   CPPScopeLevel 
   PopScope();

   void 
   PushScope(
      CPPScopeLevel scopeLevel
   );

   bool 
   EnterNamedScope(
      const JLStr& name,
      bool resolved
   );

   void 
   EnterScopeContext(
      CPPScopeRefI scope,
      bool resolved
   );

   void 
   EnterClassScope(
      CPPTypeRefI theClass,
      bool resolved
   );

   void 
   EnterResolvedGlobalScope();

   void 
   EnterLocalScope();

   void 
   EnterWorkScope();

   void 
   EnterScope(const CPPScopeRefI& scope, bool resolved);

   void 
   EnterExprScope(CPPExprRefI expr, bool resolved);

   const JLStr& 
   GetEnclosingClassName() const;

   CPPScopeRefI
   GetCurrentScope() const;

   JLStr 
   GetCurrentScopeName() const;

   void 
   SetCurrentScopeAccess(
      CPPConst::AccessSpecifier as
   );

   CPPConst::AccessSpecifier 
   GetCurrentScopeAccess() const;

   CPPScopeRefI
   CPPSymTable::GetEnclosingPermanentScope() const;

   //
   // class/struct packing
   //
   void 
   SetDefaultPacking();

   void 
   SetPacking(int packing);

   void 
   PopPacking();

   void 
   PushPacking(int packing);

   int
   GetPackingStackSize() const;

   int 
   GetCurrentPacking() const;

   //
   // Looking up stuff in the scope stack
   // 
   bool 
   LookupClass(
      const JLStr& name,
      CPPDeclRefI& declReturn,
      CPPTypeRefI& typeReturn,
      CPPScopeRefI& containingScopeReturn,
      bool searchStack = true
   ) const;

   bool 
   LookupEnum(
      const JLStr& name,
      CPPDeclRefI& declReturn,
      CPPTypeRefI& typeReturn,
      CPPScopeRefI& containingScopeReturn,
      bool searchStack = true
   ) const;

   bool 
   LookupType(
      const JLStr& name,
      CPPDeclRefI& declReturn,
      CPPTypeRefI& typeReturn,
      CPPScopeRefI& containingScopeReturn,
      bool searchStack = true
   ) const;

   bool 
   LookupScope(
      const JLStr& name,
      CPPScopeRefI& scopeReturn,
      bool searchStack = true
   ) const;

   bool
   LookupObjectDeclaration(
      const JLStr& name,
      CPPDeclRefI& firstDeclReturn,
      CPPTypeRefI& firstTypeReturn,
      CPPScopeRefI& containingScopeReturn,
      int &declCount,
      bool searchStack = true
   ) const;

   bool
   LookupAnyDeclaration(
      const JLStr& name,
      CPPDeclRefI& firstDeclReturn,
      CPPTypeRefI& firstTypeReturn,
      CPPScopeRefI& containingScopeReturn,
      int &declCount,
      bool searchStack = true
   ) const;

   //
   // Class declaration/definition
   // 
   bool
   DeclareClassType(
      const JLStr& name,
      CPPConst::TypeSpecifier ts,
      CPPScopeRefI targetScope,
      const ANTLRTokenPtr& errTok,
      CPPTypeRefI& typeReturn,
      CPPDeclRefI& declReturn,
      JLStr& classNameReturn
   );

   void 
   DefineClassType(
      const JLStr& name, 
      CPPConst::TypeSpecifier ts,
      CPPScopeRefI targetScope,
      const ANTLRTokenPtr& errTok,
      CPPTypeRefI& typeReturn,
      CPPDeclRefI& declReturn,
      JLStr& classNameReturn
   );

   bool
   DeclareClassTemplateType(
      const JLStr& name,
      CPPConst::TypeSpecifier ts, 
      CPPScopeRefI targetScope,
      TemplateParameterListRefI templateParameters,
      TemplateArgumentListRefI specializationArgList,
      const ANTLRTokenPtr& errTok,
      CPPTypeRefI& classTypeReturn,
      CPPTypeRefI& templateTypeReturn,
      CPPDeclRefI& templateDeclReturn, 
      JLStr& templateNameReturn
   );

   bool
   DefineClassTemplateType(
      const JLStr& name,
      CPPConst::TypeSpecifier ts, 
      CPPScopeRefI targetScope,
      TemplateParameterListRefI templateParameters,
      TemplateArgumentListRefI specializationArgList,
      const ANTLRTokenPtr& errTok,
      CPPTypeRefI& classTypeReturn,
      CPPTypeRefI& templateTypeReturn,
      CPPDeclRefI& templateDeclReturn, 
      JLStr& templateNameReturn
   );

   void 
   AddBaseToClass(
      CPPTypeRefI classType,
      CPPTypeRefI baseClassType,
      CPPConst::BaseSpecifier baseSpecifier
   );


   //
   // Enum declaration/definition
   // 
   bool 
   DeclareEnumType(
      const JLStr& name, 
      CPPScopeRefI targetScope,
      const ANTLRTokenPtr& errTok,
      CPPTypeRefI& typeReturn,
      CPPDeclRefI& declReturn,
      JLStr& enumNameReturn
   );

   void 
   DefineEnumType(
      const JLStr& name, 
      CPPScopeRefI targetScope,
      const ANTLRTokenPtr& errTok,
      CPPTypeRefI& typeReturn,
      CPPDeclRefI& declReturn,
      JLStr& enumNameReturn
   );

   //
   // object/type declaration/definition
   // 
   CPPDeclRefI
   DeclareObjectOrTypedef(
      const JLStr& name,
      CPPTypeRefI type,
      CPPConst::StorageClass sc,
      CPPConst::DeclModifier dm,
      CPPExprRefI initExpr,
      CPPScopeRefI scope,
      StrListRefI argNameList
   );

   bool 
   DeclareFunctionTemplate(
      const JLStr& name,
      CPPTypeRefI type,
      CPPConst::StorageClass sc,
      CPPConst::DeclModifier dm,
      CPPExprRefI initExpr,
      CPPScopeRefI scope,
      StrListRefI argNameList,
      TemplateParameterListRefI templateParameters,
      const ANTLRTokenPtr& errTok,
      CPPTypeRefI& templateTypeReturn,
      CPPDeclRefI& templateDeclReturn,
      JLStr& templateNameReturn
   );

   //
   // Lokoup of id expression 
   //
   CPPExprRefI 
   LookupExprId(
      const JLStr& name,
      const ANTLRTokenPtr& errTok
   );

   //
   // Templates
   //
   void DeclareTemplateTypeParameter(
      const JLStr& name,
      TokenListRefI tokens,
      short parameterIdx,
      short templateLevel,
      TemplateParameterRefI& templateParmReturn
   );

   void DeclareTemplateValueParameter(
      const JLStr& name,
      TokenListRefI tokens,
      short parameterIdx,
      short templateLevel,
      CPPTypeRefI parameterType,
      TemplateParameterRefI& templateParmReturn
   );

   void 
   SetTokensInTemplateDeclaration(
      CPPTypeRefI templateType,
      TemplateParameterListRefI templateParameters,
      TokenListRefI tokens
   );

   bool 
   LookupSpecialization(
      const JLStr& name, 
      TemplateArgumentListRefI argList, 
      const ANTLRTokenPtr& errTok,
      JLStr& specializationNameReturn,
      CPPTypeRefI& specializationTypeReturn
   );

   JLStr FormatClassSpecializationName(
      const JLStr& name,
      TemplateArgumentListRefI specializationArgList
   );

   JLStr FormatClassSpecializationName(
      const JLStr& name,
      TemplateParameterListRefI templateParameterList
   );

   JLStr FormatFunctionSpecializationName(
      const JLStr& name,
      TemplateParameterListRefI templateParameterList,
      TemplateArgumentListRefI specializationArgList
   );

   //
   // Template miscellany
   //
   bool 
   TypeUsesInnerTemplateParms(
      CPPTypeRefI type
   );

   //
   // Diagnostic output
   //
   void
   DumpScopes() const;

   //
   // Misc
   // 
   JLStr 
   GetUniqueName() const;

   bool
   IsGeneratedName(
      const JLStr& name
   ) const;

   static bool NameIsDecoratedName(
      const JLStr& name, 
      const JLStr& decoratedName
   );

   //
   // Methods to manage template nesting level
   //
   virtual void IncreaseTemplateLevel() { templateLevel++; }
   virtual void DecreaseTemplateLevel()  { templateLevel--; }
   virtual int GetTemplateLevel() const { return templateLevel; }

private:
   // Pointer to parser
   CPPParserSym* parser;

   //
   // Helper functions for declarations
   // 
   void
   NewClassType(
      const JLStr& name,
      const JLStr& assumedName,
      CPPConst::TypeSpecifier ts,
      CPPScopeRefI& targetScope,
      CPPTypeRefI& typeReturn,
      CPPDeclRefI& declReturn
   );

   void
   NewEnumType(
      const JLStr& name,
      CPPScopeRefI& targetScope,
      CPPTypeRefI& typeReturn,
      CPPDeclRefI& declReturn
   );


   //
   // Helper function for type modifier lookup
   //
   static CPPTypeRefI
   CPPSymTable::GetOrUseTypeModifier(
      CPPTypeRefI inputType,
      CPPTypeModifier* newModifier
   );

   //
   // Helper function for scope lookup
   //
   bool 
   LookupDeclPrivate(
      const JLStr& name,
      CPPDeclRefI& declReturn,
      CPPTypeRefI& typeReturn,
      CPPScopeRefI& containingScopeReturn,
      bool searchStack,
      bool (*predicate)(CPPDeclRefI, CPPScopeRefI),
      bool (*hidingPredicate)(CPPDeclRefI, CPPScopeRefI),
      int& declCount
   ) const;

   //
   // Helper methods for template specialization
   //
   bool 
   LookupClassSpecialization(
      const JLStr& name, 
      CPPScopeRefI containingScope,
      TemplateArgumentListRefI argList, 
      const ANTLRTokenPtr& errTok,
      JLStr& specializationNameReturn,
      CPPTypeRefI& specializationTypeReturn
   );

   bool 
   LookupFunctionSpecialization(
      const JLStr& name, 
      CPPScopeRefI containingScope,
      TemplateArgumentListRefI argList, 
      const ANTLRTokenPtr& errTok,
      JLStr& specializationNameReturn,
      CPPTypeRefI& specializationTypeReturn
   );

   bool ChooseClassPartialSpecialization(
      const JLStr& className,
      CPPScopeRefI containingScope,
      TemplateArgumentListRefI templateArgumentList, 
      CPPTypeRefI& partialSpecializationTypeReturn,
      CPPDeclRefI& partialSpecializationDeclReturn,
      TemplateParameterListRefI& partialSpecializationParameterListReturn
   );

   CPPDeclRefI GetClassPrimaryTemplate(
      const JLStr& className,
      CPPScopeRefI containingScope
   );

   TemplateArgumentListRefI
   FunctionTypeToTemplateArgList(
      CPPTypeRefI functionType,
      const ANTLRTokenPtr& errTok
   );

   TemplateArgumentListRefI
   ParameterListToSpecializationArgList(
      TemplateParameterListRefI templateParameters
   );

   //
   // Symbol table Data
   //

   // Table of builtin types, keyed by CPPParser:TypeSpecifier
   CPPBuiltinTypeTableRefI builtinTypes;
   // Global scope (root of the scope tree)
   CPPScopeRefI globalScope;
   // Current scope stack
   CPPScopeStackRefI scopeStack;
   // Empty scope for use wherever it won't be changed
   CPPScopeRefI emptyScope;
   
   //
   // Packing stack
   //
   DECLARE_INCOMPLETE(PackingStack);

   static PackingStackRefI
   NewPackingStack();

   PackingStackRefI packingStack;
   enum { DefaultPacking = 8 };

   //
   // Template placeholder management
   //

   // types
   DECLARE_INCOMPLETE(TemplatePlaceholderTypeMap);
   TemplatePlaceholderTypeMapRefI templatePlaceholderTypeMap;
   TemplatePlaceholderTypeMapRefI MakeTemplatePlaceholderTypeMap();
   CPPTypeRefI GetTemplatePlaceholderType(
      short parameterIdx, 
      short templateLevel
   );

   // values
   DECLARE_INCOMPLETE(TemplatePlaceholderValueMap);
   TemplatePlaceholderValueMapRefI templatePlaceholderValueMap;
   TemplatePlaceholderValueMapRefI MakeTemplatePlaceholderValueMap();
   CPPExprRefI GetTemplatePlaceholderValue(
      short parameterIdx, 
      short templateLevel
   );

   //
   // Private Debugging helpers
   //
   void DumpDerivedTypes(CPPTypeRefI type, int level) const;
   void DumpScope(CPPScopeRefI scope, int level) const;

   //
   // Error-reporting
   // 
   void ReportError(const JLStr& errMsg, const ANTLRTokenPtr& token);

   //
   // Substitution of declaration names
   // This is used during template expansion, to cause declarations
   // made during the expansion to have the specialization name instead
   // of the unspecialized name.
   //
   typedef pair<JLStr, JLStr> DeclNameSubst;
   DeclNameSubst declNameSubst;

   //
   // Tracking of template nesting level
   // 
   int templateLevel;
};

#endif
