// CPPSymTableClass.cpp: class type manipulation
#include "CPPSymTable.h"
#include "CPPScopeNormal.h"
#include "CPPTypeClass.h"
#include "CPPDecl.h"
#include "CPPTypeTemplated.h"
#include "TemplateParameterList.h"
#include "TemplateArgumentList.h"
#include "JLCast.h"

/////////////////////////////////////////////////////////////////////
// NewClassType: Create a new, unique class type and enter it
// into the current scope.
// 
// Inputs:
//    const JLStr&             name          The name of the class.
//    const JLStr&             assumedName   The name of the class as
//                                           found by the substitution of
//                                           declaration names for template
//                                           exmpansion.
//    CPPConst::TypeSpecifier  ts            struct/class/union
//    CPPScopeRefI&            targetScope   The scope in which the class
//                                           will be created.
// Outputs:
//    CPPTypeRefI&             typeReturn    The new class type
//    CPPDeclRefI&             declReturn    The new class type declaration
// Return: none
//
// Note: Classes *defined* in temporary scopes will have their
// declarations added to the current scope.  Classes *declared*
// in temporary scopes will have their declarations added to the
// nearest enclosing global scope.  For example:
//    void foo() {
//       class T* x;  // class declared in local scope, T in global scope
//       class T{};   // "T" defined in local scope
//       T y;         // Uses locally-defined "T"
//    }
//    T* y;           // OK because T already declared in global scope
//    T z;            // Error: T not defined in global scope.
/////////////////////////////////////////////////////////////////////
void
CPPSymTable::NewClassType(
   const JLStr& name,
   const JLStr& assumedName,
   CPPConst::TypeSpecifier ts,
   CPPScopeRefI& targetScope,
   CPPTypeRefI& typeReturn,
   CPPDeclRefI& declReturn
)
{
   // Create a scope for the class declarations
   CPPScopeRefI classScope = new CPPScopeNormal(
      CPPScope::ScopeClass, 
      assumedName, 
      (ts == CPPConst::tsCLASS) ? CPPConst::asPRIVATE : CPPConst::asPUBLIC
   );
   targetScope->AddScope(classScope);

   // Create the class type
   typeReturn = new CPPTypeClass(assumedName, ts, classScope, GetCurrentPacking());

   // Enter the declaration in the target scope
   declReturn = new CPPDeclType(
      assumedName,
      targetScope->GetCurrentAccess(),
      typeReturn, 
      false
   );
   targetScope->AddDecl(declReturn);

   // Enter the class type declaration in the class scope itself.
   // Do not use the substituted name; use the raw declaration name.
   {
      CPPDeclRefI tmpDecl = new CPPDeclType(
         name,
         CPPConst::asPUBLIC,
         typeReturn, 
         false
      );
      classScope->AddDecl(tmpDecl);
   }
}


/////////////////////////////////////////////////////////////////////
// DeclareClassType: Declare an class type.  OK if class 
// exists, but the name cannot define a different type.
// 
// Inputs:
//    const JLStr&             name    The name of the class.
//    CPPConst::TypeSpecifier  ts      struct/class/union
//    CPPScopeRefI             targetScope
//                                     The scope in which the declaration
//                                     will be made, or nil to use the
//                                     current scope context.
//    const ANTLRTokenPtr      errTok  A token for error-handling
// Outputs: 
//    CPPTypeRefI&       typeReturn    The class type
//    CPPDeclRefI&       declReturn    The class type declaration
//    JLStr&          classNameReturn  The name of the class, which is
//                                     important if the input name
//                                     was empty or if the class is a
//                                     template specialization.
// Return: true if the declaration was successful, or if a compatible
// existing declaration was found.
/////////////////////////////////////////////////////////////////////
bool
CPPSymTable::DeclareClassType(
   const JLStr& name, 
   CPPConst::TypeSpecifier ts,
   CPPScopeRefI targetScope,
   const ANTLRTokenPtr& errTok,
   CPPTypeRefI& typeReturn,
   CPPDeclRefI& declReturn,
   JLStr& classNameReturn
)
{
   bool retval = true;

   classNameReturn = (name.length() == 0) ? GetUniqueName() : name;

   // If there is a declaration name substitution in effect 
   // (for template expansion), then use the substituted name 
   // for the class entry.
   if (declNameSubst.first == classNameReturn)
   {
      classNameReturn = declNameSubst.second;
   }

   bool foundType = false;
   CPPScopeRefI containingScope;

   if (targetScope.get() == 0)
   {
      // Special consideration for declaration made in the current scope,
      // because it may find and return a declaration elsewhere on the scope stack.
      targetScope = GetCurrentScope();
      foundType = LookupType(classNameReturn, declReturn, typeReturn, containingScope);
   } else {
      // target scope was specified
      // Check for the declaration already existing in the target scope.
      declReturn = targetScope->FindTypeDeclByName(classNameReturn);
      if (declReturn.get() != 0)
      {
         foundType = true;
         typeReturn = declReturn->GetType();
         containingScope = targetScope;
      }
   }

   bool foundClass = 
      foundType &&
      !declReturn->IsTypedef() &&
      typeReturn->GetFlavor() == CPPType::Class;

   if (foundType && containingScope == targetScope && !foundClass)
   {
      // The type was found and it's not a class
      ReportError(
         JLStr("\"") + classNameReturn + "\" is already defined as something other than a class/struct/union",
         errTok
      );
      // Return the found type, for lack of anything better to do
      retval = false;
   } else if (foundType && foundClass)
   {
      // The class was found -- use the found type
   } else {
      // Declare the class in the appropriate scope
      NewClassType(name, classNameReturn, ts, targetScope, typeReturn, declReturn);
   }

   return retval;
}

/////////////////////////////////////////////////////////////////////
// DefineClassType: Define a class type.  Invalid if the
// class is already defined, but OK if it is declared.
// 
// Inputs:
//    const JLStr&             name    The name of the class (may be empty)
//    CPPConst::TypeSpecifier  ts      struct/class/union
//    CPPScopeRefI             targetScope
//                                     The scope in whcih the
//                                     definition will be made
//    const ANTLRTokenPtr      errTok  A token for error-handling
// Outputs: 
//    CPPTypeRefI&    typeReturn       The class type
//    CPPDeclRefI&    declReturn       The class declaration entry
//    JLStr&          classNameReturn  The name of the class, which is
//                                     only important if the input name
//                                     was empty.
// Return: none
/////////////////////////////////////////////////////////////////////
void
CPPSymTable::DefineClassType(
   const JLStr& name, 
   CPPConst::TypeSpecifier ts,
   CPPScopeRefI targetScope,
   const ANTLRTokenPtr& errTok,
   CPPTypeRefI& typeReturn,
   CPPDeclRefI& declReturn,
   JLStr& classNameReturn
)
{
   if (
      DeclareClassType(
         name, ts, targetScope, errTok, typeReturn, declReturn, classNameReturn
      )
   )
   {
      // make sure it was not already defined
      if (typeReturn->GetComplete())
      {
         // Error -- class already defined
         ReportError(
            JLStr("class \"") + name + "\" has already been defined ",
            errTok
         );
      } else {
         // Define it
         typeReturn->SetComplete();
      }
   }
}


/////////////////////////////////////////////////////////////////////
// AddBaseToClass: Add a base class to an already-defined class type.
// 
// Inputs:
//    CPPTypeRefI             classType        The class type to modify
//    CPPTypeRefI             baseClass        The base class to add
//    CPPConst::BaseSpecifier baseSpecifier    Base class specifiers (virtual,
//                                             public, private, protected).
// Outputs: none
// Return: none
/////////////////////////////////////////////////////////////////////
// virtual 
void 
CPPSymTable::AddBaseToClass(
   CPPTypeRefI classType,
   CPPTypeRefI baseClass,
   CPPConst::BaseSpecifier baseSpecifier
)
{
   assert(classType->GetFlavor() == CPPType::Class);
   STATIC_CAST(CPPTypeClass*, classType.get())->AddBaseClass(baseClass, baseSpecifier);
}



/////////////////////////////////////////////////////////////////////
// DeclareClassTemplateType: Declare a class template.  OK if class 
// template exists, but the name cannot define anything conflicting.
// 
// Inputs:
//    const JLStr&               name    The name of the class.
//    CPPConst::TypeSpecifier    ts      struct/class/union
//    CPPScopeRefI               targetScope
//                                       The scope in which the declaration
//                                       will be made.  Cannot be nil.
//    TemplateParameterListRefI  templateParameters
//                                       The template parameters
//    TemplateArgumentListRefI   specializationArgList
//                                       The specialization template arguments
//    const ANTLRTokenPtr        errTok  A token for error-handling
// Outputs: 
//    CPPTypeRefI&               classTypeReturn    
//                                       The class type (which will have been
//                                       attached to a template modifier).
//    CPPTypeRefI&               templateTypeReturn
//                                       The template modifier type, or the
//                                       class type for a fully specialized
//                                       template declaration.
//    CPPDeclRefI&               templateDeclReturn    
//                                       The template declaration
//    JLStr&                     templateNameReturn  
//                                       The decorated name of the template.
// Return: none
/////////////////////////////////////////////////////////////////////
bool
CPPSymTable::DeclareClassTemplateType(
   const JLStr& name,
   CPPConst::TypeSpecifier ts, 
   CPPScopeRefI targetScope,
   TemplateParameterListRefI templateParameters,
   TemplateArgumentListRefI specializationArgList,
   const ANTLRTokenPtr& errTok,
   CPPTypeRefI& classTypeReturn,
   CPPTypeRefI& templateTypeReturn,
   CPPDeclRefI& templateDeclReturn, 
   JLStr& templateNameReturn
)
{
   // Get the name of the specialization
   if (specializationArgList->empty())
   {
      templateNameReturn = FormatClassSpecializationName(name, templateParameters);
   } else {
      templateNameReturn = FormatClassSpecializationName(name, specializationArgList);
   }

   // Is this a full specialization?
   bool isFullSpecialization = templateParameters->empty();

   // Is there already a declaration under the specialized name?
   CPPScope::DeclIterator declIter;
   templateDeclReturn = targetScope->FindDeclByName(templateNameReturn, declIter);
   if (templateDeclReturn.get() != 0)
   {
      templateTypeReturn = templateDeclReturn->GetType();
      if (templateDeclReturn->IsType())
      {
         // Preexisting class template declaration
         classTypeReturn = templateTypeReturn->GetBaseType();
         return true;
      } else {
         // Conflicting function template declaration
         ReportError("Template already declared as a function", errTok);
         return false;
      }
   }

   // Look for conflicting declaration under the non-specialized name
   for (
      templateDeclReturn = targetScope->FindDeclByName(name, declIter);
      templateDeclReturn.get() != 0;
      templateDeclReturn = targetScope->NextDecl(declIter)
   )
   {
      if (templateDeclReturn->GetType()->GetFlavor() == CPPType::Templated)
      {
         if (!templateDeclReturn->IsType())
         {
            // already defined as a function template
            ReportError("'" + name + "' is already defined as a function template", errTok);
            // Return the existing entry to have something usable
            templateTypeReturn = templateDeclReturn->GetType();
            return false;
         }
      } else {
         // already defined as a non-template
         ReportError("'" + name + "' is already defined as a non-template", errTok);
         // Return the existing entry to have something usable
         templateTypeReturn = templateDeclReturn->GetType();
         return false;
      }
   }

   // Create a scope for the class member declarations
   CPPScopeRefI classScope = new CPPScopeNormal(
      CPPScope::ScopeClass, 
      templateNameReturn,
      (ts == CPPConst::tsCLASS) ? CPPConst::asPRIVATE : CPPConst::asPUBLIC
   );
   targetScope->AddScope(classScope);

   // Create the class type
   classTypeReturn = new CPPTypeClass(
      templateNameReturn, 
      ts, 
      classScope, 
      GetCurrentPacking()
   );

   if (isFullSpecialization)
   {
      // No template modifier needed
      templateTypeReturn = classTypeReturn;
   } else {
      bool isPrimary = false;
      if (specializationArgList->empty())
      {
         // If the specialization arg list was empty, then create one that
         // mirrors the formal parameters.
         isPrimary = true;
         specializationArgList = ParameterListToSpecializationArgList(templateParameters);
      }

      // Create the template modifier type
      templateTypeReturn = new CPPTypeTemplated(
         classTypeReturn,
         TokenListRefI(),
         templateParameters,
         specializationArgList,
         isPrimary,
         true      // addToBase
      );

      // Enter the template declaration in the target scope under the undecorated name
      templateDeclReturn = new CPPDeclType(
         name,
         targetScope->GetCurrentAccess(),
         templateTypeReturn, 
         false
      );
      targetScope->AddDecl(templateDeclReturn);
   }

   // Enter the templated or untemplated class declaration under the decorated name
   templateDeclReturn = new CPPDeclType(
      templateNameReturn,
      targetScope->GetCurrentAccess(),
      templateTypeReturn, 
      false
   );
   targetScope->AddDecl(templateDeclReturn);

   // Enter the class type declaration in the class scope itself.
   // Do not use the template name; use the raw declaration name.
   {
      CPPDeclRefI tmpDecl = new CPPDeclType(
         name,
         CPPConst::asPUBLIC,
         classTypeReturn, 
         false
      );
      classScope->AddDecl(tmpDecl);
   }

   return true;
}

/////////////////////////////////////////////////////////////////////
// DefineClassTemplateType: Define a class template.  OK if class 
// template exists, but the class must be incomplete and the name 
// cannot define anything else.
// 
// Inputs:
//    const JLStr&               name    The name of the class.
//    CPPConst::TypeSpecifier    ts      struct/class/union
//    CPPScopeRefI               targetScope
//                                       The scope in which the declaration
//                                       will be made.  Cannot be nil.
//    TemplateParameterListRefI  templateParameters
//                                       The template parameters
//    TemplateArgumentListRefI   specializationArgList
//                                       The specialization template arguments
//    const ANTLRTokenPtr        errTok  A token for error-handling
// Outputs: 
//    CPPTypeRefI&               classTypeReturn    
//                                       The class type (which will have been
//                                       attached to a template modifier).
//    CPPTypeRefI&               templateTypeReturn
//                                       The template modifier type
//    CPPDeclRefI&               templateDeclReturn    
//                                       The template declaration
//    JLStr&                     templateNameReturn  
//                                       The decorated name of the template.
// Return: none
/////////////////////////////////////////////////////////////////////
bool
CPPSymTable::DefineClassTemplateType(
   const JLStr& name,
   CPPConst::TypeSpecifier ts, 
   CPPScopeRefI targetScope,
   TemplateParameterListRefI templateParameters,
   TemplateArgumentListRefI specializationArgList,
   const ANTLRTokenPtr& errTok,
   CPPTypeRefI& classTypeReturn,
   CPPTypeRefI& templateTypeReturn,
   CPPDeclRefI& templateDeclReturn, 
   JLStr& templateNameReturn
)
{
   if (
      DeclareClassTemplateType(
         name,
         ts, 
         targetScope,
         templateParameters,
         specializationArgList,
         errTok,
         classTypeReturn,
         templateTypeReturn,
         templateDeclReturn, 
         templateNameReturn
      )
   )
   {
      if (classTypeReturn->GetComplete())
      {
         ReportError("'" + name + "' already has a definition", errTok);
      } else {
         classTypeReturn->SetComplete();
      }
      return true;
   } else {
      return false;
   }
}


/////////////////////////////////////////////////////////////////////
// ParameterListToSpecializationArgList: Given a template formal
// parameter list, construct the equivalent specialization argument list.
// 
// Inputs:
//    TemplateParameterListRefI  templateParameters  The list of formal parameters
// Outputs: none
// Return:
//    TemplateArgumentListRefI   The equivalent argument list.
/////////////////////////////////////////////////////////////////////
TemplateArgumentListRefI
CPPSymTable::ParameterListToSpecializationArgList(
   TemplateParameterListRefI templateParameters
)
{
   TemplateArgumentListRefI argList = new TemplateArgumentList;
   for (
      TemplateParameterList::iterator iter = templateParameters->begin();
      iter != templateParameters->end();
      iter++
   )
   {
      if ( (*iter)->GetFlavor() == TemplateParameter::Type)
      {
         argList->push_back(new TemplateArgumentType((*iter)->GetType()));
      } else {
         argList->push_back(
            new TemplateArgumentValue(
               STATIC_CAST(TemplateParameterValue*, (*iter).get())->GetValue()
            )
         );
      }
   }
   return argList;
}
