// CPPSymTableDebug.cpp: Debugging methods for the parser

#include "CPPSymTable.h"
#include "CPPType.h"
#include "CPPScope.h"
#include "CPPDecl.h"
#include "CPPExpr.h"
#include "CPPTypeContainers.h"
#include <iostream>

///////////////////////////////////////////////////////////////////////
// DumpDerivedTypes:  Dump this type and the types derived from it.
//
// Inputs:
//    const CPPTypeRefI    type        The type to dump
//    int               level       The indentation level
///////////////////////////////////////////////////////////////////////
void 
CPPSymTable::DumpDerivedTypes(
   CPPTypeRefI type,
   int level
) const
{
   // Dump the type itself
   cout << JLStr(level*3, ' ');
   cout << type->FormatName("");
   cout << endl;

   // Dump the derived types
   CPPTypePtrSetCRefI derivedTypes = type->GetDerivedTypes();
   for (
      CPPTypePtrSet::const_iterator iter = derivedTypes->begin();
      iter != derivedTypes->end();
      iter++
   )
   {
      DumpDerivedTypes(*iter, level+1);
   }
}

///////////////////////////////////////////////////////////////////////
// DumpScopes:  Dump the scope tree to cout
///////////////////////////////////////////////////////////////////////
void
CPPSymTable::DumpScopes() const
{
   DumpScope(globalScope, 1);
}

///////////////////////////////////////////////////////////////////////
// DumpScope:  Dump a scope and its children to cout
///////////////////////////////////////////////////////////////////////
void 
CPPSymTable::DumpScope(
   CPPScopeRefI scope,
   int level
) const
{
   // Dump all of the declarations in this scope
   CPPScope::AllDeclIterator declIter;
   for (
       CPPDeclRefI decl = scope->FindAllDecls(declIter);
       decl.get() != 0;
       decl = scope->NextAllDecl(declIter)
   )
   {
      cout << JLStr(level*3, ' ');
      if (decl->IsType())
      {
         if (decl->IsTypedef())
         {
            cout << "typedef " << decl->GetType()->FormatName(decl->GetName());
         } else {
            cout << "type " << decl->GetType()->FormatName(decl->GetName());
         }
      } else {
         cout << decl->GetType()->FormatName(
            decl->GetName(),
            "",
            decl->GetArgNameList()
         );
      }
      if (decl->GetInitializer().get() != 0)
      {
         cout << " = ";
         cout << decl->GetInitializer()->Format();
      }
      cout << endl;
   }

   cout << endl;
   // Dump all of the sub-scopes in this scope
   CPPScope::ScopeIterator scopeIter;
   for (
       CPPScopeRefI tmpScope = scope->FindAllScopes(scopeIter);
       tmpScope.get() != 0;
       tmpScope = scope->NextScope(scopeIter)
   ) 
   {
      cout << JLStr(level*3, ' ');
      cout << "Sub-Scope ";
      cout << tmpScope->GetName() << ":" << endl;
      DumpScope(tmpScope, level+1);
   }
}

