// CPPSymTableDecl.cpp: generic declaration methods

#include "CPPSymTable.h"
#include "CPPScope.h"
#include "CPPDecl.h"
#include "CPPType.h"
#include "CPPTypeFunction.h"
#include "CPPTypeTemplated.h"
#include "TemplateParameterList.h"
#include "TemplateArgumentList.h"

/////////////////////////////////////////////////////////////////////
// DeclareObjectOrTypedef: Enter a declaration into the given scope.
// This handles typedefs, objects, and functions.  Multiple 
// declarations under the same name are not checked.
// 
// The access to the declaration will be set to the
// scope's current default access.
//
// Inputs:
//    const JLStr&    name         The symbol name
//    CPPTypeRefI     type         The type that the symbol defines
//    StorageClass    sc           The storage class
//    DeclModifier    dm           The declaration modifiers
//    CPPExprRefI     initExpr     The initializer of the declaration,
//                                 or a zero-ref if there is none
//    CPPScopeRefI    scope        The scope in which the declaration 
//                                 is to be made.
//    StrListRefI     argNameList  A reference to the argument name list
//                                 for function declarations, or zero
//                                 for non-function declarations.
// Outputs: none
// Return: 
//    CPPDeclRefI     The new object or type declaration
/////////////////////////////////////////////////////////////////////
// virtual 
CPPDeclRefI
CPPSymTable::DeclareObjectOrTypedef(
   const JLStr& name,
   CPPTypeRefI type,
   CPPConst::StorageClass sc,
   CPPConst::DeclModifier dm,
   CPPExprRefI initExpr,
   CPPScopeRefI scope,
   StrListRefI argNameList
)
{
   CPPDeclRefI decl;
   if ((sc & CPPConst::scTYPEDEF) == CPPConst::scTYPEDEF)
   {
      // typedef
      // TODO: disallow initializer?  What about template args?
      // TODO: exclude incompatible flags
      decl = new CPPDeclType(
         name,
         scope->GetCurrentAccess(), 
         type, 
         true
      );
      scope->AddDecl(decl);
   } else {
      // object
      // TODO: exclude incompatible flags
      decl = new CPPDeclObject(
         name,
         scope->GetCurrentAccess(), 
         type, 
         sc, 
         dm, 
         initExpr,
         argNameList
      );
      scope->AddDecl(decl);
   }
   return decl;
}


/////////////////////////////////////////////////////////////////////
// DeclareFunctionTemplate: Enter a declaration into the given scope.
// This is used for declarations of template functions like:
//    template <class T> foo(T*);
// 
// The access to the declaration will be set to the
// scope's current default access.
//
// Inputs:
//    const JLStr&        name         The function name
//    CPPTypeRefI         type         The type that the symbol defines
//    StorageClass        sc           The storage class
//    DeclModifier        dm           The declaration modifiers
//    CPPExprRefI         initExpr     The initializer of the declaration,
//                                     or a zero-ref if there is none
//    CPPScopeRefI        targetScope  The scope in which the declaration 
//                                     is to be made.
//    StrListRefI         argNameList  A reference to the argument name list
//                                     for function declarations, or zero
//                                     for non-function declarations.
//    TemplateParameterListRefI templateParameters
//                                     The list of formal template parameters.
//    const ANTLRTokenPtr&  errTok     A token used for error reporting.
// Outputs:
//    CPPTypeRefI&    templateTypeReturn  The type of the declaration.
//    CPPDeclRefI&    templateDeclReturn  The declaration
//    JLStr&          templateNameReturn  The name, which may be different
//                                        than the input name if the declaration
// Outputs: none
// Return: 
//    bool    true if the declaration succeeded, false o/w
/////////////////////////////////////////////////////////////////////
bool 
CPPSymTable::DeclareFunctionTemplate(
   const JLStr& name,
   CPPTypeRefI type,
   CPPConst::StorageClass sc,
   CPPConst::DeclModifier dm,
   CPPExprRefI initExpr,
   CPPScopeRefI targetScope,
   StrListRefI argNameList,
   TemplateParameterListRefI templateParameters,
   const ANTLRTokenPtr& errTok,
   CPPTypeRefI& templateTypeReturn,
   CPPDeclRefI& templateDeclReturn,
   JLStr& templateNameReturn
)
{
   // Check that the declaration is a function
   CPPTypeRefI functionType = type->StripCVQualifiers();
   if (!functionType->IsFunctionType())
   {
      ReportError("template declaration must be a class or function", errTok);
      return false;
   }

   // Get the specialization argument list implied by the function argument list
   TemplateArgumentListRefI specializationArgList = 
      FunctionTypeToTemplateArgList(functionType, errTok);

   // Get the name of the specialization
   templateNameReturn = FormatFunctionSpecializationName(
      name, 
      templateParameters,
      specializationArgList
   );

   // Is this a full specialization?
   bool isFullSpecialization = templateParameters->empty();

   // Is there already a declaration under the specialized name?
   CPPScope::DeclIterator declIter;
   templateDeclReturn = targetScope->FindDeclByName(templateNameReturn, declIter);
   if (templateDeclReturn.get() != 0)
   {
      templateTypeReturn = templateDeclReturn->GetType();
      if (templateDeclReturn->IsType())
      {
         // Conflicting class template declaration
         ReportError("Template already declared as a class", errTok);
         return false;
      } else {
         // Preexisting function template declaration
         return true;
      }
   }

   // Look for a conflicting declaration under the non-specialized name
   for (
      templateDeclReturn = targetScope->FindDeclByName(name, declIter);
      templateDeclReturn.get() != 0;
      templateDeclReturn = targetScope->NextDecl(declIter)
   )
   {
      templateTypeReturn = templateDeclReturn->GetType();
      if (templateTypeReturn->GetFlavor() == CPPType::Templated)
      {
         if (templateDeclReturn->IsType())
         {
            // already defined as a class template
            ReportError("'" + name + "' is already defined as a class template", errTok);
            // Return the existing entry to have something usable
            templateTypeReturn = templateDeclReturn->GetType();
            return false;
         }
      } else if (!templateTypeReturn->IsFunctionType()) {
         // Already defined as a non-template, and it's not a function
         ReportError("'" + name + "' is already defined as a non-template", errTok);
         // Return the existing entry to have something usable
         templateTypeReturn = templateDeclReturn->GetType();
         return false;
      }
   }

   if (isFullSpecialization)
   {
      // No template modifier needed
      templateTypeReturn = type;
   } else {
      // Create the template modifier type
      templateTypeReturn = new CPPTypeTemplated(
         type,
         TokenListRefI(),
         templateParameters,
         specializationArgList,
         false,    // isPrimary (no function template is primary)
         true      // addToBase
      );

      // Enter the template declaration in the target scope under the undecorated name
      templateDeclReturn = new CPPDeclObject(
         name,
         targetScope->GetCurrentAccess(),
         templateTypeReturn, 
         sc,
         dm,
         initExpr,
         argNameList
      );
      targetScope->AddDecl(templateDeclReturn);
   }

   // Enter the template or function declaration under the decorated name
   templateDeclReturn = new CPPDeclObject(
      templateNameReturn,
      targetScope->GetCurrentAccess(),
      templateTypeReturn, 
      sc,
      dm,
      initExpr,
      argNameList
   );
   targetScope->AddDecl(templateDeclReturn);

   return true;
}


/////////////////////////////////////////////////////////////////////
// FunctionTypeToTemplateArgList: Given a function type, convert
// the argument list of the function to the equivalent template
// argument list, for the purpose of creating a template specialization
// entry.
//
// Inputs:
//    CPPTypeRefI           functionType   The type, which must be a function
//    const ANTLRTokenPtr&  errTok         A token used for error reporting.
// Outputs: none
// Return:
//    TemplateArgumentListRefI     The equivalent list of template arguments.
// 
// Note: All template arguments will be type arguments.
/////////////////////////////////////////////////////////////////////
TemplateArgumentListRefI
CPPSymTable::FunctionTypeToTemplateArgList(
   CPPTypeRefI functionType,
   const ANTLRTokenPtr& errTok
)
{
   assert(functionType->IsFunctionType());
   TemplateArgumentListRefI templateArgList = new TemplateArgumentList;
   do {
      switch (functionType->GetFlavor())
      {
         case CPPType::FunctionArg:
         {
            CPPTypeRefI argType = 
               STATIC_CAST(CPPTypeFunctionArg*, functionType.get())->GetArgType();
            // function type tree gets args in reverse order
            templateArgList->push_front(new TemplateArgumentType(argType));
            break;
         }
         case CPPType::FunctionReturning:
            // ignore return type
            break;
         case CPPType::FunctionEllipses:
            // TODO: how to represent this
            break;
         case CPPType::CVQualified:
            break;
         default:
            assert(0);
            break;
      }
      functionType = functionType->GetBaseType();
   } while (functionType->IsFunctionType());

   return templateArgList;
}

