// CPPSymTableLookup.cpp: Scope and declaration table lookup for CPPSymTable.

#include "CPPSymTable.h"
#include "CPPScope.h"
#include "CPPScopeContainers.h"
#include "CPPDecl.h"
#include "CPPDeclContainers.h"
#include "CPPExpr.h"
#include "CPPTypeClass.h"
#include "CPPTypeTemplatePlaceholder.h"

/////////////////////////////////////////////////////////////////////
// Predicates for classifying declarations
/////////////////////////////////////////////////////////////////////
static bool predIsClass(CPPDeclRefI decl, CPPScopeRefI) { 
   return 
      decl->IsType() && 
      decl->GetType()->GetFlavor() == CPPType::Class &&
      !decl->IsTypedef();
}

static bool predTrue(CPPDeclRefI decl, CPPScopeRefI) { return true; }

static bool predIsEnum(CPPDeclRefI decl, CPPScopeRefI) { 
   return 
      decl->IsType() && 
      decl->GetType()->GetFlavor() == CPPType::Enum &&
      !decl->IsTypedef();
}

static bool predIsObject(CPPDeclRefI decl, CPPScopeRefI) { return !decl->IsType(); }

static bool predIsType(CPPDeclRefI decl, CPPScopeRefI) { 
   return decl->IsType();
}

static bool predIsNotEnclosingClass(CPPDeclRefI decl, CPPScopeRefI scope) { 
   return 
      scope->GetFlavor() != CPPScope::ScopeClass ||
      !CPPSymTable::NameIsDecoratedName(decl->GetName(), scope->GetName());
}


/////////////////////////////////////////////////////////////////////
// LookupDeclPrivate: Find the first declaration and containing scope by name.
// This will search the scope stack for the desired name,
// and return the declaration and associated scope.
//
// Inputs:
//    const JLStr&   name              The name of the declaration
//    bool           searchStack       true to search the scope stack,
//                                     false to search just the top scope
//    bool (*)(...)  predicate         A predicate function that is true
//                                     when the declaration is the desired one
//    bool (*)(...)  hidingPredicate   A predicate function that is true
//                                     when a declaration will hide the desired one
//
// Outputs:
//    CPPDeclRefI&    declReturn             The declaration, if successful
//    CPPTypeRefI&    typeReturn             The type, if successful
//    CPPScopeRefI&   containingScopeReturn  The containing scope, if successful
//    int&            declCount              Set to the number of declartions under
//                                           that name, in the scope of the first 
//                                           declaration found.
//
// Return:
//    bool      true if a declaration of the right name was found and it
//              matched the predicate;  false otherwise
//
// Note: The scope stack is not searched past the first scope 
// containing a declaration of the right name, even if no 
// declarations match the predicate.
/////////////////////////////////////////////////////////////////////
bool 
CPPSymTable::LookupDeclPrivate(
   const JLStr& name,
   CPPDeclRefI& declReturn,
   CPPTypeRefI& typeReturn,
   CPPScopeRefI& containingScopeReturn,
   bool searchStack,
   bool (*predicate)(CPPDeclRefI, CPPScopeRefI),
   bool (*hidingPredicate)(CPPDeclRefI, CPPScopeRefI),
   int &declCount
) const
{
   declReturn = (CPPDecl*)0;
   declCount = 0;

   // Search the scope stack for the desired class declaration and class scope
   for (
      CPPScopeStack::iterator stackIter = scopeStack->begin();
      stackIter != scopeStack->end();
      stackIter++
   )
   {
      // Look for the declaration in this scope
      CPPScopeRefI scope = (*stackIter).GetScope();
      CPPScope::DeclIterator iter;
      bool foundHidingDecl = false;
      for (
          CPPDeclRefI tmpDecl = scope->FindDeclByName(name, iter);
          tmpDecl.get() != 0;
          tmpDecl = scope->NextDecl(iter)
      ) 
      {
         if (predicate(tmpDecl, scope))
         {
            declCount++;
            if (declReturn.get() == 0)
            {
               declReturn = tmpDecl;
               typeReturn = declReturn->GetType();
               containingScopeReturn = scope;
            }
         } else if (hidingPredicate(tmpDecl, scope))
         {
            foundHidingDecl = true;
         }
      }

      // Keep searching up the scope stack unless:
      //   1) We found a declaration (even one not matching the predicate), or
      //   2) The scope is a "resolved" scope, or
      //   3) searchStack is false
      if (declCount != 0 || foundHidingDecl || !searchStack || (*stackIter).GetResolved())
      {
         break;
      }
   }

   return declCount > 0;
}


/////////////////////////////////////////////////////////////////////
// LookupClass: Find a class' declaration and scope by name
// This will search the scope stack for the desired class name,
// and return it's declaration and associated scope.
//
// Inputs:
//    const JLStr&  name               The name of the class
//    bool          searchStack        true to search the scope stack,
//                                     false to search just the top scope
//
// Outputs:
//    CPPDeclRefI&    declReturn       The declaration, if successful
//    CPPTypeRefI&    typeReturn       The type, if successful
//    CPPScopeRefI&   containingScopeReturn
//                                     The containing scope. if successful
//
// Return:
//    bool      true if the name was found (successful), false otherwise
// 
// Note: This will not typedefs of classes.
/////////////////////////////////////////////////////////////////////
// virtual 
bool 
CPPSymTable::LookupClass(
   const JLStr& name,
   CPPDeclRefI& declReturn,
   CPPTypeRefI& typeReturn,
   CPPScopeRefI& containingScopeReturn,
   bool searchStack
) const
{
   int dummy;
   return LookupDeclPrivate(
         name,
         declReturn,
         typeReturn,
         containingScopeReturn,
         searchStack,
         predIsClass,
         predTrue,
         dummy
      );
}


/////////////////////////////////////////////////////////////////////
// LookupEnum: Find a enum declaration by name
// This will search the scope stack for the desired enum name,
// and return it's declaration and containing scope.
//
// Inputs:
//    const JLStr&  name               The name of the enum
//    bool          searchStack        true to search the scope stack,
//                                     false to search just the top scope
//
// Outputs:
//    CPPDeclRefI&    declReturn         The declaration, if successful
//    CPPTypeRefI&    typeReturn         The type, if successful
//    CPPScopeRefI&   containingScopeReturn
//                                    The containing scope. if successful
//
// Return:
//    bool      true if the name was found (successful), false otherwise
// 
// Note: This will not return typedefs of enums
/////////////////////////////////////////////////////////////////////
// virtual 
bool 
CPPSymTable::LookupEnum(
   const JLStr& name,
   CPPDeclRefI& declReturn,
   CPPTypeRefI& typeReturn,
   CPPScopeRefI& containingScopeReturn,
   bool searchStack
) const
{
   int dummy;
   return LookupDeclPrivate(
         name,
         declReturn,
         typeReturn,
         containingScopeReturn,
         searchStack,
         predIsEnum,
         predTrue,
         dummy
      );
}

/////////////////////////////////////////////////////////////////////
// LookupType: Find a type or typedef declaration and scope by name
// This will search the scope stack for the desired name,
// and return it's declaration and associated scope.
//
// Inputs:
//    const JLStr&  name               The name of the type
//    bool          searchStack        true to search the scope stack,
//                                     false to search just the top scope
//
// Outputs:
//    CPPDeclRefI&    declReturn         The declaration, if successful
//    CPPTypeRefI&    typeReturn         The type, if successful
//    CPPScopeRefI&   containingScopeReturn
//                                    The containing scope. if successful
//
// Return:
//    bool      true if the name was found (successful), false otherwise
/////////////////////////////////////////////////////////////////////
// virtual 
bool 
CPPSymTable::LookupType(
   const JLStr& name,
   CPPDeclRefI& declReturn,
   CPPTypeRefI& typeReturn,
   CPPScopeRefI& containingScopeReturn,
   bool searchStack
) const
{
   // Handle the special case of looking up a type in a template
   // placeholder scope, which should always produce a type.
   if (GetCurrentScope()->GetFlavor() == CPPScope::ScopeTemplate)
   {
      declReturn = GetCurrentScope()->FindTypeDeclByName(name);
      assert(declReturn.get() != 0);
      typeReturn = declReturn->GetType();
      containingScopeReturn = GetCurrentScope();
      return true;
   }

   int dummy;
   return LookupDeclPrivate(
         name,
         declReturn,
         typeReturn,
         containingScopeReturn,
         searchStack,
         predIsType,
         predTrue,
         dummy
      );
}


/////////////////////////////////////////////////////////////////////
// LookupScope: Find a scope by name
// This will search the scope stack for the desired name,
// and return the scope.
//
// Inputs:
//    const JLStr&  name               The name of the scope
//    bool          searchStack        true to search the scope stack,
//                                     false to search just the top scope
//
// Outputs:
//    CPPScopeRefI&   scopeReturn        The named scope, if found
//
// Return:
//    bool      true if the name was found (successful), false otherwise
/////////////////////////////////////////////////////////////////////
// virtual 
bool 
CPPSymTable::LookupScope(
   const JLStr& name,
   CPPScopeRefI& scopeReturn,
   bool searchStack
) const
{
   // Search the scope stack for the desired scope
   for (
      CPPScopeStack::iterator stackIter = scopeStack->begin();
      stackIter != scopeStack->end();
      stackIter++
   )
   {
      CPPScopeRefI tmpScope = (*stackIter).GetScope();
      // Look for the scope.
      scopeReturn = tmpScope->FindScope(name);
      if (scopeReturn.get() != 0)
      {
         return true;
      }

      // Failing that, look for a type/typedef of that name
      CPPDeclRefI tmpDecl = tmpScope->FindTypeDeclByName(name);

      if (tmpDecl.get() != 0)
      {
         CPPTypeRefI tmpType = tmpDecl->GetType()->StripTemplateModifiers();
         if (tmpType->GetFlavor() == CPPType::Class)
         {
            // If the name is a class or class template, 
            // then follow that to member scope.
            scopeReturn = STATIC_CAST(CPPTypeClass*, tmpType.get())->GetMemberScope();
            return true;
         }

         // Failing that, see if the name is a template placeholder type
         if (tmpType->GetFlavor() == CPPType::TemplatePlaceholder)
         {
            // If it is, then use the scope of the placeholder
            scopeReturn = STATIC_CAST(CPPTypeTemplatePlaceholder*, tmpType.get())->GetPlaceholderScope();
            return true;
         }

      }

      // Don't search past "resolved" scopes
      if (!searchStack || (*stackIter).GetResolved())
      {
         break;
      }
   }

   // scope was not found
   return false;
}



/////////////////////////////////////////////////////////////////////
// LookupObjectDeclaration: Find an object declaration given it's 
// name.  If there is more than one declaration under the name, 
// return the first one.
//
// Inputs:
//    const JLStr&  name              The name of the declaration
//    bool          searchStack       True if the scope stack is to be searched.
//
// Outputs:
//    CPPDeclRefI&  firstDeclReturn   Set to the first declaration found, if any.
//    CPPTypeRefI&  firstTypeReturn   Set to the type of the first decl found, if any.
//    CPPScopeRefI& containingScopeReturn
//                                    Set to the scope containing the declaration(s)
//    int&          declCount         Set to the number of decls under that name
//
// Return:
//    bool    true if a declaration was found
/////////////////////////////////////////////////////////////////////

bool
CPPSymTable::LookupObjectDeclaration(
   const JLStr& name,
   CPPDeclRefI& firstDeclReturn,
   CPPTypeRefI& firstTypeReturn,
   CPPScopeRefI& containingScopeReturn,
   int &declCount,
   bool searchStack
) const
{
   return LookupDeclPrivate(
      name, 
      firstDeclReturn, 
      firstTypeReturn, 
      containingScopeReturn, 
      searchStack,
      predIsObject,
      predTrue,
      declCount
   );
}


/////////////////////////////////////////////////////////////////////
// LookupAnyDeclaration: Find an object declaration given it's 
// name.  If there is more than one declaration under the name, 
// return the first one.
//
// Inputs:
//    const JLStr&  name             The name of the declaration
//    bool          searchStack      True if the scope stack is to be searched.
//
// Outputs:
//    CPPDeclRefI&  firstDeclReturn  The first declaration that was found.
//    CPPTypeRefI&  firstTypeReturn  The type of firstDeclReturn
//    CPPScopeRefI& scopeReturn      The scope in which the declaration was found
//    int&          declCount        The number of declarations under that name
//
// Return:
//    bool      true if a declaration was found, false o/w.
/////////////////////////////////////////////////////////////////////
bool 
CPPSymTable::LookupAnyDeclaration(
   const JLStr& name,
   CPPDeclRefI& firstDeclReturn,
   CPPTypeRefI& firstTypeReturn,
   CPPScopeRefI& scopeReturn,
   int &declCount,
   bool searchStack
) const
{
   return LookupDeclPrivate(
         name, 
         firstDeclReturn, 
         firstTypeReturn, 
         scopeReturn,
         searchStack,
         predTrue,
         predTrue,
         declCount
      );
}

/////////////////////////////////////////////////////////////////////
// LookupExprId: Lookup the declaration for an id expression, and
// construct an expression for that id.
//
// Inputs:
//    const JLStr&          name         Text of the id
//    const ANTLRTokenPtr&  errTok       Used for error reporting.
//
// Outputs: none
//
// Return:
//    CPPExprRefI    An expression representing the id.
/////////////////////////////////////////////////////////////////////
CPPExprRefI 
CPPSymTable::LookupExprId(
   const JLStr& name,
   const ANTLRTokenPtr& errTok
)
{
   CPPDeclRefI tmpDecl;
   CPPTypeRefI tmpType;
   CPPScopeRefI containingScope;
   int declCount;

   if (LookupObjectDeclaration(name, tmpDecl, tmpType, containingScope, declCount, true))
   {
      if (declCount == 1)
      {
         return new CPPExprId(tmpDecl);
      } else {
         // Get the list of decls matching the name
         CPPDeclListRefI declList = new CPPDeclList;
         CPPScope::DeclIterator iter;
         for (
            tmpDecl = containingScope->FindDeclByName(name, iter);
            tmpDecl.get() != 0;
            tmpDecl = containingScope->NextDecl(iter)
         )
         {
            if (!tmpDecl->IsType())
            {
               declList->push_back(tmpDecl);
            }
         }
         return new CPPExprIdLazy(name, declList);
      }
   } else {
      // Name lookup failed
      // It is OK to have undefined symbols in a template decl but nowhere else
      if (GetTemplateLevel() == 0)
      {
         ReportError("'" + name + "' is not defined", errTok);
      }
      return new CPPExprIdUndefined(name);
   }
}
