// CPPSymTablePredicates.cpp: Implementation of parser predicates.

#include "CPPSymTable.h"
#include "CPPScope.h"
#include "CPPScopeContainers.h"
#include "CPPDecl.h"

/////////////////////////////////////////////////////////////////////
// IsTypeName: See if the first declaration on the scope stack is a type.
//
// Inputs:
//    const JLStr&    name    The name to search
//
// Outputs: none
// Return:
//    bool            true if the name was found and it is a type
/////////////////////////////////////////////////////////////////////
bool 
CPPSymTable::IsTypeName(
   const JLStr& name
) const
{
   // Search the scope stack for the desired name
   for (
      CPPScopeStack::iterator stackIter = scopeStack->begin();
      stackIter != scopeStack->end();
      stackIter++
   )
   {
      CPPScopeRefI scope = (*stackIter).GetScope();

      // If there is any other declaration under that name in the
      // scope, then the name is not a type.
      // Exception: finding a declaration under the enclosing scope name
      // (i.e., ctor or class type entry) does not preclude finding a type 
      CPPScope::DeclIterator iter;
      for (
         CPPDeclRefI tmpDecl = scope->FindDeclByName(name, iter);
         tmpDecl.get() != 0;
         tmpDecl = scope->NextDecl(iter)
      )
      {
         if (
            !tmpDecl->IsType() && 
            !NameIsDecoratedName(tmpDecl->GetName(), scope->GetName())
         )
         {
            // Found a hiding declaration
            return false;
         }
      }

      // If no competing decls were found, then see if this is a type
      // Note: we could not do this check in the loop above because of
      // template placeholder scopes.
      if (scope->IsTypeName(name))
      {
         return true;
      }

      if ((*stackIter).GetResolved())
      {
         // Do not search past resolved scopes
         break;
      }
   }

   return false;
}

/////////////////////////////////////////////////////////////////////
// IsScopeName: See if the first declaration on the scope stack 
// names a scope.
//
// Inputs:
//    const JLStr&    name    The name to search
//
// Outputs: none
// Return:
//    bool            true if the name was found and it is a scope
/////////////////////////////////////////////////////////////////////
bool 
CPPSymTable::IsScopeName(
   const JLStr& name
) const
{
   // Search the scope stack for the desired name
   for (
      CPPScopeStack::iterator stackIter = scopeStack->begin();
      stackIter != scopeStack->end();
      stackIter++
   )
   {
      CPPScopeRefI scope = (*stackIter).GetScope();
      if (scope->IsScopeName(name))
      {
         return true;
      }

      // If there is any other declaration under that name in the
      // scope, then the name is not a type.
      if (
         scope->CountDeclsByName(name) != 0 ||
         (*stackIter).GetResolved()
      )
      {
         break;
      }
   }

   return false;
}

/////////////////////////////////////////////////////////////////////
// IsNamespaceName: See if the first declaration on the scope stack 
// names a scope.
//
// Inputs:
//    const JLStr&    name    The name to search
//
// Outputs: none
// Return:
//    bool            true if the name was found and it is a namespace
/////////////////////////////////////////////////////////////////////
bool 
CPPSymTable::IsNamespaceName(
   const JLStr& name
) const
{
   // Search the scope stack for the desired name
   for (
      CPPScopeStack::iterator stackIter = scopeStack->begin();
      stackIter != scopeStack->end();
      stackIter++
   )
   {
      CPPScopeRefI scope = (*stackIter).GetScope();
      if (scope->IsNamespaceName(name))
      {
         return true;
      }

      // If there is any other declaration under that name in the
      // scope, then the name is not a type.
      if (
         scope->CountDeclsByName(name) != 0 ||
         (*stackIter).GetResolved()
      )
      {
         break;
      }
   }

   return false;
}

/////////////////////////////////////////////////////////////////////
// IsTemplateName: See if the first declaration on the scope 
// stack is a templated name.
//
// Inputs:
//    const JLStr&    name    The name to search
//
// Outputs: none
// Return:
//    bool            true if the name was found and it is a template
/////////////////////////////////////////////////////////////////////
bool 
CPPSymTable::IsTemplateName(
   const JLStr& name
) const
{
   // Search the scope stack for the desired name
   for (
      CPPScopeStack::iterator stackIter = scopeStack->begin();
      stackIter != scopeStack->end();
      stackIter++
   )
   {
      CPPScopeRefI scope = (*stackIter).GetScope();
      if (scope->IsTemplatedName(name))
      {
         return true;
      }

      // If there is any other declaration under that name in the
      // scope, then the name is not a type.
      // Exception: we need to handle the case:
      //    template <class T> class A { A<T>* at; };
      // Which breaks during specialization if finding the "class A" entry
      // for the specialization hides the templated A in the enclosing scope.

      // TODO: find a better way to resolve this problem.
      // This is a hack... If it is a work scope, then disregard any
      // conflicting declarations, because template instantiation occurs
      // inside a work scope.
      if (scope->GetFlavor() != CPPScope::ScopeWork)
      {
         // If there is any other declaration under that name in the
         // scope, then the name is not a template.
         // Exception: finding a declaration under the enclosing scope name
         // (i.e., ctor or class type entry) does not preclude finding a type 
         CPPScope::DeclIterator iter;
         for (
            CPPDeclRefI tmpDecl = scope->FindDeclByName(name, iter);
            tmpDecl.get() != 0;
            tmpDecl = scope->NextDecl(iter)
         )
         {
            if (!NameIsDecoratedName(tmpDecl->GetName(), scope->GetName()))
            {
               // Found a hiding declaration
               return false;
            }
         }
      }

      if ((*stackIter).GetResolved())
      {
         break;
      }
   }

   return false;
}

/////////////////////////////////////////////////////////////////////
// IsEnclosingClassName: See if the name is equivalent to the
// name of the enclosing class.  "Equivalent to" means that a
// name is considered the same as a decorated template name using
// the same prefix.
//
// Inputs:
//    const JLStr&    name    The name to search
//
// Outputs: none
// Return:
//    bool            true if the name was found and it is a template
/////////////////////////////////////////////////////////////////////
bool 
CPPSymTable::IsEnclosingClassName(
   const JLStr& name
) const
{
   return NameIsDecoratedName(name, GetEnclosingClassName());
}
