// CPPSymTableTemplatePlaceholder.cpp: Management of template placeholder
// types and values

#include "CPPSymTable.h"
#include "CPPTypeTemplatePlaceholder.h"
#include "CPPExpr.h"
#include <map>

///////////////////////////////////////////////////////////////////////////
// Utility class for pair of shorts
///////////////////////////////////////////////////////////////////////////
typedef pair<short,short> ShortPair;

struct LessThanShortPair {
   bool operator()(const ShortPair& lhs, const ShortPair& rhs) const
   {
      return 
         lhs.first < rhs.first || 
         (lhs.first == rhs.first && lhs.second < rhs.second);
   }
};

///////////////////////////////////////////////////////////////////////////
// Definitions for placeholder types
///////////////////////////////////////////////////////////////////////////
typedef map<
   ShortPair, CPPTypeRefI, LessThanShortPair, allocator<CPPTypeRefI>
> RawTemplatePlaceholderTypeMap;

class CPPSymTable::TemplatePlaceholderTypeMap : 
   public VRefCount,
   public RawTemplatePlaceholderTypeMap
{
};

///////////////////////////////////////////////////////////////////////////
// Definitions for placeholder values
///////////////////////////////////////////////////////////////////////////
typedef map<
   ShortPair, CPPExprRefI, LessThanShortPair, allocator<CPPExprRefI>
> RawTemplatePlaceholderValueMap;

class CPPSymTable::TemplatePlaceholderValueMap : 
   public VRefCount,
   public RawTemplatePlaceholderValueMap
{
};


///////////////////////////////////////////////////////////////////////////
// MakeTemplatePlaceholderTypeMap: Create a new map between a template
// parameter index and template level pair; and a template placeholder
// type.  Used to avoid exporting the map implementation.
//
// Inputs: none
// Outputs: none
// Return:
//    CPPSymTable::TemplatePlaceholderTypeMapRefI     the new map
///////////////////////////////////////////////////////////////////////////
CPPSymTable::TemplatePlaceholderTypeMapRefI 
CPPSymTable::MakeTemplatePlaceholderTypeMap()
{
   return new TemplatePlaceholderTypeMap;
}


///////////////////////////////////////////////////////////////////////////
// GetTemplatePlaceholderType: Get a template placeholder type, either
// by finding it in the map, or creating it (and placing it in the map)
// if it does not already exist.  Note that all template placeholders 
// with the same parameter index and template level map to the same
// type pointer.
//
// Inputs: none
// Outputs: none
// Return:
//    CPPTypeRefI   smart pointer to the CPPTypeTemplatePlaceholder
///////////////////////////////////////////////////////////////////////////
CPPTypeRefI 
CPPSymTable::GetTemplatePlaceholderType(
   short parameterIdx, 
   short templateLevel
)
{
   ShortPair tmpPair(parameterIdx, templateLevel);

   TemplatePlaceholderTypeMap::iterator iter = templatePlaceholderTypeMap->find(tmpPair);

   if (iter != templatePlaceholderTypeMap->end())
   {
      return (*iter).second;
   } else {
      CPPTypeRefI newType = new CPPTypeTemplatePlaceholder(parameterIdx, templateLevel);
      templatePlaceholderTypeMap->insert(
         TemplatePlaceholderTypeMap::value_type(tmpPair, newType)
      );
      return newType;
   }
}


///////////////////////////////////////////////////////////////////////////
// MakeTemplatePlaceholderValueMap: Create a new map between a template
// parameter index and template level pair; and a template placeholder
// value.  Used to avoid exporting the map implementation.
//
// Inputs: none
// Outputs: none
// Return:
//    CPPSymTable::TemplatePlaceholderValueMapRefI     the new map
///////////////////////////////////////////////////////////////////////////
CPPSymTable::TemplatePlaceholderValueMapRefI 
CPPSymTable::MakeTemplatePlaceholderValueMap()
{
   return new TemplatePlaceholderValueMap;
}


///////////////////////////////////////////////////////////////////////////
// GetTemplatePlaceholderValue: Get a template placeholder value, either
// by finding it in the map, or by creating it (and placing it in the map)
// if it does not already exist.  Note that all template placeholders 
// with the same parameter index and template level map to the same
// expression pointer.
//
// Inputs: none
// Outputs: none
// Return:
//    CPPExprRefI   smart pointer to the CPPExprTemplatePlaceholder
///////////////////////////////////////////////////////////////////////////
CPPExprRefI 
CPPSymTable::GetTemplatePlaceholderValue(
   short parameterIdx, 
   short templateLevel
)
{
   ShortPair tmpPair(parameterIdx, templateLevel);

   TemplatePlaceholderValueMap::iterator iter = templatePlaceholderValueMap->find(tmpPair);

   if (iter != templatePlaceholderValueMap->end())
   {
      return (*iter).second;
   } else {
      CPPExprRefI newExpr = new CPPExprTemplatePlaceholder(parameterIdx, templateLevel);
      templatePlaceholderValueMap->insert(
         TemplatePlaceholderValueMap::value_type(tmpPair, newExpr)
      );
      return newExpr;
   }
}

