// CPPSymTableTypeBuiltin.cpp: Type manipulation for CPPSymTable

#include "CPPSymTable.h"
#include "CPPTypeBuiltin.h"
#include "CPPTypeContainers.h"

/////////////////////////////////////////////////////////////////////
// LookupBuiltinType: Lookup a builtin type.  Make a new entry
// for the type if necessary.
//
// Inputs:
//    CPPConst::TypeSpecifier  ts   The builtin type specifier.
//
// Outputs: none
// Return:
//    CPPTypeRefI    Pointer to the builtin type.  Valid until the
//                   parser is reset or destroyed.  Will return
//                   a type representing "void" if the type 
//                   specifiers are not valid.
/////////////////////////////////////////////////////////////////////
CPPTypeRefI
CPPSymTable::LookupBuiltinType(
   CPPConst::TypeSpecifier ts
)
{
   if (!ValidateBuiltinType(ts))
   {
      return LookupBuiltinType(CPPConst::tsVOID);
   } else {
      CPPBuiltinTypeTable::iterator iter = builtinTypes->find(ts);
      if (iter == builtinTypes->end())
      {
         return MakeBuiltinType(ts);
      } else {
         return (*iter).second.get();
      }
   }
}


/////////////////////////////////////////////////////////////////////
// ValidateBuiltinType: Validate a builtin type; i.e., make sure
// that all components of the builtin type are mutually compatible.
//
// Inputs:
//    CPPConst::TypeSpecifier  ts   The builtin type specifier.
//
// Outputs: none
// Return:
//    bool   true if the type specifiers are a valid builtin type.
/////////////////////////////////////////////////////////////////////
bool
CPPSymTable::ValidateBuiltinType(
   CPPConst::TypeSpecifier ts
)
{
   return CPPTypeBuiltin::TypeSpecifierIsValid(ts);
}


/////////////////////////////////////////////////////////////////////
// MakeBuiltinType: Make a new entry for a builtin type
//
// Inputs:
//    CPPConst::TypeSpecifier  ts   The builtin type specifier.
//
// Outputs: none
// Return:
//    CPPTypeRefI    Pointer to the builtin type.  Valid until the
//                   parser is reset or destroyed.
//
// Note: This should only be called for non-existent types.
// Call LookupBuiltinType if the type may already exist.
/////////////////////////////////////////////////////////////////////
CPPTypeRefI
CPPSymTable::MakeBuiltinType(
   CPPConst::TypeSpecifier ts
)
{
   assert(builtinTypes->find(ts) == builtinTypes->end());

   CPPTypeRefI newType = new CPPTypeBuiltin(ts);
   builtinTypes->insert(
      CPPBuiltinTypeTable::value_type(ts, newType)
   );
   return newType;
}


/////////////////////////////////////////////////////////////////
// FormatBuiltinTypeName: Format the name of the type specifiers.
//
// Inputs:
//    CPPConst::TypeSpecifier   ts   The type specifiers
// Return:
//    JLStr    The formatted type name
/////////////////////////////////////////////////////////////////
// virtual 
JLStr 
CPPSymTable::FormatBuiltinTypeName(
   CPPConst::TypeSpecifier ts
)
{
   return CPPTypeBuiltin::FormatTypeSpecifierName(ts);
}

