// CPPSymTableTypeModified.cpp: Methods to manipulate modified types

#include "CPPSymTable.h"
#include "CPPExpr.h"
#include "CPPTypes.h"
#include "CPPTypeContainers.h"
#include "JLCast.h"
#include <algorithm>

/////////////////////////////////////////////////////////////////////////
// LookupPointerToType: Given a type, get a type that represents a
// pointer to that type, making a new entry if needed
//
// Inputs:
//    CPPTypeRefI   inputType       The input type
// Return:
//    CPPTypeRefI   A type representing a pointer to the inputType
/////////////////////////////////////////////////////////////////////////
// virtual 
CPPTypeRefI 
CPPSymTable::LookupPointerToType(
   CPPTypeRefI inputType
)
{
   return GetOrUseTypeModifier(
      inputType,
      new CPPTypeIndirection(inputType, CPPTypeIndirection::IndirectPointer, false)
   );
}

/////////////////////////////////////////////////////////////////////////
// LookupReferenceToType: Given a type, get a type that represents a
// reference to that type, making a new entry if needed
//
// Inputs:
//    CPPTypeRefI   inputType       The input type
// Return:
//    CPPTypeRefI   A type representing a reference to the inputType
/////////////////////////////////////////////////////////////////////////
// virtual 
CPPTypeRefI 
CPPSymTable::LookupReferenceToType(
   CPPTypeRefI inputType
)
{
   return GetOrUseTypeModifier(
      inputType,
      new CPPTypeIndirection(inputType, CPPTypeIndirection::IndirectReference, false)
   );
}

/////////////////////////////////////////////////////////////////////////
// LookupPointerToMemberOfType: Given a type, get a type that represents a
// pointer-to-member of that type, making a new entry if needed
//
// Inputs:
//    CPPTypeRefI   inputType       The input type
//    CPPTypeRefI   memberClass     Class that the type points to a member of.
// Return:
//    CPPTypeRefI   A type representing a reference to the inputType
/////////////////////////////////////////////////////////////////////////
// virtual
CPPTypeRefI 
CPPSymTable::LookupPointerToMemberOfType(
   CPPTypeRefI inputType,
   CPPTypeRefI classType
)
{
   return GetOrUseTypeModifier(
      inputType,
      new CPPTypePointerToMember(inputType, classType, false)
   );
}

/////////////////////////////////////////////////////////////////////////
// LookupCVQualifierOfType: Given a type, get a type that represents a
// const/volatile modified version of the type, making a new entry if
// needed.  Note that there are potentially three different modifications
// of a type: const. const/volatile, and volatile.
//
// Inputs:
//    CPPTypeRefI   inputType       The input type
// Return:
//    CPPTypeRefI   A type representing a CV-qualified inputType
/////////////////////////////////////////////////////////////////////////
// virtual 
CPPTypeRefI 
CPPSymTable::LookupCVQualifierOfType(
   CPPTypeRefI inputType, 
   CPPConst::CVQualifier cv
)
{
   if (cv == CPPConst::cvNONE)
   {
      return inputType;
   }

   return GetOrUseTypeModifier(
      inputType,
      new CPPTypeCVQualified(inputType, cv, false)
   );
}


/////////////////////////////////////////////////////////////////////////
// LookupArrayOfType: Given a type, get a type that represents an
// array of the type, making a new entry if needed.  Arrays of different
// dimension are unique.
//
// Inputs:
//    CPPTypeRefI           inputType       The input type
//    CPPExprRefI           exprDimension   An expression representing the
//                                          dimension of the array, or a
//                                          zero-ref for empty dimension '[]'.
//    const ANTLRTokenPtr&  errTok          Used for error-reporting
// Return:
//    CPPTypeRefI   A type representing an array of the inputType
/////////////////////////////////////////////////////////////////////////
// virtual 
CPPTypeRefI 
CPPSymTable::LookupArrayOfType(
   CPPTypeRefI inputType, 
   CPPExprRefI exprDimension,
   const ANTLRTokenPtr& errTok
)
{
   // zero-dimension implies empty dimension []
   int dimension = 0;

   // If there is a dimension expression, then evaluate it.
   if (exprDimension.get() != 0)
   {
      double value;
      bool isInteger;
      if (!exprDimension->EvaluateConstant(value, isInteger))
      {
         ReportError("Non-constant expression for array dimension", errTok);
      } else if (!isInteger) {
         ReportError("Non-integral expression for array dimension", errTok);
      } else if (value <= 0) {
         char buf[200];
         sprintf(buf, "Array dimension [%d] is not positive", (int)value);
         ReportError(buf, errTok);
      } else {
         dimension = (int)value;
      }
   }

   return GetOrUseTypeModifier(
      inputType,
      new CPPTypeArray(inputType, dimension, false)
   );

}

/////////////////////////////////////////////////////////////////////////
// LookupFunctionReturningType: Given a type, get a type that represents a
// function returning the type, making a new entry if needed.
//
// Inputs:
//    CPPTypeRefI                       inputType   The input type
//    CPPConst::DeclarationModifier  dm          The calling convention
// Return:
//    CPPTypeRefI   A type representing a function returning the inputType
/////////////////////////////////////////////////////////////////////////
// virtual 
CPPTypeRefI 
CPPSymTable::LookupFunctionReturningType(
   CPPTypeRefI inputType,
   CPPConst::CallingConvention cc
)
{
   return GetOrUseTypeModifier(
      inputType,
      new CPPTypeFunctionReturning(inputType, cc,false)
   );
}

/////////////////////////////////////////////////////////////////////////
// LookupFunctionArgType: Given a function type, and an argument type,
// get a type that represents another function whose signature is that of
// the base function type, plus an additional argument.
//
// Inputs:
//    CPPTypeRefI   inputType     The input type (function type)
//    CPPTypeRefI   argType       The argument type
// Return:
//    CPPTypeRefI   A type representing a the inputType function signature
//               with one additional argument
/////////////////////////////////////////////////////////////////////////
// virtual 
CPPTypeRefI 
CPPSymTable::LookupFunctionArgType(
   CPPTypeRefI inputType,
   CPPTypeRefI argType
)
{
   return GetOrUseTypeModifier(
      inputType,
      new CPPTypeFunctionArg(inputType, argType, false)
   );
}

/////////////////////////////////////////////////////////////////////////
// LookupFunctionEllipsesType: Given a function type, get a
// type that represents another function whose signature is that of
// the base function type, plus an additional Ellipses argument.
//
// Inputs:
//    CPPTypeRefI   inputType     The input type (function type)
// Return:
//    CPPTypeRefI   A type representing a the inputType function signature
//               with one additional Ellipses argument
/////////////////////////////////////////////////////////////////////////
// virtual 
CPPTypeRefI 
CPPSymTable::LookupFunctionEllipsesType(
   CPPTypeRefI inputType
)
{
   return GetOrUseTypeModifier(
      inputType,
      new CPPTypeFunctionEllipses(inputType, false)
   );
}

/////////////////////////////////////////////////////////////////////////
// LookupBitfieldOfType: Given a function type, get a
// type that represents another function whose signature is that of
// the base function type, plus an additional Ellipses argument.
//
// Inputs:
//    CPPTypeRefI    inputType     The input type 
//    CPPExprRefI    exprFieldlen  An expression representing the
//                                 length of the field.
// Return:
//    CPPTypeRefI   A bitfield type derived from the input type
/////////////////////////////////////////////////////////////////////////
// virtual 
CPPTypeRefI 
CPPSymTable::LookupBitfieldOfType(
   CPPTypeRefI inputType,
   int fieldLen
)
{
   return GetOrUseTypeModifier(
      inputType,
      new CPPTypeBitfield(inputType, fieldLen, false)
   );
}


/////////////////////////////////////////////////////////////////////////
// CreateTempType: Create a temporary type that modifiers can
// be hung from.
//
// Inputs: none
// Return:
//    CPPTypeRefI   A tempory type (tsINT).
/////////////////////////////////////////////////////////////////////////
// virtual 
CPPTypeRefI 
CPPSymTable::CreateTempType()
{
   return new CPPTypeBuiltin(CPPConst::tsINT);
}

/////////////////////////////////////////////////////////////////////////
// MergeTypeModifiers: This function is used to support nested 
// declarators like:
//     int (*f)[];
// In which the declarator suffixes must be parsed before the modifiers of
// the nested delcarator can be applied.  Since the parser is LL, we apply
// the nested declarator modifiers to a temporary type, and the merge the
// nested declarator modifiers against the "real" type.
//
// Inputs:
//    CPPTypeRefI   dst    The "real" type to modify.
//    CPPTypeRefI   src    A "fake" type with a chain of type modifiers hanging 
//                        off it to be aplied to the "real" type.
// Return:
//    CPPTypeRefI   A pointer to a type representing "dst" with the 
//                  modifiers from "src" applied.
/////////////////////////////////////////////////////////////////////////
// virtual 
CPPTypeRefI
CPPSymTable::MergeTypeModifiers(
   CPPTypeRefI dst, 
   CPPTypeRefI src
)
{
   // Find the modifier to this type, if any
   CPPTypePtrSet::iterator iter = src->GetDerivedTypes()->begin();
   if (iter != src->GetDerivedTypes()->end())
   {
      // If there is already an identical modifier hanging off of
      // the destination type, then use it, otherwise add the modifier.
      CPPTypeRefI srcModifier = *iter;
      CPPTypePtrSet::iterator iter2 = dst->GetDerivedTypes()->find(srcModifier);
      if (iter2 != dst->GetDerivedTypes()->end())
      {
         // Continue down the modifier chain
         return MergeTypeModifiers(*iter2, srcModifier);
      } else {
         // Before connecting the source modifier chain to the destination, 
         // find the end of the source modifier chain, so
         // that we can return it.
         while ((*iter)->GetDerivedTypes()->begin() != (*iter)->GetDerivedTypes()->end())
         {
            iter = (*iter)->GetDerivedTypes()->begin();
         }
         CPPTypeRefI retval = *iter;

         // Add the source modifier chain to the destination derived list,
         // and disconnect the source modifier chain from the source list.
         STATIC_CAST(CPPTypeModifier*, srcModifier.get())->ChangeBaseType(dst);

         // Return the last source modifier in the chain
         return retval;
      }
   } else {
      // There are no more modifiers -- return this one
      return dst;
   }
}

/////////////////////////////////////////////////////////////////////////
//         Static functions
/////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////
// GetOrUseTypeModifier:  Search the derived type set of inputType 
// for a modifier matching the argument, and return either the existing
// modifier or add the argument to the derived modifier set.
//
// Inputs:
//    CPPTypeRefI   inputType       The base type being modified
//    CPPType*      newModifier     The modifier to search for
// Outputs:
// Return:
//    CPPTypeRefI   If the equivalent of newModifier was not in the
//                  derived type set, then newModifier is added to
//                  the derived type set and returned.
//                  Otherwise the existing modifier is returned.
/////////////////////////////////////////////////////////////////////////
//static 
CPPTypeRefI
CPPSymTable::GetOrUseTypeModifier(
   CPPTypeRefI inputType,
   CPPTypeModifier* newModifier
)
{
   CPPTypePtrSetCRefI derivedTypes = inputType->GetDerivedTypes();
   CPPTypePtrSet::const_iterator iter = derivedTypes->find(newModifier);
   if (iter != derivedTypes->end())
   {
      return *iter;
   } else {
      newModifier->AddToBase();
      return newModifier;
   }
}
