// CPPType.h: data structures to hold types.

#ifndef INCL_CPPTYPE_H
#define INCL_CPPTYPE_H

#include <assert.h>
#include "JLStr.h"
#include "JLCast.h"
#include "STLIncomplete.h"
#include "StrListIncomplete.h"
#include "CPPTypeIncomplete.h"
#include "CPPConst.h"

// NOTE: The implementation of LessThanNode() is mainly for the
// purpose of comparing modifiying types and potential candidates.
// Thus, LessThanNode() explicitly does not and should not consider
// the modified base type.

// NOTE: For any case in which a type points to something that it 
// does not "own", a "dumb" pointer is used instead of ref-counted
// pointer.  This is done to avoid ref-counted circularities, which
// would prevent types from being deleted properly.  

// Note that derived types "own" their modified types with smart pointers.
// Fundamental types DO NOT "own" their derived types and use a
// set of "dumb" pointers to the derived types.

///////////////////////////////////////////////////////////////////////
//  CPPType class
///////////////////////////////////////////////////////////////////////

class CPPType : public VRefCount
{
public:
   enum Flavor {
      Invalid,                // Invalid -- used for assertions
      // Fundamental types
      Builtin,                // Builtin (int,long, etc)
      Class,                  // Class/struct/union
      Enum,                   // Enumeration
      TemplatePlaceholder,    // Special placeholder for template type parameter
      // Modifiers for types
      CVQualified,            // const/volatile qualifier
      Indirection,            // Pointer, reference
      Array,                  // Array with or without dimensions
      PointerToMember,        // Pointer to member of class
      FunctionReturning,      // Function returning type
      FunctionArg,            // Function with another argument.  Can only
                              // modify FunctionReturning or FunctionArg
      FunctionEllipses,       // Function with an ellipses argument.  Can only
                              // modify FunctionReturning or FunctionArg
      Templated,              // Templated -- can only modify class or function
      Bitfield,               // Bitfield
      Other                   // Reserved for parser subclasses
   };

protected:
   // ctor -- Only created by derived classes
   CPPType(Flavor flavor_);

public:
   // dtor
   virtual ~CPPType();

   // Misc access methods
   Flavor GetFlavor() const { assertValid(); return flavor; }
   virtual bool IsFunctionType() const { assertValid(); return false; }
   virtual CPPTypeRefI GetFunctionReturnType() const { assertValid(); return CPPTypeRefI(); }

   // Remove template modifiers
   virtual CPPTypeRefI StripTemplateModifiers() { return this; }

   // Remove CV qualifiers
   virtual CPPTypeRefI StripCVQualifiers() { return this; }

   // Remove reference
   virtual CPPTypeRefI StripReference() { return this; }

   // Remove pointer
   virtual CPPTypeRefI StripPointer() { return this; }

   // Getting and setting the completeness for a type.
   // Only applies to Class and Enum.
   virtual bool GetComplete() const { return true; }
   virtual void SetComplete() { assert(0); }

   // Comparison of two type nodes.  Does not compare modified nodes.
   virtual bool LessThanNode(const CPPType& rhs) const
   {
      assertValid();
      // Base just compares flavors
      return (int)GetFlavor() < (int)rhs.GetFlavor();
   }
   // Equality of two nodes
   bool EqualNode(const CPPType& rhs) const
   {
      return !LessThanNode(rhs) && !rhs.LessThanNode(*this);
   }

   // Name is synthesized at run-time.
   // If declName is non-empty, then it will be inserted at the
   // proper position in the name for the declarator.
   virtual JLStr FormatName(
      const JLStr& declName,
      const JLStr& suffixStr = JLStr(),
      StrListRefI argList = StrListRefI(),
      bool forSpecialization = false
   ) const = 0;

   //
   // Derived-type management.  Each type (potentially) contains a
   // list of other derived types.  In order to save space, we don't
   // actually store the list of derived types unless we have to.
   //
   // Add a type that is derived from this one.  This is
   // called by the derived type when it is created.
   void AddDerivedType(CPPType* derivedType);
   // Remove a derived type from this type
   void RemoveDerivedType(CPPType* derivedType);
   // Get the base type for this derived type.  Returns zero for
   // fundamental types.
   virtual CPPTypeRefI GetBaseType() const { assertValid(); return CPPTypeRefI(); }
   // Determines if this class has a base type
   bool HasBaseType() const { return GetBaseType().get() != 0; }
   // Get the set of derived types -- don't change the return value!
   // Will return a static empty set if there are no derived types.
   CPPTypePtrSetCRefI GetDerivedTypes() const;

   // Search upwards in the derived-type chain (if there is one) for a 
   // template placeholder type, and get the type.  Returns true
   // if a template placeholder type was found and returned, false o/w.
   virtual bool FindTemplatePlaceholder(CPPTypeRefI& typeReturn) { return false; }

protected:
   Flavor flavor;

   // The assertValid() method is implemented by all subclasses.
   // Since it is used to catch incorrect downcasting, it should
   // be non-virtual.  I could have used RTTI, but was not convinced
   // that it was available everywhere at the time.  Besides, I
   // want the run-time checking overhead to go away when debug
   // is turned off.
   void assertValid() const { assert(magic == CPPTypeMagic); }

private:
   // disallow copy or assignment
   // If you implement these, do not copy the refcount!
   CPPType(const CPPType&);
   void operator=(const CPPType&);

   #ifndef NDEBUG
      enum { CPPTypeMagic = 0x12F0 };
      short magic;
   #endif

   // Set of dumb pointers to derived types.  Note that derived types
   // own the types they modify, not vice-versa.
   CPPTypePtrSetRefI derivedTypes;

   // Empty set of types return when there are no derived types
   static CPPTypePtrSetCRefI emptyTypeSet;
};


#endif
