// CPPTypeClass.cpp: Implementation of CPPTypeClass, which is a class
// that describes class types.

#include "CPPType.h"
#include "CPPTypeClass.h"
#include "CPPScope.h"
#include "CPPTypeClassBaseClass.h"

//////////////////////////////////////////////////////////////////////
// Class static data
//////////////////////////////////////////////////////////////////////
CPPTypeClass::BaseClassListRefI CPPTypeClass::emptyBaseClassList = new CPPTypeClass::BaseClassList;

//////////////////////////////////////////////////////////////////////
// Constructor
//////////////////////////////////////////////////////////////////////
CPPTypeClass::CPPTypeClass(
   const JLStr& name_,
   CPPConst::TypeSpecifier ts_,
   CPPScopeRefI memberScope_,
   int packing_
) : 
   CPPType(Class), 
   name(name_),
   ts(ts_),
   memberScope(memberScope_),
   complete(false),
   packing(packing_)
{
}

//////////////////////////////////////////////////////////////////////
// Destructor
//////////////////////////////////////////////////////////////////////
CPPTypeClass::~CPPTypeClass() 
{
   assertValid();
}

//////////////////////////////////////////////////////////////////////
// GetBaseClasses: Get list of base classes for a class type.
// 
// Inputs: none
// Outputs: none
// Return:
//    const JL_DList<BaseClass>&  The list of base classes.
//                            If the list is empty, then it is really
//                            a reference to a static list, so don't
//                            hang on to the reference.
//////////////////////////////////////////////////////////////////////
CPPTypeClass::BaseClassListCRefI 
CPPTypeClass::GetBaseClasses() const
{
   assertValid();
   if (baseClasses.get() == 0)
   {
      return emptyBaseClassList;
   } else {
      return baseClasses;
   }
}

//////////////////////////////////////////////////////////////////////
// AddBaseClass: Add base class to class type
// 
// Inputs:
//    const BaseClass&  baseClass   Base class type to add to
//                                  the class type.
// Outputs: none
// Return: none
//
// Note: The BaseClass struct takes care of checking that the
//       base type is really CPPTypeClass.
//////////////////////////////////////////////////////////////////////
void 
CPPTypeClass::AddBaseClass(
   CPPTypeRefI baseClass,
   CPPConst::BaseSpecifier baseSpecifier
)
{
   assertValid();
   if (baseClasses.get() == 0)
   {
      baseClasses = new BaseClassList;
   }
   baseClasses->push_back(BaseClassSpecifier(baseClass, baseSpecifier));
}


/////////////////////////////////////////////////////////////////
// FormatName: Format the name of the type, using the names of the
// modified types where needed to construct the full type name.
//
// Inputs:
//    const JLStr&  declName   If non-empty, then this is used
//                              to format the declarator name
//                              in the proper position.
//    const JLStr&  suffixStr  Any accumulated suffix string
// Outputs: none
// Return:
//    JLStr    The formatted type name
/////////////////////////////////////////////////////////////////
// virtual 
JLStr 
CPPTypeClass::FormatName(
   const JLStr& declName,
   const JLStr& suffixStr,
   StrListRefI,
   bool
) const
{
   assertValid();
   JLStr tmpStr;
   if (ts == CPPConst::tsCLASS)
   {
      tmpStr = "class ";
   } else if (ts == CPPConst::tsSTRUCT)
   {
      tmpStr = "struct ";
   } else if (ts == CPPConst::tsUNION)
   {
      tmpStr = "union ";
   } else {
      assert(0);
   }
   tmpStr += memberScope->GetParentScope()->FormatPrefix(false) + name;
   if (declName.length() != 0 || suffixStr.length() != 0)
   {
      tmpStr += " " + declName + suffixStr;
   }
   return tmpStr;
}
