// CPPTypeClass.h

#ifndef INCL_CPPTYPECLASS_H
#define INCL_CPPTYPECLASS_H

#include "CPPType.h"
#include "CPPScopeIncomplete.h"

///////////////////////////////////////////////////////////////////////
// CPPTypeClass type class.  This class holds types for classes, 
// structs, and unions.
//
// Note: For each *named* class type, there will be an entry
// in the type name map for the containing scope, that associates
// the name with the class type.
//
// Note: For each class type, there is a corresponding 
// sub-scope entered into the same scope as the class name.  That
// sub-scope holds all the class member declarations.  Even
// forward-declared classes have an associated scope (which is empty).
// The CPPScope is a ref-counted ptr, so that if the
// class' containing scope disappears (e.g., if it was a local scope),
// then the class' member scope still hangs around for the life of the
// class type (which is the life of the parser).
///////////////////////////////////////////////////////////////////////
class CPPTypeClass : public CPPType
{
public:
   // ctor/dtor
   CPPTypeClass(
      const JLStr& name_, 
      CPPConst::TypeSpecifier ts_,
      CPPScopeRefI memberScope_,
      int packing_
   );
   virtual ~CPPTypeClass();

   // access methods
   const JLStr& GetName() const { assertValid(); return name; }
   void ChangeName(const JLStr& name_) { assertValid(); name = name_; }
   CPPConst::TypeSpecifier GetTypeSpecifier() const { assertValid(); return ts; }
   CPPScopeRefI GetMemberScope() const { return memberScope; }
   int GetPacking() const { return packing; }
   virtual bool GetComplete() const { return complete; }
   virtual void SetComplete() { complete = true; }

   // Format the name of the type
   virtual JLStr FormatName(
      const JLStr& declName, 
      const JLStr& suffixStr,
      StrListRefI argList,
      bool forSpecialization
   ) const;

   class BaseClassSpecifier;
   DECLARE_INCOMPLETE(BaseClassList);

   // Comparison of two class type nodes.  
   // Must be the same node since classes are unique.
   virtual bool LessThanNode(const CPPType& rhs) const
   {
      assertValid();
      return 
         CPPType::LessThanNode(rhs) || 
         rhs.GetFlavor() == Class && this < &rhs;
   }

   // Get list of base classes -- Note that if the list of
   // base classes is empty, then this returns a global empty
   // list.  Otherwise it returns the real list.  So don't hang
   // on to the list reference returned by this method.
   BaseClassListCRefI GetBaseClasses() const;

   // Add base class to class type
   void AddBaseClass(
      CPPTypeRefI baseClass,
      CPPConst::BaseSpecifier baseSpecifier
   );

private:
   // disallow copy or assignment
   CPPTypeClass(const CPPTypeClass&);
   void operator=(const CPPTypeClass&);

   void assertValid() const { CPPType::assertValid(); assert(GetFlavor() == Class); }
   BaseClassListRefI baseClasses;

   // name of the class without scope qualifiers
   JLStr name;
   // specifies one of class/struct/union
   CPPConst::TypeSpecifier ts;

   // Scope containing the declarations of the class members
   // Will be zero reference if class is not complete
   CPPScopeRefI memberScope;

   bool complete;
   int packing;      // packing in effect when class type was declared

   // List returned when baseClasses member is nil
   static BaseClassListRefI emptyBaseClassList;
};


#endif