// CPPTypeFunction.h: definition of function-type-modifier classes

#ifndef INCL_CPPTYPEFUNCTION_H
#define INCL_CPPTYPEFUNCTION_H

#include "CPPTypeModifier.h"

///////////////////////////////////////////////////////////////////////
// CPPTypeFunctionReturning type class.  This class describes
// function types, or at least the type of a function with
// no arguments.  The modified type is the return type of the
// function.
//
// Note: This just declares a function that returns a type.  To
// declare functions taking parameters, modify one of these with
// a chain of CPPTypeFunctionArg or CPPTypeFunctionEllipes modifiers,
// one for each argument.
///////////////////////////////////////////////////////////////////////
class CPPTypeFunctionReturning : public CPPTypeModifier
{
public:
   // ctor/dtor
   CPPTypeFunctionReturning(
      CPPTypeRefI baseType_,
      CPPConst::CallingConvention cc_,
      bool addToBase
   );
   virtual ~CPPTypeFunctionReturning();

   // access methods
   CPPConst::CallingConvention GetCallingConvention() const { return cc; }
   virtual bool IsFunctionType() const { assertValid(); return true; }
   virtual CPPTypeRefI GetFunctionReturnType() const;

   // Format the name of the type
   virtual JLStr FormatName(
      const JLStr& declName, 
      const JLStr& suffixStr,
      StrListRefI argList,
      bool forSpecialization
   ) const;

   // Comparison of two function-returning nodes.
   virtual bool LessThanNode(const CPPType& rhs) const
   {
      assertValid();
      return
         CPPTypeModifier::LessThanNode(rhs) ||
         (
            rhs.GetFlavor() == FunctionReturning &&
            GetCallingConvention() < 
               STATIC_CAST(const CPPTypeFunctionReturning*, &rhs)->GetCallingConvention()
         );
   }

private:
   // disallow copy or assignment
   CPPTypeFunctionReturning(const CPPTypeFunctionReturning&);
   void operator=(const CPPTypeFunctionReturning&);

   void assertValid() const { 
      CPPTypeModifier::assertValid(); 
      assert(GetFlavor() == FunctionReturning); 
   }

   CPPConst::CallingConvention cc;
};


///////////////////////////////////////////////////////////////////////
// CPPTypeFunctionArg type class.  This class describes function
// types by modifying other function types.  The result is the
// type of a function with one more argument.  This may
// modify either CPPTypeFunctionReturning or CPPTypeFunctionArg.
//
// Note that default values for function arguments are not part
// of the function type -- they are part of the function declaration.
///////////////////////////////////////////////////////////////////////
class CPPTypeFunctionArg : public CPPTypeModifier
{
public:
   // ctor/dtor
   CPPTypeFunctionArg(
      CPPTypeRefI baseType_,
      CPPTypeRefI argType_,
      bool addToBase
   );
   virtual ~CPPTypeFunctionArg();

   // access methods
   CPPTypeRefI GetArgType() const { assertValid(); return argType; }
   virtual bool IsFunctionType() const { assertValid(); return true; }
   virtual CPPTypeRefI GetFunctionReturnType() const;

   // Format the name of the type
   virtual JLStr FormatName(
      const JLStr& declName, 
      const JLStr& suffixStr,
      StrListRefI argList,
      bool forSpecialization
   ) const;

   // Comparison of two funcionArg nodes.
   virtual bool LessThanNode(const CPPType& rhs) const
   {
      assertValid();
      return
         CPPTypeModifier::LessThanNode(rhs) ||
         (
            rhs.GetFlavor() == FunctionArg &&
            GetArgType() < STATIC_CAST(const CPPTypeFunctionArg*, &rhs)->GetArgType()
         );
   }

private:
   // disallow copy or assignment
   CPPTypeFunctionArg(const CPPTypeFunctionArg&);
   void operator=(const CPPTypeFunctionArg&);

   void assertValid() const { 
      CPPTypeModifier::assertValid(); 
      assert(GetFlavor() == FunctionArg); 
   }

   CPPTypeRefI argType;       // type of the argument
};


///////////////////////////////////////////////////////////////////////
// CPPTypeFunctionEllipses type class.  This class describes function
// types by modifying other function types.  The result is the
// type of a function with one more "ellipses" argument.  This may
// modify either CPPTypeFunctionReturning or CPPTypeFunctionArg.
///////////////////////////////////////////////////////////////////////
class CPPTypeFunctionEllipses : public CPPTypeModifier
{
public:
   // ctor/dtor
   CPPTypeFunctionEllipses(
      CPPTypeRefI baseType_,
      bool addToBase
   );
   virtual ~CPPTypeFunctionEllipses();

   // access methods
   virtual bool IsFunctionType() const { assertValid(); return true; }
   virtual CPPTypeRefI GetFunctionReturnType() const;

   // Format the name of the type
   virtual JLStr FormatName(
      const JLStr& declName, 
      const JLStr& suffixStr,
      StrListRefI argList,
      bool forSpecialization
   ) const;

   // Note: Don't need a LessThanNode() method;
   // CPPTypeModifier::LessThanNode() is enough

private:
   // disallow copy or assignment
   CPPTypeFunctionEllipses(const CPPTypeFunctionEllipses&);
   void operator=(const CPPTypeFunctionEllipses&);

   void assertValid() const { 
      CPPTypeModifier::assertValid(); 
      assert(GetFlavor() == FunctionEllipses); 
   }
};


#endif
