// CPPTypeFunctionArg.cpp: Implementation for fuction-with-argument modifier.

#include "CPPType.h"
#include "CPPTypeFunction.h"
#include "StrList.h"

///////////////////////////////////////////////////////////////////
// Constructor/destructor
///////////////////////////////////////////////////////////////////
CPPTypeFunctionArg::CPPTypeFunctionArg(
   CPPTypeRefI baseType_,
   CPPTypeRefI argType_,
   bool addToBase
) :
   CPPTypeModifier(FunctionArg, baseType_),
   argType(argType_)
{
   // Must modify a function type
   assert(
      GetBaseType()->GetFlavor() == FunctionReturning ||
      GetBaseType()->GetFlavor() == FunctionArg
   );
   assert(baseType_.get() != 0);
   if (addToBase)
   {
      AddToBase();
   }
}

//virtual
CPPTypeFunctionArg::~CPPTypeFunctionArg()
{ 
   assertValid(); 
   RemoveFromBase(); 
}


/////////////////////////////////////////////////////////////////
// FormatName: Format the name of the type, using the names of the
// modified types where needed to construct the full type name.
//
// Inputs:
//    const JLStr&  declName   If non-empty, then this is used
//                              to format the declarator name
//                              in the proper position.
//    const JLStr&  suffixStr  Any accumulated suffix string
//    bool          forSpecialization
//                             If true, then the name is formatted
//                             to be suitable in constructing a
//                             specialization scope.
// Return:
//    JLStr    The formatted type name
/////////////////////////////////////////////////////////////////
// virtual 
JLStr 
CPPTypeFunctionArg::FormatName(
   const JLStr& declName,
   const JLStr& suffixStr,
   StrListRefI argList,
   bool forSpecialization
) const
{
   assertValid();

   StrListRef newArgList;
   JLStr argStr;
   if (argList.get() != 0)
   {
      newArgList = new StrList(*argList);
      newArgList->pop_back();
      // Format the argument type with the argument name
      argStr = argType->FormatName(
         argList->back(),
         "",
         StrListRefI(),
         forSpecialization
      );
   } else {
      // Format the argument type as an abstact declarator
      argStr = argType->FormatName(
         "",
         "",
         StrListRefI(),
         forSpecialization
      );
   }

   // If this is modified by another FunctionArg or a FunctionEllipses,
   // then the suffixStr will start with "(".  If so, strip the "("
   // and prepend "(arg, ".  Oetherwise just prepend "(arg)".

   if (suffixStr.length() > 0 && suffixStr[0] == '(')
   {
      return GetBaseType()->FormatName(
         declName,
         "(" + argStr + ", " + JLStr(suffixStr.begin()+1, suffixStr.end()),
         newArgList,
         forSpecialization
      );
   } else {
      return GetBaseType()->FormatName(
         declName,
         "(" + argStr + ")" + suffixStr,
         newArgList,
         forSpecialization
      );
   }
}

/////////////////////////////////////////////////////////////////
// GetFunctionReturnType: Get the return type of the function type
//
// Inputs: none
// Outputs: none
// Return:
//    CPPTypeRefI   The return type of the function type
/////////////////////////////////////////////////////////////////
CPPTypeRefI 
CPPTypeFunctionArg::GetFunctionReturnType() const
{
   return GetBaseType()->GetFunctionReturnType();
}
