// CPPTypeFunctionReturning.cpp: Implementation of function-returning-type modifier.

#include "CPPType.h"
#include "CPPTypeFunction.h"


/////////////////////////////////////////////////////////////////
// ctor/dtor
/////////////////////////////////////////////////////////////////
CPPTypeFunctionReturning::CPPTypeFunctionReturning(
   CPPTypeRefI baseType_,
   CPPConst::CallingConvention cc_,
   bool addToBase
) :
   CPPTypeModifier(FunctionReturning, baseType_),
   cc(cc_)
{
   assert(baseType_.get() != 0);
   if (addToBase)
   {
      AddToBase();
   }
}

//virtual
CPPTypeFunctionReturning::~CPPTypeFunctionReturning()
{ 
   assertValid(); 
   RemoveFromBase(); 
}

/////////////////////////////////////////////////////////////////
// FormatName: Format the name of the type, using the names of the
// modified types where needed to construct the full type name.
//
// Inputs:
//    const JLStr&  declName   If non-empty, then this is used
//                              to format the declarator name
//                              in the proper position.
//    const JLStr&  suffixStr  Any accumulated suffix string
//    bool          forSpecialization
//                             If true, then the name is formatted
//                             to be suitable in constructing a
//                             specialization scope.
// Return:
//    JLStr    The formatted type name
/////////////////////////////////////////////////////////////////
// virtual 
JLStr 
CPPTypeFunctionReturning::FormatName(
   const JLStr& declName,
   const JLStr& suffixStr,
   StrListRefI,
   bool forSpecialization
) const
{
   assertValid();
   // This is a not-quite-right attempt to slide the function calling
   // conventions into the formatted name.  It's not right because
   // pointers-to-functions get formatted like:
   //      int (* __stdcall fp)()
   // instead of
   //      int (__stdcall * fp)()
   JLStr modifiedDeclName;
   if (cc == CPPConst::ccCDECL)
   {
      // the default
      modifiedDeclName = "";
   } else if (cc == CPPConst::ccMS_FASTCALL)
   {
      modifiedDeclName = "__fastcall ";
   } else if (cc == CPPConst::ccMS_STDCALL)
   {
      modifiedDeclName = "__stdcall ";
   } else {
      // Invalid calling convention??
      assert(0);
   }
   modifiedDeclName = modifiedDeclName + declName;
 
   // If this is modified by a FunctionArg or a FunctionEllipses,
   // then the suffixStr will start with "(".  If so, then pass
   // the suffix on down, otherwise prepend "()"
   if (suffixStr.length() > 0 && suffixStr[0] == '(')
   {
      return GetBaseType()->FormatName(
         modifiedDeclName, 
         suffixStr, 
         StrListRefI(), 
         forSpecialization
      );
   } else {
      return GetBaseType()->FormatName(
         modifiedDeclName, 
         "()" + suffixStr, 
         StrListRefI(), 
         forSpecialization
      );
   }
}

/////////////////////////////////////////////////////////////////
// GetFunctionReturnType: Get the return type of the function type
//
// Inputs: none
// Outputs: none
// Return:
//    CPPTypeRefI   The return type of the function type
/////////////////////////////////////////////////////////////////
CPPTypeRefI 
CPPTypeFunctionReturning::GetFunctionReturnType() const
{
   return GetBaseType();
}
