// CPPTypeIndirection.cpp: Implementation for indirection type modifier class.

#include "CPPType.h"
#include "CPPTypeIndirection.h"


/////////////////////////////////////////////////////////////////
// ctor/dtor
/////////////////////////////////////////////////////////////////

CPPTypeIndirection::CPPTypeIndirection(
   CPPTypeRefI baseType_,
   IndirectType indirectType_,
   bool addToBase
) : 
   CPPTypeModifier(Indirection, baseType_),
   indirectType(indirectType_)
{
   assert(baseType_.get() != 0);
   if (addToBase)
   {
      AddToBase();
   }
}

// virtual
CPPTypeIndirection::~CPPTypeIndirection()
{
   assertValid(); 
   RemoveFromBase(); 
}

/////////////////////////////////////////////////////////////////
// StripReference.  Return this type with any reference qualifiers
// removed.
// Inputs: none
// Ouputs: none
// Return:
//    This type with any reference modifiers removed
/////////////////////////////////////////////////////////////////
//virtual 
CPPTypeRefI 
CPPTypeIndirection::StripReference()
{
   if (indirectType == IndirectReference)
   {
      return GetBaseType()->StripReference();
   } else {
      return this;
   }
}

/////////////////////////////////////////////////////////////////
// StripPointer.  Return this type with any pointer qualifiers
// removed.
// Inputs: none
// Ouputs: none
// Return:
//    This type with any pointer modifiers removed
/////////////////////////////////////////////////////////////////
//virtual 
CPPTypeRefI 
CPPTypeIndirection::StripPointer()
{
   if (indirectType == IndirectPointer)
   {
      return GetBaseType()->StripPointer();
   } else {
      return this;
   }
}


/////////////////////////////////////////////////////////////////
// FormatName: Format the name of the type, using the names of the
// modified types where needed to construct the full type name.
//
// Inputs:
//    const JLStr&  declName   If non-empty, then this is used
//                              to format the declarator name
//                              in the proper position.
//    const JLStr&  suffixStr  Any accumulated suffix string
//    bool          forSpecialization
//                             If true, then the name is formatted
//                             to be suitable in constructing a
//                             specialization scope.
// Return:
//    JLStr    The formatted type name
/////////////////////////////////////////////////////////////////
// virtual 
JLStr 
CPPTypeIndirection::FormatName(
   const JLStr& declName,
   const JLStr& suffixStr,
   StrListRefI,
   bool forSpecialization
) const
{
   assertValid();

   // Format the leading */& it and pass it down to the 
   // base type with the declarator
   JLStr modifiedDeclarator;
   switch (indirectType)
   {
      case IndirectPointer:
         modifiedDeclarator = "*";
         break;
      case IndirectReference:
         modifiedDeclarator = "&";
         break;
   }
   modifiedDeclarator += declName;

   // If this modifies a postfix modifier (function/array), then
   // combine the declName and suffixes and surround them with () 
   // for indication of proper binding.
   switch (GetBaseType()->GetFlavor())
   {
      case FunctionReturning:
      case FunctionArg:
      case FunctionEllipses:
      case Array:
         return GetBaseType()->FormatName(
            JLStr("(") + modifiedDeclarator + suffixStr + JLStr(")"),
            "",
            StrListRefI(),
            forSpecialization
         );
      default:
         return GetBaseType()->FormatName(
            modifiedDeclarator, 
            suffixStr,
            StrListRefI(),
            forSpecialization
         );
   }
}

