// CPPTypeModifier.h: intermediate base class for type modifier classes

#ifndef INCL_CPPTYPEMODIFIER_H
#define INCL_CPPTYPEMODIFIER_H

#include "CPPType.h"

///////////////////////////////////////////////////////////////////////
// CPPTypeModifier type class.  This class is an intermediate
// class for all derived-class subclasses of CPPType.
///////////////////////////////////////////////////////////////////////
class CPPTypeModifier : public CPPType
{
protected:
   // only create via base class
   CPPTypeModifier(
      Flavor flavor_, 
      CPPTypeRefI baseType_
   ) : 
      CPPType(flavor_),
      baseType(baseType_) 
      #ifndef NDEBUG
         , belongsToBase(false)
      #endif
   {
      // Note: DO NOT CALL AddToBase() here.
      // The call must be made in the constructors of the most-derived 
      // subclasses of CPPTypeModifier otherwise the type will have 
      // been partially created, will fail proper comparison in 
      // the derivedTypeSet of the base type, which will end up in an
      // improper order (which makes other comparisions fail).
   }

   virtual ~CPPTypeModifier() 
   {
      // Note: DO NOT CALL RemoveFromBase() here.
      // The call must be made in the destructors of the most-derived 
      // subclasses of CPPTypeModifier otherwise the type will have 
      // been partially destroyed and will fail proper comparison in 
      // the derivedTypeSet of the base type.
   }

public:
   // access methods
   virtual CPPTypeRefI GetBaseType() const { assertValid(); return baseType; }

   // Search for template placeholder type somewhere in the chain
   virtual bool FindTemplatePlaceholder(CPPTypeRefI& typeReturn) 
   {
      return baseType->FindTemplatePlaceholder(typeReturn);
   }

   // Add this modification to the derived-type list of the base type.
   // Don't do this more than once.  This is typically used when determining
   // whether a new modifier is needed for a type.  You build a test modifier,
   // serach for it in the derived set for the base, and if not found,
   // add it to the base.
   void AddToBase() {
      assertValid();
      assert(!belongsToBase);
      baseType->AddDerivedType(this);
      #ifndef NDEBUG
         belongsToBase = true;
      #endif
   }

   // Remove a type modifier from its base type.  This is usually called during
   // destruction of the most-derived subclasses of CPPTypeModifier
   void RemoveFromBase() {
      assertValid();
      assert(belongsToBase);
      baseType->RemoveDerivedType(this);
      #ifndef NDEBUG
         belongsToBase = false;
      #endif
   }

   // Change the base type of a modifier class, adding it to the 
   // new base derived type list.  This is typically used when processing
   // function-pointers, because the LL parser gets some of the type
   // modifiers out-of-sequence and has to rearrange the type modifiers.
   void ChangeBaseType(CPPTypeRefI newBase) 
   {
      // Add the derived type to the new base before removing it from the
      // old base, so that the ref-counting does not delete it.
      // (Actually this reason is obsolete now because the type-modifiers own
      // their base types, but it won't hurt).
      assertValid();
      assert(belongsToBase);
      newBase->AddDerivedType(this);
      baseType->RemoveDerivedType(this);
      baseType = newBase;
   }

protected:
   void assertValid() const {
      CPPType::assertValid();
      assert(
         flavor == CVQualified ||
         flavor == Indirection ||
         flavor == Array ||
         flavor == PointerToMember ||
         flavor == FunctionReturning ||
         flavor == FunctionArg ||
         flavor == FunctionEllipses ||
         flavor == Templated ||
         flavor == Bitfield
      );
   }
private:
   // disallow copy or assignment
   CPPTypeModifier(const CPPTypeModifier&);
   void operator=(const CPPTypeModifier&);

   // Base type pointer.
   // This can be a smart pointer because base types do not own 
   // their modifiers; modifiers own the base types.
   CPPTypeRefI baseType;

   #ifndef NDEBUG
      // true if the type has been added to the derived type list of the base.
      bool belongsToBase;
   #endif
};



#endif
