// CPPTypePointerToMember.cpp: Implementation for ptr-to-member type modifier class.

#include "CPPType.h"
#include "CPPTypePointerToMember.h"
#include "CPPTypeClass.h"
#include "JLCast.h"

//////////////////////////////////////////////////////////////////////
// Constructor/destroctor
//////////////////////////////////////////////////////////////////////
CPPTypePointerToMember::CPPTypePointerToMember(
   CPPTypeRefI baseType_,
   CPPTypeRefI ptrToMemberClass_,
   bool addToBase
) :
   CPPTypeModifier(PointerToMember, baseType_),
   ptrToMemberClass(ptrToMemberClass_)
{
   // ptrToMemberClass must be a class type
   assert(ptrToMemberClass->GetFlavor() == Class);

   assert(baseType_.get() != 0);
   if (addToBase)
   {
      AddToBase();
   }
}


CPPTypePointerToMember::~CPPTypePointerToMember()
{ 
   assertValid(); 
   RemoveFromBase(); 
}

/////////////////////////////////////////////////////////////////
// FindTemplatePlaceholder: Search the type chain for a template
// placeholder and return it if found.
//
// Inputs: none
// Outputs:
//    CPPTypeRefI&   typeReturn   The type if found
// Return:
//    bool     true if a template placeholder type was found
/////////////////////////////////////////////////////////////////
//virtual 
bool 
CPPTypePointerToMember::FindTemplatePlaceholder(
   CPPTypeRefI& typeReturn
)
{
   return
      CPPTypeModifier::FindTemplatePlaceholder(typeReturn) ||
      ptrToMemberClass->FindTemplatePlaceholder(typeReturn);
}

/////////////////////////////////////////////////////////////////
// FormatName: Format the name of the type, using the names of the
// modified types where needed to construct the full type name.
//
// Inputs:
//    const JLStr&  declName   If non-empty, then this is used
//                              to format the declarator name
//                              in the proper position.
//    const JLStr&  suffixStr  Any accumulated suffix string
//    bool          forSpecialization
//                             If true, then the name is formatted
//                             to be suitable in constructing a
//                             specialization scope.
// Return:
//    JLStr    The formatted type name
/////////////////////////////////////////////////////////////////
// virtual 
JLStr 
CPPTypePointerToMember::FormatName(
   const JLStr& declName,
   const JLStr& suffixStr,
   StrListRefI,
   bool forSpecialization
) const
{
   assertValid();

   // Format the leading "T::*" it and pass it down to the 
   // base type with the declarator
   JLStr modifiedDeclarator = 
      STATIC_CAST(const CPPTypeClass*,ptrToMemberClass.get())->GetName();
   modifiedDeclarator += "::* ";
   modifiedDeclarator += declName;

   // If this modifies a postfix modifier (function/array), then
   // combine the declName and suffixes and surround them with () 
   // for indication of proper binding.
   switch (GetBaseType()->GetFlavor())
   {
      case FunctionReturning:
      case FunctionArg:
      case FunctionEllipses:
      case Array:
         return GetBaseType()->FormatName(
            JLStr("(") + modifiedDeclarator + suffixStr + JLStr(")"),
            "",
            StrListRefI(),
            forSpecialization
         );
      default:
         return GetBaseType()->FormatName(
            modifiedDeclarator, 
            suffixStr,
            StrListRefI(),
            forSpecialization
         );
   }
}
