// CPPTypeTemplatePlaceholder.h: definition of template-placeholder-type class

#ifndef INCL_CPPTYPETEMPLATEPLACEHOLDER_H
#define INCL_CPPTYPETEMPLATEPLACEHOLDER_H

#include "CPPType.h"
#include "CPPScopeIncomplete.h"

///////////////////////////////////////////////////////////////////////
// CPPTypeTemplatePlaceholder type class.  This class describes special "types"
// that serve as placeholders for template parameters that are types.
// For example, in:
//     template <class T, int I> class A{};
// The "T" template parameter is a type argument.
//
// This type does not exist in the scope containing the template.  
// Instead, it resides in the TemplateParameterList.  For example, in:
//     template <class T> int foo(T t);
// when "int foo(T t);" is parsed, the "T" is bound to the
// CPPTypeTemplatePlaceholder stored in the TemplateParameterList
// of the CPPTypeTemplated modifying the function type.
///////////////////////////////////////////////////////////////////////
class CPPTypeTemplatePlaceholder : public CPPType
{
public:
   // ctor/dtor
   CPPTypeTemplatePlaceholder(
      short parameterIdx_,
      short templateLevel_
   ) : 
      CPPType(TemplatePlaceholder),
      parameterIdx(parameterIdx_),
      templateLevel(templateLevel_)
   {}
   virtual ~CPPTypeTemplatePlaceholder() {}

   // Search for template placeholder type somewhere in the chain
   virtual bool FindTemplatePlaceholder(CPPTypeRefI& typeReturn);

   // Return a template placeholder scope
   CPPScopeRefI GetPlaceholderScope() const { return thePlaceholderScope; }

   // Format the name of the type.
   // If declName is non-empty, then it will be inserted at the
   // proper position in the name for the declarator.
   virtual JLStr FormatName(
      const JLStr& declName,
      const JLStr& suffixStr,
      StrListRefI argList,
      bool forSpecialization
   ) const;

   // Equality of two template type nodes.
   // Must be the same node since template placeholders are unique
   virtual bool LessThanNode(const CPPType& rhs) const
   {
      assertValid();
      return 
         CPPType::LessThanNode(rhs) || 
         (rhs.GetFlavor() == TemplatePlaceholder && this < &rhs);
   }

   // Accessors
   short GetParameterIdx() const { assertValid(); return parameterIdx; }
   short GetTemplateLevel() const { assertValid(); return templateLevel; }

private:
   // disallow copy or assignment
   CPPTypeTemplatePlaceholder(const CPPTypeTemplatePlaceholder&);
   void operator=(const CPPTypeTemplatePlaceholder&);

   void assertValid() const { 
      CPPType::assertValid(); 
      assert(GetFlavor() == TemplatePlaceholder);
   }

   // The index of the parameter in its list (zero-based)
   short parameterIdx;
   // The level of template nesting of the parameter (zero-based)
   short templateLevel;
   // A placeholder scope used when looking up members of placeholders
   static CPPScopeRefI thePlaceholderScope;
};


#endif
