// CPPTypeTemplated.cpp: Implementation of "templated" modifier

#include "CPPType.h"
#include "CPPTypeTemplated.h"
#include "TemplateParameterList.h"
#include "TokenList.h"

/////////////////////////////////////////////////////////////////
// Constructor
// Inputs:
//    CPPTypeRefI                baseType_  Type that this modifies
//    TokenListRefI              tokens_    List of tokens that comprise
//                                          the body of the template.  The
//                                          list will be copied.
//    TemplateParameterListRefI  parameterList_
//                                          List of template parameters.
//    TemplateArgumentListRefI   specializationArgList_
//                                          List of template specialization arguments.
//                                          This mirrors the formal parameters for
//                                          primary templates, and is equivalent to the
//                                          function arg list for functions.
//    bool                       isPrimary_ true if this is a primary template
//                                          declaration
//    bool                       addToBase  If true, then this modifier
//                                          is added to the list of modifiers
//                                          for the base type.
/////////////////////////////////////////////////////////////////
CPPTypeTemplated::CPPTypeTemplated(
   CPPTypeRefI baseType_,
   TokenListRefI tokens_,
   TemplateParameterListRefI parameterList_,
   TemplateArgumentListRefI specializationArgList_,
   bool isPrimary_,
   bool addToBase
) :
   CPPTypeModifier(Templated, baseType_),
   isPrimary(isPrimary_),
   parameterList(parameterList_),
   specializationArgList(specializationArgList_)
{
   // Create an empty token list if the input list is nil.
   if (tokens_.get() == 0)
   {
      tokens = new TokenList;
   } else {
      tokens = new TokenList(*tokens_);
   }

   assert(baseType_.get() != 0);
   if (addToBase)
   {
      AddToBase();
   }
}


/////////////////////////////////////////////////////////////////
// Destructor
/////////////////////////////////////////////////////////////////
// virtual 
CPPTypeTemplated::~CPPTypeTemplated() 
{
   assertValid();
   RemoveFromBase();
}


/////////////////////////////////////////////////////////////////
// SetTokens: Replace the template body tokens.  This is used for
// the body of a class template declaration, because the template
// declaration is made before the body is parsed and the template
// tokens must be set later.
//
// Inputs:
//    TokenListRefI     tokens_     List of tokens that comprise
//                                  the body of the template.  The
//                                  list will be copied.
/////////////////////////////////////////////////////////////////
void 
CPPTypeTemplated::SetTokens(
   TokenListRefI tokens_
)
{
   *tokens = *tokens_;
}


/////////////////////////////////////////////////////////////////
// FormatName: Format the name of the type, using the names of the
// modified types where needed to construct the full type name.
//
// Inputs:
//    const JLStr&  declName   If non-empty, then this is used
//                             to format the declarator name
//                             in the proper position.
//    const JLStr&  suffixStr  Any accumulated suffix string
//    bool          forSpecialization
//                             If true, then the name is formatted
//                             to be suitable in constructing a
//                             specialization scope.
// Return:
//    JLStr    The formatted type name
/////////////////////////////////////////////////////////////////
// virtual 
JLStr 
CPPTypeTemplated::FormatName(
   const JLStr& declName,
   const JLStr& suffixStr,
   StrListRefI argNameList,
   bool forSpecialization
) const
{
   assertValid();

   JLStr templateParmStr = "template <";
   bool first = true;
   for (
      TemplateParameterList::iterator iter = parameterList->begin();
      iter != parameterList->end();
      iter++
   )
   {
      if (first)
      {
         first = false;
      } else {
         templateParmStr += ", ";
      }
      if ((*iter)->GetFlavor() == TemplateParameter::Type)
      {
         templateParmStr += "typename ";
         templateParmStr += (*iter)->GetName();
      } else {
         TemplateParameter *parm = *iter;
         templateParmStr += STATIC_CAST(TemplateParameterValue*, parm)->GetType()->FormatName(
            (*iter)->GetName(), 
            "", 
            StrListRefI(),
            forSpecialization
         );
      }
   }
   templateParmStr += "> " + suffixStr;

   JLStr retStr;
   retStr = 
      templateParmStr + 
      GetBaseType()->FormatName(declName, suffixStr, argNameList, forSpecialization);

   // Turn this code on if you want the token list appended to the
   // end of templated types
   #if 0
      // Append the token list to the output
      retStr += " TOKENS =";
      for (
         TokenList::iterator tokenIter = tokens->begin();
         tokenIter != tokens->end();
         tokenIter++
      )
      {
         retStr += " ";
         retStr += (*tokenIter)->getText();
      }
   #endif

   return retStr;
}


/////////////////////////////////////////////////////////////////////
// StripTemplateModifiers: Remove template modifiers from a type and
// return the resulting type.
//
// Inputs: none
// Outputs: none
// Return:
//    CPPTypeRefI   The type with any templated modifiers removed
/////////////////////////////////////////////////////////////////////
CPPTypeRefI
CPPTypeTemplated::StripTemplateModifiers()
{
   return GetBaseType()->StripTemplateModifiers();
}

