// cplusplus.g: PCCTS-based C++ grammar

/////////////////////////////////////////////////////////////////////////
// NOTES AND CONVENTIONS:
//
// 1) The "ANTLR" used to compile this grammar is ***NOT*** a stock
//    ANTLR by any stretch of the imagination.  Don't even try to
//    compile this with a stock ANTLR.  The modified ANTLR will:
//      (A) Generate try/catch instead of setjmp/longjmp for backtracking
//      (B) Execute all actions while in "guess mode"
//      (C) Only execute a syntactic predicate once when successful.
//      (D) throw exceptions for syntax and semantic errors.
//      (E) Generate try/catch blocks for parser exception clauses.
//
// 2) There are places in the grammar (mostly declarations) where actions
//    are executed during "guess mode" and must be backed out if the guess
//    fails.  The backing out is handled by the destructor of an object
//    created on the call stack.  Such backtracking objects are clearly
//    marked.  There are few places where this is necessary, because most
//    things are returned from rules by ref-counted pointers, so an exception
//    automatically cleans up any constructed information.  The only things
//    that need to be explicitly backed out are symbol table changes.
//
// 3) Whenever a straight C++ pointer is returned by a rule, 
//    it is a pointer into a data structure kept by the parser.
//    These return values should never be explicitly deleted.
//    The parser takes care of any cleanup.  Most values are
//    returned by reference-counted pointers, even if they point
//    to items that are stored in the symbol table.
//
// 4) The declaration rule hierarchy passes around a CPPTypeRefI that is
//    successively modified by each additional element in the
//    declaration-specifiers or declarators.  The types themselves are
//    owned by data structures in the parser.
//
// 5) The rules qualified_id and scope_override
//    have the side-effect of altering the current scope stack.  
//    For example, once scope_override is called,
//    the scope stack is left in a state where looked-up identifiers
//    will be found in the overriding class/namespace instead of the
//    default scope.  Callers of these rules must take care
//    to use SaveScope/RestoreScope to manage the scope stack properly.
//    The above rules all employ backtracking to undo scope stack
//    changes when a guess fails (or when any error is encountered,
//    for that matter).
//
// 6) Where rules process and return values, local variables and
//    return values are always preferable to global variables, class
//    static variables, and even per-parser-object class data members.
//    It should be obvious that using class statics and global variables 
//    breaks when multiple instances of the parser are created.  More 
//    subtle is the fact that using per-parser-object data members will
//    break when grammar rules using those data members are called 
//    recursively.  For example, the declaration hierarchy can be called
//    recursively when processing function arguments, especially when
//    processing function arguments that are pointers-to-functions.
//    The only per-parser data structures should be of the same nature
//    as the symbol table -- they exist for collecting persistent information.
//
// 7) Sometimes it is impossible or undesireable to pass a value directly,
//    perhaps because the value is complex (like a container) and we 
//    don't want to clutter the grammar public interface with too many
//    dependencies on complex types.  In these cases, a data structure
//    should be accessed via the parser subclass (CPPParserSym) to
//    contain the complex information.  The parser grammar base class should
//    contain pure-virtual methods for accessing the data structure. This
//    is the case with the scope, type, and declaration hierarchy.
//    If such a value is processed in recursive rules, then the data 
//    structure should allow for that.
//
// 8) Rules never return values that must be explicitly freed.  
//    If a return value is allocated from the heap, it is returned as a
//    reference-counted pointer so that the caller can safely ignore it,
//    and so that a parser exception will automatically free it.
//
// 9) When a smart pointer is returned from a rule (e.g., CPPTypeRefI), every
//    attempt should be made to return a non-zero pointer.  This is
//    usually acheived automatically because syntax errors throw
//    exceptions so bad values resulting from syntax errors are 
//    never returned.  If other errors occur that do not throw exceptions,
//    such as when name lookup fails, then a usable placeholder is 
//    returned (e.g. an "int" type is returned when type lookup fails).
//
// 10) templates are difficult and nasty to process.  The crux of most
//    difficulty stems from the fact that the body of the template
//    function or class must be parsed even though the nature of the
//    parameterized types is not yet known.  This poses difficulty when
//    disambiguating between types and identifiers, for example:
//       template <class T> class A {
//          T::I foo();   // is T::I a type or an identifier?
//       }
//    To aid template parsing, the keyword 'typename' is used in
//    front of qualified names that use a parameterized type as
//    a scope to indiciate that it is a type, for example:
//       template <class T> class A {
//          typename T::I foo();   // T::I is assumed to be a type
//          int i = T::foo();      // T::foo is an identifier
//       }
//    Thus, any symbol qualified by "T::" where T is a parameterized
//    type, is assumed to be an identifier unless preceded by 'typename'.
//    However, things like:
//       T::TT::x
//    are still nasty, because "TT::" must be treated as a type, or at
//    least a scope, even though there is no 'typename'.  In addition,
//    in order not to distort or replicate the grammar for dealing with
//    templates, there are special "scopes" that represent the parameterized
//    types, that are able to cough up any declaration upon demand.  This
//    makes dealing with "T::TT::i" nice and uniform, because the "T::" and
//    "TT::" are the special scopes that claim to have every symbol possible.
//    
//    Unfortunately, it is even more complex than that.  Recall that T::I
//    is treated as an identifier unless preceeded by 'typename'.  So a
//    "special" scope, when asked if an identifier is a type, always 
//    reports "no".  But, if asked to look up type for any identifier
//    it must dutifully produce one.  For example:
//       template <class T> class A {
//          T::I foo();
//       }
//    is a syntax error because T::I is not a type, and the "scope" for
//    "T" always says that any identifier its not a type.  However, in:
//       template <class T> class A {
//          typename T::I foo();
//       }
//    T::I is assumed to be a type because of the 'typename' keyword,
//    so when asked to cough up a CPPTypeRefI for the symbol, the special
//    "T" scope is obliged to do so.
//
// 11) The grammar rules *always* prevent hoisting of predicates by
//     starting with two actions (init-action and first real action).
//     ANTLR is a little odd about hoisting predicates, and we don't
//     need it to anyway.  All rules (and some subrules) start with:
//         <<;>>  <<;>>
//
// 12) The scope-searching methods Lookup* are implemented in the parser
//     subclass by calling the analogous method of the symbol table object.
//     However, the parser subclass' methods always return a usable value,
//     even in the event of error.  The symbol table methods may return
//     empty references.
/////////////////////////////////////////////////////////////////////////


#header <<
   #ifdef _MSC_VER
      // Disable some warnings that appear commonly in Rogue Wave STL
      #pragma warning (disable:4786)    // long debug names
      #pragma warning (disable:4146)    // unary minus operator applied to unsigned type
      #pragma warning (disable:4018)    // '>' : signed/unsigned mismatch
   #endif
   #include "CPPCommon.h"
>>

/* Do not define any tokens in this grammar.  All tokens should be defined
 * in the preprocessor grammar.  The "tokens.h" used here is generated from
 * the preprocessor grammar.
 */
#tokdefs "../cpre/tokens.h"

// Token class for operator types
#tokclass OPTOR_SIMPLE_TOKCLASS {
   TOK_PLUS TOK_MINUS TOK_STAR TOK_DIVIDE TOK_MOD TOK_BITWISEXOR TOK_AMPERSAND 
   TOK_BITWISEOR TOK_TILDE TOK_NOT 
   TOK_SHIFTLEFT TOK_SHIFTRIGHT
   TOK_ASSIGNEQUAL TOK_TIMESEQUAL TOK_DIVIDEEQUAL TOK_MODEQUAL 
   TOK_PLUSEQUAL TOK_MINUSEQUAL 
   TOK_SHIFTLEFTEQUAL TOK_SHIFTRIGHTEQUAL TOK_BITWISEANDEQUAL 
   TOK_BITWISEXOREQUAL TOK_BITWISEOREQUAL
   TOK_EQUAL TOK_NOTEQUAL TOK_LESSTHAN TOK_GREATERTHAN TOK_LESSTHANOREQUALTO 
   TOK_GREATERTHANOREQUALTO TOK_OR TOK_AND
   TOK_PLUSPLUS TOK_MINUSMINUS TOK_COMMA TOK_POINTERTO TOK_POINTERTOMBR
}

// Token class for assignment tokens
#tokclass ASSIGN_TOKCLASS {
   TOK_ASSIGNEQUAL TOK_TIMESEQUAL TOK_DIVIDEEQUAL 
   TOK_MODEQUAL TOK_PLUSEQUAL TOK_BITWISEXOREQUAL
   TOK_MINUSEQUAL TOK_SHIFTLEFTEQUAL 
   TOK_SHIFTRIGHTEQUAL TOK_BITWISEANDEQUAL TOK_BITWISEOREQUAL
}

//
// expression operator token classes
//
#tokclass UNARYOP_TOKCLASS {
   TOK_AMPERSAND TOK_STAR TOK_PLUS TOK_MINUS TOK_TILDE TOK_NOT
   TOK_PLUSPLUS TOK_MINUSMINUS
}
#tokclass ADDITIVE_TOKCLASS { TOK_PLUS TOK_MINUS }
#tokclass EQUALITY_TOKCLASS { TOK_NOTEQUAL TOK_EQUAL }
#tokclass MULTIPLICATIVE_TOKCLASS { TOK_STAR TOK_DIVIDE TOK_MOD }
#tokclass PM_TOKCLASS { TOK_DOTMBR TOK_POINTERTOMBR }
#tokclass RELATIONAL_TOKCLASS {
   TOK_LESSTHAN TOK_GREATERTHAN
   TOK_LESSTHANOREQUALTO TOK_GREATERTHANOREQUALTO
}
#tokclass SHIFT_TOKCLASS { TOK_SHIFTLEFT TOK_SHIFTRIGHT }

// token class for literal tokens
#tokclass CONSTANT_TOKCLASS {
   TOK_OCTALINT TOK_DECIMALINT TOK_HEXADECIMALINT TOK_CHARACTER
   TOK_FLOATONE TOK_FLOATTWO  TOK_BOOLTRUE TOK_BOOLFALSE
}

// token class for builtin type specifiers
#tokclass BUILTIN_TOKCLASS {
   TOK_VOID TOK_CHAR TOK_WCHAR_T TOK_SHORT TOK_INT
   TOK_LONG TOK_FLOAT TOK_DOUBLE TOK_SIGNED TOK_UNSIGNED
   TOK_BOOL
   TOK_MSEXT_INT8 TOK_MSEXT_INT16 TOK_MSEXT_INT32 TOK_MSEXT_INT64
}

// token class for new cast operators
#tokclass CAST_TOKCLASS { 
   TOK_CONST_CAST TOK_DYNAMIC_CAST TOK_REINTERPRET_CAST TOK_STATIC_CAST
}

// Token class for curly braces of either type
#tokclass CURLY_SEMICOLON_TOKCLASS { TOK_LCURLY TOK_RCURLY TOK_SEMICOLON }
#tokclass CURLY_TOKCLASS { TOK_LCURLY TOK_RCURLY }
#tokclass RCURLY_EOF_TOKCLASS { TOK_RCURLY TOK_Eof }

// Token classes for error synchronization
#tokclass EXTERNAL_SYNC_TOKCLASS { TOK_RCURLY TOK_SEMICOLON TOK_Eof }
#tokclass EXTERNAL_SYNC2_TOKCLASS { TOK_RCURLY }
#tokclass MEMBER_SYNC_TOKCLASS { TOK_RCURLY TOK_SEMICOLON TOK_Eof }
#tokclass STATEMENT_SYNC_TOKCLASS { TOK_LCURLY TOK_RCURLY TOK_SEMICOLON TOK_Eof }

// Token class for tokens ambiguous between declaration-specifiers and declarators
#tokclass DECLSPEC_DECLARATOR_AMBIG_TOKCLASS { TOK_SCOPE TOK_TYPENAME TOK_ID }


/////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////
//  Parser class definition
/////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////

class CPPParser {
<<
   #include "CPPParserMembers.h"

protected:
   #define PARSER_PURE_VIRTUAL = 0
   #include "CPPParserVirtual.h"
   #undef PARSER_PURE_VIRTUAL
>>

/////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////
//  GRAMMAR: rules organized alphabetically
/////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////
// Rule: access_specifier.  Matches one of public/protected/private.
// Inputs: none
// Outputs:
//    CPPConst::AccessSpecifier   as   The access-specifier
/////////////////////////////////////////////////////////////////////////
access_specifier > [ CPPConst::AccessSpecifier as ] :
   << $as = CPPConst::asNONE; >>
      TOK_PUBLIC     << $as = CPPConst::asPUBLIC; >>
   |  TOK_PROTECTED  << $as = CPPConst::asPROTECTED; >>
   |  TOK_PRIVATE    << $as = CPPConst::asPRIVATE; >>
   ;

/////////////////////////////////////////////////////////////////////////
// Rule: base_clause.  Matches a list of base-class specifiers in a class
//    definition.  It will add the matched base classes to the list
//    of base classes for the class being defined.
//
// Inputs:
//    CPPTypeRefI   classType   The type of the class being defined
// Outputs: none
/////////////////////////////////////////////////////////////////////////
base_clause [ CPPTypeRefI classType ] :
   <<;>>
   <<;>>
   TOK_COLON 
   base_specifier [ $classType ]
   (
      TOK_COMMA 
      base_specifier [ $classType ]
   )*
   ;

/////////////////////////////////////////////////////////////////////////
// Rule: base_specifier.  Matches a base-class specifier in a class
//    definition.  It will add the matched base class to the list
//    of base classes for the class being defined.
//
// Inputs:
//    CPPTypeRefI   classType   The type of the class being defined
// Outputs: none
/////////////////////////////////////////////////////////////////////////
base_specifier [ CPPTypeRefI classType ] :
   << 
      CPPConst::BaseSpecifier bs;
      CPPTypeRefI baseClassType;
      CPPScopeRefI scope;
   >>
   <<;>>
   (
         TOK_PUBLIC      << bs |= CPPConst::bsPUBLIC; >>
      |  TOK_PROTECTED   << bs |= CPPConst::bsPROTECTED; >>
      |  TOK_PRIVATE     << bs |= CPPConst::bsPRIVATE; >>
      |  TOK_VIRTUAL     << bs |= CPPConst::bsVIRTUAL; >>
   )*
   qualified_type > [ baseClassType ]
   <<  AddBaseToClass($classType, baseClassType, bs); >>
   ;

/////////////////////////////////////////////////////////////////////////
// Rule: builtin_type_specifier.  Matches one of the built-in type
//      specifiers (int/float/bool/etc).
// Outputs:
//    CPPConst::TypeSpecifier  ts  One of the type specifier flags.
/////////////////////////////////////////////////////////////////////////
builtin_type_specifier > [CPPConst::TypeSpecifier ts] :
   << $ts = CPPConst::tsVOID;>>
   <<;>>
   tok:BUILTIN_TOKCLASS
   <<
      switch ($tok->getType())
      {
      case TOK_VOID:
         $ts = CPPConst::tsVOID;
         break;
      case TOK_CHAR:
         $ts = CPPConst::tsCHAR;
         break;
      case TOK_WCHAR_T:
         $ts = CPPConst::tsWCHAR_T;
         break;
      case TOK_SHORT:
         $ts = CPPConst::tsSHORT;
         break;
      case TOK_INT:
         $ts = CPPConst::tsINT;
         break;
      case TOK_LONG:
         $ts = CPPConst::tsLONG;
         break;
      case TOK_FLOAT:
         $ts = CPPConst::tsFLOAT;
         break;
      case TOK_DOUBLE:
         $ts = CPPConst::tsDOUBLE;
         break;
      case TOK_SIGNED:
         $ts = CPPConst::tsSIGNED;
         break;
      case TOK_UNSIGNED:
         $ts = CPPConst::tsUNSIGNED;
         break;
      case TOK_BOOL:
         $ts = CPPConst::tsBOOL;
         break;

      // Compiler-specific extensions
      case TOK_MSEXT_INT8:
         $ts = CPPConst::tsCHAR;
         break;
      case TOK_MSEXT_INT16:
         $ts = CPPConst::tsSHORT;
         break;
      case TOK_MSEXT_INT32:
         $ts = CPPConst::tsLONG;
         break;
      case TOK_MSEXT_INT64:
         $ts = CPPConst::tsINT64;
         break;
      default:
         assert(0);
      }
   >>
   ;

/////////////////////////////////////////////////////////////////////////
// Rule: class_specifier.  Match things like "class A", "class A {}", and
// "struct A : public B {};
//
// Inputs: none
// Outputs: 
//   CPPTypeRefI               classType    The type for the class, or
//                                          a "void" type if an invalid
//                                          class name was specified.
//   CPPConst::TypeSpecifier   ts           specifies class/struct/union
/////////////////////////////////////////////////////////////////////////
class_specifier > [ CPPTypeRefI classType, CPPConst::TypeSpecifier ts ] :
   << 
      ScopePopper scopePopper(this); 
      $classType = (CPPType*)0;
      $ts = CPPConst::tsNONE;
      bool dummy;
   >>
   <<;>>
   (
         TOK_STRUCT <<$ts = CPPConst::tsSTRUCT;>>
      |  TOK_UNION  <<$ts = CPPConst::tsUNION;>>
      |  TOK_CLASS  <<$ts = CPPConst::tsCLASS;>>
   )

   // Allow stuff like "class A<int>::B::C"
   scope_override > [ dummy ]

   (
         // This alternate handles either a named class definition like:
         //     class A : public B {... }
         // or a named class declaration, potentially templated like:
         //     class A
         //     class A<int>

         id:TOK_ID

         (
         <<;>>  // so next action binds to alternate

               // Named class definition or declaration
               << JLStr className; >>

               // Optional template specialization arguments
               << 
                  TemplateArgumentListRefI argList; 
                  bool haveArgList = false;
               >>
               {
                  <<;>>
                  // This is tricky, because the ID must be looked up
                  // with the current scope stack, but the
                  // template arguments must be looked up in the
                  // scope context designated by the scopeState argument.
                  // So we save the scope state and restore it.
                  <<
                     bool tmpBool;
                     ScopeSaver scopeSaver(this, scopePopper.Get());
                     scopePopper.Restore();
                     haveArgList = true;
                  >>

                  TOK_LESSTHAN 
                  template_argument_list > [ argList, tmpBool ]
                  TOK_GREATERTHAN

                  <<
                     scopeSaver.Restore();
                     scopeSaver.Commit();
                  >>
               }


               (
               <<;>>

                     // Definition
                     << 
                        JLStr tmpName = $id->getText();
                        if (haveArgList)
                        {
                           tmpName = FormatClassSpecializationName(tmpName, argList);
                        }

                        CPPDeclRefI tmpDecl;
                        DefineClassType(
                           tmpName,
                           $ts, 
                           GetCurrentScope(), 
                           $id, 
                           $classType,
                           tmpDecl,
                           className
                        );
                     >>

                     { 
                     <<;>>
                        // It can't be anything but a class_spec if we get here
                        // Warning: there is an implicit assumption that:
                        // 1) class_specifier is always inside (...)?, OR
                        // 2) the class_specifier is *not* part of a larger enclosing (...)?
                        << CommitGuess(); >>
                        base_clause [ $classType ]
                     }

                     TOK_LCURLY
                     // It can't be anything but a class_spec if we get here
                     // Warning: there is an implicit assumption that:
                     // 1) class_specifier is always inside (...)?, OR
                     // 2) the class_specifier is *not* part of a larger enclosing (...)?
                     << CommitGuess(); >>

                     <<
                        // enter new class scope
                        EnterNamedScope(className, false);
                     >>
                     (
                        declaration_member
                        exception <<
                           ReportExceptionError(ex);
                           consume();
                           consumeUntil(MEMBER_SYNC_TOKCLASS_set);
                        >>
                     )*
                     TOK_RCURLY
                  |
                     // Declaration
                     <<
                        if (haveArgList)
                        {
                           // Class template specialization
                           LookupSpecialization(
                              $id->getText(),
                              argList,
                              $id,
                              className,
                              $classType
                           );
                        } else {
                           // Non-templated class declaration
                           // Look up the class declaration if it already exists, or create 
                           // an (incomplete) class declaration if it does not exist.
                           // scope in which the decl is made is only specified if is an override
                           CPPScopeRefI declScope;
                           if (scopePopper.Get() != SaveScope())
                           {
                              declScope = GetCurrentScope();
                           }
                           CPPDeclRefI tmpDecl;
                           DeclareClassType(
                              $id->getText(), 
                              $ts, 
                              declScope,
                              $id, 
                              $classType, 
                              tmpDecl,
                              className
                           );
                        }
                     >>
               )
         )

      |

         // Unnamed class definition, not templated
         lcurly:TOK_LCURLY
         <<
            JLStr className;
            CPPDeclRefI tmpDecl;
            DefineClassType(
               "", 
               $ts, 
               GetCurrentScope(),
               $lcurly, 
               $classType, 
               tmpDecl,
               className
            ); 
            // enter new (unnamed) class scope
            EnterNamedScope(className, false);
         >>
         (
            declaration_member
            exception <<
               ReportExceptionError(ex);
               consume();
               consumeUntil(MEMBER_SYNC_TOKCLASS_set);
            >>
         )*
         TOK_RCURLY
   )
   ;

/////////////////////////////////////////////////////////////////////////
// Rule: class_specifier_simple.  Match a class specifier without a 
// definition (i.e., no base classes, no body).
//
// Inputs: none
// Outputs: 
//   CPPTypeRefI               classType    The type for the class, or
//                                          a "void" type if an invalid
//                                          class name was specified.
//   CPPConst::TypeSpecifier   ts           specifies class/struct/union
/////////////////////////////////////////////////////////////////////////
class_specifier_simple > [ CPPTypeRefI classType, CPPConst::TypeSpecifier ts ] :
   << 
      ScopePopper scopePopper(this); 
      bool dummy;
   >>
   <<;>>
   (
         TOK_STRUCT <<$ts = CPPConst::tsSTRUCT;>>
      |  TOK_UNION  <<$ts = CPPConst::tsUNION;>>
      |  TOK_CLASS  <<$ts = CPPConst::tsCLASS;>>
   )

   // Allow stuff like "class A<int>::B::C"
   scope_override > [ dummy ]

   id:TOK_ID

   (
         // This predicate resolves an ambiguity arising from things like:
         //     operator class A < X 
         // where it is syntactically legal to use the operator function
         // address in a relational expression.  However, this is semantically
         // invalid, so we are OK in always binding the "<" to the class
         (TOK_LESSTHAN)? => <<1>>?

         (
            << 
               JLStr specializedClassName; 
               bool tmpBool;
            >>
            // Templated named class specifier (class specialization).
            template_specialization 
               [ $id->getText(), scopePopper.Get(), $id ] > 
               [ specializedClassName, $classType, tmpBool ]
         )
      |
         // Untemplated named class specifier.  
         // Look up the class declaration if it already exists, or create 
         // an (incomplete) class declaration if it does not exist.
         << 
            // scope in which the decl is made is only specified if is an override
            CPPScopeRefI declScope;
            if (scopePopper.Get() != SaveScope())
            {
               declScope = GetCurrentScope();
            }
            JLStr tmpName;
            CPPDeclRefI tmpDecl;
            DeclareClassType(
               $id->getText(), 
               $ts, 
               declScope,
               $id, 
               $classType, 
               tmpDecl,
               tmpName
            );
         >>
   )
   ;

/////////////////////////////////////////////////////////////////////////
// Rule: compiler_specific_calling_convention.  Match an optional 
// non-standard calling convention modifier.
//
// Outputs:
//    CPPConst::CallingConvention cc    The calling convention modifier.
//
// Note: This gets used in a number of places, not all of which are
// semantically correct.  It should be checked that this does not get
// used for data-declarators.
/////////////////////////////////////////////////////////////////////////
compiler_specific_calling_convention > [ CPPConst::CallingConvention cc ] :
   <<;>>
   <<;>>
   (
         TOK_MSEXT_CDECL
         << $cc = CPPConst::ccCDECL; >>
      |  
         TOK_MSEXT_FASTCALL
         << $cc = CPPConst::ccMS_FASTCALL; >>
      |  
         TOK_MSEXT_STDCALL
         << $cc = CPPConst::ccMS_STDCALL; >>
      |  
         // empty -- default is same as __cdecl
         << $cc = CPPConst::ccCDECL; >>
   )
   ;

///////////////////////////////////////////////////////////////////////
// Rule: constant.  Matches scalar constant values.
// Outputs:
//    ANTLRTokenPtr   tokOut   The constant token
///////////////////////////////////////////////////////////////////////
constant > [ ANTLRTokenPtr tokOut ] :
   << $tokOut = (ANTLRAbstractToken*)0; >>
   <<;>>
   tok:CONSTANT_TOKCLASS
   << $tokOut = $tok; >>
   ;

/////////////////////////////////////////////////////////////////////////
// Rule ctor_declarator:  Matches the declarator for a constructor,
//    which is like a normal function declarator. but has an optional
//    trailing member initializer list.
//
// Inputs: none
// Outputs:
//    JLStr           declName     The name of the declarator.
//    StrListRefI     argNameList  The argument names for a function declarator.
//                                 If there is no function suffix, then this
//                                 is a zero-ref.
//    CPPScopeRefI    declScope    The scope in which the declarator should
//                                 be declared.  For normal declarations, this
//                                 is always the current scope.  It may be a
//                                 different scope for explicitly-qualified
//                                 definitions.
//    CPPScopeRefI    argScope     If not nil, this scope contains the function
//                                 parameter declarations.
//    CPPTypeRefI     outputType   The type of the ctor 
//    bool            usedInnerTemplateParms 
//                                 true if the scope override used any of
//                                 the template parameters from the
//                                 innermost template head.
/////////////////////////////////////////////////////////////////////////
ctor_declarator > 
[ 
   JLStr declName, 
   StrListRefI argNameList,
   CPPScopeRefI declScope,
   CPPScopeRefI argScope,
   CPPTypeRefI outputType,
   bool usedInnerTemplateParms 
] :

   << 
      CPPConst::CallingConvention cc;
      ScopePopper scopePopper(this);
   >>
   <<;>>
   compiler_specific_calling_convention > [ cc ]
   (
         (
            qualified_id [ false, true, false ] > [ $declName, $usedInnerTemplateParms ]

            // Validation predicate: this cannot be a ctor unless the
            // name of the qualified_id matches the name of the enclosing
            // class scope.  We use this predicate to bail out of the ctor
            // alternative as soon as possible.
            << IsEnclosingClassName($declName) >>?

            <<
               // If there was a scope override in the qualified id, then
               // save the overridden scope, pop the scope stack, and re-enter
               // the scope, pulling in its context (namespace/base classes).
               $declScope = GetCurrentScope();
               if (SaveScope() != scopePopper.Get())
               {
                  scopePopper.Restore();
                  EnterScopeContext($declScope, false);
               }
            >>

            // If we get to here and the next token is a "(", 
            // then we know it ctor and we can commit to the ctor alt.
            // Warning: there is an implicit assumption that:
            // 1) ctor decl is always inside (...)?, OR
            // 2) the ctor decl is *not* part of a larger enclosing (...)?
            << 
               if (LA(1) == TOK_LPAREN) 
               {
                  CommitGuess(); 
               }
            >>

            // ctor is entered as function returning 'void'
            // TODO: use an "undefined" type?
            declarator_suffix_function 
               [ LookupVoidType(), CPPConst::ccCDECL ] > 
               [ $argNameList, $argScope, $outputType ]
         )
      |
         // Allow for "class A { (A()); };"
         TOK_LPAREN 
         ctor_declarator > 
            [ $declName, $argNameList, $declScope, $argScope, $outputType, $usedInnerTemplateParms]
         TOK_RPAREN
   )
   ;


/////////////////////////////////////////////////////////////////////////
// Rule ctor_initializer:  Matches the initializer list for a constructor.
/////////////////////////////////////////////////////////////////////////
ctor_initializer :
   <<;>>
   <<;>>
   TOK_COLON 
   ctor_member_init 
   (
      TOK_COMMA 
      ctor_member_init
   )*
   ;

/////////////////////////////////////////////////////////////////////////
// Rule: ctor_member_init.  Matches a member initializer for a
// constructor, like 
//   A::A() : 
//      x(1)   <-- member initializer
/////////////////////////////////////////////////////////////////////////
ctor_member_init :
   << 
      ScopePopper scopePopper(this);
      JLStr memberInitName;
      bool tmpBool;
   >>
   <<;>>

   // Must allow types for qualified_id because of base class initialization
   qualified_id [ true, false, false ] > [ memberInitName, tmpBool ]

   << scopePopper.Restore(); >>
   TOK_LPAREN 
   {
      expression_list
   } 
   TOK_RPAREN
   ;

/////////////////////////////////////////////////////////////////////////
// Rule: cv_qualifier.  Matches one of 'const' or 'virtual'.
// Inputs: none
// Outputs:
//    CPPConst::CVQualifier  cv    The const/virtual qualifier
/////////////////////////////////////////////////////////////////////////
cv_qualifier > [CPPConst::CVQualifier cv] :
   <<;>>
   <<;>>
      TOK_CONST      <<$cv = CPPConst::cvCONST;>>
   |  TOK_VOLATILE   <<$cv = CPPConst::cvVOLATILE;>>
   ;

/////////////////////////////////////////////////////////////////////////
// Rule: cv_qualifier_seq.  Matches zero-or-more 'const' or 'virtual'.
// Inputs: none
// Outputs:
//    CPPConst::CVQualifier  cv    cvNONE, or an ORed combination of
//                                 cvCONST or cvVOLATILE
/////////////////////////////////////////////////////////////////////////
cv_qualifier_seq > [CPPConst::CVQualifier cv] : 
   << $cv = CPPConst::cvNONE; >>
   <<;>>
   (
         TOK_CONST      <<$cv |= CPPConst::cvCONST;>>
      |  TOK_VOLATILE   <<$cv |= CPPConst::cvVOLATILE;>>
   )*
   ;

/////////////////////////////////////////////////////////////////////////
// Rule declaration_external:  Match an external declaration.
// Inputs: none
// Outputs: none
/////////////////////////////////////////////////////////////////////////
declaration_external :
   <<;>> 
   <<;>>
   (
         // Special hack for MSVC++ headers, which contain:
         //     typedef unsigned short wchar_t
         //     typedef int bool
         // Harmless elsewhere.
         <<GetOption(OptMSTypedefHack)>>?
         (
            TOK_TYPEDEF
            (
                  TOK_UNSIGNED TOK_SHORT TOK_WCHAR_T
               |
                  TOK_INT TOK_BOOL
            )
            TOK_SEMICOLON
         )?
      |
         // There is an LL(1) ambiguity here regarding TOK_EXTERN.
         // because declarations can also start with TOK_EXTERN.
         // This predicate looks for a linkage spec iff the next token
         // is TOK_STRING
         (TOK_EXTERN)? => <<LT(2)->getType() == TOK_STRING>>?

         // like extern "C" ...
         linkage_specification
      |
         declaration_external_or_member [ false ]
      |
         pragma_line
   )
   ;


/////////////////////////////////////////////////////////////////////////
// Rule: declaration_external_or_member:  Matches common elements of 
//   declaration_member and declaration_external.
//
// Inputs:
//    bool    memberDecl     true if the declaration is inside a class
//
// Note: JEL has rewritten this considerably, opting for a
// grammar that allows many invalid constructs that are
// to be checked semantically.  This makes the grammar
// simpler, and changes syntax errors into semantic errors,
// which are easier to diagnose.  In addition, it gets rid
// of some backtracking.   The grammar for declarations
// is extensively left-factored, so dissimilar rules may be
// collected under one alternative.
/////////////////////////////////////////////////////////////////////////
declaration_external_or_member [ bool memberDecl ] :
   << ScopePopper scopePopper(this); >> 
   << ; >>

      template_declaration [ GetCurrentScope(), memberDecl ]

   |
      // "Normal" declarations
      // Start with storage class and/or declaration modifiers
      <<
         CPPConst::StorageClass sc;
         CPPConst::DeclModifier dm;
      >>
      declaration_sc_or_dm_seq > [ sc, dm ]
      
      (
            // The leading decl-specs are confusing in LL with the class name
            // of the ctor, for example:
            //     class A {
            //        A();   // ctor
            //        (A()); // ctor
            //        A* ap; // data member
            // So backtrack out of this if it fails. 
            // We cannot do the decl-specs first because:
            //     class A { A(X); };
            // looks like a declaration of X as type A
            (
               // Ctor declaration or definition.

               << 
                  JLStr ctorName;
                  CPPDeclRefI ctorDecl; 
                  CPPScopeRefI declScope;
                  CPPScopeRefI argScope;
                  CPPTypeRefI ctorType;
                  bool tmpBool;
                  StrListRefI argNameList;
               >>

               ctor_declarator > 
                  [ ctorName, argNameList, declScope, argScope, ctorType, tmpBool ]

               <<
                  // DWP says that a ctor is not entered into the class scope under
                  // any name that can be searched.  So we use "$ctor".
                  ctorDecl = DeclareObjectOrTypedef(
                     "$ctor",
                     ctorType,
                     sc,
                     dm,
                     CPPExprRefI(),
                     declScope,
                     argNameList
                  );
               >>

               (
                     // ctor definition
                     (
                        <<
                           if (declScope != GetCurrentScope())
                           {
                              EnterScopeContext(declScope, false);
                           }
                           EnterScope(argScope, false);
                        >>
                        { ctor_initializer }
                        // Choose generic_block if this is inside a class,
                        // because not all the fields have been seen yet.
                        (
                              (TOK_LCURLY)? => <<memberDecl>>?
                              // We don't actually parse the function body
                              // until later to get the member definitions.
                              generic_block
                              // TODO: use the collected tokens
                           |
                              statement_compound
                        )
                     )
                  |
                     // ctor declaration
                     TOK_SEMICOLON
               )
            )?
         |
            // Non-templated declarations and definitions.
            // This includes functions, operators and dtors.
            (
               <<;>>
               << CPPTypeRefI declSpecType; >>

               // Optional declaration specifiers
               (
                     // Backtrack out of decl-specs; it's too hard to write
                     // a predicate that parses template specifiers.  However,
                     // we should only backtrack for the ambiguous tokens.
                  
                     // Alt for non-ambiguous declaration-specifiers to get
                     // non-backtracking error-reporting for most cases.
                     // The warning produced here is OK.
                     <<!set_el(LA(1), DECLSPEC_DECLARATOR_AMBIG_TOKCLASS_set)>>?
                     declaration_specifiers [ sc, dm ] > [ declSpecType, sc, dm ]
                  |
                     // Alt for cases that we must backtrack out of
                     (
                        declaration_specifiers [ sc, dm ] > [ declSpecType, sc, dm ]
                     )?
                  |
                     // empty
                     // The decl-spec type defaults to "void".  This is only
                     // valid for ctor/dtor/operator.  Old-style functions should be
                     // caught with a warning.  Other declarations are disallowed.
                     <<declSpecType = LookupVoidType();>>
               )

               (
                     // No declarators at all.
                     // If there were declaration specifiers. then this is
                     // only valid if they were class decl/def like 
                     // "class A { ... }" or enum decl/def like "enum E {...}".
                     // If there are no declaration-specifiers, then this is an 
                     // empty declaration.
                     TOK_SEMICOLON
                  |
                     // A function declarator with body, or a declarator-init-list
                     declarator_function_or_init_list [ declSpecType, sc, dm, memberDecl ]
               )

            )
      )
   ;
   // end declaration_external_or_member

/////////////////////////////////////////////////////////////////////////
// Rule: declaration_member.  Matches declarations that are allowed
// in struct/class/union.
//
// Inputs: none
// Outputs: none
/////////////////////////////////////////////////////////////////////////
declaration_member :
   <<;>>
   <<;>>
   (
         declaration_external_or_member [ true ]
      |
         (
            << CPPConst::AccessSpecifier as; >>
            access_specifier > [as]
            << SetCurrentScopeAccess(as); >>
            TOK_COLON
         )
      |
         pragma_line
   )
   ;

/////////////////////////////////////////////////////////////////////////
// Rule declaration_modifier: match any one of the possible
// declaration modifiers.  Semantic checks must be done to
// determine validity of the modifier with a particular declaration.
//
// Inputs: none
// Outputs:
//    CPPConst::DeclModifier   dm  The flag of the matched declaration 
//                                 modifier.
/////////////////////////////////////////////////////////////////////////
declaration_modifier > [ CPPConst::DeclModifier dm ] :
   <<;>>
   <<;>>
      TOK_INLINE        << $dm = CPPConst::dmINLINE; >>
   |  TOK_VIRTUAL       << $dm = CPPConst::dmVIRTUAL; >>
   |  TOK_FRIEND        << $dm = CPPConst::dmFRIEND; >>
   |  TOK_EXPLICIT      << $dm = CPPConst::dmEXPLICIT; >>
   |  TOK_MSEXT_INLINE  << $dm = CPPConst::dmINLINE; >>
   ;


/////////////////////////////////////////////////////////////////////////
// Rule declaration_sc_or_dm_seq: match a possibly-empty sequence
// of storage class and declaration modifiers
//
// Inputs: none
// Outputs:
//    CPPConst::StorageClass   sc  Any storage class
//    CPPConst::DeclModifier   dm  Any declaration modifiers
/////////////////////////////////////////////////////////////////////////
declaration_sc_or_dm_seq > [ CPPConst::StorageClass sc, CPPConst::DeclModifier dm ] :
   <<
      CPPConst::StorageClass sc2;
      CPPConst::DeclModifier dm2;
   >>
   (
         // auto/register/static/extern/typedef
         storage_class_specifier > [sc2]
         << $sc |= sc2; >>
      |  
         // inline/virtual/friend
         declaration_modifier > [ dm2 ]
         << $dm |= dm2; >>
   )*
   ;

/////////////////////////////////////////////////////////////////////////
// Rule: declaration_specifiers.  Match the part of the declaration
// that specifies the fundamental type of the declaration.
//
// JEL changed this to disallow the worst incompatible
// type sequences.  For example, "class A {} int" should not
// be allowed.  More importantly, we must be able to parse:
//       typedef struct foo {} foo;
// And not drag in the second "foo" as part of the type because
// this makes things parse incorrectly.  Be careful modifying this;
// it took a long time to get right.
// 
// Inputs:
//    CPPConst::StorageClass  scin    The storage class (auto/register/etc).
//    CPPConst::DeclModifier  dmin    The declaration modifiers (inline/friend/etc).
// Outputs:
//    CPPTypeRefI             type    A pointer to the type entry defined
//                                    by the declaration specifiers.
//    CPPConst::StorageClass  scout   The augmented storage class
//    CPPConst::DeclModifier  dmout   The augmented declaration modifiers
/////////////////////////////////////////////////////////////////////////
declaration_specifiers 
[ CPPConst::StorageClass scin, CPPConst::DeclModifier dmin ] > 
[ CPPTypeRefI type, CPPConst::StorageClass scout, CPPConst::DeclModifier dmout ] :
   <<
      CPPConst::TypeSpecifier ts = CPPConst::tsNONE;
      CPPConst::TypeSpecifier ts2;
      CPPConst::CVQualifier cv = CPPConst::cvNONE;
      CPPConst::CVQualifier cv2;
      CPPConst::DeclModifier dm2;
      $scout = $scin;
      $dmout = $dmin;
   >>
   <<;>>

   // Start with an optional series of cv-qualifiers, but not storage class
   // or declaration_modifier because that is picked up elsewhere
   (
      // const/volatile
      cv_qualifier > [cv2]
      << cv |= cv2; >>
   )*

   // Next, match one of the mutually-exclusive parts
   // along with possibly more qualifiers.
   (
         (
            << ANTLRTokenPtr tok = LT(1); >>
            // Match a sequence of built-in types interspersed with qualifiers.
            builtin_type_specifier > [ts2]
            << ts |= ts2; >>

            (
                  // like int,long,double,bool
                  builtin_type_specifier > [ts2]
                  << ts |= ts2; >>
               |
                  declaration_specifiers_qualifier [cv, $scout, $dmout ] > [ cv, $scout, $dmout ]
            )*

            << 
               $type = LookupBuiltinType(ts, tok);
            >>
         )
      |
         // Match class/type/enum, of which there may only be one
         (
               // Any named typedef/class/enum like A, E, A<T1>::B, etc
               qualified_type > [ $type ]
               << ts = CPPConst::tsTYPEID; >>
            |  
               // like "class A", "struct B", "class C {...}", or "class {...}"
               class_specifier > [ $type, ts ]
            |  
               // like "enum E", "enum E {...}" or "enum {...}"
               enum_specifier > [ $type ]
               << ts = CPPConst::tsENUM; >>
         )

         // class/type/enum may be followed by yet more qualifiers
         (
            declaration_specifiers_qualifier [cv, $scout, $dmout ] > [ cv, $scout, $dmout ]
         )*
   )

   // Process the const/volatile decl-specs
   << $type = LookupCVQualifierOfType($type, cv); >>
   ;


/////////////////////////////////////////////////////////////////////////
// Rule: declaration_specifiers_qualifier.  Match the part of the 
// declaration specifiers that is just a qualifier
//
// Inputs:
//    CPPConst::CVQualifier   cvin    Any const/volatile qualifiers
//    CPPConst::StorageClass  scin    The storage class (auto/register/etc).
//    CPPConst::DeclModifier  dmin    The declaration modifiers (inline/friend/etc).
// Outputs:
//    CPPConst::CVQualifier   cvout   Any const/volatile qualifiers
//    CPPConst::StorageClass  scout   The storage class (auto/register/etc).
//    CPPConst::DeclModifier  dmout   The declaration modifiers (inline/friend/etc).
/////////////////////////////////////////////////////////////////////////
declaration_specifiers_qualifier
[ CPPConst::CVQualifier cvin, CPPConst::StorageClass scin, CPPConst::DeclModifier dmin ] >
[ CPPConst::CVQualifier cvout, CPPConst::StorageClass scout, CPPConst::DeclModifier dmout ] :
   <<
      CPPConst::CVQualifier cv2;
      CPPConst::StorageClass sc2;
      CPPConst::DeclModifier dm2;
      // Initialize results
      $scout = $scin;
      $dmout = $dmin;
      $cvout = $cvin;
   >>
   <<;>>
      // auto/register/static/extern/typedef
      storage_class_specifier > [sc2]
      << $scout |= sc2; >>
   |  
      // const/volatile
      cv_qualifier > [cv2]
      << $cvout |= cv2; >>
   |  
      // inline/virtual/friend
      declaration_modifier > [ dm2 ]
      << $dmout |= dm2; >>
   ;

/////////////////////////////////////////////////////////////////////////
// Rule: declaration_specifiers_simple.  Match the part of the declaration
// that specifies the fundamental type of the declaration.  Like 
// declaration_specifiers, but does not allow enums or classes with
// explicit bodies like "enum {...}" or "class {...}"
//
// Inputs:
//    CPPConst::StorageClass  scin    The storage class (auto/register/etc).
//    CPPConst::DeclModifier  dmin    The declaration modifiers (inline/friend/etc).
// Outputs:
//    CPPTypeRefI             type    A pointer to the type entry defined
//                                    by the declaration specifiers.
//    CPPConst::StorageClass  scout   The augmented storage class 
//    CPPConst::DeclModifier  dmout   The augmented declaration modifiers
/////////////////////////////////////////////////////////////////////////
declaration_specifiers_simple
[ CPPConst::StorageClass scin, CPPConst::DeclModifier dmin ] > 
[ CPPTypeRefI type, CPPConst::StorageClass scout, CPPConst::DeclModifier dmout ] :
   <<
      CPPConst::TypeSpecifier ts = CPPConst::tsNONE;
      CPPConst::TypeSpecifier ts2;
      CPPConst::CVQualifier cv = CPPConst::cvNONE;
      CPPConst::CVQualifier cv2;
      CPPConst::DeclModifier dm2;
      $scout = $scin;
      $dmout = $dmin;
   >>
   <<;>>

   // Start with an optional series of cv-qualifiers, but not storage class
   // or declaration_modifier because that is picked up elsewhere
   (
      // const/volatile
      cv_qualifier > [cv2]
      << cv |= cv2; >>
   )*

   // Next, match one of the mutually-exclusive parts
   // along with possibly more qualifiers.
   (
         (
            << ANTLRTokenPtr tok = LT(1); >>
            // Match a sequence of built-in types interspersed with qualifiers.
            builtin_type_specifier > [ts2]
            << ts |= ts2; >>

            (
                  // like int,long,double,bool
                  builtin_type_specifier > [ts2]
                  << ts |= ts2; >>
               |
                  declaration_specifiers_qualifier [cv, $scout, $dmout ] > [ cv, $scout, $dmout ]
            )*

            << 
               $type = LookupBuiltinType(ts, tok);
            >>
         )
      |
         // Match class/type/enum, of which there may only be one
         (
               // Any named typedef/class/enum like A, E, A<T1>::B, etc
               qualified_type > [ $type ]
               << ts = CPPConst::tsTYPEID; >>
            |  
               // like "class A", "struct B", "class C {...}", or "class {...}"
               class_specifier_simple > [ $type, ts ]
            |  
               // like "enum E", "enum E {...}" or "enum {...}"
               enum_specifier_simple > [ $type ]
               << ts = CPPConst::tsENUM; >>
         )

         // class/type/enum may be followed by yet more qualifiers
         (
            declaration_specifiers_qualifier [cv, $scout, $dmout ] > [ cv, $scout, $dmout ]
         )*
   )

   // Process the const/volatile decl-specs
   << $type = LookupCVQualifierOfType($type, cv); >>
   ;


/////////////////////////////////////////////////////////////////////////
// Rule: declarator.  Matches a named declarator.
//
// Inputs:
//    CPPTypeRefI                  inputType  The type of the declaration specifiers.
//    CPPConst::CallingConvention  cc         The calling convention
// Outputs:
//    JLStr           declName     The name of the declarator.
//    StrListRefI     argNameList  The argument names for a function declarator.
//                                 If there is no function suffix, then this
//                                 is a zero-ref.
//    CPPScopeRefI    declScope    The scope in which the declarator should
//                                 be declared.  For normal declarations, this
//                                 is always the current scope.  It may be a
//                                 different scope for explicitly-qualified
//                                 definitions.
//    CPPScopeRefI    argScope     If not nil, this scope contains the function
//                                 parameter declarations.
//    CPPTypeRefI     outputType   The composite of the inputType and the
//                                 declarator's modifiers.
//    bool            usedInnerTemplateParms     
//                                 true if the scope override used any of
//                                 the template parameters from the
//                                 innermost template head.
/////////////////////////////////////////////////////////////////////////
declarator 
[ CPPTypeRefI inputType, CPPConst::CallingConvention cc ] >
[ 
   JLStr declName, 
   StrListRefI argNameList, 
   CPPScopeRefI declScope, 
   CPPScopeRefI argScope, 
   CPPTypeRefI outputType,
   bool usedInnerTemplateParms
] :

   << 
      $outputType = $inputType;
      CPPConst::CallingConvention cc2;
      ANTLRTokenPtr tok = LT(1);
   >>
   <<;>>

   compiler_specific_calling_convention > [ cc2 ]
   << 
      if (cc != CPPConst::ccCDECL && cc2 != CPPConst::ccCDECL)
      {
         ReportError("Calling convention specified more than once", tok);
      }
      cc = cc2;
   >>

   (
         // Backtrack to differentiate between normal declaration
         // and ptr-to-member qualifier on declaration:
         //    A::foo    // normal decl
         //    A::* foo  // ptr-to-member qualifier precedes decl
         // Since ptr-to-member is rare, make it the second alt.
         (
            declarator_direct 
               [ $outputType, cc ] > 
               [ $declName, $argNameList, $declScope, $argScope, $outputType, $usedInnerTemplateParms ]
         )?
      |
         declarator_ptr [ $outputType ] > [ $outputType ]
         declarator 
            [ $outputType, cc ] > 
            [ $declName, $argNameList, $declScope, $argScope, $outputType, $usedInnerTemplateParms ]
   )
   ;


////////////////////////////////////////////////////////////////////////////////////////
// Rule: declarator_direct.  Matches a named declarator without prefixes.
// This will match type names to handle odd cases like:
//    typedef struct S { } S;
// This will match ctors to handle ctor declarations.
//
// Inputs:
//    CPPTypeRefI             inputType    The type of the declaration specifiers
//                                         modified by any declarator prefixes.
//    CPPConst::CallingConvention  cc      The calling convention
// Outputs:
//    JLStr           declName     The name of the declarator.
//    StrListRefI     argNameList  The argument names for a function declarator.
//                                 If there is no function suffix, then this
//                                 is a zero-ref.
//    CPPScopeRefI    declScope    The scope in which the declarator should
//                                 be declared.  For most declarations, this
//                                 is the current scope.  It may be a
//                                 different scope for explicitly-qualified
//                                 definitions.
//    CPPScopeRefI    argScope     If not nul, this scope contains the function
//                                 parameter declarations.
//    CPPTypeRefI     outputType   The composite of the inputType and the
//                                 declarator's modifiers.
//    bool            usedInnerTemplateParms     
//                                 true if the scope override used any of
//                                 the template parameters from the
//                                 innermost template head.
////////////////////////////////////////////////////////////////////////////////////////
declarator_direct 
[ CPPTypeRefI inputType, CPPConst::CallingConvention cc ] > 
[ 
   JLStr declName, 
   StrListRefI argNameList, 
   CPPScopeRefI declScope, 
   CPPScopeRefI argScope, 
   CPPTypeRefI outputType,
   bool usedInnerTemplateParms
] :

   << 
      // Initialize values to something usable
      $outputType = $inputType;
      $declScope = GetCurrentScope(); 
      ScopePopper scopePopper(this); 
   >>
   << ; >>

   (
         // Must allow types and ctors
         qualified_id [ true, true, false ] > [ $declName, $usedInnerTemplateParms ]
         << 
            // If there was a scope override, then save the override scope,
            // pop the scope stack, and re-enter the override scope,
            // pulling in context (namespaces/base classes).
            if (SaveScope() != scopePopper.Get())
            {
               $declScope = GetCurrentScope();
               scopePopper.Restore(); 
               EnterScopeContext($declScope, false);
            }
         >>
         declarator_suffixes [ $outputType, cc ] > [ $argNameList, $argScope, $outputType ]
      |
         (
            << 
               CPPTypeRefI tmpType1;
               CPPTypeRefI tmpType2;
               StrListRefI argNameList2;
            >>
            // Parenthesized declarators are hard for LL parsers,
            // because we really need the type modification resulting
            // from the suffixes before parsing the sub-declarator.
            // Instead, we fake it by starting with a fake type entry,
            // applying the sub-declarator modifiers to it, and
            // then re-applying the collected type modifiers to the
            // "real" suffix-modified type.
            //
            // To further complicate matters, both the nested declarator
            // and the declarator_suffixes may declare a function suffix,
            // as in:
            //     int (*foo())(int)  // function returning pointer to function
            // The rule to use here is that if the nested declarator is
            // "nil", e.g.:
            //     int (foo)(int)
            // then use the exterior arguments, otherwise use the nested arguments.
            TOK_LPAREN
            << 
               tmpType1 = CreateTempType(); 
               // ignore nested-declarator function parms scope
               CPPScopeRefI tmpScope;  
            >>
            declarator 
               [ tmpType1, cc ] > 
               [ $declName, $argNameList, $declScope, tmpScope, tmpType2, $usedInnerTemplateParms ]

            << 
               if ($declScope != GetCurrentScope())
               {
                  EnterScopeContext($declScope, false);
               }
            >>
            TOK_RPAREN 
            declarator_suffixes [ $outputType, cc ] > [ argNameList2, $argScope, $outputType ]

            <<
               // Figure out where the function arguments come from
               if (tmpType2 == tmpType1)
               {
                  // exterior function suffix
                  $argNameList = argNameList2;
               }

               // Apply all of the type modifiers hanging off of tmpType1
               // to the outputType created by declarator_suffixes
               $outputType = MergeTypeModifiers($outputType, tmpType1);
            >>
         )
   )
   << scopePopper.Commit(); >>
   ;


/////////////////////////////////////////////////////////////////////////
// Rule: declarator_function_or_init_list.  Matches either a single
// declarator with a function body, a list of optionally-initialized
// declarators, or a single semicolon
//
// Inputs:
//    CPPTypeRefI             inputType  The type defined by the declaration specifiers.
//    CPPConst::StorageClass  sc         The storage class (auto/register/etc).
//    CPPConst::DeclModifier  dm         The declaration modifiers (inline/friend/etc).
//    bool                    memberDecl The declaration is inside a class
// 
// Note: We separate an initialized declarator
// list from a function definition.  The reason is that
// combining them yields wierd ambiguous situations like:
//    int foo() = new enum E { enumerators } 
// The ambiguity is betweeen this interpretation and 
//    int foo() = new enum E { function body} 
// To avoid this problem, we syntactically exclude something
// from being both initialized and having a function body.
////////////////////////////////////////////////////////////////////////
declarator_function_or_init_list
[ 
   CPPTypeRefI declSpecType, 
   CPPConst::StorageClass sc, 
   CPPConst::DeclModifier dm,
   bool memberDecl
] :

   << 
      ScopePopper scopePopper(this);
      CPPTypeRefI declaratorType; 
      CPPExprRefI initExpr;
      CPPScopeRefI declScope;
      CPPScopeRefI argScope;
      JLStr declName;
      StrListRefI argNameList;
      bool tmpBool;
   >>
   <<;>>

   // Left-factor: Both alternatives start with a single declarator
   declarator 
      [ $declSpecType, CPPConst::ccCDECL ] > 
      [ declName, argNameList, declScope, argScope, declaratorType, tmpBool ]


   (
         (
            // Declarator init list alternative
            // Note that this alt also handles pure-virtual functions.

            // Optional initializer for the first declarator
            {  
                  TOK_ASSIGNEQUAL 
                  initializer > [ initExpr ]
               |  
                  TOK_LPAREN 
                  expression_list > [ initExpr ] 
                  TOK_RPAREN
            }

            <<
               DeclareObjectOrTypedef(
                  declName,
                  declaratorType,
                  $sc,
                  $dm,
                  initExpr,
                  declScope,
                  argNameList
               );
            >>
         
            {
               // Process more optional init-declarators
               TOK_COMMA
               declarator_init_list [ $declSpecType, $sc, $dm ]
            }
            TOK_SEMICOLON
         )
      |
         // Declarator with function body.
         <<
            DeclareObjectOrTypedef(
               declName,
               declaratorType,
               $sc,
               $dm,
               CPPExprRefI(),
               declScope,
               argNameList
            );
         >>

         // Choose statement_compound for external declarations, but use
         // generic_block for member declarations because the fields have
         // not all been seen yet.
         (
               (TOK_LCURLY)? => <<memberDecl>>?
               // We don't actually parse the function body
               // until later to get the member definitions.
               generic_block
               // TODO: use the collected tokens
            |
               <<
                  if (declScope != GetCurrentScope())
                  {
                     // Re-enter the decl's scope override for the function body
                     EnterScopeContext(declScope, false);
                  }
                  EnterScope(argScope, false);
               >>
               statement_compound
         )
   )
   ;

//////////////////////////////////////////////////////////////////////////////////////
// Rule declarator_init.  Matches a declarator with initializers, and
//    enters the declaration in the current scope.
//
// Inputs:
//    CPPTypeRefI             inputType  The type defined by the declaration specifiers.
//    CPPConst::StorageClass  sc         The storage class (auto/register/etc).
//    CPPConst::DeclModifier  dm         The declaration modifiers (inline/friend/etc).
//////////////////////////////////////////////////////////////////////////////////////
declarator_init 
[ CPPTypeRefI inputType, CPPConst::StorageClass sc, CPPConst::DeclModifier dm  ] :

   << 
      CPPTypeRefI declaratorType; 
      CPPExprRefI initExpr;
      CPPScopeRefI declScope;
      CPPScopeRefI argScope;
      JLStr declName;
      StrListRefI argNameList;
      bool tmpBool;
   >>
   <<;>>

   declarator 
      [ $inputType, CPPConst::ccCDECL ] > 
      [ declName, argNameList, declScope, argScope, declaratorType, tmpBool ]

   {  
         TOK_ASSIGNEQUAL 
         initializer > [ initExpr ]
      |  
         TOK_LPAREN 
         expression_list > [ initExpr ]
         TOK_RPAREN
   }

   <<
      DeclareObjectOrTypedef(
         declName,
         declaratorType,
         sc,
         dm,
         initExpr,
         declScope,
         argNameList
      );
   >>
   ;

/////////////////////////////////////////////////////////////////////////
// Rule: declarator_init_list.  Matches a list of optionally-initialized
// declarators.
/////////////////////////////////////////////////////////////////////////
declarator_init_list 
[ CPPTypeRefI declSpecType, CPPConst::StorageClass sc, CPPConst::DeclModifier dm  ] :

   << ; >>
   << ; >>
   declarator_init [ declSpecType, sc, dm ]
   (
      TOK_COMMA
      declarator_init [ declSpecType, sc, dm  ]
   )*
   ;

////////////////////////////////////////////////////////////////////////
// Rule: declarator_opt_abstract. A declarator that is optionally an 
// abstract declarator (i.e., one with no declarator TOK_ID).  
// Used for function parameters.
//
// Inputs:
//    CPPTypeRefI  inputType    The type of the declaration specifiers.
// Outputs:
//    JLStr        declName     The name of the declarator, or a 
//                              blank string if the declarator is abstract.
//    CPPTypeRefI  outputType   The composite of the inputType and the
//                              declarator's modifiers.
////////////////////////////////////////////////////////////////////////
declarator_opt_abstract 
[ CPPTypeRefI inputType ] > [ JLStr declName, CPPTypeRefI outputType ] :

   << 
      StrListRefI dummyNameList;
      CPPConst::CallingConvention cc;
      ScopePopper scopePopper(this);
      $outputType = $inputType;
   >>
   <<;>>
   compiler_specific_calling_convention > [ cc ]
   (
         // The TOK_ID ambiguity here arises between cases like:
         //    A::*        // ptr-to-member declarator
         //    A<int>::*   // ptr-to-member declarator
         // and
         //    A           // normal declarator
         // This is an LL(1) ambiguity but not LL(2).
         // In order to be a declarator_ptr, the TOK_ID must be followed
         // by TOK_SCOPE or TOK_LESSTHAN

         (TOK_ID)? => <<LT(2)->getType() == TOK_SCOPE || LT(2)->getType() == TOK_LESSTHAN>>?
         declarator_ptr [ $outputType ] > [ $outputType ]
         declarator_opt_abstract  [ $outputType ] > [ $declName, $outputType ]
      |  
         // Only (...) and TOK_ID may be followed by function suffixes
         (
               (
                  << 
                     CPPTypeRefI tmpType1;
                     CPPTypeRefI dummyType;
                  >>
                  // See note under declarator_direct for what is going on here
                  TOK_LPAREN
                  << 
                     tmpType1 = CreateTempType(); 
                     CPPScopeRefI argScope;
                  >>
                  declarator_opt_abstract [ tmpType1 ] > [ $declName, dummyType ]
                  TOK_RPAREN 
                  (
                        declarator_suffix_function 
                           [ $outputType, cc ] > 
                           [ dummyNameList, argScope, $outputType ]
                     |
                        declarator_suffix_array [ $outputType ] > [ $outputType ]
                     |
                        /* no suffix */
                  )
                  <<
                     $outputType = MergeTypeModifiers($outputType, tmpType1);
                  >>
               )
            |
               // ID cannot be qualified for declarator_opt_abstract
               id:TOK_ID
               << 
                  $declName = $id->getText(); 
                  CPPScopeRefI argScope;
               >>

               (
                     declarator_suffix_function 
                        [ $outputType, cc ] > 
                        [ dummyNameList, argScope, $outputType ]
                  |
                     declarator_suffix_array [ $outputType ] > [ $outputType ]
                  |
                     /* no suffix */
               )
         )
      |  
         // This is just a plain array declarator, like
         //    int *[3]
         declarator_suffix_array [ $outputType ] > [ $outputType ]
      |
         /* empty declarator */
   )
   ;

/////////////////////////////////////////////////////////////////////////
// declarator_ptr:  Matches a pointer or reference.
//
// Inputs:
//    CPPTypeRefI    inputType    Type before modification by indirection.
// Outputs:
//    CPPTypeRefI    outputType   Type after modification by indirection.
/////////////////////////////////////////////////////////////////////////
declarator_ptr [ CPPTypeRefI inputType ] > [ CPPTypeRefI outputType ] :
   << CPPConst::CVQualifier cv; >>
   << $outputType = $inputType; >>
   (
         TOK_AMPERSAND
         << $outputType = LookupReferenceToType($inputType); >>
      |  
         TOK_STAR
         << $outputType = LookupPointerToType($inputType); >>
      |  
         declarator_ptr_to_member [ $outputType ] > [ $outputType ]
   )

   cv_qualifier_seq > [ cv ]
   << $outputType = LookupCVQualifierOfType($outputType, cv); >>
   ;

/////////////////////////////////////////////////////////////////////////
// declarator_ptr_to_member.  Match T::TT::*
//
// Inputs:
//    CPPTypeRefI    inputType    Type before modification by indirection.
// Outputs:
//    CPPTypeRefI    outputType   Type after modification by indirection.
/////////////////////////////////////////////////////////////////////////
declarator_ptr_to_member [ CPPTypeRefI inputType ] > [ CPPTypeRefI outputType ] :
   << 
      ScopePopper scopePopper(this);
      $outputType = $inputType;
      CPPTypeRefI classType;
      ANTLRTokenPtr errTok;
   >>
   << ; >>

   // This code is a bit nasty because we want to process the scope prefixes
   // as they are recognized, except that the last scope prefix must be a
   // class and we want to know what class it is.

   {
      TOK_SCOPE
      << EnterResolvedGlobalScope(); >>
   }

   << 
      errTok = LT(1); 
      bool tmpBool;
   >>
   scope_segment [ scopePopper.Get(), false, true ] > [ classType, tmpBool ]

   (
      // This predicate validates that the following tokens 
      // syntactically match a scope segment.
      << IsScopeSegment(1, true) >>?

      // template keyword indicates that id is a template
      << bool isTemplate = false; >>
      { TOK_TEMPLATE << isTemplate = true;>> }
      << errTok = LT(1); >>
      scope_segment [ scopePopper.Get(), isTemplate, true ] > [ classType, tmpBool ]
   )*

   TOK_STAR
   <<
      // It is possible for classType to be zero if the last scope_segment
      // specified a namespace instead of a scope.
      if (classType.get() == 0)
      {
         ReportError(JLStr("'") + errTok->getText() + "' is not a class name", errTok);
      } else {
         $outputType = LookupPointerToMemberOfType($inputType, classType, errTok);
      }
   >>
   ;


/////////////////////////////////////////////////////////////////////////
// Rule: declarator_suffix_array.  Matches array suffixes on a
// declarator, like "[ expr ]"
//
// Inputs:
//    CPPTypeRefI   inputType    The declarator type before encountering the
//                               array suffix.
// Outputs:
//    CPPTypeRefI   outputType   The declarator type after modification by the
//                               array suffix.
/////////////////////////////////////////////////////////////////////////
declarator_suffix_array [ CPPTypeRefI inputType ] > [ CPPTypeRefI outputType ] :
   << 
      $outputType = $inputType;
      CPPExprRefI exprDimension;
      ANTLRTokenPtr errTok = LT(1);
   >>
   << ; >>

   // Array dimensions.  The first dimension is an optional constant expression
   // but the following dimensions are not optional

   TOK_LSQUARE 
   { 
      << errTok = LT(1); >>
      expr_constant > [ exprDimension ]
   }
   TOK_RSQUARE
   << $outputType = LookupArrayOfType($outputType, exprDimension, errTok); >>

   (
      TOK_LSQUARE 
      << errTok = LT(1); >>
      (
            expr_constant > [ exprDimension ]
         |
            // Allow empty [] syntactically but report error semantically
            <<
               ReportError("empty [] only allowed for first dimension", errTok);
               exprDimension = MakeExprLiteral(TOK_DECIMALINT, "1");
            >>
      )
      TOK_RSQUARE
      << $outputType = LookupArrayOfType($outputType, exprDimension, errTok); >>
   )*
   ;

/////////////////////////////////////////////////////////////////////////
// Rule: declarator_suffix_bitfield.  Matches bitfield suffixes on a
// declarator ( : expr ).
// Inputs:
//    CPPTypeRefI   inputType    The declarator type before encountering the
//                               bitfield suffix.
// Outputs:
//    CPPTypeRefI   outputType   The declarator type after modification by the
//                               bitfield suffix.
/////////////////////////////////////////////////////////////////////////
declarator_suffix_bitfield [ CPPTypeRefI inputType ] > [ CPPTypeRefI outputType ] :
   << 
      ANTLRTokenPtr errTok;
      CPPExprRefI exprFieldlen;
      $outputType = $inputType;
   >>
   << ; >>

   // bitfield
   TOK_COLON 
   << errTok = LT(1); >>
   expr_constant > [ exprFieldlen ]
   << $outputType = LookupBitfieldOfType($inputType, exprFieldlen, errTok); >>
   ;

/////////////////////////////////////////////////////////////////////////
// Rule: declarator_suffix_function.  Matches argument suffixes on a
// declarator (arg, arg, ...)
// Inputs:
//    CPPTypeRefI            inputType    The declarator type before 
//                                        encountering the function suffix.
//    CPPConst::CallingConvention cc      The calling convention
// Outputs:
//    StrListRefI    argNameList  List of argument names in the order
//                                that they were declared.  Abstract
//                                declarator arguments have empty name.
//    CPPScopeRefI   argScope     Contains the parameter declarations
//    CPPTypeRefI    outputType   The declarator type after modification by the
//                                function suffix.
/////////////////////////////////////////////////////////////////////////
declarator_suffix_function 
[ CPPTypeRefI inputType, CPPConst::CallingConvention cc ] > 
[ StrListRefI argNameList, CPPScopeRefI argScope, CPPTypeRefI outputType ] :

   << 
      CPPConst::CVQualifier cv;
      $outputType = LookupFunctionReturningType($inputType, cc); 
      ScopePopper scopePopper(this);
      EnterLocalScope();
      $argScope = GetCurrentScope();
   >>
   << ; >>
   TOK_LPAREN 
   parameter_list_opt [ $outputType ] > [ $argNameList, $outputType ]
   TOK_RPAREN 

   cv_qualifier_seq > [ cv ]
   << $outputType = LookupCVQualifierOfType($outputType, cv); >>
   ;

/////////////////////////////////////////////////////////////////////////
// Rule: declarator_suffixes.  general-purpose declarator suffixes.
// Not intended for abstract declarators.
// Inputs:
//    CPPTypeRefI             inputType  The declarator type before 
//                                       encountering the suffixes.
//    CPPConst::CallingConvention cc     The calling convention
// Outputs:
//    StrListRefI    argNameList  With a function suffix, this is the 
//                                list of argument names in the order
//                                that they were declared.  Abstract
//                                declarator arguments have empty name.
//                                Zero-ref if there is no function suffix.
//    CPPScopeRefI   argScope     If not nil, contains the parameter declarations.
//    CPPTypeRefI    outputType   The declarator type after modification by the
//                                function suffix.
/////////////////////////////////////////////////////////////////////////
declarator_suffixes 
[ CPPTypeRefI inputType, CPPConst::CallingConvention cc ] > 
[ StrListRefI argNameList, CPPScopeRefI argScope, CPPTypeRefI outputType ] :
   << 
      $outputType = $inputType;
   >>
   <<;>> 
   (
         declarator_suffix_array [ $inputType ] > [ $outputType ]
      |  
         // function arguments

         // This backtracking resolves the ambiguity encountered in the
         // "declarator_init" rule:
         //     declarator_init: declarator { ... | TOK_LPAREN expression_list TOK_RPAREN }
         // The ambiguity is between function arguments in the declarator_suffixes
         // of the declarator rule, and a function-like initializer list in 
         // the declarator_init.  For example:
         //      int foo(int x);     // argument list -- function decl
         //      int foo(int(x));    // ambiguous -- resolve to function decl
         //      int foo(int(3));    // object definition & intializer list
         // DWP May96 s8.2 states that in the second case, the
         // declaration is a function.  The rule is to resolve these as
         // function declarations if possible.  So we attempt function decl,
         // and if it fails we back out and take the empty suffix alternate
         (
            declarator_suffix_function [ $inputType, cc ] > [ $argNameList, $argScope, $outputType ]
         )?
      |
         declarator_suffix_bitfield [ $inputType ] > [ $outputType ]
      |
         /* empty */
         << $outputType = $inputType; >>
   )
   ;

///////////////////////////////////////////////////////////////////////
// Rule: enum_specifier.  Matches an enum-type specifier, which may
//      be a declaration, definition, or simply "enum E", where E is
//      a previously-declared enum.  For example:
//            enum E ;       enum declaration
//            enum E ...     previously-declared enum
//            enum E {...}   enum definition
// Outputs: 
//    CPPTypeRefI  enumType   The enumeration type.
/////////////////////////////////////////////////////////////////////////
enum_specifier > [ CPPTypeRefI enumType ] :
   << 
      CPPTypeRefI constEnumType; 
      ScopePopper scopePopper(this);
      bool dummy;
   >>
   <<;>> 

   TOK_ENUM
   (
         // Unnamed enumeration
         TOK_LCURLY
         <<
            // Cannot be an init-action
            JLStr tmpName;
            CPPDeclRefI tmpDecl;
            DefineEnumType("", GetCurrentScope(), LT(1), $enumType, tmpDecl, tmpName);
            constEnumType = LookupCVQualifierOfType($enumType, CPPConst::cvCONST);
         >>
         enumerator_list [ constEnumType ]
         TOK_RCURLY
      |

         // allow for scope_override in enum declaration (not definition)
         scope_override > [ dummy ]

         id:TOK_ID

         (
               // enum definition
               
               TOK_LCURLY 
               << 
                  // Report error for explicit qualification
                  if (scopePopper.Get() != SaveScope())
                  {
                     ReportError("Enum definitions cannot be explicitly-qualified", $id);
                  }

                  JLStr tmpName;
                  CPPDeclRefI tmpDecl;
                  DefineEnumType(
                     $id->getText(), 
                     GetCurrentScope(), 
                     $id, 
                     $enumType, 
                     tmpDecl, 
                     tmpName
                  ); 
                  constEnumType = LookupCVQualifierOfType($enumType, CPPConst::cvCONST);
               >>
               enumerator_list [ constEnumType ]
               TOK_RCURLY 
            |
               // no trailing {}.  This is tricky, because if the next
               // token is ';' it is a declaration of the enum type,
               // otherwise it's an error if the enum is not
               // already defined in some other scope, as per 
               // WP 28May96 s3.4.3.2
               << 
                  if (LT(1)->getType() == TOK_SEMICOLON)
                  {
                     // Declaration

                     // Specify scope for decl iff there is an override, otherwise
                     // use implicit scope stack lookup.
                     CPPScopeRefI declScope;
                     if (scopePopper.Get() != SaveScope())
                     {
                        declScope = GetCurrentScope();
                     }

                     JLStr tmpName;
                     CPPDeclRefI tmpDecl;
                     DeclareEnumType(
                        $id->getText(), 
                        declScope,
                        $id, 
                        $enumType, 
                        tmpDecl, 
                        tmpName
                     ); 
                  } else {
                     // Lookup of previously-declared enumeration
                     // TODO: lookup explicitly-qualified enum
                     $enumType = LookupEnum($id->getText(), $id);
                  }
               >>
         )
   )
   ;

///////////////////////////////////////////////////////////////////////
// Rule: enum_specifier_simple.  Matches an enum-type specifier, without
// allowing a definition (i.e., {...}).
// Outputs: 
//    CPPTypeRefI  enumType   The enumeration type.
/////////////////////////////////////////////////////////////////////////
enum_specifier_simple > [ CPPTypeRefI enumType ] :
   << 
      CPPTypeRefI constEnumType; 
      ScopePopper scopePopper(this);
      bool dummy;
   >>
   <<;>> 

   TOK_ENUM

   scope_override > [ dummy ]
   id:TOK_ID

   // This is tricky, because if the next
   // token is ';' it is a declaration of the enum type,
   // otherwise it's an error if the enum is not
   // already defined in some other scope, as per 
   // WP 28May96 s3.4.3.2
   << 
      if (LT(1)->getType() == TOK_SEMICOLON)
      {
         // Declaration
         JLStr tmpName;
         CPPDeclRefI tmpDecl;
         DeclareEnumType($id->getText(), GetCurrentScope(), $id, $enumType, tmpDecl, tmpName); 
      } else {
         // Lookup of previously-declared enumeration
         $enumType = LookupEnum($id->getText(), $id);
      }
   >>
   ;

///////////////////////////////////////////////////////////////////////
// Rule: enumerator.  Matches an enumerator for an enumeration.
//     Enters the enumerator as a declaration of the enumeration type.
// Inputs: 
//    CPPTypeRefI  enumType  The enumeration type under which the enumerators
//                           will be declared.  This is actually a "const"
//                           modifier of the enum type.
//    CPPExprRefI  exprIn    The constant value to be assigned to the 
//                           enumerator if it has no explicit value.
// Outputs:
//    CPPExprRefI  exprOut   The value actually used for the enumerator
/////////////////////////////////////////////////////////////////////////
enumerator [ CPPTypeRefI enumType, CPPExprRefI exprIn ] > [ CPPExprRefI exprOut ] :
   << ANTLRTokenPtr errTok = LT(1); >>
   <<;>>
   id:TOK_ID
   (
         // constant expression assignment
         TOK_ASSIGNEQUAL 
         << errTok = LT(1); >>
         expr_constant > [ $exprOut ]
      |
         // empty -- use the input value expression
         << $exprOut = $exprIn; >>
   )
   << DeclareEnumerator($id->getText(), $enumType, $exprOut, errTok); >>
   ;

///////////////////////////////////////////////////////////////////////
// Rule: enumerator_list.  Matches a list of enumerators for an enumeration.
// Inputs: 
//    CPPTypeRefI  enumType   The enumeration type under which the enumerators
//                         will be declared.  This is actually a "const"
//                         modifier of the type.
/////////////////////////////////////////////////////////////////////////
enumerator_list [ CPPTypeRefI enumType ] :
   << CPPExprRefI exprValue = MakeExprLiteral(TOK_DECIMALINT, "0"); >>
   <<;>> 

   enumerator [ $enumType, exprValue ] > [ exprValue ]

   // Create an expression which is the previous value plus one, to be
   // used for the next value if nothing is specified.
   <<
      exprValue = MakeExprBinary(
         exprValue,
         MakeExprLiteral(TOK_DECIMALINT, "1"),
         TOK_PLUS
      );
   >>

   (
      // Predicate says that TOK_COMMA TOK_ID is always another enumerator
      (TOK_COMMA)? => <<LT(2)->getType() == TOK_ID>>?

      TOK_COMMA 
      enumerator [ $enumType, exprValue ] > [ exprValue ]
      <<
         exprValue = MakeExprBinary(
            exprValue,
            MakeExprLiteral(TOK_DECIMALINT, "1"),
            TOK_PLUS
         );
      >>
   )*

   // Allow optional trailing comma at end of enumerator list
   // This is not standard, but we allow it because a lot of old
   // machine-generated code looks like this.
   { TOK_COMMA }
   ;

///////////////////////////////////////////////////////////////////////
// Rule: expression.  This is really the "comma" expression.
//       For purposes of constant expression evaluation,
//       it only returns the last expression in the list.
//       When we have complete expression semantics, this should
//       build a list of expressions.
// Outputs:
//    CPPExprRefI    expr     The resulting ref-counted expression
///////////////////////////////////////////////////////////////////////
expression > [ CPPExprRefI expr ] :
   <<;>> 
   <<;>> 
   expr_assignment > [ $expr ]
   (
      TOK_COMMA 
      <<CPPExprRefI expr2;>> 
      expr_assignment > [ expr2 ]
      << $expr = MakeExprBinary($expr, expr2, TOK_COMMA); >>
   )*
   ;

///////////////////////////////////////////////////////////////////////
// Rule: expression_list.  Used where a list of expressions is needed.
//       Syntacically identical to a comma-expression but semantically
//       different.
// Outputs:
//    CPPExprRefI    expr     The resulting ref-counted expression
///////////////////////////////////////////////////////////////////////
expression_list > [ CPPExprRefI expr ] :
   expression > [ $expr ]
   ;

///////////////////////////////////////////////////////////////////////
// Rule: expr_additive: add/subtract expression
// Outputs:
//    CPPExprRefI    expr     The resulting ref-counted expression
// Note: See comment for expr_multiplicative regarding #pragma 
///////////////////////////////////////////////////////////////////////
expr_additive > [ CPPExprRefI expr ] :
   <<;>> 
   <<;>> 
   expr_multiplicative > [ $expr ]
   (
      << CPPExprRefI expr2; >>
      tok:ADDITIVE_TOKCLASS
      expr_multiplicative > [ expr2 ]
      << $expr = MakeExprBinary($expr, expr2, $tok->getType()); >>
   )*
   ;

///////////////////////////////////////////////////////////////////////
// Rule: expr_and: bitwise-and expression
// Outputs:
//    CPPExprRefI    expr     The resulting ref-counted expression
///////////////////////////////////////////////////////////////////////
expr_and > [ CPPExprRefI expr ] :
   <<;>> 
   <<;>> 
   expr_equality > [ $expr ]
   (
      << CPPExprRefI expr2; >>
      TOK_AMPERSAND 
      expr_equality > [ expr2 ]
      << $expr = MakeExprBinary($expr, expr2, TOK_AMPERSAND); >>
   )*
   ;

///////////////////////////////////////////////////////////////////////
// Rule: expr_assignment: assignment expression
// Outputs:
//    CPPExprRefI    expr     The resulting ref-counted expression
// Note: association is right-to-left for assignment op, which is why
// expr_conditional comes first.
///////////////////////////////////////////////////////////////////////
expr_assignment > [ CPPExprRefI expr ] :
   <<;>> 
   <<;>> 
   expr_conditional > [ $expr ]
   {  
      tok:ASSIGN_TOKCLASS
      <<CPPExprRefI expr2;>> 
      expr_assignment > [ expr2 ]
      << $expr = MakeExprBinary($expr, expr2, $tok->getType()); >>
   }
   ;

///////////////////////////////////////////////////////////////////////
// Rule: expr_cast: cast expression
// Outputs:
//    CPPExprRefI    expr     The resulting ref-counted expression
///////////////////////////////////////////////////////////////////////
expr_cast > [ CPPExprRefI expr ] :
   <<;>> 
   <<;>> 
   (
         ////////////////////////////////////////////////////////////////
         // There is a nasty ambiguity (from an LL standpoint), between 
         // old casts and function-like casts surrounded by parentheses:
         //     (type_name)   vs   (type_name(...))
         // For example:
         //     (int)3   vs   (int(3))
         //     (T)x     vs   (T(x))
         // Since type_name can be an arbitralily complex thing, the 
         // only simple way to resolve this is through backtracking (sigh).
         // So we try the traditional cast, and failing that revert to 
         // the unary expression, which eventually resolves to ( expr ).
         ///////////////////////////////////////////////////////////////////////
         (
            << CPPTypeRefI type; >>
            TOK_LPAREN 
            type_name > [ type ]
            TOK_RPAREN 
            expr_cast > [ $expr ]
            << $expr = MakeExprTraditionalCast(type, $expr); >>
         )?
      |  
         expr_unary > [ $expr ]
   )
   ;

///////////////////////////////////////////////////////////////////////
// Rule: expr_conditional: conditional expression
// Outputs:
//    CPPExprRefI    expr     The resulting ref-counted expression
///////////////////////////////////////////////////////////////////////
expr_conditional > [ CPPExprRefI expr ] :
   <<;>> 
   <<;>> 
   expr_logical_or > [ $expr ]
   { 
      << CPPExprRefI lhs, rhs; >>
      TOK_QUESTIONMARK 
      expression > [ lhs ]
      TOK_COLON 
      expr_logical_or > [ rhs ]
      << $expr = MakeExprConditional($expr, lhs, rhs); >>
   }
   ;

///////////////////////////////////////////////////////////////////////
// Rule: expr_constant: constant expression
// Outputs:
//    CPPExprRefI    expr     The resulting ref-counted expression
//
// Used in contexts where a constant expression is needed,
// The semantic phase still needs to check that this is 
// really a constant expression by calling EvaluateConstant on
// the returned expression node.
///////////////////////////////////////////////////////////////////////
expr_constant > [ CPPExprRefI expr ] :
   <<;>> 
   <<;>> 
   expr_conditional > [ $expr ]
   ;

///////////////////////////////////////////////////////////////////////
// Rule: expr_delete: delete expression
// Inputs:
//    bool          global      true if the delete expression was
//                              preceded by "::"
// Outputs:
//    CPPExprRefI   expr        A ref-counted pointer to the
//                              expression object
///////////////////////////////////////////////////////////////////////
expr_delete [ bool global ] > [ CPPExprRefI expr ] :
   <<bool deleteArray = false;>> 
   <<;>> 
   TOK_DELETE 
   { 
      TOK_LSQUARE 
      TOK_RSQUARE 
      << deleteArray = true; >>
   } 

   expr_cast > [ $expr ]

   << $expr = MakeExprDelete($expr, deleteArray); >>
   ;

///////////////////////////////////////////////////////////////////////
// Rule: expr_equality: equality-comparison expression
// Outputs:
//    CPPExprRefI    expr     The resulting ref-counted expression
///////////////////////////////////////////////////////////////////////
expr_equality > [ CPPExprRefI expr ] :
   <<;>> 
   <<;>> 
   expr_relational > [ $expr ]
   (
      << CPPExprRefI expr2; >>
      tok:EQUALITY_TOKCLASS
      expr_relational > [ expr2 ]
      << $expr = MakeExprBinary($expr, expr2, $tok->getType()); >>
   )*
   ;

///////////////////////////////////////////////////////////////////////
// Rule: expr_exclusive_or: bitwise exclusive-OR expression
// Outputs:
//    CPPExprRefI    expr     The resulting ref-counted expression
///////////////////////////////////////////////////////////////////////
expr_exclusive_or > [ CPPExprRefI expr ] :
   <<;>> 
   <<;>> 
   expr_and > [ $expr ]
   (
      << CPPExprRefI expr2; >>
      TOK_BITWISEXOR
      expr_and > [ expr2 ]
      << $expr = MakeExprBinary($expr, expr2, TOK_BITWISEXOR); >>
   )*
   ;

///////////////////////////////////////////////////////////////////////
// Rule: expr_id: id expression
//    Matches identifiers, dtors, and operators
// Inputs:
//    CPPExprRefI  objectExpr   If non-null, this expression is the
//                              left side of '.' or '->'.
// Outputs:
//    CPPExprRefI  expr         The resulting ref-counted expression
///////////////////////////////////////////////////////////////////////
expr_id [ CPPExprRefI objectExpr ] > [ CPPExprRefI expr ] :
   << 
      ScopePopper scopePopper(this);
      JLStr exprIdName;
      ANTLRTokenPtr errTok = LT(1);
      bool tmpBool;
   >>
   <<;>> 

   // Enter the scope of the object expression, if any, because
   // things like p->mbr must look in the scope of the type of "p".
   << 
      if ($objectExpr.get() != 0)
      {
         EnterExprScope($objectExpr, true);
      }
   >>

   // No ctors or types in id expressions, but we allow dtors
   // Allow a dtor that is not found in the current scope iff objectExpr
   // is non-null.  This supports pseudo-destructor calls like:
   //     A *xp;
   //     xp->~A();
   qualified_id [ false, false, $objectExpr.get()!=0 ] > [ exprIdName, tmpBool ]

   <<
      $expr = LookupExprId(exprIdName, errTok);
   >>
   ;

///////////////////////////////////////////////////////////////////////
// Rule: expr_inclusive_or: bitwise-OR expression
// Outputs:
//    CPPExprRefI    expr     The resulting ref-counted expression
///////////////////////////////////////////////////////////////////////
expr_inclusive_or > [ CPPExprRefI expr ] :
   <<;>> 
   <<;>> 
   expr_exclusive_or > [ $expr ]
   (
      << CPPExprRefI expr2; >>
      TOK_BITWISEOR 
      expr_exclusive_or > [ expr2 ]
      << $expr = MakeExprBinary($expr, expr2, TOK_BITWISEOR); >>
   )*
   ;

///////////////////////////////////////////////////////////////////////
// Rule: expr_literal: a literal expression
// Outputs:
//    CPPExprRefI    expr     The resulting ref-counted expression
///////////////////////////////////////////////////////////////////////
expr_literal > [ CPPExprRefI expr ] :
   <<;>> 
   <<;>> 
      (
         << ANTLRTokenPtr literalTok; >>
         constant > [ literalTok ]
         << 
            // token may be empty if there was a syntax error
            if (literalTok.operator->() != 0)
            {
               $expr = MakeExprLiteral(literalTok->getType(), literalTok->getText());
            }
         >>
      )
   |
      (
         << JLStr str; >>
         (
            strTok:TOK_STRING
            << str += $strTok->getText(); >>
         )+
         << $expr = MakeExprLiteral(TOK_STRING, str.c_str()); >>
      )
   ;

///////////////////////////////////////////////////////////////////////
// Rule: expr_logical_and:  logical-AND expression
// Outputs:
//    CPPExprRefI    expr     The resulting ref-counted expression
///////////////////////////////////////////////////////////////////////
expr_logical_and > [ CPPExprRefI expr ] :
   <<;>> 
   <<;>> 
   expr_inclusive_or > [ $expr ]
   (
      << CPPExprRefI expr2; >>
      TOK_AND 
      expr_inclusive_or > [ expr2 ]
      << $expr = MakeExprBinary($expr, expr2, TOK_AND); >>
   )*
   ;

///////////////////////////////////////////////////////////////////////
// Rule: expr_logical_or:  logical-OR expression
// Outputs:
//    CPPExprRefI    expr     The resulting ref-counted expression
///////////////////////////////////////////////////////////////////////
expr_logical_or > [ CPPExprRefI expr ] :
   <<;>> 
   <<;>> 
   expr_logical_and > [ $expr ]
   (
      << CPPExprRefI expr2; >>
      TOK_OR 
      expr_logical_and > [ expr2 ]
      << $expr = MakeExprBinary($expr, expr2, TOK_OR); >>
   )*
   ;

///////////////////////////////////////////////////////////////////////
// Rule: expr_multiplicative: multiply/divide/modulus expression
// Outputs:
//    CPPExprRefI    expr     The resulting ref-counted expression
///////////////////////////////////////////////////////////////////////
expr_multiplicative > [ CPPExprRefI expr ] :
   <<;>> 
   <<;>> 
   expr_pm > [ $expr ]
   (
      << CPPExprRefI expr2; >>
      tok:MULTIPLICATIVE_TOKCLASS
      expr_pm > [ expr2 ]
      << $expr = MakeExprBinary($expr, expr2, $tok->getType()); >>
   )*
   ;

///////////////////////////////////////////////////////////////////////
// Rule: expr_new: memory-allocation expression
//
// Inputs:
//    bool          global      true if the new expressions was
//                              preceded by "::"
// Outputs:
//    CPPExprRefI   expr        A ref-counted pointer to the
//                              expression object
///////////////////////////////////////////////////////////////////////
expr_new [ bool global ] > [ CPPExprRefI expr ] :
   <<
      CPPExprRefI placementExpr;
      CPPExprRefI arraySizeExpr;
      CPPExprRefI initExpr;
      CPPTypeRefI newType;
   >> 
   <<;>> 
   TOK_NEW

   { 
      // This backtracking used to resolve placement syntax
      //     new (expr) (type)
      // vs. non-placement syntax
      //     new (type)
      // because both (expr) and (type) look identical until you've
      // seen the whole thing.
      (
         TOK_LPAREN 
         expression_list > [ placementExpr ]
         TOK_RPAREN
      )? 
   }


   (  
        expr_new_type_id > [ newType, arraySizeExpr ]
     |  
        TOK_LPAREN 
        type_name  > [ newType ]
        TOK_RPAREN
   )

   { 
      // expr_new_initializer conflicts with declarator initializers
      // when the bitfield field length expression is an expr_new, which
      // is semantically bogus but allowed by the syntax:
      //     int x : new T ();
      // Where the () could be either part of the expr_new or
      // could be the initializer list for the declarator.  Since
      // this usage is bogus, we just always bind the (...) to the
      // expr_new as an initializer.
      (TOK_LPAREN)? => <<1>>?
      expr_new_initializer > [ initExpr ]
   }

   << $expr = MakeExprNew(placementExpr, newType, arraySizeExpr, initExpr, LT(1)); >>
   ;

///////////////////////////////////////////////////////////////////////
// Rule: expr_new_initializer.  Matches () or (expr).
// Inputs: none
// Ouputs:
//    CPPExprRefI   initExpr   The initializer expression list, or
//                             a zero-ref if the expression list is empty.
///////////////////////////////////////////////////////////////////////
expr_new_initializer > [ CPPExprRefI initExpr ] :
   TOK_LPAREN 
   { 
      expression_list > [ $initExpr ]
   } 
   TOK_RPAREN
   ;

///////////////////////////////////////////////////////////////////////
// Rule: expr_new_type_id.  Matches a type identifier, with optional
// non-constant array dimensions, for an expr_new.
//
// Inputs: none
// Outputs:
//    CPPTypeRefI   cppType          Type of the new expression
//    CPPExprRefI   exprArraySize    Size of the array, or a zero-ref
//                                   for no array subscript.  Note that an
//                                   empty array subscript is subsumed by
//                                   the cppType.
///////////////////////////////////////////////////////////////////////
expr_new_type_id > [ CPPTypeRefI cppType, CPPExprRefI exprArraySize ] :
   <<
      CPPConst::StorageClass sc;
      CPPConst::DeclModifier dm;
      ANTLRTokenPtr errTok = LT(1);
   >>
   <<;>> 

   declaration_specifiers_simple [ CPPConst::scNONE, CPPConst::dmNONE ] > [ $cppType, sc, dm ]

   <<
      if (sc != CPPConst::scNONE)
      {
         ReportError("Storage class not allowed for new type id", errTok);
      }
      if (dm != CPPConst::dmNONE)
      {
         ReportError("Declaration modifier not allowed for new type id", errTok);
      }
   >>

   // Note that declarator_ptr includes references, which must be
   // rejected semantically later.
   // The predicate here groups all of the potential pointer-operators
   // with the declarator, as opposed to interpreting it as a
   // binary or unary operator in an expression.  For example:
   //     new int * 5;
   // means (new int*) 5, which is a syntax error, as opposed to the expression
   //     (new int) * 5

   (
      (TOK_AMPERSAND | TOK_STAR)? => <<1>>?
      declarator_ptr [ $cppType ] > [ $cppType ]
   )*

   {
      // Optional array dimensions.  The first dimension is an optional expression
      // but the following dimensions are non-optional constant expressions.  This
      // allows constructs like:
      //     new char[][5];      // allocate one pointer-to-array-of-five-char
      //     new char[x=10][5];  // allocate array of ten arrays-of-five-char

      TOK_LSQUARE 
      << errTok = LT(1); >>
      (
            // expression indicates size of array
            expression > [ $exprArraySize ]
         |
            // it is an empty array type e.g. char[][5]
            << $cppType = LookupArrayOfType($cppType, CPPExprRefI(), errTok); >>
      )
      TOK_RSQUARE

      // Trailing dimensions all must be constant expressions.  
      // They indicate dimensions of the allocated type.
      (
         TOK_LSQUARE 
         << 
            errTok = LT(1); 
            CPPExprRefI exprDimension;
         >>
         expr_constant > [ exprDimension ]
         TOK_RSQUARE
         << $cppType = LookupArrayOfType($cppType, exprDimension, errTok); >>
      )*
   }
   ;

///////////////////////////////////////////////////////////////////////
// Rule: expr_pm: pointer-to-member expression
// Outputs:
//    CPPExprRefI    expr     The resulting ref-counted expression
///////////////////////////////////////////////////////////////////////
expr_pm > [ CPPExprRefI expr ] :
   <<;>> 
   <<;>> 
   expr_cast > [ $expr ]
   (
      << CPPExprRefI expr2; >>
      tok:PM_TOKCLASS
      expr_cast > [ expr2 ]
      << $expr = MakeExprBinary($expr, expr2, $tok->getType()); >>
   )*
   ;

///////////////////////////////////////////////////////////////////////
// Rule: expr_postfix: postfix expression, handling array subscript,
//    function-call, member access, increment, decrement
// Outputs:
//    CPPExprRefI    expr     The resulting ref-counted expression
///////////////////////////////////////////////////////////////////////
expr_postfix > [ CPPExprRefI expr ] :
   <<;>> 
   <<;>> 

   (  
   <<;>>
         // The backtracking allows ANTLR to search for a leading *type*
         // to start a function-style cast, as opposed to a leading
         // *identifier*, which may start some other type of expression.
         // ANTLR will backtrack if the leading thing is not a type
         // For example:
         //    struct A { struct B {}; static void f(); };
         //    A::B(1)     // function-style cast
         //    A::f(1)     // id-expression (which is part of primary-expression)
         << CPPTypeRefI castType; >>
         (
            simple_type_specifier > [ castType ]
            << CommitGuess(); >>
            TOK_LPAREN 
            { 
               expression_list > [ $expr ]
            }
            TOK_RPAREN
         )?
         << $expr = MakeExprFunctionLikeCast(castType, $expr); >>
      |  
         expr_primary > [ $expr ]
      |
         // new-style casts are grammatically postfix expressions,
         // even though semantically they are cast expressions
         << CPPTypeRefI castType; >>
         ct:CAST_TOKCLASS
         TOK_LESSTHAN
         type_name > [ castType ]
         TOK_GREATERTHAN
         TOK_LPAREN
         expression > [ $expr ]
         TOK_RPAREN
         <<
            // Create the appropriate cast expression
            $expr = MakeExprNewCast($ct->getType(), castType, $expr);
         >>
   )

   // A postfix expression may be followed by a chain of trailing modifiers
   (
      // This predicate handles the TOK_LPAREN ambiguity.  This
      // arises the because an expr_constant is used in the bitfield
      // suffix.  Since the bitfield suffix can syntactically be
      // followed by "(...)" as an initializer list:
      //    declarator(initializers)
      // Then it is syntacticially possible to write, for example:
      //    int x : f(1)
      // Which is either a bitfield of size "f" initialized to "1", 
      // or a bitfield whose size is the return value of "f(1)".
      // Since bitfields cannot be initialized using that syntax, 
      // either interpretation is an error, so we always bind 
      // trailing (...) to the expression.
      (TOK_LPAREN)? => <<1>>?

      << CPPExprRefI expr2; >>
      (
            TOK_LSQUARE 
            expression > [ expr2 ]
            TOK_RSQUARE
            << $expr = MakeExprArraySubscript($expr, expr2); >>
         |  
            TOK_LPAREN 
            { 
               expression_list > [ expr2 ]
            }
            TOK_RPAREN
            << $expr = MakeExprFunctionCall($expr, expr2); >>
         |  
            TOK_DOT 
            expr_id [ $expr ] > [ expr2]
            << $expr = MakeExprBinary($expr, expr2, TOK_DOT); >>
         |  
            TOK_POINTERTO 
            expr_id [ $expr ] > [ expr2]
            << $expr = MakeExprBinary($expr, expr2, TOK_POINTERTO); >>
         |  
            TOK_PLUSPLUS
            << $expr = MakeExprPostop($expr, TOK_PLUSPLUS); >>
         |  
            TOK_MINUSMINUS
            << $expr = MakeExprPostop($expr, TOK_MINUSMINUS); >>
      )
   )*
   ;

///////////////////////////////////////////////////////////////////////
// Rule: expr_primary: primary expression.  The "most atomic" expression.
// Outputs:
//    CPPExprRefI    expr     The resulting ref-counted expression
///////////////////////////////////////////////////////////////////////
expr_primary > [ CPPExprRefI expr ] :
   <<;>> 
   <<;>> 
      expr_id  [ CPPExprRefI() ] > [ $expr ]
   |  
      expr_literal > [ $expr ]
   |  
      TOK_THIS
      << $expr = MakeExprThis(); >>
   |  
      TOK_LPAREN 
      expression > [ $expr ]
      TOK_RPAREN
   ;

///////////////////////////////////////////////////////////////////////
// Rule: expr_relational: relational (compare) expression.
// Outputs:
//    CPPExprRefI    expr     The resulting ref-counted expression
///////////////////////////////////////////////////////////////////////
expr_relational > [ CPPExprRefI expr ] :
   <<;>> 
   <<;>> 
   expr_shift > [ $expr ]
   (  
      << CPPExprRefI expr2; >>
      tok:RELATIONAL_TOKCLASS
      expr_shift > [ expr2 ]
      << $expr = MakeExprBinary($expr, expr2, tok->getType()); >>
   )*
   ;

///////////////////////////////////////////////////////////////////////
// Rule: expr_shift: bitwise-shift expression.
// Outputs:
//    CPPExprRefI    expr     The resulting ref-counted expression
///////////////////////////////////////////////////////////////////////
expr_shift > [ CPPExprRefI expr ] :
   <<;>> 
   <<;>> 
   expr_additive > [ $expr ]
   (
      << CPPExprRefI expr2; >>
      tok:SHIFT_TOKCLASS
      expr_additive > [ expr2 ]
      << $expr = MakeExprBinary($expr, expr2, tok->getType()); >>
   )*
   ;

///////////////////////////////////////////////////////////////////////
// Rule: expr_unary: unary expression.  Handles pre-increment,
//        pre_decrement, as well as the usual unary operations.
// Outputs:
//    CPPExprRefI    expr     The resulting ref-counted expression
///////////////////////////////////////////////////////////////////////
expr_unary > [ CPPExprRefI expr ] :
   <<;>> 
   <<;>> 

   (  
         // Syntactic predicate disambiguates the cases that can be 
         // either a unary TILDE operator or a dtor:
         //    class A;
         //    ~A
         (
            expr_postfix > [ $expr ]
         )?
      |  
         tok:UNARYOP_TOKCLASS 
         expr_cast > [ $expr ]
         << $expr = MakeExprUnary($expr, tok->getType()); >>
      |  
         TOK_SIZEOF
         (  
               // Resolves ambiguity between (type) and (expr)
               (
                  << CPPTypeRefI sizeofType; >>
                  TOK_LPAREN 
                  type_name > [ sizeofType ]
                  TOK_RPAREN
                  << $expr = MakeExprSizeofType(sizeofType); >>
               )?
            |  
               expr_unary > [ $expr ]
               << $expr = MakeExprSizeofExpr($expr); >>
         )
      |  
         // New and delete are left-factored for "::"
         (
            << bool global = false; >>
            { 
               TOK_SCOPE
               << global = true; >>
            }
            (
                  expr_new [global] > [ $expr ]
               |  
                  expr_delete [global] > [ $expr ]
            )
         )
   )
   ;

/////////////////////////////////////////////////////////////////////////
// Rule: generic_block.  Matches nested blocks of {...}.  Used for
// function bodies where the parse of the body must be deferred until
// later, such as for class methods specified in the class body.
//
// Inputs: none
// Outputs: none
//
// Note: Use a TokenCollector before calling this rule to collect the
// tokens matched by the rule.
/////////////////////////////////////////////////////////////////////////
generic_block :
   <<;>>
   <<;>>
   TOK_LCURLY
   (
         ~CURLY_TOKCLASS
      |
         generic_block
   )*
   TOK_RCURLY
   ;

/////////////////////////////////////////////////////////////////////////
// Rule: initializer.  Matches an initialization expression or
// list of expressions in curly braces.
//
// Inputs: none
// Outputs:
//    CPPExprRefI    expr     The expression or expression list.
/////////////////////////////////////////////////////////////////////////
initializer > [ CPPExprRefI expr ] :
   <<;>> 
   <<;>> 
      expr_assignment > [ $expr ]
   |  
      TOK_LCURLY 
      expression_list > [ $expr ]
      TOK_RCURLY
   ;

/////////////////////////////////////////////////////////////////////////
// Rule: linkage_specification.  Matches stuff like
//    extern "C" declaration
// or 
//    extern "C" { declarations }
/////////////////////////////////////////////////////////////////////////
linkage_specification :
   <<;>> 
   <<;>> 
   TOK_EXTERN
   TOK_STRING
   (  
         TOK_LCURLY 
         (
            // Tell ANTLR to shut up about this intentional ambiguity
            #pragma approx
            (
                  declaration_external
               |
                  // This is only here for error synchronization
                  ~RCURLY_EOF_TOKCLASS
                  << fail_syntax(TOK_SEMICOLON);>>
            )
            exception <<
               ReportExceptionError(ex);
               consume();
               consumeUntil(EXTERNAL_SYNC_TOKCLASS_set);
               if (set_el(LA(1), EXTERNAL_SYNC2_TOKCLASS_set))
               {
                  consume();
               }
            >>
         )* 
         TOK_RCURLY
      |  
         declaration_external
   )
   ;

///////////////////////////////////////////////////////////////////////
// Rule: optor_conversion.  Match a user-defined type conversion specifier
// after an "operator" token.
// 
// Inputs: none
// Outputs:
//    CPPTypeRefI   conversionType   The conversion type.
///////////////////////////////////////////////////////////////////////
optor_conversion > [ CPPTypeRefI conversionType ]:
   <<
      CPPConst::StorageClass sc;
      CPPConst::DeclModifier dm;
      ANTLRTokenPtr errTok = LT(1);
   >>
   <<;>> 
   declaration_specifiers_simple 
      [ CPPConst::scNONE, CPPConst::dmNONE ] > 
      [ $conversionType, sc, dm ]

   <<
      // Disallow storage class of decl modifier
      if (sc != CPPConst::scNONE || dm != CPPConst::dmNONE)
      {
         ReportError("Storage class or declaration modifier not allowed in operator", errTok);
      }
   >>

   ( 
      // The predicate is here to resolve a nasty ambiguity arising
      // from things like "operator A*", which is a valid optor,
      // but which causes problems when used like:
      //      B::operatorA*  foo;
      // which can be either (B::operatorA) * foo -- multiply
      // or (B::operator A*) foo -- syntax error.
      // The pragma always binds the "*|&" with the operator, thus leading
      // to declarations, as required by WP May96 s8.2.1
      (TOK_AMPERSAND | TOK_STAR)? => <<1>>?

      declarator_ptr [ $conversionType ] > [ $conversionType ]
   )*
   ;

///////////////////////////////////////////////////////////////////////
// Rule: optor_simple.  Match any of the stock operator types, but 
// not the user-defined type conversion operators.
// 
// Inputs: none
// Outputs:
//    JLStr    opName     The name of the operator.
///////////////////////////////////////////////////////////////////////
optor_simple > [ JLStr opName ]:
   <<;>>
   <<;>> 
      TOK_NEW
      << $opName = " new"; >>
      // Since the name of an operator function is a valid
      // id-expression, we tell ANTLR to bind any "[]" with 
      // the name of the operator instead of treating it
      // as part of a postfix-expression.
      { 
         (TOK_LSQUARE)? => <<LT(2)->getType() == TOK_RSQUARE>>?
         TOK_LSQUARE TOK_RSQUARE
         << $opName = "[]"; >>
      }
   |   
      TOK_DELETE
      << $opName = " delete"; >>

      // Since the name of an operator function is a valid
      // id-expression, we tell ANTLR to bind any "[]" with 
      // the name of the operator instead of treating it
      // as part of a postfix-expression.
      { 
         (TOK_LSQUARE)? => <<LT(2)->getType() == TOK_RSQUARE>>?
         TOK_LSQUARE TOK_RSQUARE 
         << $opName = "[]"; >>
      }
   |  
      TOK_LPAREN TOK_RPAREN
      << $opName = "()"; >>
   |  
      TOK_LSQUARE TOK_RSQUARE
      << $opName = "[]"; >>
   |  
      os:OPTOR_SIMPLE_TOKCLASS
      << $opName = $os->getText(); >>
   ;

/////////////////////////////////////////////////////////////////////////
// Rule: parameter_declaration.  Matches an optionally-assignment-
// initialized declaration for a function argument.
//
// Inputs:
//    CPPTypeRefI   inputType    Type of the function without this parameter.
// Outputs:
//    JLStr      argName      The name of the parameter.
//    CPPTypeRefI   outputType   Function type modified by this parameter.
/////////////////////////////////////////////////////////////////////////
parameter_declaration 
[ CPPTypeRefI inputType ] > [ JLStr argName, CPPTypeRefI outputType ] :
   << 
      CPPTypeRefI argType; 
      CPPConst::StorageClass sc; 
      CPPConst::DeclModifier dm;
      CPPExprRefI initExpr;
      ANTLRTokenPtr tok = LT(1);
      $outputType = $inputType; 
   >>
   <<;>> 

   declaration_sc_or_dm_seq > [ sc, dm ]
   declaration_specifiers_simple [ sc, dm ] > [ argType, sc, dm ]

   <<
      // Handle error for invalid storage class or decl-modifier
      if (
         (
            sc & 
            (
               CPPConst::scAUTO | CPPConst::scSTATIC | CPPConst::scEXTERN | 
               CPPConst::scTYPEDEF | CPPConst::scMSDLLImport | 
               CPPConst::scMSDLLExport | CPPConst::scMSBased
            )
         ) != CPPConst::scNONE
      )
      {
         ReportError("Invalid storage class for function parameter", tok);
      }
      if (dm != CPPConst::dmNONE)
      {
         ReportError("Declaration modifier not allowed for function parameter", tok);
      }
   >>
   declarator_opt_abstract [ argType ] > [ $argName, argType ]
   { 
      TOK_ASSIGNEQUAL 
      expr_assignment > [ initExpr ]
   }
   << 
      $outputType = LookupFunctionArgType($inputType, argType, tok); 
      if ($argName.length() != 0)
      {
         DeclareObjectOrTypedef(
            $argName,
            argType,
            sc,
            CPPConst::dmNONE,
            initExpr,
            GetCurrentScope(),
            StrListRefI()
         );
      }
   >>
   ;

/////////////////////////////////////////////////////////////////////////
// Rule: parameter_declaration_list.  Matches a comma-separated list 
// of optionally-assignment-initialized declarations.
//
// Inputs:
//    CPPTypeRefI    inputType    Type of the function without parameters.
// Outputs:
//    StrListRefI    argNameList  List of argument names in the order
//                                that they were declared.  Abstract
//                                declarator arguments have empty name.
//    CPPTypeRefI    outputType   Function type modified by parameters.
/////////////////////////////////////////////////////////////////////////
parameter_declaration_list
[ CPPTypeRefI inputType ] > 
[ StrListRefI argNameList, CPPTypeRefI outputType ] :

   << 
      JLStr argName;
      $argNameList = MakeStrList(); 
      $outputType = $inputType;
   >>
   <<;>>

   parameter_declaration [ $inputType ] > [ argName, $outputType ]
   << AddStrToList($argNameList, argName); >>
   ( 
      // An ambiguity regarding TOK_COMMA arises because the TOK_COMMA
      // may either start another parameter:
      //     foo(int x, int y);
      // or it may separate the trailing ellipsis:
      //     foo(int x, ...)
      // This predicate checks which case it is

      (TOK_COMMA)? => <<LT(2)->getType() != TOK_ELLIPSIS>>?
      TOK_COMMA
      parameter_declaration [ $outputType ] > [ argName, $outputType ]
      << AddStrToList($argNameList, argName); >>
   )*
   ;

/////////////////////////////////////////////////////////////////////////
// Rule: parameter_list.  Matches a comma-separated list of optionally-
// assignment-initialized declarations, with an optional trailing ellipses.
//
// Inputs:
//    CPPTypeRefI    inputType    Type of the function with return.
// Outputs:
//    StrListRefI    argNameList  List of argument names in the order
//                                that they were declared.  Abstract
//                                declarator arguments have empty name.
//    CPPTypeRefI    outputType   Function type modified by arguments and
//                                ellipses, if present.
/////////////////////////////////////////////////////////////////////////
parameter_list 
[ CPPTypeRefI inputType ] > 
[ StrListRefI argNameList, CPPTypeRefI outputType ] :
   << $outputType = $inputType; >>
   <<;>>

      parameter_declaration_list [ $inputType ] > [ $argNameList, $outputType ]
      { 
         {TOK_COMMA} TOK_ELLIPSIS
         << $outputType = LookupFunctionEllipsesType($outputType); >>
      }
   |  
      TOK_ELLIPSIS
      << $outputType = LookupFunctionEllipsesType($inputType); >>
   ;

/////////////////////////////////////////////////////////////////////////
// Rule: parameter_list_opt.  Matches a an optional parameter_list
//
// Inputs:
//    CPPTypeRefI    inputType    Type of the function with return.
// Outputs:
//    StrListRefI    argNameList  List of argument names in the order
//                                that they were declared.  Abstract
//                                declarator arguments have empty names.
//    CPPTypeRefI    outputType   Function type modified by arguments and
//                                ellipses, if present.
/////////////////////////////////////////////////////////////////////////
parameter_list_opt
[ CPPTypeRefI inputType ] > 
[ StrListRefI argNameList, CPPTypeRefI outputType ] :
   <<;>>
   <<;>>

   (
         // This predicate resolves the ambiguity upon seeing TOK_VOID,
         // which can either start:
         //    foo(void)      // empty parameter list
         // or
         //    foo(void *p)   // parameter declaration
         (TOK_VOID)? => <<LT(2)->getType() == TOK_RPAREN>>?

         { TOK_VOID }
         << 
            $outputType = $inputType;
            $argNameList = MakeStrList(); 
         >>
      |
         parameter_list [ $inputType ] > [ $argNameList, $outputType ]
   )
   ;

/////////////////////////////////////////////////////////////////////////
// Rule: pragma_line.  Matches #pragma "line" forwarded from preprocessor
/////////////////////////////////////////////////////////////////////////
pragma_line :
   <<;>>
   <<;>>

   // Action called before pragma token is consumed
   // HandlePragma() directly processes LT(i).
   << HandlePragma(); >>
   TOK_PNDPRAGMA
   (~TOK_PRAGMA_END)*
   TOK_PRAGMA_END        // token synthesized by preprocessor
   ;

/////////////////////////////////////////////////////////////////////////
// Rule: qualified_id.  Matches a qualified identifier ::T::B::foo
// (including dtor, "operator").  If allowType is true, then this 
// will also match type names and ctors.  If allowCtor is true then 
// it will match ctors.  dtors are always matched.
//
// Inputs:
//    bool   allowType     If true, then the qualified ID may be a type (or ctor)
//    bool   allowCtor     If true, then the qualified ID may be a ctor if it
//                         matches the enclosing class name.
//    bool   allowAnyDtor  If true, then the qualified ID may be a dtor even
//                         if it does not match the enclosing class name.
// Outputs:
//    JLStr  name          The name part of the qualified ID.
//    bool   usedInnerTemplateParms     
//                         true if the scope override used any of
//                         the template parameters from the
//                         innermost template head.
//
// Modified states: If there is an explicit scope-resolution present,
//    then the new scope will be pushed in the stack in "resolved"
//    mode.  The caller should use SaveScope()/RestoreScope().
/////////////////////////////////////////////////////////////////////////
qualified_id 
[ bool allowType, bool allowCtor, bool allowAnyDtor ] > 
[ JLStr name, bool usedInnerTemplateParms ] :

   << 
      // In case of error, this will rollback scope modifications
      ScopePopper scopePopper(this); 
   >>
   << ; >>

   // Match optional typename iff allowType is true
   {
      <<allowType>>?
      TOK_TYPENAME
   }

   scope_override > [ $usedInnerTemplateParms ]

   (
         <<bool isDtor = false;>>
         { TOK_TILDE << isDtor = true; >> }

         id:TOK_ID
         
         (
               // This predicate is to disambiguate cases like:
               //      A::foo<X> x;
               // which should be parsed as a relational expression
               // only if A::foo is not a templated name.
               (TOK_LESSTHAN)? => <<IsTemplateName($id->getText())>>?

               // Save the scope stack before processing template arguments,
               // because template_specialization pops the scope stack.
               << 
                  ScopeSaver scopeSaver(this, scopePopper.Get());
                  CPPTypeRefI specializationType; 
                  bool tmpBool;
               >>

               // Match a specialized class or function template
               template_specialization 
                  [ $id->getText(), scopePopper.Get(), $id ] > 
                  [ $name, specializationType, tmpBool ]
            |
               // normal id -- no template specialization
               << $name = $id->getText(); >>
         )

         << 
            bool qualifiedIdValid = 
               (!isDtor && allowType && IsTypeName($name)) ||
               (isDtor && (allowAnyDtor || IsEnclosingClassName($name))) ||
               ((isDtor || allowCtor) && IsEnclosingClassName($name)) ||
               (!isDtor && !IsScopeName($name));
         >>

         // Validating predicate rejects types unless allowType is set,
         // rejects ctors unless allowType or allowCtor is set.
         // Namespace names are always rejected
         << qualifiedIdValid >>?

         // If it was a dtor then construct the dtor name
         <<
            if (isDtor) {
               $name = "~" + $name;
            }
         >>

      |  
         TOK_OPERATOR
         (
         <<;>>
               << JLStr tmpName; >>
               optor_simple > [ tmpName ]
               << $name = "operator" + tmpName; >>
            |
               (
                  // NOTE: The () enclosing this alt is important, because
                  // the ScopeSaver restores the scope stack when this alt ends.

                  // Save any scope override, but pop it while looking up the
                  // conversion type.  For example:
                  //     A::operator B
                  //                 ^
                  //                 do not look for B in A::
                  << 
                     ScopeSaver scopeSaver(this, scopePopper.Get());
                     scopePopper.Restore();
                     CPPTypeRefI conversionType; 
                  >>

                  optor_conversion > [ conversionType ]
                  << $name = "operator " + FormatTypeName(conversionType); >>
               )
         )
   )

   // Prevent scope rollback on success
   << scopePopper.Commit(); >>
   ;

/////////////////////////////////////////////////////////////////////////
// Rule: qualified_type.  Matches a type, optionally qualified by
// scope-resolution prefixes.
//
// Inputs: none
// Outputs:
//    CPPTypeRefI   type         The type of the recognized identifier.
//                               or the "void" type if no valid type 
//                               was found.
/////////////////////////////////////////////////////////////////////////
qualified_type > [ CPPTypeRefI type ] :
   << 
      ScopePopper scopePopper(this); 
      bool hasTypenameKeyword = false;
      bool dummy;
   >>
   <<;>>

   // Anything preceded by 'typename' is assumed to be a type
   { 
      TOK_TYPENAME 
      << hasTypenameKeyword = true; >>
   }
   
   scope_override > [ dummy ]

   id:TOK_ID

   // Validating predicate
   << hasTypenameKeyword || IsTypeName($id->getText()) >>?

   (
         // This predicate resolves an ambiguity arising from things like:
         //     operator A < X 
         // where it is syntactically legal to use the operator function
         // address in a relational expression.  However, this is semantically
         // invalid, so we are OK in always binding the "<" to the class
         (TOK_LESSTHAN)? => <<1>>?
         
         (
            // trailing template arguments -- create/find a class specialization
            << 
               JLStr specializationName; 
               bool tmpBool;
            >>
            template_specialization 
               [ $id->getText(), scopePopper.Get(), $id ] > 
               [ specializationName, $type, tmpBool ]
         )
      |
         // empty
         <<
            // Lookup the type corresponding to the name
            // This will work even if IsTypeName() returned false,
            // as long as the scope override is a special template scope.
            $type = LookupType($id->getText(), $id);
         >>
   )
   ;

/////////////////////////////////////////////////////////////////////////
// scope_override: Match the A::B::C:: or nothing
//
// Inputs: none
// Outputs:
//    bool   usedInnerTemplateParms   true if the scope segment used any of
//                                    the template parameters from the
//                                    innermost template head.
//
// Modified states: If there is an explicit scope-resolution present,
//    then the new scope will be pushed on the stack in "resolved"
//    mode.  The caller should use SaveScope()/RestoreScope().
/////////////////////////////////////////////////////////////////////////
scope_override > [ bool usedInnerTemplateParms ] :
   <<
      CPPScopeState scopeState = SaveScope();
      $usedInnerTemplateParms = false;
   >>
   <<;>>
   {
      TOK_SCOPE
      << EnterResolvedGlobalScope(); >>
   }

   {
      // This predicate validates that the following tokens 
      // syntactically match a scope segment.
      << IsScopeSegment(1, true) >>?
      scope_segment_list [ scopeState, true ] > [ $usedInnerTemplateParms ]
   }
   ;


/////////////////////////////////////////////////////////////////////////
// scope_segment: Match a scope override segment like A:: or A<int>::
//
// Inputs:
//    ScopeState   scopeState   The "base" scope state to which the
//                              scope stack should be popped before
//                              entering the new resolved scope.
//    bool         isTemplate   If true, then this rule was preceded
//                              by the "template" keyword.
//    bool         isVerified   true if the caller of this method has
//                              already verified the scope segment syntax
//                              by calling IsScopeSegment()
// Outputs:
//    CPPTypeRefI  classType    If the scope segment matched is a 
//                              class name or class template name, 
//                              then this contains the type of
//                              the recognized class.
//                              ****OTHERWISE THIS IS ZERO****
//    bool         usedInnerTemplateParms
//                              true if the scope segment used any of
//                              the template parameters from the
//                              innermost template head.
//
// Modified states: The scope stack will enter the specified scope
// and any base-class scopes in "resolved" mode.
// The caller should use SaveScope()/RestoreScope().
/////////////////////////////////////////////////////////////////////////
scope_segment 
   [ CPPScopeState scopeState, bool isTemplate, bool isVerified ] > 
   [ CPPTypeRefI classType, bool usedInnerTemplateParms ] :

   <<JLStr scopeName;>>
   <<;>>

   id:TOK_ID

   (
         // Templated class specifier (class specialization).
         (
            // validating predicate
            <<$isTemplate || IsTemplateName($id->getText()) >>?

            // We need this predicate (which is purely a syntax check)
            // to make sure that the nesting <> are followed by ::, because
            // otherwise template_specialization tries to specialize the
            // template name even when it's not a scope, which is wrong
            // in the declaration of partial template specializations, e.g.:
            //    template <class T> class A<T*> {};
            //                                 |
            //                        don't want to specialize here
            << isVerified || NestedBracketsAreFollowedByScope() >>?

            template_specialization
               [ $id->getText(), scopeState, $id ] > 
               [ scopeName, $classType, $usedInnerTemplateParms ]
            TOK_SCOPE
         )
      |
         // Non-templated scope name
         TOK_SCOPE
         << scopeName = $id->getText(); >>

         // Lookup the class type if it's not a namespace.
         // We cannot perform a positive test like "isClassName"
         // because template placeholder scopes are weird.
         << 
            if (!IsNamespaceName(scopeName))
            {
               // Even though this may find a non-class type, it is OK
               // because the scope entry will fail and reset $classType.
               $classType = LookupType(scopeName, $id);
            }
         >>
   )

   <<
      // Find the scope to enter
      CPPScopeRefI scope = LookupScope(scopeName, $id);
      // Pop the scope stack and enter the new scope
      RestoreScope(scopeState);
      EnterScope(scope, true);
   >>
   ;


/////////////////////////////////////////////////////////////////////////
// scope_segment_list: Match A::B::C::, where all but the first
// scope segment may have a preceding "template" specifier
//
// Inputs:
//    ScopeState   scopeState   The "base" scope state to which the
//                              scope stack should be popped before
//                              entering the new resolved scope.
//    bool         isVerified   true if the caller of this method has
//                              already verified the scope segment syntax
//                              by calling IsScopeSegment()
// Outputs:
//    bool         usedInnerTemplateParms
//                              true if the scope segment used any of
//                              the template parameters from the
//                              innermost template head.
//
// Modified states: The scope stack will enter the specified scope
// and any base-class scopes in "resolved" mode.
// The caller should use SaveScope()/RestoreScope().
/////////////////////////////////////////////////////////////////////////
scope_segment_list 
[ CPPScopeState scopeState, bool isVerified ] > 
[ bool usedInnerTemplateParms ] :

   <<
      ScopePopper scopePopper(this);
      CPPTypeRefI tmpType;
   >>
   <<;>>

   scope_segment [ scopeState, false, isVerified ] > [ tmpType, $usedInnerTemplateParms ]

   (
      // This predicate validates that the following tokens 
      // syntactically match a scope segment.
      << IsScopeSegment(1, true) >>?

      // template keyword indicates that id is a template
      << 
         bool isTemplate = false; 
         bool tmpUsedInnerTemplateParms;
      >>
      { TOK_TEMPLATE << isTemplate = true;>> }

      scope_segment [ scopeState, isTemplate, true ] > [ tmpType, tmpUsedInnerTemplateParms ]
      << $usedInnerTemplateParms |= tmpUsedInnerTemplateParms; >>
   )*
   <<scopePopper.Commit();>>
   ;


/////////////////////////////////////////////////////////////////////////
// Rule: simple_type_specifier.  Match a type specifier that can be
// used before a function-like cast.
//
// Inputs: none
// Outputs:
//    CPPTypeRefI   cppType   The type specifier
/////////////////////////////////////////////////////////////////////////
simple_type_specifier > [ CPPTypeRefI cppType ] :
   <<;>>
   <<;>>

   (
   <<;>>
         << 
            CPPConst::TypeSpecifier ts; 
            ANTLRTokenPtr errTok = LT(1);
         >>
         builtin_type_specifier > [ ts ]
         << $cppType = LookupBuiltinType(ts, errTok); >>
      |  
         qualified_type > [ $cppType ]
   )
   ;

///////////////////////////////////////////////////////////////////////
// There are a host of ambiguities between declaration and expression.
// These are resolved by backtracking out of declaration if it
// doesn't work, because declaration requires non-empty 
// declaration-specifiers.  For example:
//      int f();   // declaration
//      int a;     // declaration
//      f();       // expression-statement
//      a;         // expression-statement
// The standard (WP May96 s6.8.4) says to treat would-be
// declarations that are lacking any declaration-specifiers as
// statements.
//
// Note that the plain ";" statement is handled by statement_declaration
///////////////////////////////////////////////////////////////////////
statement :
   <<;>>
   <<;>>
      statement_compound
   |  statement_selection
   |  statement_iteration
   |  statement_jump
   |  statement_try_block
   |  statement_throw
   |  statement_asm

   |
      // There is a TOK_ID ambiguity arising between the label syntax
      // and the remaining rules.  This predicate checks for TOK_COLON
      // following the label id.
      (TOK_ID)? => <<LT(2)->getType() == TOK_COLON>>?
      statement_labeled
   |
      // Expressions and declarations are ambiguous to an arbitrary lookahead.
      // Try the declaration and backtrack out if it is not acceptable.
      // We would like to do the expression first since it is more common, but
      // there is a rule that says "anything that can be a declaration should 
      // be parsed as a declaration".
      ( statement_declaration )?
   |  
      { expression } TOK_SEMICOLON
   ;

///////////////////////////////////////////////////////////////////////
// Rule: statement_asm.  Match an assembly-language "statement", like
//   asm {
//       ...any tokens...
//   }
//
// or 
//   asm ...instructions...
//
// The last case is really hard because (apparently) either end-of-line
// or a closing curly (and maybe other stuff!?!?) can end the statement
// since we don't have end-of-line here, we punt on the second case
// and just go until we hit a curly or semicolon, which at least accounts
// for the common cases of:
//     void foo() { asm mov eax, ebx }
// or
//     void foo() {
//         asm mov aex,abx;
//         int x = 3;
//     }
///////////////////////////////////////////////////////////////////////
statement_asm :
   <<;>>
   <<;>>
   ( TOK_ASM | TOK_MSEXT_ASM )

   (
         // Predicate tells ANTLR to always treat a "{" following
         // an asm keyword as part of the asm statement.
         (TOK_LCURLY)? => <<1>>?
         TOK_LCURLY
         ( ~TOK_RCURLY )*
         TOK_RCURLY 
      |
         // #pragma approx tells ANTLR to ignore ambiguities and assume
         // that any non-curly/semicolon continues the asm statement
         #pragma approx
         (~CURLY_SEMICOLON_TOKCLASS)*
   )
   ;

///////////////////////////////////////////////////////////////////////
// Rule: statement_catch.  Matches a catch block following a 
// try block, like:
//      catch (expression) statement
///////////////////////////////////////////////////////////////////////
statement_catch :
   TOK_CATCH
   TOK_LPAREN 
   statement_exception_declaration 
   TOK_RPAREN
   statement_compound
   ;

///////////////////////////////////////////////////////////////////////
// Rule: statement_compound.  Matches a block statement like:
//   {
//       ...statements...
//   }
//
// Inputs: none
// Outputs: none
///////////////////////////////////////////////////////////////////////
statement_compound :
   << 
      ScopePopper scopePopper(this);
      EnterLocalScope();
   >>
   <<;>>
   TOK_LCURLY 
   { statement_list }
   TOK_RCURLY 
   ;

/////////////////////////////////////////////////////////////////////////
// Rule statement_declaration: Matches the simplest form of declaration.
// Inputs: none
// Outputs: none
/////////////////////////////////////////////////////////////////////////
statement_declaration :
   <<
      CPPTypeRefI cppType;
      CPPConst::StorageClass sc;
      CPPConst::DeclModifier dm;
      ScopePopper scopePopper(this);
   >>
   << ; >>

   // Get leading storage class and modifiers
   declaration_sc_or_dm_seq > [ sc, dm ]

   // decl-specs are required for declarations in a code block
   declaration_specifiers [ sc, dm ] > [ cppType, sc, dm ]

   { 
      declarator_init_list [ cppType, sc, dm ]
   } 
   TOK_SEMICOLON
   ;

///////////////////////////////////////////////////////////////////////
statement_exception_declaration :
   <<;>>
   <<;>>
   (
         // Not yet implemented
         parameter_declaration_list [ CPPTypeRefI() ]
      |  
         TOK_ELLIPSIS
   )
   ;

///////////////////////////////////////////////////////////////////////
statement_iteration :
   <<;>>
   <<;>>
      TOK_WHILE TOK_LPAREN expression TOK_RPAREN statement
   |  
      TOK_DO statement TOK_WHILE TOK_LPAREN expression TOK_RPAREN TOK_SEMICOLON
   |  
      TOK_FOR TOK_LPAREN
      (
            ( 
               statement_declaration 
            )?
         |  
            { expression } TOK_SEMICOLON 
      )
      { expression } 
      TOK_SEMICOLON
      { expression }
      TOK_RPAREN 
      statement
   ;

///////////////////////////////////////////////////////////////////////
statement_jump :
   <<;>>
   <<;>>
      TOK_GOTO TOK_ID TOK_SEMICOLON
   |  TOK_CONTINUE TOK_SEMICOLON
   |  TOK_BREAK TOK_SEMICOLON
   |  TOK_RETURN { expression } TOK_SEMICOLON
   ;

///////////////////////////////////////////////////////////////////////
statement_labeled :  
   <<;>>
   <<;>>
      TOK_ID TOK_COLON statement
   |  TOK_CASE expr_constant TOK_COLON statement
   |  TOK_DEFAULT TOK_COLON statement
   ;

///////////////////////////////////////////////////////////////////////
statement_list :
   <<;>>
   <<;>>
   (
      statement
      exception <<
         ReportExceptionError(ex);
         consume();
         consumeUntil(STATEMENT_SYNC_TOKCLASS_set);

         // This is a hack to avoid infinite loop -- should do better
         if (LA(1) == TOK_Eof) throw;
      >>
   )+
   ;

///////////////////////////////////////////////////////////////////////
// NOTE: cannot remove TOK_ELSE ambiguity, but it parses correctly.
// The warning is removed with the #pragma, which tells ANTLR to
// always bind "else" to the nearest "if"
///////////////////////////////////////////////////////////////////////
statement_selection :
   <<;>>
   <<;>>
      TOK_IF 
      TOK_LPAREN 
      expression 
      TOK_RPAREN 
      statement

      // This #pragma approx tells ANTLR to ignore the "dangling else"
      // ambiguity.  It is correctly resolved and bound to the nearest
      // enclosing if(...)
      #pragma approx
      { 
         TOK_ELSE 
         statement 
      }
   |
      TOK_SWITCH 
      TOK_LPAREN 
      expression 
      TOK_RPAREN 
      statement
   ;

///////////////////////////////////////////////////////////////////////
// This is an expression of type void according to the ARM, which
// to me means "statement"; it removes some ambiguity to put it in
// as a statement also.
///////////////////////////////////////////////////////////////////////
statement_throw :
   <<;>>
   <<;>>
   TOK_THROW 
   { expr_assignment } 
   TOK_SEMICOLON
   ;

///////////////////////////////////////////////////////////////////////
statement_try_block :
   <<;>>
   <<;>>
   TOK_TRY statement_compound (statement_catch)*
   ;

/////////////////////////////////////////////////////////////////////////
storage_class_specifier > [CPPConst::StorageClass sc] :
   <<;>>
   <<;>>
      TOK_AUTO                             <<$sc = CPPConst::scAUTO;>>
   |  TOK_REGISTER                         <<$sc = CPPConst::scREGISTER;>>
   |  TOK_STATIC                           <<$sc = CPPConst::scSTATIC;>>
   |  TOK_EXTERN                           <<$sc = CPPConst::scEXTERN;>>
   |  TOK_TYPEDEF                          <<$sc = CPPConst::scTYPEDEF;>>
   // Compiler-specific extensions
   |  TOK_MSEXT_BASED TOK_LPAREN TOK_ID TOK_RPAREN     <<$sc = CPPConst::scMSBased; >>
   |  
      TOK_MSEXT_DECLSPEC 
      TOK_LPAREN 
      declSpecId:TOK_ID
      <<
         if (strcmp($declSpecId->getText(), "dllimport") == 0)
         {
            $sc = CPPConst::scMSDLLImport;
         } else if (strcmp($declSpecId->getText(), "dllexport") == 0)
         {
            $sc = CPPConst::scMSDLLExport;
         } else {
            $sc = CPPConst::scNONE;
         }
      >>
      TOK_RPAREN
   ;

/////////////////////////////////////////////////////////////////////////
// Rule: template_argument.   Matches a single argument used during
// template instantiation.  This is either a typename or an expression.
//
// Here expr_assignment was changed to expr_shift to rule out
//  x< 1<2 > which causes ambiguities. As a result, these can be used only
//  by enclosing parentheses x<(1<2)>.
//
// Inputs: none
// Outputs: 
//    bool   usedInnerTemplateParms    true if the scope segment used any 
//                                     of the template parameters from the
//                                     innermost template head.
/////////////////////////////////////////////////////////////////////////
template_argument > 
[ TemplateArgumentRefI argRef, bool usedInnerTemplateParms ] :

   <<$usedInnerTemplateParms = false;>> 
   <<;>>

      // A simple semantic predicate here is insufficient to 
      // disambiguate the nasty declaration/expression ambiguities
      // because they are arbitrarily complex, so we backtrack.
      // We put the type_name alternate first because of the rule that
      // "anything that can be a type should be treated as such".
      // for example:
      //     T<bool(x)>    // expression
      //     T<bool(*)>    // type: ptr-to-bool
      (
         << CPPTypeRefI type; >>
         type_name > [ type ]
         << 
            // Check to see if the type argument uses a template placeholder
            // from the innermost template head.
            $usedInnerTemplateParms = TypeUsesInnerTemplateParms(type);

            // Create the template argument
            $argRef = MakeTemplateArgumentType(type); 
         >>
      )?
   |
      (
         << CPPExprRefI expr; >>

         // Use expr_shift to avoid conflict with closing ">" in template
         expr_shift > [ expr ]
         << $argRef = MakeTemplateArgumentValue(expr); >>
      )
   ;

/////////////////////////////////////////////////////////////////////////
// Rule: template_argument_list.  Matches a comma-separated list
// of actual template parameters (i.e., those used during template
// instantiation).
//
// Inputs: none
// Outputs:
//    TemplateArgumentListRefI   argList  List of actual template arguments.
//    bool                       usedInnerTemplateParms     
//                                        true if the scope segment used any of
//                                        the template parameters from the
//                                        innermost template head.
/////////////////////////////////////////////////////////////////////////
template_argument_list  > 
[ TemplateArgumentListRefI argList, bool usedInnerTemplateParms ] :

   << 
      TemplateArgumentRefI argRef;
      $argList = MakeTemplateArgumentList(); 
   >>
   <<;>>

   template_argument > [ argRef, $usedInnerTemplateParms ]
   << AddTemplateArgumentToList($argList, argRef); >>
   (
      TOK_COMMA
      << bool tmpUsedInnerTemplateParms; >>
      template_argument > [ argRef, tmpUsedInnerTemplateParms ]
      << 
         AddTemplateArgumentToList($argList, argRef); 
         $usedInnerTemplateParms |= tmpUsedInnerTemplateParms;
      >>
   )*
   ;

/////////////////////////////////////////////////////////////////////////
// Rule: template_declaration.  Matches "template<...>" followed by
// a class or function declaration.  Well, not quite -- there are some
// special cases/restrictions on templates that don't apply to normal
// declarations.
//
// Inputs:
//    CPPScopeRefI  preTemplateScope  scope that encapsulates any nested
//                                    template scopes outside of this one
//    bool          memberDecl        true if the declaration is inside a class
// Outputs: none
/////////////////////////////////////////////////////////////////////////
template_declaration [ CPPScopeRefI preTemplateScope, bool memberDecl ] :
   << 
      ScopePopper scopePopper(this);
      TemplateParameterListRefI templateParameters;
      CPPDeclRefI rawDecl;
   >> 
   <<;>>

   template_head > [ templateParameters ]

   << 
      // Remember the template parameter scope
      ScopePopper templateScopePopper(this);
      CPPScopeRefI templateParameterScope = GetCurrentScope(); 

      // Bump template level; restore it upon return/exception
      TemplateLevelIncrementer templateLevelIncrementer(this);
   >>

   (
         // Nested template declaration, like:
         //    template<class T1> template <class T2> class A<T1>::B<T2> {...};

         template_declaration [ preTemplateScope, memberDecl ]
      |
         // There is an ambiguity between this alternate
         // and the declaration_specifiers of a function template
         // or class template static data declaration.
         // The ambiguity is between the a class declaration
         //    template <class T> class A {...};
         // and a function or static data member declaration
         //    template <class T> class A foo();
         //    template <class T> class A C::a;
         //
         // We wrap this alternate in a syntactic predicate and backtrack
         // out if it fails.  Since the non-template-class forms starting
         // with class/struct/union are rare, the backtracking is
         // seldom done.

         (
            template_declaration_class [ 
               preTemplateScope, 
               templateParameters, 
               memberDecl, 
               templateScopePopper.Get() 
            ]
         )?
      |
         template_declaration_function_or_data [ 
            preTemplateScope, 
            templateParameters, 
            memberDecl, 
            templateScopePopper.Get(),
            templateParameterScope
         ]
   )
   ;



/////////////////////////////////////////////////////////////////////////
// Rule: template_declaration_class.  Matches a class declaration following
// a template_head, like
//     template <class T> struct A;               // declaration
//     template <class T> class A : public B {};  // definition
//     template <class T> class A<T*> {};         // specialization
//
// Inputs:
//    CPPScopeRefI               preTemplateScope     scope that encapsulates any nested
//                                                    template scopes outside of this one
//    TemplateParameterListRefI  templateParameters   The template parameter list 
//    bool                       memberDecl           true if the declaration is inside a class
//    CPPScopeState              templateScopeState   The scope stack state of the template
//                                                    parameter scope.
// Outputs: none
/////////////////////////////////////////////////////////////////////////
template_declaration_class
[ 
   CPPScopeRefI preTemplateScope, 
   TemplateParameterListRefI templateParameters,
   bool memberDecl,
   CPPScopeState templateScopeState
] :

   << 
      ANTLRTokenPtr errTok = LT(1);
      CPPTypeRefI classType;
      CPPDeclRefI templateDecl;
      CPPTypeRefI templateType;
      JLStr templateName;
      CPPConst::TypeSpecifier ts;
      // Create a token collector to grab the body of the declaration.
      TokenCollector tokenCollector(this); 
      bool usedInnerTemplateParms;
      TemplateArgumentListRefI specializationArgList = MakeTemplateArgumentList();
   >>

   (
         TOK_STRUCT <<ts = CPPConst::tsSTRUCT;>>
      |  TOK_UNION  <<ts = CPPConst::tsUNION;>>
      |  TOK_CLASS  <<ts = CPPConst::tsCLASS;>>
   )

   // We must know if an inner template parameter was used in the scope
   // override, because if it was then the declaration is a class template
   // member, *not* itself a template.
   scope_override > [ usedInnerTemplateParms ] 

   // If there was a scope override and the
   // override "consumes" the template parameters at this nesting
   // level, then the declaration is not a template declaration, but
   // rather a member of a class template.

   // If there was a scope override, then use the overridden
   // scope as the declaration scope, otherwise use the
   // pre-template scope.  
   << 
      CPPScopeRefI declContextScope;
      if ($templateScopeState != SaveScope())
      {
         declContextScope = GetCurrentScope(); 
         // Pop scope stack to process template specialization args
         RestoreScope($templateScopeState);
      } else {
         declContextScope = $preTemplateScope;
      }
   >>

   // Name of the class
   id:TOK_ID

   {
      // Optional specialization arguments
      TOK_LESSTHAN
      << bool tmpBool; >>
      template_argument_list > [ specializationArgList, tmpBool ]
      TOK_GREATERTHAN
   }

   (
   <<;>>
         // Definition
         << 
            // If we get here then we are committed to the class-template alt
            CommitGuess();

            bool classDefined;
            if (usedInnerTemplateParms)
            {
               // Plain class decl inside class template
               DefineClassType(
                  $id->getText(), 
                  ts, 
                  declContextScope,
                  errTok,
                  classType,
                  templateDecl,
                  templateName
               );
               classDefined = true;
            } else {
               // Define the class template.
               classDefined = DefineClassTemplateType(
                  $id->getText(), 
                  ts, 
                  declContextScope,
                  $templateParameters,
                  specializationArgList,
                  errTok,
                  classType,
                  templateType,
                  templateDecl, 
                  templateName
               );
            }
         >>

         {
            base_clause [ classType ]
         }

         <<
            // Enter the class template scope
            if (classDefined)
            {
               EnterClassScope(classType, false);
            }
         >>

         TOK_LCURLY
         (
            declaration_member
            exception <<
               ReportExceptionError(ex);
               consume();
               consumeUntil(MEMBER_SYNC_TOKCLASS_set);
            >>
         )*
         TOK_RCURLY

         // subrule forces consume of semicolon before getting token list
         (TOK_SEMICOLON)

         << 
            // If the definition was successful, *and* it's a real template 
            // declaration, then set the template tokens.
            // TODO: we must set the token list for plain old class template members,
            if (classDefined && !usedInnerTemplateParms)
            {
               SetTokensInTemplateDeclaration(
                  templateType, 
                  $templateParameters, 
                  tokenCollector.Get()
               );
            }
         >>
      |
         // declaration

         // If we get here then we are committed to template-class
         <<CommitGuess();>>

         // subrule forces consume of semicolon before getting token list
         (TOK_SEMICOLON)

         <<
            // Declare the class template.  Note that the declaration scope
            // is always well-defined -- there is no searching the scope stack
            // for declarations in containing scopes.
            if (usedInnerTemplateParms)
            {
               // Plain old class decl inside class template (this is invalid I think)
               DeclareClassType(
                  $id->getText(), 
                  ts, 
                  declContextScope,
                  errTok,
                  templateType,
                  templateDecl, 
                  templateName
               );
            } else {
               // Declare class template
               DeclareClassTemplateType(
                  $id->getText(), 
                  ts, 
                  declContextScope,
                  $templateParameters,
                  specializationArgList,
                  errTok,
                  classType,
                  templateType,
                  templateDecl, 
                  templateName
               ); 

               SetTokensInTemplateDeclaration(
                  templateType, 
                  $templateParameters, 
                  tokenCollector.Get()
               );
            }
         >>
   )
   ;


/////////////////////////////////////////////////////////////////////////
// Rule: template_declaration_function_or_data.  Matches a function
// or data declaration following a template_head
//
// Inputs:
//    CPPScopeRefI               preTemplateScope     scope that encapsulates any nested
//                                                    template scopes outside of this one
//    TemplateParameterListRefI  templateParameters   The template parameter list 
//                                                    from template_head.
//    bool                       memberDecl           true if the declaration is inside a class
//    CPPScopeState              templateScopeState   The scope stack state of the template
//                                                    parameter scope.
//    CPPScopeRefI               templateParameterScope
//                                                    The scope in which the template
//                                                    parameters are declared.
// Outputs: none
/////////////////////////////////////////////////////////////////////////
template_declaration_function_or_data
[ 
   CPPScopeRefI preTemplateScope, 
   TemplateParameterListRefI templateParameters,
   bool memberDecl,
   CPPScopeState templateScopeState,
   CPPScopeRefI templateParameterScope
] :

   <<
      ANTLRTokenPtr errTok = LT(1);
      // Create a token collector to grab the body of the declaration.
      TokenCollector tokenCollector(this); 

      // Information used to make the declaration
      CPPConst::StorageClass sc;
      CPPConst::DeclModifier dm;
      JLStr declName;
      StrListRefI argNameList;
      CPPScopeRefI declScope;
      CPPScopeRefI argScope;
      CPPTypeRefI declType;
      CPPExprRefI initExpr;
      bool usedInnerTemplateParms;
   >>

   // Start with common storage class and declaration modifiers
   declaration_sc_or_dm_seq > [ sc, dm ]

   (
   <<;>>
         // Predicate for ctors.  ctors are not valid as member decls.
         // They are only valid as external decls, e.g.:
         //     template <class T> A<T>::A() {}
         <<!memberDecl>>?

         // ctors are confusing with decl-specs because the leading
         // type name could be part of decl-specs.  So backtrack out.
         // We cannot do the decl-specs first because:
         //     class A { A(X); };
         // looks like a declaration of X as type A
         (
            // Ctor definition (declarations cannot be templated)
            ctor_declarator > 
               [ declName, argNameList, declScope, argScope, declType, usedInnerTemplateParms ]

            <<
               ScopePopper ctorScopePopper(this);
               if (declScope != GetCurrentScope())
               {
                  EnterScopeContext(declScope, false);
               }
               // DWP says that ctor is not entered in the class scope under
               // any name that can be uttered.  So we use "$ctor" instead.
               declName = "$ctor";
               EnterScope(argScope, false);
            >>
            { ctor_initializer }
            statement_compound
         )?
      |
         (
            // Optional declaration specifiers to handle cases like:
            //    template <class T> A<T>::operator X() {}

            (
                  // Backtrack out of decl-specs; it's too hard to write
                  // a predicate that parses template specifiers
                  (
                     declaration_specifiers_simple [ sc, dm ] > [ declType, sc, dm ]
                  )?
               |
                  /* empty */
                  // The decl-spec type defaults to "void".  This is only
                  // valid for operator.  Old-style functions should be
                  // caught with a warning.  Other declarations are disallowed.
                  <<declType = LookupVoidType();>>
            )

            // Both function and init-declarator start with a declarator
            declarator 
               [ declType, CPPConst::ccCDECL ] > 
               [ declName, argNameList, declScope, argScope, declType, usedInnerTemplateParms ]

            (
            <<;>>
                  // init-declarator alternative
                  // Note that this alt also handles pure-virtual functions.

                  // Optional initializer
                  {  
                        TOK_ASSIGNEQUAL 
                        initializer > [ initExpr ]
                     |  
                        TOK_LPAREN 
                        expression_list > [ initExpr ] 
                        TOK_RPAREN
                  }
                  TOK_SEMICOLON
               |
                  // Declarator with function body.

                  // Choose statement_compound for external declarations, but use
                  // generic_block for member declarations because the fields have
                  // not all been seen yet.
                  (
                        (TOK_LCURLY)? => <<memberDecl>>?
                        // We don't actually parse the function body
                        // until later to get the member definitions.
                        generic_block
                        // TODO: use the collected tokens
                     |
                        <<
                           // Set up the scope for the function body
                           ScopePopper functionScopePopper(this);
                           if (declScope != GetCurrentScope())
                           {
                              EnterScopeContext(declScope, false);
                           }
                           EnterScope(argScope, false);
                        >>
                        statement_compound
                  )
            )
         )
   )

   <<
      // If there was a scope override, then use the overridden
      // scope as the declaration scope, otherwise use the
      // pre-template scope.
      CPPScopeRefI declContextScope;
      if (templateParameterScope != declScope)
      {
         declContextScope = declScope; 
      } else {
         declContextScope = $preTemplateScope;
      }

      // Declare the object template.  Note that the declaration scope
      // is always well-defined -- there is no searching the scope stack
      // for declarations in containing scopes.

      CPPTypeRefI templateType;
      CPPDeclRefI templateDecl;
      JLStr templateName;

      if (usedInnerTemplateParms)
      {
         // untemplated member of class-template declaration
         // TODO: assign template token list
         templateDecl = DeclareObjectOrTypedef(
            declName,
            declType,
            sc,
            dm,
            initExpr,
            declContextScope,
            argNameList
         );
      } else {
         // template function declaration.  DeclareFunctionTemplate will check that
         // the declarator is a function
         if (DeclareFunctionTemplate(
               declName,
               declType,
               sc,
               dm,
               initExpr,
               declContextScope,
               argNameList,
               $templateParameters,
               errTok,
               templateType,
               templateDecl, 
               templateName
            )
         )
         {
            SetTokensInTemplateDeclaration(
               templateType, 
               $templateParameters, 
               tokenCollector.Get()
            );
         }
      }
   >>
   ;


/////////////////////////////////////////////////////////////////////////
// Rule template_head.  Matches the <> and formal parameter list for
// a template declaration.
//
// Inputs: none
// Outputs: 
//    TemplateParameterListRefI  parmList         Template parameters
//
// If successful, the scope stack will be left with an extra temporary
// scope in which the template parameters are declared.  
// Use SaveScope/RestoreScope to restore the scope stack.
/////////////////////////////////////////////////////////////////////////
template_head > [ TemplateParameterListRefI parmList ] :
   <<
      // Rollback scope stack in case of syntax error...
      ScopePopper scopePopper(this);
      EnterLocalScope();
   >>
   <<;>>
   TOK_TEMPLATE
   TOK_LESSTHAN 
   template_parameter_list > [ $parmList ]
   TOK_GREATERTHAN

   // OK to keep scope on stack now.
   <<scopePopper.Commit();>>
   ;

/////////////////////////////////////////////////////////////////////////
// Rule: template_parameter.  Matches a formal parameter
// of a template_head.  Each element of the list may be of the form 
// "class T", or it may be a normal function-parameter declaration.  
// Either form may have an initializer.
// 
// Inputs:
//    short           parameterIdx   The index of the template
//                                   parameter in the list.
// Outputs: 
//    TemplateParameterRefI   templateParm   The template parameter
/////////////////////////////////////////////////////////////////////////
template_parameter 
   [ short parameterIdx ] > 
   [ TemplateParameterRefI templateParm ] :

   << 
      ANTLRTokenPtr tok = LT(1);
      JLStr argName;
   >>
   << ; >>
   (
         // Type template parameter

         // This predicate tells ANTLR to always resolve the "class X" 
         // ambiguity in favor of the first alternate (i.e., type name).
         ( TOK_CLASS | TOK_TYPENAME )? => <<1>>?

         (
            ( TOK_CLASS | TOK_TYPENAME )
            { 
               id:TOK_ID 
               << argName = $id->getText(); >>
            }
            << TokenCollector tokenCollector(this); >>
            {
               TOK_ASSIGNEQUAL

               // We can't really parse the default value because it may
               // depend on previous template parameters, so we just collect
               // the tokens in a list and parse it later if needed.
               << 
                  CPPTypeRefI typeIgnored;
                  tokenCollector.Clear(); 
               >>

               type_name > [ typeIgnored ]
            }
            <<
               // Declare the type parameter as CPPTypeTemplatePlaceholder in 
               // the template scope, and create a scope associated with the 
               // template parameter.
               // This allows the template body to be parsed using the 
               // template parameter name.
               DeclareTemplateTypeParameter(
                  argName,
                  tokenCollector.Get(),
                  tok,
                  $parameterIdx,
                  GetTemplateLevel(),
                  $templateParm
               );
            >>
         )
      |
         // Value template parameter.  These do not require the special treatment
         // of a template type parameter, since they have a well-defined type.
         (
            << 
               CPPTypeRefI argType; 
               CPPConst::StorageClass sc; 
               CPPConst::DeclModifier dm;
               CPPExprRefI dummyExpr;
            >>

            // TODO: disallow use of previous template type parameters for
            // the type in value parameters.
            declaration_specifiers_simple  
               [ CPPConst::scNONE, CPPConst::dmNONE ] > 
               [ argType, sc, dm ]
            <<
               // Disallow storage class of decl modifier
               if (sc != CPPConst::scNONE || dm != CPPConst::dmNONE)
               {
                  ReportError("Storage class or declaration modifier not allowed for template argument", tok);
               }
            >>

            declarator_opt_abstract [ argType ] > [ argName, argType ]

            << TokenCollector tokenCollector(this); >>

            {
               TOK_ASSIGNEQUAL 

               // Since we can't really parse the default value because it may
               // depend on previous template parameters, we just collect the
               // tokens in a list and parse it later if needed.
               << tokenCollector.Clear(); >>

               // Use expr_shift to avoid conflict with closing ">" in template
               expr_shift > [ dummyExpr ]
            }

            <<
               // Declare the value parameter as CPPExprTemplatePlaceholder 
               // in the template scope.
               // This allows the template body to be parsed using the 
               // template parameter name.
               DeclareTemplateValueParameter(
                  argName,
                  tokenCollector.Get(),
                  tok,
                  $parameterIdx,
                  GetTemplateLevel(),
                  argType,
                  $templateParm
               );
            >>
         )
   )
   ;

/////////////////////////////////////////////////////////////////////////
// Rule: template_parameter_list.  Matches a comma-separated list
// of formal template parameters (i.e., those used during template
// declaration).  The list may be empty.
//
// Inputs: none
// Outputs: 
//    TemplateParameterListRefI   parmList   The template parameters
/////////////////////////////////////////////////////////////////////////
template_parameter_list > [ TemplateParameterListRefI parmList ] :
   << 
      TemplateParameterRefI parmRef;
      $parmList = MakeTemplateParameterList(); 
      short parameterIdx = 0;
   >>
   <<;>>

   {
      template_parameter [ parameterIdx ] > [ parmRef ]
      << 
         AddTemplateParameterToList($parmList, parmRef);
         parameterIdx++;
      >>
      ( 
         TOK_COMMA 
         template_parameter [ parameterIdx ] > [ parmRef ]
         << 
            AddTemplateParameterToList($parmList, parmRef);
            parameterIdx++;
         >>
      )*
   }
   ;


/////////////////////////////////////////////////////////////////////////
// Rule: template_specialization.  Matches a template argument list
// enclosed by <>, and creates a specialization for the named class
// or function template.
// Assumes that the name can be found somewhere on the scope stack.
// The specialization is created in the same scope as the 
// template declaration.
//
// Inputs:
//    const JLStr&          name        The name of the class or function to 
//                                      specialize.
//    CPPScopeState         scopeState  A scope state to which the scope stack
//                                      is restored after looking up the name,
//                                      and before looking up any symbols for
//                                      the template arguments.
//    const ANTLRTokenPtr&  errTok      A token used for error-reporting
// Outputs: 
//    JLStr         specializationName  The decorated name of the specialization.
//    CPPTypeRefI   specializationType  The type for the specialization,
//                                      or some usable type if the specialization
//                                      could not be created (in which case
//                                      an error will have already been reported).
//    bool   usedInnerTemplateParms     true if the scope segment used any of
//                                      the template parameters from the
//                                      innermost template head.
/////////////////////////////////////////////////////////////////////////
template_specialization
[ const JLStr& name, CPPScopeState scopeState, const ANTLRTokenPtr& errTok  ] > 
[ JLStr specializationName, CPPTypeRefI specializationType, bool usedInnerTemplateParms ] :
   << TemplateArgumentListRefI argList; >>
   <<;>>

   // This is tricky, because the ID must be looked up
   // with the current scope stack, but the
   // template arguments must be looked up in the
   // scope context designated by the scopeState argument.
   // So we save the scope state and restore it.
   <<
      ScopeSaver scopeSaver(this, scopeState);
      RestoreScope(scopeState);
   >>

   TOK_LESSTHAN 
   template_argument_list > [ argList, $usedInnerTemplateParms ]
   TOK_GREATERTHAN

   <<
      scopeSaver.Restore();
      scopeSaver.Commit();
      LookupSpecialization(
         $name,
         argList, 
         $errTok,
         $specializationName, 
         $specializationType
      );
   >>
   ;

/////////////////////////////////////////////////////////////////////////
// Rule: template_specialization_reparse.  Matches an external declaration
// for the purpose of reparsing the tokens of a template specialization.
//
// Inputs: none
// Outputs: none
/////////////////////////////////////////////////////////////////////////
template_specialization_reparse :
   << ANTLRTokenPtr errTok = LT(1); >>
   << ;>>
   declaration_external_or_member [ false ]
   ;
   exception <<
      ReportExceptionError(ex);
      ReportError("Error parsing template specialization", errTok);
   >>

/////////////////////////////////////////////////////////////////////////
// Rule: translation_unit.  The whole mess.
/////////////////////////////////////////////////////////////////////////
translation_unit :
   <<;>>
   <<;>>
   (
      // Tell ANTLR to shut up about this intentional ambiguity
      #pragma approx
      (
            declaration_external
         |
            // This is only here for error synchronization
            ~TOK_Eof
            << fail_syntax(TOK_SEMICOLON); >>
      )
      exception <<
         ReportExceptionError(ex);
         consume();
         consumeUntil(EXTERNAL_SYNC_TOKCLASS_set);
         if (set_el(LA(1), EXTERNAL_SYNC2_TOKCLASS_set))
         {
            consume();
         }
      >>
   )* 
   TOK_Eof
   ;


/////////////////////////////////////////////////////////////////////////
// Rule: type_name.  Matches a (potentially complex) type name, which
// is basically declaration-specifiers followed by an abstract declarator.
//
// Inputs: none
// Outputs:
//    CPPTypeRefI   outputType    The recognized type.  If a syntax error
//                             occurred, this will be type "void".
/////////////////////////////////////////////////////////////////////////
type_name > [ CPPTypeRefI outputType ]:
   <<
      CPPConst::StorageClass sc;
      CPPConst::DeclModifier dm;
      JLStr declName;
      ANTLRTokenPtr tok;
   >>
   <<;>>

   declaration_specifiers_simple 
      [ CPPConst::scNONE, CPPConst::dmNONE ] > 
      [ $outputType, sc, dm ]

   // TODO: validate StorageClass and DeclModifier
   << tok = LT(1); >>
   declarator_opt_abstract [ $outputType ] > [ declName, $outputType ]
   <<
      // validate that there is no declarator name
      if (!declName.empty())
      {
         ReportError("No declarator name is allowed in a type name", tok);
      }
   >>
   ;


}
