// JLTokenBuffer.cpp: mostly-readable token buffer implementation

#include "JLTokenBuffer.h"
#include "ATokenStream.h"

JLTokenBuffer::JLTokenBuffer(
   ANTLRTokenStream *input_,
   int k_
) :
   input(input_),
   k(k_),
   #ifndef NDEBUG
      magic(Magic),
   #endif
   num_markers(0),
   offset(0)
{
}

JLTokenBuffer::~JLTokenBuffer() 
{
   validate();
   #ifndef NDEBUG
      magic = 0;
   #endif
}

// Consume LT(1) and get another token if needed
// Note that when marking (guess mode), no tokens are deleted from the
// front of the token list.  Instead, the offset is incremented.
void
JLTokenBuffer::consume()
{
   validate();

   // Fill the buffer sufficiently
   FillBuffer(k+offset+1);

   // Remove token or increase offset
   if (num_markers > 0)
   {
      // guess mode -- leave leading tokens and bump offset.
      offset++;
   } else {
      // normal mode -- remove first token
      tokens.next();
   }
}

// Return LT(i), where i is one-based
// If guessing, then it uses the offset
ANTLRTokenPtr 
JLTokenBuffer::LT(int i)
{
   validate();

   // Make sure there are enough tokens
   FillBuffer(offset + i);
   return tokens[offset+i-1];
}

// Make sure the token buffer is filled with enough tokens
void
JLTokenBuffer::FillBuffer(
   int amount
)
{
   while (tokens.entries() < amount)
   {
      // Append a new token
      tokens.append(input->getToken());
   }
}


// Between calls to mark() and rewind(), which may be nested,
// token pointers read by the token buffer are buffered 
// up instead of being deleted from the front of the list.
// The value returned by this method should be passed to rewind()
// to restore the token buffer to it's previous state, or to
// unmark() to allow the tokens at the list front to be discarded
int JLTokenBuffer::mark()
{
   validate();
   assert(num_markers != 0 || offset == 0);
   num_markers++;
   return offset;
}

// This method reverses the affect of mark(), by restoring the
// token buffer to it's state before mark() was called.
void JLTokenBuffer::rewind(int pos)
{
   validate();
   offset = pos;
   num_markers--;
   assert(num_markers != 0 || offset == 0);
}

// This method aborts a mark(), by decrementing the
// number of markers, but leaving the buffer the same.
// If the unmark reduces the number of marks to zero, then
// the buffered tokens at the list front are discarded.
void JLTokenBuffer::unmark(int pos)
{
   validate();
   num_markers--;
   if (num_markers == 0)
   {
      // No more markers -- Not in guess mode anymore
      assert(pos==0);
      // Discard the saved tokens at the front of the buffer
      while (offset > 0)
      {
         offset--;
         tokens.next();
      }
   }
}

