/*
 * Main program to test C++ grammar and preprocessor
 */

#include "JLStr.h"
#include "tokens.h"
#include "DLGLexer.h"
#include "BufferedCPreParser.h"
#include "CPreToCPPBuffer.h"
#include "JLTokenBuffer.h"
#include "CPPParserSym.h"
#include <iostream>

static void usage(char* progname);

int main(int argc, char *argv[])
{
   // For reporting memory usage
   char *p = new char[100000];
   long heap1 = (long)(void*)p;
   delete [] p;

   // Parse command-line options
   JLStr includePath;
   JLStr definitions;
   FILE *inputFile = stdin;
   bool doTraceParse = false;
   bool doTraceInclude = false;
   bool doDump = false;
   char *progname = *argv;

   argc--;
   argv++;
   while (argc > 0)
   {
      if (argv[0][0] == '-')
      {
         switch (argv[0][1])
         {
            case 'I':
               if (argv[0][2] != 0)
               {
                  // argument is part of "-I"
                  includePath += ';';
                  includePath += &argv[0][2];
               } else if (argc > 1)
               {
                  argc--;
                  argv++;
                  includePath += ';';
                  includePath = *argv;
               } else {
                  usage(progname);
               }
               break;
            case 'd':
               if (strcmp(argv[0], "-dump") == 0)
               {
                  doDump = true;
               }
               break;
            case 'D':
               if (argv[0][2] != 0)
               {
                  // argument is part of "-D"
                  definitions += ';';
                  definitions += &argv[0][2];
               } else if (argc > 1)
               {
                  // argument follows "-D"
                  argc--;
                  argv++;
                  definitions += ';';
                  definitions = *argv;
               } else {
                  usage(progname);
               }
               break;
            case 't':
               if (strcmp(*argv, "-traceParse") == 0)
               {
                  doTraceParse = true;
               } else if (strcmp(*argv, "-traceInclude") == 0)
               {
                  doTraceInclude = true;
               } else {
                  usage(progname);
               }
               break;
            default:
               usage(progname);
               break;
         }
      } else {
         // must be the input filename
         if (inputFile != stdin)
         {
            // already have one
            usage(progname);
         }
         FILE *fp = fopen(*argv, "r");
         if (fp == NULL)
         {
            fprintf(stderr, "%s: cannot open %s for input\n", progname, *argv);
            exit(0);
         }
         inputFile = fp;
      }
      argc--;
      argv++;
   }

   // Create input stream, lexer
   DLGFileInput input(inputFile);
   DLGLexer scanner(&input);
   FastToken tok;
   scanner.setToken(&tok);

   // Create preprocessor parser.  Note that the preprocessor parser has
   // a built-in token buffer in the form of an input stack.
	BufferedCPreParser preprocessor(&scanner);
	preprocessor.init();

   // set include path and defines directly for debugging
   if (includePath.length() == 0)
   {
      includePath = "\\msdev\\devstudio\\vc\\include;\\msdev\\devstudio\\vc\\mfc\\include";
   }
   if (definitions.length() == 0)
   {
      definitions = 
         "__cplusplus;"
         "__DATE__=\"date\";"
         "__FILE__=\"filename\";"
         "__LINE__=1;"
         "__TIME__=\"time\";"
         "__TIMESTAMP__=\"timestamp\";"
         "_M_IX86=400;"
         "_MSC_VER=1100;"
         "_WIN32;"
         "_INTEGRAL_MAX_BITS=64";
   }

   // Set preprocessor options
   preprocessor.SetIncludePath(includePath);
   preprocessor.SetDefinitions(definitions);

   // Set options in parser to make it act standard with MS Extensions
   preprocessor.SetOption(CPreParserImp::OptMSExtensions, true);
   preprocessor.SetOption(CPreParserImp::OptBool, true);
   preprocessor.SetOption(CPreParserImp::OptWCharT, true);
   preprocessor.SetOption(CPreParserImp::OptPragmas, true);
   preprocessor.SetOption(CPreParserImp::OptExpandPragmas, true);
   preprocessor.SetOption(CPreParserImp::OptTrackInclude, doTraceInclude);

   // Buffer to store tokens output by preprocessor
   CPreToCPPBuffer pipe2(&preprocessor);

   // Connect preprocessor to second token buffer so that the
   // preprocessor can unilaterally squirt new tokens into the buffer
   preprocessor.SetBuffer(&pipe2);

   // Token buffer for the usual backtracking, etc.
   JLTokenBuffer pipe3(&pipe2, CPPParserSym::ConstLLK);

   // Create the C++ parser
   CPPParserSym parser(&pipe3);

   // Set parser options to look like MSVC++
   parser.SetOption(CPPParserSym::OptPragmaMSPacking, true);
   parser.SetOption(CPPParserSym::OptMSTypedefHack, true);

   //preprocessor.doTrace(doTraceParse);
   parser.doTrace(doTraceParse);

   // Process top-level rule
   parser.init();
   parser.translation_unit();

   // Close the input file stream
   if (inputFile != stdin)
   {
      fclose(inputFile);
   }

   // For reporting memory usage
   // For reporting memory usage
   p = new char[100000];
   long heap2 = (long)(void*)p;
   delete [] p;
   cout << "Approx heap usage before dump: " << (heap2 - heap1) << endl;

   if (doDump)
   {
      // Dump the scope hierarchy
      cout << "Dump of scope hierarchy" << endl;
      parser.DumpScopes();
      cout << endl;
   }

   // For reporting memory usage
   // For reporting memory usage
   p = new char[100000];
   long heap3 = (long)(void*)p;
   delete [] p;
   cout << "Approx heap usage after dump: " << (heap3 - heap1) << endl;

   return 0;
}

static void usage(char* progname)
{
   fprintf(
      stderr, 
      "usage: %s [-I directories] [-D definitions] [-traceParse] [ -traceInclude ] [-o ofile] [ifile]\n"
      "    -I directories: semicolon-separated list of include directories\n"
      "    -D definitions: semicolon-separated list of definitions, like:\n"
      "                    symbol1;symbol2=value2\n"
      "    -traceParse: output debugging information\n"
      "    -traceInclude: output trace of include stack\n"
      "    -dump: dump type and scope hierarchy\n"
      "    ifile: name of input file (otherwise it uses stdin)\n",
      progname
   );
   exit(1);
}
