// CPreInputStack.h:  Input token buffer object for the preprocessor.
// This token buffer object was created to support a stack of lexers
// which in turn support the processing of #include directives.
// The token buffer must be able to:
//    1) Allow the "pushing" of another input token stream onto a
//       stack of input tokens.
//    2) Automatically pop the topmost input stream when
//       Eof is returned from the input token stream.
//
// Note that this token buffer does not support mark(), and does not
// support LT(i) for i!=1.

#ifndef INCL_CPreInputStack_H
#define INCL_CPreInputStack_H

#include "tokens.h"
#include "ATokenBuffer.h"
#include "assert.h"
#include "STLIncomplete.h"
using namespace std;

class CPreParser;
class ANTLRTokenStream;

class CPreInputStack : public ANTLRTokenBuffer
{
public:
   // ctor pushes the stream it was created with
   CPreInputStack(
      ANTLRTokenStream *in
   );
   virtual ~CPreInputStack();


   // ANTLRTokenBuffer interface overrides

   // Consume the current token and shift the circular buffer
	virtual void consume();

   // Unsupported interface
	virtual int mark() { assert(0); return 0; }
	virtual void rewind(int pos) { assert(0); }
	virtual void unmark(int pos) { assert(0); }

   // Get LT(i) -- valid only for i==1
	virtual ANTLRTokenPtr LT(int i)
   {
      assert(i==1);
      return LT1;
   }

   // Push new token stream
   void PushStream(ANTLRTokenStream *in);

   // Called to signal stream pop when Eof token is encountered.
   // Designed to let some other object be notified when EOF is
   // reached on a #include.
   virtual void OnIncludeEof();

   // Set the parser to notify about EOF
   void SetParser(CPreParser* parser_) { parser = parser_; }

private:
   typedef stack<
      ANTLRTokenStream*, 
      deque<ANTLRTokenStream*, allocator<ANTLRTokenStream*> >
   > StreamStack;

   // Remove the top stream from the input stack
   void PopStream();
   
   // LT(1)
   ANTLRTokenPtr LT1;
   
   // Get the next token directly from the input stack
   ANTLRTokenPtr CPreInputStack::GetToken();

   // Stack of token stream records
   StreamStack *streamStack;

   // The parser to notify at and of include files
   CPreParser *parser;

   // magic number and method to catch any incorrect downcasting
   #ifdef _DEBUG
      enum { Magic = 0x3A4F };
      short magic;
      void validate() const { assert(magic == Magic); }
   #else
      void validate() const {}
   #endif
};

#endif
