// CPreParserImp.h: Implementation class for CPreParser
// This class isolates the implementation details of the
// class from the grammar file, so that changing the actions
// of the grammar does not require recompilation of the grammar
// and subsequent rebuilding of the world.

#ifndef INCL_CPREPARSERIMP_H
#define INCL_CPREPARSERIMP_H

#include "tokens.h"
#include "CPreParser.h"
#include "JLRefPtrIncomplete.h"
#include "STLIncomplete.h"
#include "tokenList.h"
#include "CPreInputStack.h"

// Extensive use is made of incomplete types (where possible) to
// decouple the implementation of various tables from the parser
// class.  This speeds compilation and makes for easier replacement
// of the implemenations

// Incomplete types -- simple cases
template <class T> class MacroSymbolTable;
class CPreMacroSet;
class IfBlockStack;
struct IncludeRecord;
class MacroEntrySimple;
class MacroEntryFunction;
typedef refcnt_ptr<MacroEntrySimple> MacroEntrySimpleRef;
typedef refcnt_ptr<MacroEntryFunction> MacroEntryFunctionRef;

// Incomplete types -- complex cases
typedef stack<
   IncludeRecord, 
   deque<IncludeRecord, allocator<IncludeRecord> >
> IncludeRecordStack;
typedef list<JLStr, allocator<JLStr> > IncludePath;
typedef set<int, less<int>, allocator<int> > OptionSet;

class CPreParserImp : public CPreParser
{
public:
   CPreParserImp(ANTLRTokenStream *input_);
   virtual ~CPreParserImp();

   // do not define a ctor -- ANTLR makes one for you
   // init is called from base class ctor
   virtual void init();

   // Methods used to initialize parser
   // Semicolon-separated include directory path
   void SetIncludePath(const JLStr& path);
   // Semicolon-separated list of IDENT=VALUE pairs
   void SetDefinitions(const JLStr& definitions);
   // Set FILE* for default token output
   void SetOutputFile(FILE* outputFile_) { outputFile = outputFile_; }

   // Get/set preprocessor options
   virtual void SetOption(Option option, bool value);
   virtual bool GetOption(Option option) const;

   // File inclusion notification
   virtual void OnIncludeEof();

   // Various statistics
   #ifdef _DEBUG
      void DumpSymtableStats();
   #endif

protected:
   //
   // temp token list and temp identifier list operations
   //
   virtual void ClearTokenList(TokenList* tokenList);
   virtual void AppendToTokenList(TokenList* tokenList, const ANTLRTokenPtr& tokPtr);

   // 
   // Symbol table management
   //
   // Implicitly uses tempTokenList
   virtual void EnterSimpleMacro(const ANTLRTokenPtr& name);

   // Implicitly uses tempIdentifierList and tempTokenList
   virtual void EnterFunctionMacro(const ANTLRTokenPtr& name);

   // Macro definition
   virtual void UndefineMacro(const ANTLRTokenPtr& name);
   virtual bool MacroIsDefined(const ANTLRTokenPtr& name);

   // Look ahead to see if the coming tokens form a function-like macro
   // Implicitly uses LT(1..N) for lookahead
   virtual bool IsFunctionLikeMacro(const ANTLRTokenPtr& name);

   // Called when a macro is expanded so that a subclass can keep
   // usage information
   virtual void MarkMacroAsUsed(const ANTLRTokenPtr& tok) {}

   //
   // Symbol table data.  Uses reference counting because table
   // returns entries by value.
   MacroSymbolTable<MacroEntrySimpleRef> *simpleSymtable;
   MacroSymbolTable<MacroEntryFunctionRef> *functionSymtable;

   //
   // #if block tracking.
   //
   // Stack is used to maintain stack of nested #if blocks and their status.
   IfBlockStack *ifBlockStack;
   // Methods to control #if block stack
   virtual void EnterIfBlock(bool isTrue, const ANTLRTokenPtr& token);
   virtual void EnterElifBlock(bool isTrue, const ANTLRTokenPtr& token);
   virtual void EnterElseBlock(const ANTLRTokenPtr& token);
   virtual void ExitIfBlock(const ANTLRTokenPtr& token);
   virtual bool IfBlockTrue() const;
   virtual bool EnclosingIfBlockTrue() const;
   virtual bool InsideElsePart() const;
   virtual bool InsideIfPart() const;

   // Symbol from "command line" -- like "symbol" or "symbol=value"
   bool EnterSymbolDefinition(const JLStr& symbolDef);

   //
   // Evaluates the expression represented by tempTokenList
   virtual bool EvaluateExpression();

   //
   // #include processing
   // 
   virtual void IncludeFile(const ANTLRTokenPtr& includeToken);
   IncludeRecordStack *includeStack;
   #ifdef _DEBUG
      void DumpIncludeStack();
   #endif

   IncludePath* includePath;
   JLStr pathsepStr;

   //
   // Debugging and diagnostics
   //
   void DumpTokenList(const TokenList *list) const;
   void DumpTempTokenList() const { DumpTokenList(tempTokenList); }
   void DumpTempIdentifierList() const { DumpTokenList(tempIdentifierList); }

   //
   // Low-level debug and trace
   //
   virtual void panic(char *);

   // Methods for errors detected in code
   virtual void HandlePndError();
   virtual void HandleError(const JLStr& msg, const ANTLRTokenPtr& token);
   virtual void FatalError(const JLStr& msg);

   // Output of tokens
   virtual void OutputTokenBase(ANTLRTokenPtr& token) const;
   virtual void ExpandAndOutputTokenList(TokenList *list);
   virtual void OutputToken(const ANTLRTokenPtr& token) const;
   virtual void OutputTokenList(const TokenList *list) const;

   // Struct used as return value for ExpandTokenList
   struct MacroExpansionStatus {
      bool success;
      bool failedFromIncompleteFunctionMacro;
      int incompleteFunctionMacroPosition;
   };

   // Macro-expand token list inline
   void ExpandTokenList(
      TokenList *list,
      TokenList::iterator position,
      bool ignoreIncompleteFunctionMacros = false,
      MacroExpansionStatus& expandStatReturn = MacroExpansionStatus()
   );

   // Processing of ## tokens
   void ReplacePndPndInTokenList(TokenList *list);

   // Convert token list to string token
   ANTLRTokenPtr StringizeTokenList(const TokenList *list);

   // Called when a #pragma is processed -- implicitly uses tempTokenList.
   virtual void HandlePragma();

   //
   // Misc
   //
   bool TokenCanBeDefined(const ANTLRTokenPtr& token);

private:
   // Stack of input streams.  
   // Also implements reduced-function token buffer.
   CPreInputStack inputStack;

   // map for holding option values
   OptionSet *options;

   // Replace single token in tokenList with replacement list
   pair<TokenList::iterator, TokenList::iterator>
   CPreParserImp::ReplaceTokensWithList(
      TokenList *tokenList,
      TokenList::iterator iter,
      int nbrTokensToReplace,
      const TokenList *replacementList
   );

   // Struct used as return value for GetFunctionMacroReplacement
   struct MacroReplacementStatus {
      bool success;
      bool failedFromIncompleteFunctionMacro;
      bool failedFromNoLeadingParenFunctionMacro;
      JLStr errMsg;
      int nbrTokensToReplace;
      TokenList tokenList;
   };

   // Macro expand function-like macro in token list
   void
   GetFunctionMacroReplacement(
      const TokenList *tokenListIn,
      TokenList::const_iterator iter,
      MacroReplacementStatus& statusReturn
   );

   // Extract macro-argument token list from macro invocation.  
   // Return true on success.
   bool ExtractMacroArgument(
      const TokenList* tokenList,
      TokenList::const_iterator& iter,
      TokenList* listResult
   );

   // Dictionary tracks tokens that are in the process of expansion, to avoid
   // infinite recursion during expansion
   CPreMacroSet *macrosBeingExpanded;

   // output for standard (text) output
   FILE* outputFile;
};


#endif
