// cpre_expend.cpp: methods for macro-expanding token lists

#include "CPreParserImp.h"
#include "tokenList.h"
#include "cpre_stringstream.h"
#include "DLGLexer.h"
#include "cpre_symtable.h"
#include <algorithm>

///////////////////////////////////////////////////////////////////////
//                    Manipulation of tokens                         //
///////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////
// CPreParserImp::ReplaceTokensWithList -- replace the token(s) 
//     "pointed" to by the iterator with a list of tokens.
//
// Inputs:
//    TokenList*           tokenList            List of tokens to be modified
//    TokenList::iterator  iter                 Iterator pointing at tokens 
//                                              to replace
//    int                  nbrTokensToReplace   Number of tokens to replace
//    const TokenList*     replacementList      List of replacement tokens
// Outputs: none
//
// Return:
//    pair<TokenList::iterator, TokenList::iterator>
//    
//           .first         iterator pointing to the first replacement
//                          token.  If the replacement token list was
//                          empty, then the iterator points to the first
//                          token following the replaced token, which
//                          may be equal to tokenList->end().
//
//           .second        iterator pointing to the token following the
//                          replacement tokens, which
//                          may be equal to tokenList->end().
///////////////////////////////////////////////////////////////////////
pair<TokenList::iterator, TokenList::iterator>
CPreParserImp::ReplaceTokensWithList(
   TokenList *tokenList,
   TokenList::iterator iter,
   int nbrTokensToReplace,
   const TokenList *replacementList
)
{
   pair<TokenList::iterator, TokenList::iterator> returnVal;
   // Insert tokens after the token to be replaced
   TokenList::iterator iter2(iter);
   advance(iter2, nbrTokensToReplace);

#ifdef _DEBUG
   // Verify that there are nbrTokensToReplace tokens to be replaced
   TokenList::iterator testIter(iter);
   for (int i = 0; i < nbrTokensToReplace; i++, testIter++)
   {
      assert (testIter!= tokenList->end());
   }
#endif

   // Insert the replacement list
   (void)copy(
      replacementList->begin(), 
      replacementList->end(), 
      inserter(*tokenList, iter2)
   );

   // Return pointer to token following replacement tokens
   returnVal.second = iter2;

   // Remove token(s) and point to next token
   while (nbrTokensToReplace > 0)
   {
      iter = tokenList->erase(iter);
      nbrTokensToReplace--;
   }

   // Return pointer to first replacement token (if any)
   returnVal.first = iter;

   return returnVal;
}

///////////////////////////////////////////////////////////////////////
//            Token list output for text-oriented output             //
///////////////////////////////////////////////////////////////////////
void CPreParserImp::ExpandAndOutputTokenList(TokenList *tokenList)
{
   ExpandTokenList(tokenList, tokenList->begin());
   OutputTokenList(tokenList);
   fputc('\n', outputFile);
}

// Predicate to find ## token
class FindPndPnd {
public:
   bool operator()(const ANTLRTokenPtr& tok)
   {
      return tok->getType() == TOK_PNDPND;
   }
} findPndPnd;

///////////////////////////////////////////////////////////////////////
// CPreParserImp::ReplacePndPndInTokenList -- Replaces occurances of
//       t1 ## t2 with t1t2, where t1 and t2 are arbitrary tokens.
//       re-lexes the concatenated text t1t2 to form a new token.
//
// Inputs:
//    TokenList*           list      List of tokens to process
// Outputs:
//    TokenList*  list   List with t1 ## t2 replaced by concatenation
// Return: none
///////////////////////////////////////////////////////////////////////
void CPreParserImp::ReplacePndPndInTokenList(
   TokenList *tokenList
)
{
// printf("ExpandTokenList: "); DumpTokenList(list); printf("\n");

   TokenList::iterator iter;
   while (
      (iter = find_if(tokenList->begin(), tokenList->end(), findPndPnd)) != 
      tokenList->end()
   )
   {
      // It is invalid to have ## at beginning or end of list
      TokenList::iterator tmpIter(iter);
      if (iter == tokenList->begin() || ++tmpIter == tokenList->end())
      {
         HandleError("## cannot begin or end macro replacement list", *iter);
         return;
      }

      // Save position of first token
      iter--;
      TokenList::iterator leftTokenIter(iter);

      // Concatentate tokens on either side of ##
      JLStr tokenText = (*iter)->getText();
      iter++;
      iter++;
      tokenText += (*iter)->getText();
      // Append newline to delimit token for scanning
      JLStr origText = tokenText;
      tokenText += '\n';

      // Scan text to get the concatenation token
      FastToken tok;   // Token to hand to lexer so it can make more
      CPreStringInput symbolInput(tokenText);
      DLGLexer symbolScan(&symbolInput);
      symbolScan.setToken(&tok);
      ANTLRTokenStream* symStream = &symbolScan;    // Upcast because getToken() is hidden
      ANTLRTokenPtr newToken = symStream->getToken();
      newToken->setLine((*leftTokenIter)->getLine());

      // TODO: Flag an error if the scanned result is not a valid token.

      // Replace the three tokens with the new one
      leftTokenIter = tokenList->erase(leftTokenIter);
      leftTokenIter = tokenList->erase(leftTokenIter);
      *leftTokenIter = newToken;
   }
}

///////////////////////////////////////////////////////////////////////
// CPreParserImp::ExpandTokenList -- Expands all of the macro invocations
//    found in a token list.  Perform the expansion inline by replacing
//    the token(s) that are expanded.  Rescans new tokens for further 
//    replacement.  Keeps track of tokens in the process of being
//    expanded, to avoid infinite recursion.
//
// Inputs:
//    TokenList*           list      List of tokens to expand
//
//    TokenList::iterator  position  Starting position for processing
//
//    bool                 ignoreIncompleteFunctionMacros
//                                   true if errors resulting from
//                                   attempts to expand incomplete
//                                   function macros should be ignored.
//                                   This is typically true during the
//                                   re-expansion of replacement lists.
// Outputs:
//    TokenList*           list      Input list with expanded macro 
//                                   tokens replaced with their 
//                                   replacement lists
//
//    MacroExpansionStatus&  expandStatReturn
//                                   Fields of the status are as follows:
//
//         bool    success           true if the expansion succeeded.  false
//                                   if it failed due to malformed function macro
//
//         bool    failedFromIncompleteFunctionMacro
//                                   if (!success) then this is set to true
//                                   when the failure was due to an unterminated
//                                   function macro (as opposed to other errors)
//
//         int     incompleteFunctionMacroPosition
//                                   if (failedFromIncompleteFunctionMacro), then
//                                   this is set to the offset of the failing
//                                   function id token from the 'position' argument.
//       
// Return: none
//
// Note: ignoreIncompleteFunctionMacros is set to true during 
//       re-expansion to account for cases like:
//            #define FOO GLARP
//            #define GLARP(x) x
//            FOO(x)
///////////////////////////////////////////////////////////////////////
void CPreParserImp::ExpandTokenList(
   TokenList *tokenList,
   TokenList::iterator position,
   bool ignoreIncompleteFunctionMacros,
   MacroExpansionStatus& expandStatReturn
)
{
   // remember beginning position
   TokenList::iterator iterStart(position);

   // Initialize return
   expandStatReturn.success = true;

   // put out here to avoid ctor/dtor in loop
   MacroReplacementStatus replStat;

   for (
      TokenList::iterator iter(position);
      iter != tokenList->end();
      // iterator increment handled in body
   )
   {
      ANTLRTokenPtr macroNameToken(*iter);

      // Don't expand a macro that is already being expanded.  The draft standard
      // states that an attempt to do so is not an error -- you just leave the
      // macro name token unexpanded.
      if (macrosBeingExpanded->contains(macroNameToken))
      {
         iter++;
         continue;
      }

      bool haveReplacement = false;

      // Attempt simple macro expansion first
      if (simpleSymtable->contains(macroNameToken))
      {
         // Get a copy of the simple macro replacement list
         haveReplacement = true;
         MacroEntrySimpleRef entryRef;
         entryRef = (*simpleSymtable->find(macroNameToken)).second;
         replStat.success = true;
         replStat.nbrTokensToReplace = 1;
         entryRef->GetReplacementTokenList().CopyToTokenList(replStat.tokenList);
      } else if (functionSymtable->contains(macroNameToken))
      {
         // Process the function-invication tokens and create a
         // replacement token list.
         GetFunctionMacroReplacement(tokenList, iter, replStat);

         // Make substitution only when function macro parse succeeds.
         haveReplacement = replStat.success;

         if (!replStat.success)
         {
            // Getting the macro replacement list failed (could not parse
            // function-like macro invocation).
            expandStatReturn.success = false;

            if (replStat.failedFromIncompleteFunctionMacro)
            {
               expandStatReturn.failedFromIncompleteFunctionMacro = true;
               if (ignoreIncompleteFunctionMacros)
               {
                  // Error was due to incomplete function macro and we are 
                  // ignoring such errors due to re-expansion.  In this case
                  // just ignore the error and return, since the occurance
                  // of such an error means we've hit the end of the list.
                  // Return the offset of the error from the position argument
                  // so that the caller can reprocess the failed portion of
                  // the list with additional tokens appended.
                  // This is needed to handle cases like:
                  //    #define FOO GLARP
                  //    #define GLARP(x) x
                  //    FOO(a)
                  // because the FOO gets expanded to GLARP, and then we try
                  // to re-expand GLARP, which fails because it is missing
                  // it's arguments.  The caller, however, will be able to
                  // replace FOO with GLARP, and then re-process GLARP(a)
                  expandStatReturn.incompleteFunctionMacroPosition = distance(iterStart, iter);
                  // break out of loop and return
                  break;
               } else {
                  // Error was due to incomplete function macro and we are 
                  // not ignoring such errors.  However, we *still* ignore
                  // errors due to no leading "(" in a function-like macro
                  if (!replStat.failedFromNoLeadingParenFunctionMacro)
                  {
                     expandStatReturn.failedFromIncompleteFunctionMacro = false;
                     HandleError(replStat.errMsg, *iter);
                  }
               }
            } else {
               // Some other error was encountered when expanding the
               // function macro, like not having enough arguments.
               // Report this as a normal error.
               expandStatReturn.failedFromIncompleteFunctionMacro = false;
               HandleError(replStat.errMsg, *iter);
            }
         }
      }

      if (haveReplacement)
      {
         // Process ## operators in replacement list
         ReplacePndPndInTokenList(&replStat.tokenList);

         // Add macro to set of macros being expanded
         macrosBeingExpanded->insert(macroNameToken);

         // Tell subclass that a macro was used
         MarkMacroAsUsed(macroNameToken);

         // Re-expand the replacement list, not reporting any errors
         // that result from incomplete function macros, since
         // these may be completed by subsequent tokens in the 
         // rest of the token list, and re-scanned later.
         MacroExpansionStatus expandStat;
         ExpandTokenList(&replStat.tokenList, replStat.tokenList.begin(), true, expandStat);

         // Pair of iterators denote start, end of replaced tokens in tokenList;
         pair<TokenList::iterator, TokenList::iterator> replacedPosition;

         // Replace token(s) with expanded replacement list
         replacedPosition = ReplaceTokensWithList(
            tokenList, iter, replStat.nbrTokensToReplace, &replStat.tokenList
         );

         // Check for the "incomplete function macro" error in the expansion
         if (!expandStat.success && expandStat.failedFromIncompleteFunctionMacro)
         {
            // We replaced tokens with an incomplete function invocation.
            // Point the iterator to the incomplete function macro TOK_ID token 
            // so that expansion of the incomplete function can continue 
            // on the next iterator of the loop.
            iter = replacedPosition.first;
            advance(iter, expandStat.incompleteFunctionMacroPosition);
         } else {
            // Point iterator past the replaced tokens
            iter = replacedPosition.second;
         }


         // Remove macro from set of macros being expanded
         macrosBeingExpanded->remove(macroNameToken);
      } else {
         // no replacement -- advance iterator to next token
         iter++;
      }
   }
}

///////////////////////////////////////////////////////////////////////
// CPreParserImp::GetFunctionMacroReplacement -- Gets the replacement
//    token list for a function-like macro invocation.  Rescans the 
//    resulting replacement list for expansion of ## operators, but
//    *not* for # operators.
//
// Inputs:
//    const TokenList*      tokenList   Input list of tokens to scan
//    TokenList::iterator   iter        Points at the TOK_ID token in the
//                                      token list that presumably starts
//                                      a function-like macro invocation.
//
// Outputs:
//    MacroReplacementStatus& replStat    Fields of the status are set as follows:
// 
//       bool       success               whether valid function macro was scanned
// 
//       bool       failedFromIncompleteFunctionMacro
//                                        if (!success), this indicates that
//                                        the failure was due to an
//                                        incomplete function macro
//
//       bool       failedFromNoLeadingParenFunctionMacro
//                                        if (!success), this indicates that
//                                        the failure was due to a function
//                                        macro with no leading "("
//
//       JLStr     errmsg                if !success, this is the error
//                                        in a user-viewable form
//
//       int        nbrTokensToReplace    number of tokens to replace
//                                        in the input list (i.e., number
//                                        of tokens comprising the function
//                                        macro invocation.
//
//       TokenList  tokenList             replacement tokens.
//
//  Note: When a function-like macro TOK_ID token is encountered with no
//        following (...), then the TOK_ID is to be left unchanged.  To
//        support that behavior, this method will create a replacement
//        token list containing the original token.
///////////////////////////////////////////////////////////////////////
void
CPreParserImp::GetFunctionMacroReplacement(
   const TokenList* tokenList,
   TokenList::const_iterator iter,
   MacroReplacementStatus& replStat
)
{
   // Remember leading position
   TokenList::const_iterator iterStart(iter);

   // Initialize return
   replStat.success = false;
   replStat.failedFromIncompleteFunctionMacro = false;
   replStat.failedFromNoLeadingParenFunctionMacro = false;
   replStat.nbrTokensToReplace = 0;
   replStat.tokenList.clear();

   // Get the function macro entry.
   assert(functionSymtable->contains(*iter));
   MacroEntryFunctionRef entryRef = (*functionSymtable->find(*iter)).second;

   // Skip macro name token 
   iter++;

   // Skip any newlines or comments
   while (
      iter != tokenList->end() &&
      ((*iter)->getType() == TOK_SLCOMMENT || (*iter)->getType() == TOK_NEWLINE))
   {
      iter++;
   }

   // Check to see if end of list was hit
   if (iter == tokenList->end() || (*iter)->getType() != TOK_LPAREN)
   {
      // No '(' after macro name -- this is OK, according to
      // the standard, and also matches MSVC++ and Borland behavior.
      replStat.success = false;
      replStat.failedFromIncompleteFunctionMacro = true;
      replStat.failedFromNoLeadingParenFunctionMacro = true;
      return;
   }

   // Advance past '('
   if (++iter == tokenList->end())
   {
      // Error: no tokens after leading '('
      replStat.failedFromIncompleteFunctionMacro = true;
      replStat.errMsg = "Incomplete function-like macro";
      return;
   }

   // Make a table associating each formal argument name with a 
   // token list containing each actual macro argument.
   typedef map<JLStr, TokenListRef, less<JLStr>, allocator<TokenListRef> > ArgumentTable;
   ArgumentTable argumentTable;

   // For each formal argument extract a replacement token list 
   // corresponding to the formal argument, and enter the 
   // formal argument name and the token list into a table.
   TokenList formalArgList;
   entryRef->GetArgumentTokenList().CopyToTokenList(formalArgList);
   for (
      TokenList::iterator formalArgIter(formalArgList.begin());
      formalArgIter != formalArgList.end();
      formalArgIter++
   )
   {
      TokenList *argumentTokens = new TokenList;
      if (!ExtractMacroArgument(tokenList, iter, argumentTokens))
      {
         // Error: More formal arguments than actual arguments
         replStat.failedFromIncompleteFunctionMacro = (iter == tokenList->end());
         replStat.errMsg = "Not enough actual arguments for macro";
         return;
      }

      // Expand actual argument token list
      ExpandTokenList(argumentTokens, argumentTokens->begin());

      // Add expanded argument token list to the table
      argumentTable.insert(
         ArgumentTable::value_type((*formalArgIter)->getText(), argumentTokens)
      );
   }

   // Check for end of token list -- missing ')'
   if (iter == tokenList->end())
   {
      // Error: list terminated before ')'
      replStat.failedFromIncompleteFunctionMacro = true;
      replStat.errMsg = "Incomplete function-like macro";
      return;
   }

   // Check that there are no more arguments by verifying trailing ')'
   if ((*iter)->getType() != TOK_RPAREN)
   {
      replStat.errMsg = "Too many actual arguments for macro";
      return;
   }

   // Advance iterator past ')'
   iter++;

   // Calculate number of tokens to be replaced from iterator positions
   replStat.nbrTokensToReplace = distance(iterStart, iter);

   // Copy the replacement list for the macro
   entryRef->GetReplacementTokenList().CopyToTokenList(replStat.tokenList);

   // Replace tokens in replacement token list with formal argument token lists
   // by looping over all tokens in the replacement list and replacing any
   // that have entries in the argumentTable.
   for (
      TokenList::iterator replacementIter(replStat.tokenList.begin());
      replacementIter != replStat.tokenList.end();
      // iterator is incremented in the loop
   )
   {
      bool processPnd = false;

      // Check for special # token processing
      if ((*replacementIter)->getType() == TOK_PND)
      {
         // Skip # token and set flag to process next token specially
         processPnd = true;
         replacementIter = replStat.tokenList.erase(replacementIter);
         // replacementIter points to next token
         if (replacementIter == replStat.tokenList.end())
         {
            replStat.errMsg = "# token must be followed by formal argument";
            replStat.tokenList.clear();       // For easier error recovery
            break;
         }
      }

      // Lookup next token in token replacement list in the argumentTable
      JLStr key((*replacementIter)->getText());
      ArgumentTable::const_iterator argTableIter;
      if ((argTableIter = argumentTable.find(key)) != argumentTable.end())
      {
         // Token in replacement list matches formal argument name.
         TokenListRef formalArgTokens((*argTableIter).second);
         if (processPnd)
         {
            // Replace formal argument name token with stringized argument list
            *replacementIter = StringizeTokenList(&*formalArgTokens);
            replacementIter++;
         } else {
            // Replace formal argument name with actual argument token list
            replacementIter = ReplaceTokensWithList(
               &replStat.tokenList, 
               replacementIter,
               1,
               &*formalArgTokens
            ).second;
            // Do not increment iterator -- it already points to next token
         }
      } else if (processPnd)
      {
         // Error if the previous token was # -- it must be followed
         // by a formal argument name.
         replStat.errMsg = "# in replacement list must be followed by formal argument";
         replStat.tokenList.clear();       // For easier error recovery
         break;
      } else {
         replacementIter++;
      }
   }

   // Process ## in resulting replacement list
   ReplacePndPndInTokenList(&replStat.tokenList);

   replStat.success = true;
}

///////////////////////////////////////////////////////////////////////
// CPreParserImp::ExtractMacroArgument -- Scans the token list for a
//    function-like macro invocation, and extracts a single list
//    of tokens representing the argument.  Handles nested parentheses.
//
// Inputs:
//    const TokenList*            tokenList   List of argument tokens.
//    TokenList::const_iterator&  iter        Points to the first token in the
//                                            token list for the arguments.
// Outputs:
//    TokenList::const_iterator&  iter        Points at the first token beyond the
//                                            extracted argument.  That token will
//                                            be ')' or ',' unless an error occurred.
//    TokenList*                  listResult  List of tokens that comprise the
//                                            next argument of the macro.  This list
//                                            does not change if there are no more
//                                            arguments.
//
// Return: true if an argument is extracted.  false if no further
//         arguments are available, or if an error occured.
//
// Note: If there are no errors encountered, then a false return will
//       leave the iterator pointed at the closing ')' token
///////////////////////////////////////////////////////////////////////
bool CPreParserImp::ExtractMacroArgument(
   const TokenList* tokenList,
   TokenList::const_iterator& iter,
   TokenList* listResult
)
{
   listResult->clear();
   int parenLevel = 0;
   if ( iter == tokenList->end() || (*iter)->getType() == TOK_RPAREN)
   {
      // no tokens available for argument
      return false;
   }

   bool done = false;
   while ( iter != tokenList->end() && !done)
   {
      switch ((*iter)->getType())
      {
         case TOK_RPAREN:
            if (parenLevel > 0)
            {
               parenLevel--;
            } else {
               done = true;
               // do not skip over rparen token
            }
            break;
         case TOK_LPAREN:
            parenLevel++;
            break;
         case TOK_COMMA:
            if (parenLevel == 0)
            {
               done = true;
               // skip over comma token
               iter++;
            }
            break;
         default:
            break;
      }
      if (!done)
      {
         listResult->push_back(*iter);
         iter++;
      }
   }

   return true;
}

///////////////////////////////////////////////////////////////////////
// CPreParserImp::StringizeTokenList -- Concatenates the text of a each
//     token in a token list to form a new string token.
//
// Inputs:
//    const TokenList&   list   List of tokens to stringize
// Return:
//   ANTLRTokenPtr    The new string token
///////////////////////////////////////////////////////////////////////
ANTLRTokenPtr CPreParserImp::StringizeTokenList(
   const TokenList *tokenList
)
{
   JLStr str("\"");
   int first = true;
   int line = 1;

   // Cast away const of list -- really, we won't change it!
   for (
      TokenList::const_iterator iter(tokenList->begin());
      iter != tokenList->end();
      iter++
   )
   {
      if (first)
      {
         first = false;
         line = (*iter)->getLine();
      } else {
         // separate tokens by a space
         str += ' ';
      }
      for (const char *p = (*iter)->getText(); *p != 0; p++)
      {
         if (*p == '\\' || *p == '"')
         {
            // escape some characters of stringized tokens
            str += '\\';
         }
         str += *p;
      }
   }
   str += "\"";
   
   // Create a TOK_STRING token
   return ANTLRTokenPtr(FastToken::newToken(TOK_STRING, str.c_str(), line));
}

