// cpre_expr_eval.cpp: methods to evaluate expressions for #if and #elif

#include "CPreParserImp.h"
#include "CPreExpr.h"
#include "TokenStream.h"
#include "TokenKey.h"
#include "JLTokenBuffer.h"

///////////////////////////////////////////////////////////////////////
// CPreParserImp::EvaluateExpression -- Evaluate the expression contained
//    in the tempTokenList member.
//
// Members used:
//    TokenList*    tempTokenList     The list of tokens that is 
//                                    accumulated by the token_list* rules.
//
// Return: 
//    bool      true if the expression evaluates non-zero, false otherwise
///////////////////////////////////////////////////////////////////////
bool CPreParserImp::EvaluateExpression()
{
   // Replace "defined name" and "defined(name)" with integer
   // "1" token if defined or "0" otherwise
   for(
      TokenList::iterator iter(tempTokenList->begin());
      iter != tempTokenList->end();
      iter++
   )
   {
      ANTLRTokenKey definedToken(*iter);
      if (definedToken->getType() == TOK_DEFINED)
      {
         // remember the position of the 'defined' token
         TokenList::iterator iterDefined(iter);

         // There must be at least one token following 'defined'
         if (++iter == tempTokenList->end())
         {
            HandleError("Missing identifier or '(' following 'defined'", definedToken);
            return false;
         }

         bool idDefined;
         int nbrTokens;
         if ((*iter)->getType() == TOK_LPAREN)
         {
            nbrTokens = 4;
            // Look for "defined ( id )"
            // Skip over '(' token
            if (++iter == tempTokenList->end())
            {
               HandleError("Malformed 'defined' expression", definedToken);
               return false;
            }
            // Examine next token
            if (!TokenCanBeDefined(*iter))
            {
               HandleError("Expected defined(id)", definedToken);
               return false;
            }
            idDefined = MacroIsDefined(*iter);
            // Verify that following token is ')'
            if (++iter == tempTokenList->end() || (*iter)->getType() != TOK_RPAREN)
            {
               HandleError("Expected defined(id)", definedToken);
               return false;
            }
         } else {
            nbrTokens = 2;
            // Look for "defined id"
            if (!TokenCanBeDefined(*iter))
            {
               HandleError("Expected 'defined id' or defined(id)", definedToken);
               return false;
            }
            idDefined = MacroIsDefined(*iter);
         }

         // Point past "defined id" or "defined(id)"
         iter++;

         // Insert numeric token
         iter = tempTokenList->insert(
            iter, 
            FastToken::newToken(
               TOK_DECIMALINT, 
               idDefined ? "1" : "0", 
               definedToken->getLine()
            )
         );

         // Remove "defined..." tokens
         while (nbrTokens > 0)
         {
            iterDefined = tempTokenList->erase(iterDefined);
            assert(iterDefined != tempTokenList->end());
            nbrTokens--;
         }
      }
   }

   // Expand the remaining macros in the token list
   ExpandTokenList(tempTokenList, tempTokenList->begin());

   // Replace all remaining non-numeric tokens with integer "0" tokens
   for (
      iter = tempTokenList->begin();
      iter != tempTokenList->end();
      iter++
   )
   {
      if (TokenCanBeDefined(*iter))
      {
         // Replace the current token with an integer "0" token
         (*iter) = FastToken::newToken(TOK_DECIMALINT, "0", (*iter)->getLine());
      }
   }

   // Parse and evaluate the resulting expression by piping the list
   // of expression tokens into the expression parser.
   TokenListStream stream(*tempTokenList);
   JLTokenBuffer pipe(&stream, CPreExpr::ConstLLK);
   CPreExpr exprParser(&pipe);
   exprParser.doTrace(trace_on);
   exprParser.init();

   CPreValue val = exprParser.constant_expression();
   return (long)val;
}

