// cpre_include.cpp: methods to handle file inclusion stack

#include "CPreParserImp.h"
#include "cpre_include.h"
#include "tokenList.h"
#include "CPreInputStack.h"
#include "DLGLexer.h"
#include <iostream>

///////////////////////////////////////////////////////////////////////
//                        File inclusion                             //
///////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////
// CPreParserImp::IncludeFile -- Handle #include directive.
//    The included file is opened, a lexer is created to parse its
//    contents, and the parser is connected to the new lexer until
//    EOF is encountered.
//
// Inputs:
//    const ANTLRTokenPtr&  includeToken   #include token
// members used:
//    TokenList             tempTokenList  Token list that comprises the filename.
//
// Return: none
///////////////////////////////////////////////////////////////////////
void CPreParserImp::IncludeFile(const ANTLRTokenPtr& includeToken)
{
   // Expand macros in the filename token list
   ExpandTokenList(tempTokenList, tempTokenList->begin());

   // Concatentate the tokens into a (presumed) filename
   JLStr filename;
   for(
      TokenList::iterator tokenIter(tempTokenList->begin());
      tokenIter != tempTokenList->end();
      tokenIter++
   )
   {
      filename += (*tokenIter)->getText();
   }

   // Remove leading/trailing spaces
   while (filename.length() > 0 && filename[0] == ' ')
   {
      filename.erase(filename.begin());
   }
   while (filename.length() > 0 && *(filename.end()-1) == ' ')
   {
      filename.erase(filename.end()-1);
   }

   // See if the resulting filename looks like "filename" or <filename>
   size_t length = filename.length();
   if (
      length < 3 ||
      !(
         filename[(size_t)0] == '"' && filename[length-1] == '"' ||
         filename[(size_t)0] == '<' && filename[length-1] == '>'
      )
   )
   {
      JLStr msg("Invalid filename for #include: ");
      msg += filename;
      msg += ": ";
      HandleError(msg, includeToken);
      return;
   }

   // Remove "" or <>
   filename = JLStr(filename.begin()+1, filename.end()-1);

   // Open include file
   // Since I really don't want to build in knowledge of the path
   // separator string, this tries both '/' and '\' and determines
   // which one works, and then uses that for all other #includes
   FILE *fp = NULL;
   JLStr chosenPathsepStr;
   for (
      IncludePath::iterator pathIter(includePath->begin());
      fp == NULL && pathIter != includePath->end();
      pathIter++
   )
   {
      if (pathsepStr.length() == 0)
      {
         static char* pathsepTypes[] = { "/", "\\", NULL };
         // Path separator has not been determined yet -- try them all
         for (int sepIdx = 0; fp == NULL && pathsepTypes[sepIdx] != NULL; sepIdx++)
         {
            // Construct path
            chosenPathsepStr = pathsepTypes[sepIdx];
            JLStr path = *pathIter + chosenPathsepStr + filename.data();
            fp = fopen(path.c_str(), "r");
         }
      } else {
         // Construct path with known path separator 
         JLStr path = *pathIter + pathsepStr + filename.data();
         fp = fopen(path.c_str(), "r");
      }
   }

   // If no include path prefix worked, then treat it like an absolute path
   if (fp == NULL)
   {
      fp = fopen(filename.data(), "r");
   }

   if (fp == NULL)
   {
      JLStr msg = "#include: cannot open file ";
      msg += filename;
      HandleError(msg, includeToken);
      return;
   }

   if (pathsepStr.length() == 0)
   {
      // File opened successfully -- set path separator to one that works.
      pathsepStr = chosenPathsepStr;
   }

   // Create a new input stream and lexer for opened file
   DLGFileInput *stream = new DLGFileInput(fp);
   DLGLexer *scan = new DLGLexer(stream);
   ANTLRAbstractToken *tok = FastToken::newToken();
   scan->setToken(tok);

   // Squirrel away the new input classes, so that we
   // can close/delete them when EOF is encountered for
   // the input file.  The inputTokens object will call
   // the OnIncludeEof() method when the include file ends.
   IncludeRecord incRec(fp, stream, scan, tok);
   incRec.filename = filename;
   includeStack->push(incRec);
   #ifdef _DEBUG
      if (GetOption(OptTrackInclude)) DumpIncludeStack();
   #endif

   // Process the file by installing a new lexer and input stream
   // into the CPreInputStack and continuing the parse.
   // NOTE: This code assumes that k==1 and that the current lookahead
   // LT(1) will be consumed *after* this method returns.

   inputStack.PushStream(scan);
}

///////////////////////////////////////////////////////////////////////
// CPreParserImp::OnIncludeEof -- method called by the token buffer
//    when the end of a #included file is reached.  This cleans
//    up by closing the file and destroying the created lexer.
//
// Inputs:
//    ANTLRTokenPtr   includeToken     #include token
// members used:
//    TokenList       tempTokenList    Token list that comprises the filename.
//
// Return: none
///////////////////////////////////////////////////////////////////////
void CPreParserImp::OnIncludeEof()
{
   IncludeRecord incRec = includeStack->top();
   includeStack->pop();
   delete incRec.scan;
   delete incRec.stream;
   delete incRec.tok;
   fclose(incRec.fp);
   #ifdef _DEBUG
      if (GetOption(OptTrackInclude)) DumpIncludeStack();
   #endif
}

#ifdef _DEBUG
   // Debug include stack
   void CPreParserImp::DumpIncludeStack()
   {
      // Since we cannot travese the stack directly,
      // copy the stack contents onto another stack and then put it back.
      IncludeRecordStack tmpStack;

      cout << "#include stack: ";
      while (!includeStack->empty())
      {
         IncludeRecord tmpRec = includeStack->top();
         includeStack->pop();
         tmpStack.push(tmpRec);
      }
      while (!tmpStack.empty())
      {
         IncludeRecord tmpRec = tmpStack.top();
         tmpStack.pop();
         cout << tmpRec.filename << " ";
         includeStack->push(tmpRec);
      }
      cout << endl;
   }
#endif


///////////////////////////////////////////////////////////////////////
// CPreParserImp::SetIncludePath -- set the list of directories to
//     search when a #include directive is processed.
//
// Inputs:
//    const JLStr&   path    A list of fully-specified or relative
//                               pathnames, in the format of the local
//                               filesystem, separated by semicolons.
// Return: none
///////////////////////////////////////////////////////////////////////
void CPreParserImp::SetIncludePath(
   const JLStr& path
)
{
   JLStr tmpPath(path);

   // Initialize path with '.'
   includePath->clear();
   includePath->push_back(".");

   JLStr::size_type pos;
   // Extract the semicolon-separated path portions
   while ((pos = tmpPath.find(';')) != JLStr::npos)
   {
      if (pos > 0)
      {
         // Add the ';'-delimitied substring to the path
         includePath->push_back(JLStr(tmpPath.begin(), tmpPath.begin()+pos));
      }
      // Remove the leading substring and ';' from the path
      tmpPath = JLStr(tmpPath.begin()+pos+1, tmpPath.end());
   }

   if (!tmpPath.empty())
   {
      // Add the final segment of the path
      includePath->push_back(tmpPath);
   }
}


