// cpre_symtable.h:  Auxilliary classes for the C preprocessor symbol table.

#ifndef INCL_CPRE_SYMTABLE_H
#define INCL_CPRE_SYMTABLE_H

#include <map>
#include <set>
#include <JL_DList.h>
#include "JLRefptr.h"
#include "JLFreeList.h"
#include "tokenList.h"
#include "TokenKey.h"

///////////////////////////////////////////////////////////////////////
// Specialized "token" for macro replacement lists
// This is designed to save memory, since it is significantly
// smaller than a "normal" token.  However, the line and column
// information must be gleaned from some other token, such as the
// macro name token used as the key for the symbol tables.
///////////////////////////////////////////////////////////////////////
class MacroReplacementToken : public RefCount
{
public:
   MacroReplacementToken(ANTLRTokenType tt_ = TOK_Eof, const char* text_ = "");
   ~MacroReplacementToken();

   // Copy ctor and assignment op necessary to avoid copying refcount
   MacroReplacementToken(const MacroReplacementToken& rhs);
   MacroReplacementToken& operator=(const MacroReplacementToken& rhs);

   // Memory allocation
   void* operator new(size_t size)
   {
      // Subclasses should overide operators new and delete
      assert(size == sizeof(MacroReplacementToken));
      return freeList.New();
   }
   void operator delete(void* ptr, size_t size)
   {
      // Subclasses should overide operators new and delete
      assert(size == sizeof(MacroReplacementToken));
      freeList.Delete(ptr);
   }

   // Data access
   const char* getText() const { return text; }
   ANTLRTokenType getType() const { return (ANTLRTokenType)tokenType; }

   // Comparison for ordering in maps and sets
   bool operator < (const MacroReplacementToken& rhs) const
   {
      return strcmp(text, rhs.text) < 0;
   }

   // Comparison for equality
   bool operator == (const MacroReplacementToken& rhs) const
   {
      return tokenType == rhs.tokenType && strcmp(text, rhs.text) == 0;
   }

private:
   char *text;
   short tokenType;

   enum { MemPoolLimit = 36, MemPoolBlockSize = 50, FreeListBlockSize = 200 };
   static JLMemPool memPool;
   static JLFreeList freeList;
};

typedef refcnt_ptr<MacroReplacementToken> MacroReplacementTokenRef;


///////////////////////////////////////////////////////////////////////
// List of reduced-size tokens.  This is used for symbol table storage.
// Knows how to convert to/from the normal TokenList, which is used 
// for processing.
///////////////////////////////////////////////////////////////////////
class MacroReplacementTokenList : 
   public JL_DList<MacroReplacementTokenRef, allocator<MacroReplacementTokenRef> >
{
public:
   // ctor/dtor
   MacroReplacementTokenList() {}
   MacroReplacementTokenList(const TokenList* tokenList);

   void CopyToTokenList(TokenList& outputList) const;

   // Special equality test compares each token value
   bool contentEquals(const MacroReplacementTokenList& rhs) const;
private:
   class CmpReplacementTokenRefs {
   public:
      bool operator()(
         const MacroReplacementTokenRef& tok1, 
         const MacroReplacementTokenRef& tok2
      )
      {
         return *tok1 == *tok2;
      }
   };
};


///////////////////////////////////////////////////////////////////////
// Simple (no argument) macro entry
///////////////////////////////////////////////////////////////////////
class MacroEntrySimple : public RefCount
{
public:
   // ctor/dtor
   MacroEntrySimple(
      const TokenList *replacementTokenList_
   ) :
      replacementTokenList(replacementTokenList_)
   {}
   virtual ~MacroEntrySimple() {}

   // Memory allocation
   void* operator new(size_t size)
   {
      // Subclasses should overide operators new and delete
      assert(size == sizeof(MacroEntrySimple));
      return freeList.New();
   }
   void operator delete(void* ptr, size_t size)
   {
      // Subclasses should overide operators new and delete
      assert(size == sizeof(MacroEntrySimple));
      freeList.Delete(ptr);
   }

   // Comparison
   bool operator == (const MacroEntrySimple& rhs) const 
   { 
      // use special "contentEquals" method, since default list
      // comparison will just compare ref-pointers
      return replacementTokenList.contentEquals(rhs.replacementTokenList);
   }
   bool operator != (const MacroEntrySimple& rhs) const { return !(*this == rhs); }

   // Data member access
   const MacroReplacementTokenList &GetReplacementTokenList() const 
   { return replacementTokenList; }
   // misc
   virtual bool IsFunctionLike() { return false; }
private:
   MacroReplacementTokenList replacementTokenList;
   enum { FreeListBlockSize = 100 };
   static JLFreeList freeList;

   // Do not allow copy/assignment
   // If you implement these, do not copy the refcount!
   MacroEntrySimple(const MacroEntrySimple& rhs);
   MacroEntrySimple& operator=(const MacroEntrySimple& rhs);
};
typedef refcnt_ptr<MacroEntrySimple> MacroEntrySimpleRef;


///////////////////////////////////////////////////////////////////////
// Function (with arguments) macro entry
///////////////////////////////////////////////////////////////////////
class MacroEntryFunction : public MacroEntrySimple
{
public:
   // ctor/dtor
   MacroEntryFunction(
      const TokenList *argumentTokenList_,
      const TokenList *replacementTokenList_
   ) :
      MacroEntrySimple(replacementTokenList_),
      argumentTokenList(argumentTokenList_)
   {}
   virtual ~MacroEntryFunction() {}

   // Memory allocation
   void* operator new(size_t size)
   {
      // Subclasses should overide operators new and delete
      assert(size == sizeof(MacroEntryFunction));
      return freeList.New();
   }
   void operator delete(void* ptr, size_t size)
   {
      // Subclasses should overide operators new and delete
      assert(size == sizeof(MacroEntryFunction));
      freeList.Delete(ptr);
   }

   // Comparison
   bool operator == (const MacroEntryFunction& rhs) const
   { 
      // use special "contentEquals" method, since default list
      // comparison will just compare ref-pointers
      return 
         argumentTokenList.contentEquals(rhs.argumentTokenList) &&
         GetReplacementTokenList().contentEquals(rhs.GetReplacementTokenList());
   }
   bool operator != (const MacroEntryFunction& rhs) const { return !(*this == rhs); }

   // data access
   const MacroReplacementTokenList & GetArgumentTokenList() const 
   { return argumentTokenList; }

   // misc
   virtual bool IsFunctionLike() { return true; }

private:
   MacroReplacementTokenList argumentTokenList;
   enum { FreeListBlockSize = 50 };
   static JLFreeList freeList;
};
typedef refcnt_ptr<MacroEntryFunction> MacroEntryFunctionRef;


///////////////////////////////////////////////////////////////////////
// Symbol table template.  T should be either MacroEntrySimpleRef 
// or MacroEntryFunctionRef
///////////////////////////////////////////////////////////////////////
template <class T> class MacroSymbolTable : 
   public map<ANTLRTokenKey, T, less<ANTLRTokenKey> >
{
public:
   MacroSymbolTable() { }
   bool contains(const ANTLRTokenKey& key) { return find(key) != end(); }
};
typedef MacroSymbolTable<MacroEntrySimpleRef> MacroSymbolTableSimple;
typedef MacroSymbolTable<MacroEntryFunctionRef> MacroSymbolTableFunction;


///////////////////////////////////////////////////////////////////////
// Class used to track macros in the process of expansion
///////////////////////////////////////////////////////////////////////
class CPreMacroSet
{
public:
   CPreMacroSet() { }
   void insert(const ANTLRTokenKey& key) { names.insert(key); }
   bool contains(const ANTLRTokenKey& key) { return names.find(key) != names.end(); }
   void remove(const ANTLRTokenKey& key) 
   { 
      assert(contains(key));
      names.erase(names.find(key));
   }
private:
   set<ANTLRTokenKey, less<ANTLRTokenKey> > names;
};

///////////////////////////////////////////////////////////////////////
// Hash objects for keys -- unused
///////////////////////////////////////////////////////////////////////
#if 0
struct TokenHash
{
   unsigned long operator()(const ANTLRTokenKey& tok) const
   {
      register unsigned val = 0;
      for (
         register const char *s = tok->getText();
         *s != 0;
         s++
      )
      {
         val = (val << 2) + *s;
      }
      return val;
   }
};

#endif


#endif    //INCL_CPRE_SYMTABLE_H
