#include "tokens.h"
#include "CPreParserImp.h"
#include "DLGLexer.h"
#include <string>
#include <iostream>

static void usage(char* progname);

int main(int argc, char *argv[])
{
   // Testing for amount of memory used.  I wish I knew how to query the heap size.
   char* p_heap1 = new char[100000];

   // Parse command-line options
   string includePath;
   string definitions;
   FILE *inputFile = stdin;
   FILE *outputFile = stdout;
   bool doTrace = false;
   bool doTrackInclude = false;
   bool doHeapUsage = false;
   char *progname = *argv;
   argc--;
   argv++;
   while (argc > 0)
   {
      if (argv[0][0] == '-')
      {
         switch (argv[0][1])
         {
            case 'D':
               if (argv[0][2] != 0)
               {
                  // argument is part of "-D"
                  definitions += ';';
                  definitions += &argv[0][2];
               } else if (argc > 1)
               {
                  // argument follows "-D"
                  argc--;
                  argv++;
                  definitions += ';';
                  definitions = *argv;
               } else {
                  usage(progname);
               }
               break;
            case 'h':
               if (strcmp(argv[0], "-heap") == 0)
               {
                  doHeapUsage = true;
               }
               break;
            case 'I':
               if (argv[0][2] != 0)
               {
                  // argument is part of "-I"
                  includePath += ';';
                  includePath += &argv[0][2];
               } else if (argc > 1)
               {
                  argc--;
                  argv++;
                  includePath += ';';
                  includePath = *argv;
               } else {
                  usage(progname);
               }
               break;
            case 'o':
            {
               char *ofilename = "";
               if (argv[0][2] != 0)
               {
                  // argument is part of "-o"
                  ofilename = &argv[0][2];
               } else if (argc > 1)
               {
                  argc--;
                  argv++;
                  ofilename = *argv;
               } else {
                  usage(progname);
               }
               outputFile = fopen(ofilename, "w");
               if (outputFile == NULL)
               {
                  fprintf(stderr, "%s: cannot open %s for output\n", progname, ofilename);
                  exit(1);
               }
            }
               break;
            case 't':
               if (strcmp(*argv, "-trace") == 0)
               {
                  doTrace = true;
               } else if (strcmp(*argv, "-trackInclude") == 0)
               {
                  doTrackInclude = true;
               } else {
                  usage(progname);
               }
               break;
            default:
               usage(progname);
               break;
         }
      } else {
         // must be the input filename
         if (inputFile != stdin)
         {
            // already have one
            usage(progname);
         }
         FILE *fp = fopen(*argv, "r");
         if (fp == NULL)
         {
            fprintf(stderr, "%s: cannot open %s for input\n", progname, *argv);
            exit(0);
         }
         inputFile = fp;
      }
      argc--;
      argv++;
   }

   // Create input stream
	DLGFileInput input(inputFile);

   // Create lexer
	DLGLexer scan(&input);
	FastToken tok;
	scan.setToken(&tok);

   // Create preprocessor parser.  Note that the preprocessor parser has
   // a built-in token buffer in the form of an input stack.
	CPreParserImp parser(&scan);
	parser.init();

   // for debugging
   if (includePath.length() == 0)
   {
      includePath = "\\msdev\\devstudio\\vc\\include;\\msdev\\devstudio\\vc\\mfc\\include";
   }
   if (definitions.length() == 0)
   {
      definitions = 
         "__cplusplus;"
         "__DATE__=\"date\";"
         "__FILE__=\"filename\";"
         "__LINE__=1;"
         "__TIME__=\"time\";"
         "__TIMESTAMP__=\"timestamp\";"
         "_M_IX86=400;"
         "_MSC_VER=1100;"
         "_WIN32;"
         "_INTEGRAL_MAX_BITS=64";
   }

   // Preset some information in the parser to make it act like MSVC++
   parser.SetIncludePath(includePath);
   parser.SetDefinitions(definitions);
   parser.SetOutputFile(outputFile);
   parser.doTrace(doTrace);

   // Set options in parser to make it act like MSVC++ 4.2
   parser.SetOption(CPreParserImp::OptMSExtensions, true);
   parser.SetOption(CPreParserImp::OptBool, true);
   parser.SetOption(CPreParserImp::OptWCharT, true);
   parser.SetOption(CPreParserImp::OptPragmas, true);
   parser.SetOption(CPreParserImp::OptExpandPragmas, true);
   parser.SetOption(CPreParserImp::OptTrackInclude, doTrackInclude);

   // Process top-level rule
   while (parser.preprocessing_chunk())
   {}

   // Close the input file stream
   if (inputFile != stdin)
   {
      fclose(inputFile);
   }

   if (doHeapUsage)
   {
      // For reporting memory usage
      char* p_heap2 = new char[100000];
      cout << "Approx heap usage: " << (long)(p_heap2 - p_heap1) << endl;
      delete [] p_heap2;
   }

   return 0;
}

static void usage(char* progname)
{
   fprintf(
      stderr, 
      "usage: %s [-I directories] [-D definitions] [-trace] [-o ofile] [ifile]\n"
      "    -I directories: semicolon-separated list of include directories\n"
      "    -D definitions: semicolon-separated list of definitions, like:\n"
      "                    symbol1;symbol2=value2\n"
      "    -trace: output trace of recursive-descent parse\n"
      "    -trackInclude: output #include tracking info\n"
      "    -heap: report heap usage\n"
      "    -o ofile: name of output file (otherwise it uses stdout)\n"
      "    ifile: name of input file (otherwise it uses stdin)\n",
      progname
   );
   exit(1);
}

