// FastToken.h:  Implementation of a token with quick allocation and
// low memory usage.

#ifndef INCL_FASTTOKEN_H
#define INCL_FASTTOKEN_H

#include "AToken.h"
#include "JLFreelist.h"

//////////////////////////////////////////////////////////////////////
//      FastToken: base class for the FastToken family of classes   //
//////////////////////////////////////////////////////////////////////
class FastToken : public ANTLRAbstractToken
{
public:
	FastToken(
      ANTLRTokenType t = (ANTLRTokenType)0,
      int line_ = 0
   ) :
      line(line_)
   {
      setType(t);
   }

   // Line number interface
   protected:
	   short line;
   public:
	   virtual int getLine() const { return line; }
	   virtual void setLine(int line_) { line = (short)line_; }

   // Static method used to make tokens
	static ANTLRAbstractToken *newToken(
      ANTLRTokenType tt = (ANTLRTokenType)0, 
      const char *txt = "", 
      int line = 0
   );

   // Text interface -- here only to allow instantiation so we can pass
   // one to the lexer.  Will fail if called.
 	virtual const ANTLRChar *getText() const { assert(0); return 0; }
	virtual void setText(const char *) { assert(0); }


   // Virtual method to make token (used by the lexer)
	virtual ANTLRAbstractToken *makeToken(
      ANTLRTokenType tt,
		ANTLRChar *txt,
		int line
   )
	{
      return newToken(tt, txt, line);
	}
};

//////////////////////////////////////////////////////////////////////
//      FastTokenVar: Token that can hold variable-length text      //
//////////////////////////////////////////////////////////////////////
class FastTokenVar : public FastToken
{
public:
   // ctor/dtor
   FastTokenVar(
      ANTLRTokenType tt,
		const char *text_,
		int line_
   ) :
      FastToken(tt, line_)
   {
      InitBuffer(text_);
   }
   virtual ~FastTokenVar() { delete [] text; }

   // Memory allocation
   void* operator new(size_t size)
   {
      // Subclasses should overide operators new and delete
      assert(size == sizeof(FastTokenVar));
      return freeList.New();
   }
   void operator delete(void* ptr, size_t size)
   {
      // Subclasses should overide operators new and delete
      assert(size == sizeof(FastTokenVar));
      freeList.Delete(ptr);
   }

   // Text interface
   protected:
	   char *text;
      int bufSize;
      void InitBuffer(const char* s)
      {
         bufSize = strlen(s) + 1;
         text = new char[bufSize];
         strcpy(text, s);
      }
   public:
	   virtual const ANTLRChar *getText() const { return text; }
	   virtual void setText(const char *s);

   // Fast allocator pool
   enum { FreeListBlockSize = 200 };
   static JLFreeList freeList;
};


/////////////////////////////////////////////////////////////////////
// Fixed-length token template.
// This is used to create a family of fixed-length-buffer tokens
// that can quickly allocate a token of the appropriate size.
/////////////////////////////////////////////////////////////////////
template <int sz> class FastTokenFixed : public FastToken
{
public:
   FastTokenFixed(
      ANTLRTokenType tt,
		const char *text_,
		int line_
   ) :
      FastToken(tt, line_)
   {
      assert(strlen(text_) < sizeof(text));
      strcpy(text, text_);
   }

   virtual ~FastTokenFixed() {}

   // Memory allocation
   void* operator new(size_t size)
   {
      // Subclasses should overide operators new and delete
      assert(size == sizeof(FastTokenFixed<sz>));
      return freeList.New();
   }
   void operator delete(void* ptr, size_t size)
   {
      // Subclasses should overide operators new and delete
      assert(size == sizeof(FastTokenFixed<sz>));
      freeList.Delete(ptr);
   }

   // Text interface
   protected:
	   char text[sz];
   public:
	   virtual const ANTLRChar *getText() const { return text; }
	   virtual void setText(const ANTLRChar *s)
      {
         assert(strlen(s) < sizeof(text));
         strcpy(text, s);
      }

private:
   // Fast allocator pool
   enum { FreeListBlockSize = 200 };
   static JLFreeList freeList;
};


#endif
