// JLSmartFlag.h: Template for creating smart enumerated flag sets.

#ifndef INCL_JLSMARTFLAG_H
#define INCL_JLSMARTFLAG_H

#include <assert.h>

/////////////////////////////////////////////////////////////////
// Template for creating smart enumerated flag sets.  They are
// smart because:
//     1) They are always initialized.
//     2) With debug on, they will check assignments
//     3) They implement the logical and bitwise operations,
//        but exclude other (math and relational)
//
// They are not completely general in that they are limited
// to the range of an unsigned long integer.  This is really
// intended to be a safe but efficient replacement for
// straight ORed flags.
//
// The second template parameter is there to give a unique
// signature to the template instantiation, which allows the
// creation of two identical flag ranges without allowing
// assignment between the two.  For example, these two
// flag sets are identical types and are operationally compatible:
//    typedef SmartFlag<0x20, 1> A, B;
// but these two are incompatible:
//    typedef SmartFlag<0x20, 0> A;
//    typedef SmartFlag<0x20, 1> B;
//
// Example: create a flag set whose max flag is 0x20:
//     typedef SamrtFlag(0x20) MyFlag;
/////////////////////////////////////////////////////////////////

template <int maxFlag, int signature=0> class SmartFlag {
public:
   enum { 
      MaxFlag = maxFlag,
      MaxValue = (maxFlag << 1) - 1 
   };

   // ctor
   SmartFlag(unsigned long value_ = 0) : value(value_)
   { 
      assert(IsSingleBit(MaxFlag));
      assert(value <= MaxValue); 
   }
   SmartFlag(const SmartFlag& rhs) : value(rhs.value) {}

   // Assignment
   SmartFlag& operator=(const SmartFlag& rhs)
   { 
      value = rhs.value; 
      return *this; 
   }
   SmartFlag& operator|=(const SmartFlag& rhs)
   { 
      value |= rhs.value; 
      return *this; 
   }
   SmartFlag& operator&=(const SmartFlag& rhs)
   { 
      value &= rhs.value; 
      return *this; 
   }
   SmartFlag& operator^=(const SmartFlag& rhs)
   { 
      value ^= rhs.value; 
      return *this; 
   }

   // Comparision
   bool operator==(const SmartFlag& rhs) const
   {
      return value == rhs.value;
   }
   bool operator!=(const SmartFlag& rhs) const
   {
      return value != rhs.value;
   }

   // Unary Operations
   SmartFlag operator~() const
   {
      return SmartFlag(~value & MaxValue);
   }

   // Binary Operations
   SmartFlag operator&(const SmartFlag& rhs) const
   { 
      return SmartFlag(value & rhs.value); 
   }
   SmartFlag operator|(const SmartFlag& rhs) const
   { 
      return SmartFlag(value | rhs.value); 
   }
   SmartFlag operator^(const SmartFlag& rhs) const
   { 
      return SmartFlag(value ^ rhs.value); 
   }

   // Conversion to integer.  For safety, don't make this
   // an implicit conversion operator
   unsigned int GetValue() const { return value; }

   // Ordering.  This is defined solely so that the less<T> functor
   // can be used to order maps with this as the key
   bool operator < (const SmartFlag& rhs) const
   {
      return value < rhs.value;
   }

protected:
   static int BitCount(unsigned long value)
   {
      unsigned long tmp = value;
      int bitCount = 0;
      while (tmp != 0) 
      {
         if ( (tmp & 1) != 0 )
         {
            bitCount++;
         }
         tmp >>= 1;
      }
      return bitCount;
   }

   static bool IsSingleBit(unsigned long value) { return BitCount(value) == 1; }

private:
   unsigned long value;

};

#endif
