// STLIncomplete.h: 
//
// This header file serves seveal purposes:
// 1) Forword declarations for STL classes
// 2) Processing of STLI_* macros that configure the STL implementation.
// 3) defines a DEFAULT_ALLOCATOR() macro that can be used for the
//    Allocator template argument. 

#ifndef INCL_STL_INCOMPLETE_H
#define INCL_STL_INCOMPLETE_H

// The following definitions can be set on the command-line, in the
// project, or in some header included before this header:
//    STLI_DEFAULT_ARGS: compiler accepts default template arguments
//    STLI_COMPLEX_DEFAULT_ARGS: compiler accepts default template 
//         arguments that depend on other template arguments
//    STLI_STANDARD_ALLOCATOR: your STL supports a full standard allocator
//    STLI_ALLOCATOR_IS_TEMPLATE: your STL implements allocators as templates.
//    STLI_BYTE_ALLOCATE: when STLI_STANDARD_ALLOCATOR is false, then 
//         this macro is used to allocate arbitrarily-sized blocks.
//    STLI_BYTE_DEALLOCATE: when STLI_STANDARD_ALLOCATOR is false, then 
//         this macro is used to free arbitrarily-sized blocks.
//
// If you don't define any of these, then this header file will set
// the #defines for a maximally-compliant STL implementation.

//
// Some predefined sets of defines for specific compiler/STL combinations
//
#define STLI_IMP_STANDARD             1   // Standard-compliant STL
#define STLI_IMP_MSVC_ROGUE           2   // MSVC++ 4.2 with Rogue wave STL
#define STLI_IMP_MSVC42_MSVC42        3   // MSVC++ 4.2 with internal STL

// Attempt to choose an implementation based on compiler constants
#ifndef STLI_IMP
   #if defined(_MSC_VER) && _MSC_VER == 1020
      // The default implementation is set for MSVC++ 4.2 using internal STL
      #define STLI_IMP STLI_IMP_MSVC42_MSVC42
   #endif
#endif

#if STLI_IMP == STLI_IMP_MSVC42_ROGUE
   #define STLI_DEFAULT_ARGS 1
   #define STLI_COMPLEX_DEFAULT_ARGS 0
   #define STLI_STANDARD_ALLOCATOR 0
   #define STLI_ALLOCATOR_IS_TEMPLATE 0
   #define STLI_MEMBER_TEMPLATES 0
#elif STLI_IMP == STLI_IMP_MSVC42_MSVC42
   #define STLI_DEFAULT_ARGS 1
   #define STLI_COMPLEX_DEFAULT_ARGS 0
   #define STLI_STANDARD_ALLOCATOR 0
   #define STLI_ALLOCATOR_IS_TEMPLATE 1
   #define STLI_MEMBER_TEMPLATES 0
   #define STLI_BYTE_ALLOCATE(allocator, amount) (void*)allocator.Charalloc(amount)
#endif

//
// Set defines to maximally-compliant STL if not already set
//
#if !STLI_DEFAULT_ARGS
   // Cannot have complex default args if there are no default args
   #undef STLI_COMPLEX_DEFAULT_ARGS
   #define STLI_COMPLEX_DEFAULT_ARGS 0
#endif
#ifndef STLI_DEFAULT_ARGS
   #define STLI_DEFAULT_ARGS 1
#endif
#ifndef STLI_COMPLEX_DEFAULT_ARGS
   #define STLI_COMPLEX_DEFAULT_ARGS 1
#endif
#ifndef STLI_STANDARD_ALLOCATOR
   #define STLI_STANDARD_ALLOCATOR 1
#endif
#ifndef STLI_ALLOCATOR_IS_TEMPLATE
   #define STLI_ALLOCATOR_IS_TEMPLATE 0
#endif
#ifndef STLI_MEMBER_TEMPLATES
   #define STLI_MEMBER_TEMPLATES 1
#endif


// Define some macros to handle the two types of default template arguments
#if STLI_DEFAULT_ARGS
   #define STLI_SIMPLE_DEFAULT_ARG(a) = a
   #if STLI_ALLOCATOR_IS_TEMPLATE && !STLI_COMPLEX_DEFAULT_ARGS
      // When the allocator is a template, then even simple default
      // args cannot be used because the allocator is complex
      #define STLI_SIMPLE_DEFAULT_ARG_WITH_ALLOC(a)
   #else
      #define STLI_SIMPLE_DEFAULT_ARG_WITH_ALLOC(a) = a
   #endif
#else
   #define STLI_SIMPLE_DEFAULT_ARG(a)
   #define STLI_SIMPLE_DEFAULT_ARG_WITH_ALLOC(a)
#endif

#if STLI_COMPLEX_DEFAULT_ARGS
   #define STLI_COMPLEX_DEFAULT_ARG(a) = a
#else
   #define STLI_COMPLEX_DEFAULT_ARG(a)
#endif


// Macro for building the default allocator argument given type T
#if STLI_ALLOCATOR_IS_TEMPLATE
   template <class T> class allocator;
   #define STLI_DEFAULT_ALLOCATOR(T) allocator<T>
#else
   class allocator;
   #define STLI_DEFAULT_ALLOCATOR(T) allocator
#endif
// Macro used for the default allocator argument
#define STLI_DEFAULT_ALLOCATOR_ARG(T) STLI_COMPLEX_DEFAULT_ARG(DEFAULT_ALLOCATOR(T))


// Macro used to allocate arbitrary bytes via the allocator interface 
#ifndef STLI_BYTE_ALLOCATE
   #if STLI_STANDARD_ALLOCATOR
      // Use the template member function
      #define STLI_BYTE_ALLOCATE(allocator, amount) (void*)allocator.allocate<char,char>(amount)
   #else
      #if STLI_ALLOCATOR_IS_TEMPLATE
         #error "You must define a byte-allocation macro for your STL"
      #else
         // This works for Rogue Wave, and probably others
         #define STLI_BYTE_ALLOCATE(allocator, amount) allocator.allocate(amount)
      #endif
   #endif
#endif


// Macro used to free arbitrary bytes via the allocator interface 
#ifndef STLI_BYTE_DECALLOCATE
   #if STLI_STANDARD_ALLOCATOR
      // Use the template member function
      #define STLI_BYTE_DEALLOCATE(allocator, ptr) allocator.deallocate<char>(ptr)
   #else
      // Call the non-templated deallocate; this works for many STLs
      #define STLI_BYTE_DEALLOCATE(allocator, ptr) allocator.deallocate(ptr)
   #endif
#endif


// These are the common and simple ones
template <class T> struct less;

// Templates that use simple default arguments, unless the allocator
// is a template in which case they are complex default arguments
template <class T, class Allocator STLI_DEFAULT_ALLOCATOR_ARG(T) > class deque;
template <class T, class Allocator STLI_DEFAULT_ALLOCATOR_ARG(T) > class list;
template <class T, class Allocator STLI_DEFAULT_ALLOCATOR_ARG(T) > class JL_DList;
template <class T, class Allocator STLI_DEFAULT_ALLOCATOR_ARG(T) > class vector;

// Templates that use complex default arguments
template <
   class T, 
   class Imp STLI_COMPLEX_DEFAULT_ARG(deque<T>),
   class Allocator STLI_DEFAULT_ALLOCATOR_ARG(T)
> class stack;

template <
   class Key, 
   class T, 
   class Compare STLI_COMPLEX_DEFAULT_ARG(less<Key>),
   class Allocator STLI_DEFAULT_ALLOCATOR_ARG(T)
> class map;

template <
   class Key, 
   class T, 
   class Compare STLI_COMPLEX_DEFAULT_ARG(less<Key>),
   class Allocator STLI_DEFAULT_ALLOCATOR_ARG(T)
> class multimap;

template <
   class Key, 
   class Compare STLI_COMPLEX_DEFAULT_ARG(less<Key>),
   class Allocator STLI_DEFAULT_ALLOCATOR_ARG(T)
> class set;

#endif

