// DebugOutput.cpp -- implementation of CDebugOutput, CDebugTracker

#include <stdio.h>
#include "DebugOutput.h"
#include <afxwin.h>

///////////////////////////////////////////////////////////////////////
//                          DebugOutputImp declaration               //
///////////////////////////////////////////////////////////////////////

class MyWnd : public CWnd
{
public:
   MyWnd();
   virtual ~MyWnd();
   CEdit& GetEdit() { return cedit; }
protected:
	// Generated message map functions
	//{{AFX_MSG(MyWnd)
	afx_msg void OnSize(UINT nType, int cx, int cy);
   afx_msg int OnCreate(LPCREATESTRUCT lpCreateStruct);
	//}}AFX_MSG
	DECLARE_MESSAGE_MAP()
private:
   CEdit cedit;
};

class CDebugOutputImp
{
public:
   // If outputFilename is non-null, then debug output also goes to the file.
   CDebugOutputImp(
      const char* outputFilename_ = NULL,
      int indentSpacing = 3
   );
   ~CDebugOutputImp();

   // These methods are forwarded to implementation class
   void Indent();
   void Unindent();
   void Output(const CString& str);

private:
   // CWnd sub-class used to implement the output window
   MyWnd* mywnd;

   void Init();

   // misc data
   int indentSpacing;
   int indentLevel;
   CString outputFilename;
   FILE* fp;

   // constants
   enum { 
      MaxLines = 500      // max lines in edit control -- file is unlimited
   };
};


///////////////////////////////////////////////////////////////////////
//                             DebugOutput                           //
///////////////////////////////////////////////////////////////////////

CDebugOutput::CDebugOutput(
   const char* outputFilename, 
   int indentSpacing
)
{
   imp = new CDebugOutputImp(outputFilename, indentSpacing);
}

CDebugOutput::~CDebugOutput() { delete imp; }

void CDebugOutput::Indent()      { imp->Indent(); }
void CDebugOutput::Unindent()    { imp->Unindent(); }

void CDebugOutput::Output(
   const CString& str
)
{
   imp->Output(str);
}

void CDebugOutput::Output(
   const char *str
)
{
   imp->Output(str);
}


///////////////////////////////////////////////////////////////////////
//                             CDebugTracker                         //
///////////////////////////////////////////////////////////////////////

CDebugTracker::CDebugTracker(
   const CString& methodName_
) :
   methodName(methodName_)
{
   CString msg("Entering ");
   msg += methodName;
   output.Output(msg);
   output.Indent();
}

CDebugTracker::~CDebugTracker()
{
   CString msg("Exiting ");
   msg += methodName;
   output.Unindent();
   output.Output(msg);
}

CDebugOutput CDebugTracker::output("DebugOut.txt");

///////////////////////////////////////////////////////////////////////
//                 DebugOutputImp implementation                     //
///////////////////////////////////////////////////////////////////////

CDebugOutputImp::CDebugOutputImp(
   const char* outputFilename_,
   int indentSpacing_
) : 
   fp(0),
   indentSpacing(indentSpacing_),
   indentLevel(0),
   mywnd(0),
   outputFilename(outputFilename_)
{
}

CDebugOutputImp::~CDebugOutputImp()
{
   if (fp != 0)
   {
      fclose(fp);
   }
   if (mywnd != 0)
   {
      delete mywnd;
   }
}

void CDebugOutputImp::Indent()
{
   indentLevel++;
}

void CDebugOutputImp::Unindent()
{
   indentLevel--;
}

void CDebugOutputImp::Output(
   const CString& str
)
{
   Init();

   // Construct an output string with the right indent level
   CString outstr(' ', indentLevel * indentSpacing);
   // Append message
   outstr += str;

   // Output to file
   if (fp != 0)
   {
      fputs(outstr, fp);
      fputc('\n', fp);
   }

   CEdit& edit = mywnd->GetEdit();

   // Remove first line if needed
   if (edit.GetLineCount() > MaxLines)
   {
      int firstLinePos = edit.LineIndex(1);
      edit.SetSel(0, firstLinePos, TRUE);
      edit.ReplaceSel("");
   }

   // Add to edit control
   outstr += "\r\n";
   int textEndPos = edit.LineIndex(edit.GetLineCount()-1);
   edit.SetSel(textEndPos, textEndPos);
   edit.ReplaceSel(outstr);
}

void
CDebugOutputImp::Init()
{
   if (mywnd != 0)
   {
      return;
   }

   mywnd = new MyWnd();
   mywnd->CreateEx(
      0,                           // dwExStyle
      ::AfxRegisterWndClass(0),    // class name
      "Debug Output",
      WS_VISIBLE | WS_OVERLAPPEDWINDOW,
      0, 
      0,
      ::GetSystemMetrics(SM_CXFULLSCREEN) / 3,
      ::GetSystemMetrics(SM_CYFULLSCREEN),
      0,                      // parent Cwnd
      0                       // menu id
   );

   if (outputFilename.GetLength() != 0)
   {
      fp = fopen((const char*)outputFilename, "w");
      if (fp == NULL)
      {
         CString msg("Error: cannot open file for debug output: ");
         msg += outputFilename;
         Output(msg);
      }
   }
}


BEGIN_MESSAGE_MAP(MyWnd, CWnd)
	//{{AFX_MSG_MAP(MyWnd)
	ON_WM_CREATE()
	ON_WM_SIZE()
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

MyWnd::MyWnd() 
{}

MyWnd::~MyWnd()
{}

void 
MyWnd::OnSize(
   UINT nType, 
   int cx, 
   int cy
)
{
   cedit.MoveWindow(0, 0, cx, cy);
}

int 
MyWnd::OnCreate( 
   LPCREATESTRUCT lpCreateStruct
)
{
   CWnd::OnCreate(lpCreateStruct);
   cedit.Create(
      WS_CHILD | WS_VISIBLE | ES_LEFT | ES_MULTILINE | ES_AUTOVSCROLL | ES_READONLY | ES_WANTRETURN,
      CRect(), 
      this, 
      1           // id
   );

   return 0;
}
