// JLString.cpp: misc string operations

#include "JLStr.h"
#include <ctype.h>

//////////////////////////////////////////////////////////////////////
// Escape: convert '\' escapes to non-printing characters
// The following escapes are supported:
// 
//     \r     carriage return (13)
//     \n     newline (10)
//     \t     tab (8)
//     \"     quote (")
//     \\     backslash (\)   (unless keepDoubleBackslash is true)
//     \0nnn  octal escape
//     \xddd  hex escape
//
// Other backslashes are left untouched
//////////////////////////////////////////////////////////////////////
JLStr
JLStr::Escape(
   bool keepDoubleBackslash
) const
{
   assertValid();

   // Loop to remove escapes, stripping leading and trailing quote
   JLStr newStr;
   for (
      JLStr::const_iterator iter(begin());
      iter != end();
      iter++
   )
   {
      if (*iter == '\\')
      {
         iter++;
         if (iter != end())
         {
            switch (*iter)
            {
            case 'r':
               newStr += '\r';
               break;
            case 'n':
               newStr += '\n';
               break;
            case 't':
               newStr += '\t';
               break;
            case '\\':
               if (keepDoubleBackslash)
               {
                  newStr += '\\';
               }
               newStr += '\\';
               break;
            case '"':
               newStr += '"';
               break;
            case '0':
               {
                  // octal escape code
                  int value = 0;
                  iter++;
                  while (
                     iter != end() && 
                     isdigit(*iter) && *iter != '8' && *iter != '9'
                  )
                  {
                     value = value * 8 + *iter - '0';
                     iter++;
                  }
                  iter--;
                  newStr += (char)value;
               }
               break;
            case 'x':
               {
                  // hexadecimal escape code
                  int value = 0;
                  iter++;
                  while (iter != end() && isxdigit(*iter))
                  {
                     if (isdigit(*iter))
                     {
                        value = value * 16 + *iter - '0';
                     } else {
                        value = value * 16 + 10 + tolower(*iter) - 'a';
                     }
                     iter++;
                  }
                  iter--;
                  newStr += (char)value;
               }
               break;
            default:
               // leave backslash untouched otherwise
               newStr += '\\';
               newStr += *iter;
               break;
            }
         } else {
            // backslash at end of string
            newStr += '\\';
         }
      } else {
         newStr += *iter;
      }
   }
   return newStr;
}


//////////////////////////////////////////////////////////////////////
// UnEscape: convert non-printing characters to '\' escapes:
// The following escapes are supported:
// 
//     \r     carriage return (13)
//     \n     newline (10)
//     \t     tab (8)
//     \"     quote (")
//     \\     backslash (\)
//     \0nnn  octal escape
//     \xddd  hex escape
//////////////////////////////////////////////////////////////////////
JLStr 
JLStr::UnEscape() const
{
   assertValid();

   // Loop to remove escapes, stripping leading and trailing quote
   JLStr newStr;
   for (
      JLStr::const_iterator iter(begin());
      iter != end();
      iter++
   )
   {
      if (!isprint(*iter))
      {
         switch (*iter)
         {
         case '\r':
            newStr += "\\r";
            break;
         case '\n':
            newStr += "\\n";
            break;
         case '\t':
            newStr += "\\t";
            break;
         case '\\':
            newStr += "\\\\";
            break;
         case '\"':
            newStr += "\\\"";
            break;
         default:
            {
               char buf[10];
               sprintf(buf, "\\x%x", (unsigned char)*iter);
               newStr += buf;
            }
               break;
         }
      } else {
         newStr += *iter;
      }
   }
   return newStr;
}


// Remove leading chars from string
JLStr
JLStr::StripLeading(
   char stripChar,
   int maxToStrip
) const
{
   assertValid();

   for (
      JLStr::const_iterator iter = begin();
      iter != end() && *iter == stripChar && maxToStrip > 0;
      iter++, maxToStrip--
   )
   {}
   return JLStr(iter, end());
}

// Remove trailing chars from string
JLStr
JLStr::StripTrailing(
   char stripChar,
   int maxToStrip
) const
{
   assertValid();

   for (
      JLStr::const_iterator iter = end();
      iter != begin() && *(iter-1) == stripChar && maxToStrip > 0;
      iter--, maxToStrip--
   )
   {}
   return JLStr(begin(), iter);
}


/////////////////////////////////////////////////////////////////////////
// Truncate: Extract a substring of a string, ignoring boundary errors.
//
// Inputs:
//    const JLStr&    str     The input string
//    int              offset  The offset into the string
//    int              maxlen  The maximum resulting string length
// Outputs: none
// Return:
//    JLStr     The desired substring
/////////////////////////////////////////////////////////////////////////
JLStr
JLStr::Truncate(
   int offset,
   int maxlen
) const
{
   assertValid();

   if (end() - begin() > offset)
   {
      JLStr tmpStr(begin()+offset, end());
      if ((int)tmpStr.size() > maxlen)
      {
         tmpStr.resize(maxlen);
      }
      return tmpStr;
   } else {
      return JLStr();
   }
}

// Trim characters from front and back
JLStr 
JLStr::Trim(
	char charToTrim,
   int maxToStrip
) const
{
   assertValid();

	return StripLeading(charToTrim, maxToStrip).StripTrailing(charToTrim, maxToStrip);
}


// Empty string
JLStr JLStr::Empty;
