/* AParser.h
 *
 * Define the generic ANTLRParser superclass, which is subclassed to
 * define an actual parser.
 *
 * Before entry into this file: ANTLRTokenType must be set.
 *
 * SOFTWARE RIGHTS
 *
 * We reserve no LEGAL rights to the Purdue Compiler Construction Tool
 * Set (PCCTS) -- PCCTS is in the public domain.  An individual or
 * company may do whatever they wish with source code distributed with
 * PCCTS or the code generated by PCCTS, including the incorporation of
 * PCCTS, or its output, into commerical software.
 * 
 * We encourage users to develop software with PCCTS.  However, we do ask
 * that credit is given to us for developing PCCTS.  By "credit",
 * we mean that if you incorporate our source code into one of your
 * programs (commercial product, research project, or otherwise) that you
 * acknowledge this fact somewhere in the documentation, research report,
 * etc...  If you like PCCTS and have developed a nice tool with the
 * output, please mention that you developed it using PCCTS.  In
 * addition, we ask that this header remain intact in our source code.
 * As long as these guidelines are kept, we expect to continue enhancing
 * this system and expect to make other tools available as they are
 * completed.
 *
 * ANTLR 1.33
 * Terence Parr
 * Parr Research Corporation
 * with Purdue University and AHPCRC, University of Minnesota
 * 1989-1995
 */

#ifndef INCL_APARSER_H
#define INCL_APARSER_H

#include <stdio.h>
#include "AToken.h"
#include "ATokPtr.h"

typedef unsigned char SetWordType;

/* Define external bit set stuff (for SetWordType) */
#define EXT_WORDSIZE   (sizeof(char)*8)
#define EXT_LOGWORDSIZE   3

/* a n t l r  p a r s e r  d e f */

struct ANTLRParserState {
   /* class variables */
   int guessing;
};

class ANTLRTokenBuffer;

/* notes:
 *
 * multiple inheritance is a cool way to include what stuff is needed
 * in this structure (like guess stuff).  however, i'm not convinced that
 * multiple inheritance works correctly on all platforms.  not that
 * much space is used--just include all possibly useful members.
 *
 * the class should also be a template with arguments for the lookahead
 * depth and so on.  that way, more than one parser can be defined (as
 * each will probably have different lookahead requirements).  however,
 * am i sure that templates work?  no, i'm not sure.
 *
 * no attributes are maintained and, hence, the 'asp' variable is not
 * needed.  $i can still be referenced, but it refers to the token
 * associated with that rule element.  question: where are the token's
 * stored if not on the software stack?  in local variables created
 * and assigned to by antlr.
 */
class ANTLRParser {

   // Give tracer access to protected methods
   class Tracer;
   friend class ANTLRParser::Tracer;

protected:
   /* class variables */
   static SetWordType bitmask[sizeof(SetWordType)*8];

protected:
   int LLk;                    // number of lookahead symbols
   ANTLRTokenType eofToken;    // when do I stop during resync (always==1)
   int tokenSetSize;           // size of bitsets created by ANTLR in
                               // units of SetWordType

   ANTLRTokenBuffer *inputTokens;   //place to get input tokens

   int guessing;               // if guessing (using (...)? predicate)

   const ANTLRChar * const *token_tbl;      // pointer to table of token type strings

   // fast reference cache of token.getType()
   // ref cache is sized for largest expected K
   enum {MAX_LLK=32};
   ANTLRTokenType token_type[MAX_LLK];

protected:

   //
   // Syntax and predicate exceptions
   // 

   // type thrown for syntax/predicate exceptions
   struct AParserEX {
      enum EXType { Syntax, Predicate };
      virtual EXType GetEXType() const = 0;
   };

   struct AParserEXSyntax : public AParserEX {
      AParserEXSyntax(SetWordType* set_) : 
         set(set_)
      {};
      virtual EXType GetEXType() const { return Syntax; }
      // data
      SetWordType* set;
   };

   struct AParserEXPredicate : public AParserEX {
      AParserEXPredicate(const char* msg_) : msg(msg_) {}
      virtual EXType GetEXType() const { return Predicate; }
      // data
      const char* msg;
   };

   // Set used to construct syntax set on the fly for single-token
   // syntax error.  This is allocated at runtime.
   SetWordType *tempTokenSet;

   // Input token buffer management for guess mode
   virtual int MarkInputTokens();
   virtual void RewindInputTokens(int marker);
   virtual void UnmarkInputTokens(int marker);

   // Commit to an alternate.  This is used for an alt with backtracking.
   // It simply stops the guess, so that errors become real errors instead
   // of backtracking.
   virtual void CommitGuess() { guessing = 0; }

   // Methods to throw syntax and predicate exceptions
   void fail_syntax(ANTLRTokenType tt);
   void fail_syntax(SetWordType* set_);
   void fail_predicate(const char* msg);

   // Match the tokentype or set; throw an exception on failure
   void _match(ANTLRTokenType);
   void _setmatch(SetWordType *);

   virtual void consume();
   void prime_lookahead();


   //
   // Trace output
   //

   // Used for output when trace is on -- base class goes to cout
   virtual void trace_output(const char* msg);

   // Trace object to be placed on stack
   class Tracer {
   public:
      Tracer(ANTLRParser* parser_, const char* msg_);
      ~Tracer();
   private:
      void OutputStuff(const char* msg) const;
      void tabs() const;
      ANTLRParser* parser;
      static int level;
      const char* msg;
   };
   // Trace mode
   int trace_on;

   unsigned MODWORD(unsigned x) const {return x & (EXT_WORDSIZE-1);}   // x % EXT_WORDSIZE
   unsigned DIVWORD(unsigned x) const {return x >> EXT_LOGWORDSIZE;}   // x / EXT_WORDSIZE
   int set_el(ANTLRTokenType, SetWordType *) const;

public:
   ANTLRParser(
      ANTLRTokenBuffer *inputTokens_,
      int k=1,
      int use_inf_look=0,  // unused
      int dlook=0,  // unused
      int tokenSetSize_=1
   );
   virtual ~ANTLRParser();

   virtual void init();
   
   ANTLRTokenType LA(int i)
   {
      return token_type[i-1];
   }
   ANTLRTokenPtr LT(int i);

   void saveState(ANTLRParserState *);
   void restoreState(ANTLRParserState *);

   virtual void panic(char *msg);

   void consumeUntil(SetWordType *st);
   void consumeUntilToken(int t);

   int isTracing() const { return trace_on; }
   void doTrace(int trace_on_) { trace_on = trace_on_; }
};


#define zzRULE   // empty in case we need it for something later


#endif
