/* DLGLexerBase.h
 *
 * SOFTWARE RIGHTS
 *
 * We reserve no LEGAL rights to the Purdue Compiler Construction Tool
 * Set (PCCTS) -- PCCTS is in the public domain.  An individual or
 * company may do whatever they wish with source code distributed with
 * PCCTS or the code generated by PCCTS, including the incorporation of
 * PCCTS, or its output, into commerical software.
 * 
 * We encourage users to develop software with PCCTS.  However, we do ask
 * that credit is given to us for developing PCCTS.  By "credit",
 * we mean that if you incorporate our source code into one of your
 * programs (commercial product, research project, or otherwise) that you
 * acknowledge this fact somewhere in the documentation, research report,
 * etc...  If you like PCCTS and have developed a nice tool with the
 * output, please mention that you developed it using PCCTS.  In
 * addition, we ask that this header remain intact in our source code.
 * As long as these guidelines are kept, we expect to continue enhancing
 * this system and expect to make other tools available as they are
 * completed.
 *
 * ANTLR 1.33
 * Terence Parr
 * Parr Research Corporation
 * with Purdue University and AHPCRC, University of Minnesota
 * 1989-1995
 */

#ifndef INCL_DLGLEXERBASE_H
#define INCL_DLGLEXERBASE_H

#include "AToken.h"
#include "ATokenStream.h"
#include <stdio.h>

/* must define what a char looks like; can make this a class too */
typedef char DLGChar;

/* user must subclass this */
class DLGInputStream {
public:
   virtual ~DLGInputStream() {}
	virtual int nextChar() = 0;

   // Override this if the stream has an efficient block read function
   virtual int readChars(char *buf, int size) 
   {
      int c = nextChar();
      if (c != EOF)
      {
         *buf = (char)c;
         return 1;
      } else {
         return 0;
      }
   }
};

// Character buffer internal to lexer avoids virtual
// call for each char from input stream
class DLGCharBuffer 
{
public:
	DLGCharBuffer(DLGInputStream* input_) : input(input_)
   {
      bufend = buf;
      next = buf;
   }
   int nextChar() 
   {
      if (next < bufend)
      {
         return *next++;
      } else {
         // get more input
         bufend = buf + input->readChars(buf, BufSize);
         if (bufend == buf)
         {
            return EOF;
         } else {
            next = buf;
            return *next++;
         }
      }
	}
private:
   enum { BufSize = 1024 };
	char buf[BufSize];
   char *bufend;
   char *next;
	DLGInputStream* input;
};

///////////////////////////////////////////////////////////
// Token stream input from FILE
class DLGFileInput : public DLGInputStream {
private:
	int found_eof;
	FILE *input;
public:
	DLGFileInput(FILE *f) { input = f; found_eof = 0; }

   // Single-character read interface
	int nextChar() 
   {
		int c;

      // Keep track of EOF because some broken implementation don't keep returning EOF.
		if ( found_eof ) 
      {
         return EOF;
      } else {
			c=getc(input);
			if ( c==EOF ) found_eof = 1;
			return c;
		}
	}

   // Block read interface
   virtual int readChars(char *buf, int size)
   {
		if ( found_eof ) 
      {
         // Because fread() will retry stdin
         return 0;
      } else {
         int retval = (int)fread(buf, 1, size, input);
         if (feof(input))
         {
            found_eof = 1;
         }
         return retval;
      }
   }
};


///////////////////////////////////////////////////////////
// Class to save/restore lexer state
class DLGState {
public:
	DLGInputStream *input;
	int auto_num;
	int add_erase;
	int lookc;
	int char_full;
	int begcol, endcol;
	int line;
	DLGChar *lextext, *begexpr, *endexpr;
	int bufsize;
	DLGChar *nextpos;
	DLGChar *lastpos;
	int class_num;
};

///////////////////////////////////////////////////////////
// Base lexer class.
class DLGLexerBase : public ANTLRTokenStream {
public:
   // ctor/dtor
	DLGLexerBase(
      DLGInputStream *in
   );
	virtual ~DLGLexerBase() { delete [] _lextext; }

   // External interface
	virtual ANTLRTokenType erraction();
	virtual void advance(void) = 0;
	void	skip(void);		/* erase lextext, look for antoher token */
	void	more(void);		/* keep lextext, look for another token */
	void	mode(int k);	/* switch to automaton 'k' */
	void	saveState(DLGState *);
	void	restoreState(DLGState *);
	virtual ANTLRTokenType nextTokenType(void)=0;/* get next token */
	void	replchar(DLGChar c);	/* replace last recognized reg. expr. with
									 a character */
	void	replstr(DLGChar *s);	/* replace last recognized reg. expr. with
									 a string */

   //
   // error reporting
   //
	int err_in();
	void errstd(char *);
   virtual void errout(char*);
	virtual void panic(char *msg);

   // Current line
	int line() { return _line; }
	virtual void newline() { _line++; }

   // Access to current token text
	DLGChar *lextext() { return _lextext; }

   // Current Column
	int begcol() { return _begcol; }
	int endcol() { return _endcol; }
	void set_begcol(int a) { _begcol=a; }
	void set_endcol(int a) { _endcol=a; }

   // Access to begin/end of current match
	DLGChar *begexpr() { return _begexpr; }
	DLGChar *endexpr() { return _endexpr; }

   // Get the size of current token text buffer
	int bufsize() { return _bufsize; }

   // Set the token used to clone other tokens
	void	setToken(ANTLRAbstractToken *t)	{ token_to_fill = t; }

   // Change the input char stream
	void	setInputStream(DLGInputStream *);

protected:
	DLGInputStream *input;
	int interactive;
	DLGChar	*_lextext;	/* token text buffer */
	int	_bufsize;		/* number of characters in lextext */
	DLGChar *nextpos;	/* points to next available position in lextext*/
	DLGChar *lastpos;    /* last space reserved for the null char */

	DLGChar	*_begexpr;	/* beginning of last reg expr recogn. */
	DLGChar	*_endexpr;	/* beginning of last reg expr recogn. */
	int	_begcol;		   /* column that first character of token is in*/
	int	_endcol;		   /* column that last character of token is in */
	int	_line;			/* line current token is on */
	int	ch;				/* character to determine next state */
	int	charfull;
	int	cl;
	int automaton;
	int	add_erase;
	DLGChar ebuf[70];

   // Token used to clone more tokens
	ANTLRAbstractToken* token_to_fill;

	virtual ANTLRTokenPtr getToken();

   // Extend the _lextext buffer to hold more text
   void extend_lextext();

   // Intermediate buffer to speed up character reads
   DLGCharBuffer buffer;

   // Copy next character into buffer
   void CopyCharToLexText()
   {
      if (nextpos+1 >= lastpos)
      {
         extend_lextext();
      }
      *nextpos = ch;
      nextpos++;
   }

   enum { DefaultBufSize = 32 };
};

#endif
